/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

// Streams.cxx: implementation of the CStreams class.
//
//////////////////////////////////////////////////////////////////////

#include "geniedefs.h"

#include <stdlib.h>

#ifdef _WIN32
#   include <io.h>
#   include <fcntl.h>
#endif

#include "Streams.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

// Undefine the override definitions in geniedefs.h to allow us to use the streams
#undef cout
#undef cin
#undef cerr

#ifdef _WIN32
// Fix up console handles with I/O streams in terminal mode
static void fix_handle(FILE* stream, DWORD handle, const char* opts);
#endif

unsigned CStreams::s_dwSessionID = 0;


// Note that setting bInf or bDebug to false will just disable O/P for that stream for
// the life of this streams object, GENIE actually uses finer control for switching
// informational messages on/off.
CStreams::CStreams(bool bTerm, bool bInf, bool bDebug)
{
    if (s_dwSessionID != 0)
    {
	// this should be OK as we have at least one set of streams already
	gerr << "Not allowed to create more than one streams object per process - aborting" << endl;
	exit(1);
    }
    else
    {
	s_dwSessionID++;
	
	// If we are using terminal I/O, initialize that now
	if (bTerm)  // Things not to forget if we are interacting on a console
	{
	    
	    // allocate a console for I/O (only on WIN32)
#if defined(_WIN32) && defined(_DEBUG)
	    AllocConsole();				// Trap debuggin I/O
	    fix_handle(__si, STD_INPUT_HANDLE, "rt");	// Fix C stdio handles for _WIN32
	    fix_handle(__so, STD_OUTPUT_HANDLE, "wt");	// note: ( __si is saved stdin, __so stdout and __se stderr )
	    fix_handle(__se, STD_ERROR_HANDLE, "wt");
#endif
	    
	    // Make sure C and C++ I/O work together
	    ios::sync_with_stdio();
	    
	    // Tie streams together so output gets flushed before input requested
	    cin.tie(&cout);	    
	}
	
	// Create the streams
	m_pgoIn = new istream(cin.rdbuf()),
	m_pgoOut = new genie_ostream(cout, NEUTRAL, bTerm, 0);
	m_pgoInf = new genie_ostream(cout, BLUE, bInf & bTerm, 0);
	m_pgoErr = new genie_ostream(cerr, RED, bTerm, 0);
	m_pgoDbg = new genie_ostream(cout, PINK, bDebug & bTerm, 0);
	
	// Now record these global streams for use by the rest of GENIE
	g_in = m_pgoIn;
	g_out = m_pgoOut;
	g_inf = m_pgoInf;
	g_err = m_pgoErr;
	g_dbg = m_pgoDbg;
    }
}

CStreams::~CStreams()
{
    // delete the streams
    delete m_pgoIn;
    delete m_pgoOut;
    delete m_pgoInf;
    delete m_pgoErr;
    delete m_pgoDbg;
    // Do this last in case something bombs in the kill
    s_dwSessionID--;
}

const char* CStreams::Out()
{
    return m_pgoOut->str();
}

const char* CStreams::Inf()
{
    return m_pgoInf->str();
}

const char* CStreams::Err()
{
    return m_pgoErr->str();
}

const char* CStreams::Dbg()
{
    return m_pgoDbg->str();
}


// WIN32 utility for fixing up I/O streams with a generated console
#ifdef _WIN32
static void fix_handle(FILE* stream, DWORD handle, const char* opts)
{
    int hCrt;
    FILE* hf;
    hCrt = _open_osfhandle((long)GetStdHandle(handle), _O_TEXT);
    hf = _fdopen(hCrt, opts);
    *stream = *hf;
    setvbuf(stream, NULL, _IONBF, 0);
}
#endif

void CStreams::ClearAll()
{
    // Just accessing clears the buffers as the next access will refill
    // the buffers from scratch
    Out();
    Inf();
    Err();
    Dbg();
}
