/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

/*
 *      Module:         genie_callout_list
 *      Author:         Freddie Akeroyd, ISIS
 *      Purpose:        Decode encapsulated parameters passed out from Smalltalk in geniecall()
 *
 */
#include "geniedefs.h"
#include <stdlib.h>
#include <fstream.h>
#include <iostream.h>
#include <string.h>

#include "smalltalk.h"

#include "g_types.h"
#include "genie_base.h"
#include "parse_funcs.h"
#include "callout_list.h"

#if defined(__ultrix) || defined(_WIN32)
#	include "strcasecmp.h"
#endif

//#define DISPLAY_PROCESSING

static const char* rcsid = "@(#)$Id: callout_list.cxx,v 1.18 1998/01/09 14:05:09 cmm Exp $";
USE_RCSID(rcsid);

// Construct a list of Smalltalk variable names from a "geniecall" parameters OOP
GenieCalloutList::GenieCalloutList(OOP parameters) : m_names(0), m_count(0), m_paramValues(nilOOP)
{
    int len;
    if (!validParameters(parameters))
    {
	gerr << "GenieCalloutList(invalid parameters)" << endl;
    }
    else
    {
        OOP paramNames = arrayAt(parameters, 1);
        m_count = numIndexableFields(paramNames);
        m_names = new char*[m_count+1];
        m_names[0] = m_names[1] = 0;			// just for safety!
        char* item_name;
	OOP symbol;
// Parse parameter names into an array
        for (int i=2; i <= m_count; i++) 
	{
            symbol = arrayAt(paramNames,i);             /*** symbol containing field name ***/
	    item_name = OOPToString(symbol);
	    len = strlen(item_name);
	    m_names[i] = new char[len + 1];
	    strcpy(m_names[i], item_name);
            UnGarble(m_names[i]);
#ifdef DISPLAY_PROCESSING
	    gerr << "GenieCalloutList(GCL parameter " << i << " = \"" << m_names[i] << "\")" << endl;
#endif	/* DISPLAY_PROCESSING */
	    free(item_name);
        }    
	m_paramValues = arrayAt(parameters, 2);
    }
}

// Look up variable "name" is the list, and then initialise the
// GenieVariable with this OOP
// Return: 0 if variable is found, 1 otherwise
int GenieCalloutList::associate(const char* name, GenieVariableBase* gvp) const
{
    OOP value;
    int i, location;
#ifdef DISPLAY_PROCESSING
    gerr << "GenieCalloutList(looking up \"" << name << "\")" << endl;
#endif
    for(i=2; i<=m_count; i++)
    { 
    	if (GStrcasecmp(m_names[i], name) == 0)
	{
	    break;
	}
    }
    if (i <= m_count)
    {
        value = instVarAt(m_paramValues, i);           /*** field value being passed ***/
	location = i;
#ifdef DISPLAY_PROCESSING
        gerr << "GenieCalloutList(name \"" << name << "\" found at position " << location << ")" << endl;
#endif
    }
    else
    {
	gerr << "GenieCalloutList(name \"" << name << "\" not found)" << endl;
	value = nilOOP;
	location = 0;		// This variable will not be written back by the GenieVariable class destructor
    }
    gvp->init(value, m_paramValues, location);
    return (location != 0 ? 0 : 1);
}

GenieCalloutList::~GenieCalloutList()
{
    if (m_names != 0)
    {
        for(int i=2; i<=m_count; i++)
        {
	    delete[] m_names[i];
        }
        delete[] m_names;
    }
}

/*
 * Simple heuristic check to ensure parameters are being passed from GENIE correctly
 */
int GenieCalloutList::validParameters(OOP parameters) 
{
    int is_oop, is_array, field_count_ok;

    is_oop = isOOP(parameters);                                 /*** First check proper OOP, ***/
    is_array = isAKindOf(oopClass(parameters),arrayClass);      /*** then for an array, ***/
    field_count_ok = numIndexableFields(parameters) == 2;       /*** and finally for correct number of elements ***/

    return is_oop && is_array && field_count_ok;
}
