/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "geniedefs.h"
#include "Pginclude.h"
#include "g_types.h"
#include "coord_trans_func.h"

// This is done assuming a C style array i.e. first element is 0 for matrix
static const char* rcsid = "@(#)$Id: coord_trans_func.cxx,v 1.10 1997/10/21 14:36:29 cmm Exp $";
USE_RCSID(rcsid);

CoordinateTransformFunction::CoordinateTransformFunction(int dimension, g_real* forward_matrix) :
	m_dimension(0), m_forward_matrix(0), m_reverse_matrix(0), m_forward_function(0), m_reverse_function(0)
{
    setMatrix(dimension, forward_matrix);
}

CoordinateTransformFunction::CoordinateTransformFunction(int dimension, coord_tran_func_t forward_func, 
		coord_tran_func_t reverse_func) : m_dimension(0), m_forward_matrix(0), m_reverse_matrix(0)
{
    setFunction(dimension, forward_func, reverse_func);
}

void CoordinateTransformFunction::setMatrix(int dimension, g_real* forward_matrix)
{
    freeUp();
    m_dimension = dimension;
    int size = dimension * (dimension + 1);
    m_forward_matrix = new pg_float[size];
    if (m_forward_matrix != 0)
    {
        for(int i=0; i<size; i++)
        {
	    m_forward_matrix[i] = F(forward_matrix[i]);
        }
    }
// compute inverse matrix transform
    m_reverse_matrix = new pg_float[size];
    if (m_reverse_matrix != 0 && dimension == 2) // only works for 2d at moment!
    {
	double det = m_forward_matrix[1] * m_forward_matrix[5] - m_forward_matrix[2] * m_forward_matrix[4];
	m_reverse_matrix[0] = (m_forward_matrix[2] * m_forward_matrix[3] - m_forward_matrix[5] * m_forward_matrix[0]) / F(det);
	m_reverse_matrix[1] = + m_forward_matrix[5] / F(det);
	m_reverse_matrix[2] = - m_forward_matrix[2] / F(det);
	m_reverse_matrix[3] = (m_forward_matrix[4] * m_forward_matrix[0] - m_forward_matrix[1] * m_forward_matrix[3]) / F(det);
	m_reverse_matrix[4] = - m_forward_matrix[4] / F(det);
	m_reverse_matrix[5] = + m_forward_matrix[1] / F(det);
    }
    else
    {
	gerr << "CoordinateTransformFunction::setMatrix - only handle 2D at moment!" << endl;
    }
#if 0
    gerr << "Forward and reverse matrices" << endl;
    for(int i=0; i<6; i++)
    {
	gerr << m_forward_matrix[i] << " " << m_reverse_matrix[i] << endl;
    }
#endif
}

void CoordinateTransformFunction::freeUp()
{
    delete []m_forward_matrix; 
    m_forward_matrix = 0; 
    delete []m_reverse_matrix; 
    m_reverse_matrix = 0; 
    m_forward_function = 0; 
    m_reverse_function = 0; 
    m_dimension = 0; 
}

// prototype should agree with "coord_tran_func_t" typedef 
int CoordinateTransformFunction::forwardFunction(int* flags, const g_real* val_in, g_real* val_out)
{
    if (m_forward_matrix != 0)
    {
	return matrixFunction(m_forward_matrix, val_in, val_out);
    }
    else
    {
        return (*m_forward_function)(flags, val_in, val_out);
    }
}
    
// prototype should agree with "coord_tran_func_t" typedef 
int CoordinateTransformFunction::reverseFunction(int* flags, const g_real* val_in, g_real* val_out)
{
    if (m_reverse_matrix != 0)
    {
	return matrixFunction(m_reverse_matrix, val_in, val_out);
    }
    else
    {
        return (*m_reverse_function)(flags, val_in, val_out);
    }
}

int CoordinateTransformFunction::matrixFunction(const pg_float* matrix, const g_real* val_in, g_real* val_out)
{
    int start(0); // first element of row of matrix
    for(int i=0; i<m_dimension; i++)
    {
	val_out[i] = matrix[start];
   	for(int j=1; j<=m_dimension; j++)
	{
	    val_out[i] += (matrix[start+j] * val_in[j-1]);
	}
	start += (1 + m_dimension);
    }
    return 0;
}

void CoordinateTransformFunction::setFunction(int dimension, coord_tran_func_t forward_function, coord_tran_func_t reverse_function)
{
    m_dimension = dimension;
    m_forward_function = forward_function;
    m_reverse_function = reverse_function;
}
