/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

/*
 *      Module:         derived_unit
 *      Author:         Freddie Akeroyd, ISIS, 1/3/96
 *      Purpose:        Allow definition of a new unit, which is composed of basic SI units
 */

#include "geniedefs.h"
#include <stdio.h>
#include <math.h>
#include <gnustring.h>
#include "derived_unit.h"
#include "genievars.h"

static const char* rcsid = "@(#)$Id: derived_unit.cxx,v 1.10 1997/10/28 17:56:48 cmm Exp $";
USE_RCSID(rcsid);

/*
 * name: 	The name of the unit e.g. "electron volt"
 * label: 	The graphical character representing the unit e.g. "eV"
 * components:	A comma separated list of SI unit building blocks, made up of basic unit names 
 *		(see SIUnit::CodeName) and a power e.g. "coulomb,1" or "ampere,1,second,1"
 *		unit names can be prefixed (see SIUnit::PrefixName) e.g. tera-watt
 * scale:	A additional multiplier e.g. "1.602-19"
 */
DerivedUnit::DerivedUnit(const char* name, const char* label, const char* components, double scale)
	: m_name(name), m_label(label), m_components(components), m_scale(scale)
{
    setup();
}

DerivedUnit::DerivedUnit(const GenieWorkspace& gw) : 
	m_name((GenieString)gw("name")), m_label((GenieString)gw("label")), m_components((GenieString)gw("components"))//, 
//	m_scale(gw("scale"))
{
	m_scale = (GenieReal)gw("scale");
	setup();
}

DerivedUnit::operator GenieWorkspace() const
{
    GenieWorkspace gw("GXUnit");
//    GenieWorkspace gw(5);
    gw("name", m_name);
    gw("label", m_label);
    gw("components", m_components);
    gw("scale", m_scale);
    gw("offset", 0.0);
    return gw;
}
    
void DerivedUnit::setup()
{
    m_unit_list.reset();		// Clear out any old items
#if defined(__DECCXX) || defined(_MSC_VER)
    const int MAXCOMPONENTS = 20;
#else
    const int MAXCOMPONENTS = 1 + m_components.freq(',');
#endif
    String items[MAXCOMPONENTS];
    const int nitems = split(m_components, items, MAXCOMPONENTS, ',');
    SIUnit::Prefix prefix;
    SIUnit::Code code;
    double exponent;
    for(int i=0; i<nitems; i += 2)
    {
	if (items[i].contains('-'))		// Does the unit contain a prefix?
	{
	    prefix = SIUnit::lookupPrefix(String(items[i].before('-')));
	    code = SIUnit::lookupCode(String(items[i].after('-')));
	}
	else
	{
	    prefix = SIUnit::NoPrefix;
	    code = SIUnit::lookupCode(items[i]);
	}
	exponent = atof(items[i+1]);
	m_unit_list.push(DerivedUnitElement(code, prefix, exponent));
    }
    return;
}


double DerivedUnit::convertTo(const DerivedUnit& d) const
{
    if (dimensions() != d.dimensions())
    {
	return 0.0;	// Dimensions do not match, so no conversion possible
    }
    int i, j;
    double conv(1.0), conv_temp, conv_try;
// This will look at all cross products of the SI unit componments. Invalid combinations will return 0.0, but
// we can ignore these as we have already checked that the resulting combination is OK
    for(i=0; i<count(); i++)
    {
	conv_temp = 1.0;
	for(j=0; j<d.count(); j++)
	{
	    conv_try = list(i).unit.convertTo(d.list(j).unit);
	    if (conv_try != 0.0)
	    {
	        conv_temp *= pow(conv_try, list(j).exponent);
	    }
	}
	conv *= pow(conv_temp, list(i).exponent);
    }
    return conv * (d.m_scale / m_scale);
}

DerivedUnit::DerivedUnit(const char* name, const SIUnitElement* element_list, int nlist, double scale) :
	m_name(name), m_scale(scale)
{
    for(int i=0; i<nlist; i++)
    {
	DerivedUnitElement element(element_list[i].code, element_list[i].prefix, element_list[i].exponent);
	m_unit_list.push(element);
//	SIUnit::prefixName(element_list[i].prefix) + '-' + etc
//	m_components += 
    }
}
    
ostream& DerivedUnit::print(ostream& os) const
{
    os << "Dimensions" << endl;
    os << dimensions() << endl;
    return os;
}
    
SIDimensionalExponents DerivedUnit::dimensions() const
{
    SIDimensionalExponents exponents;
    for(int i=0; i<m_unit_list.count(); i++)
    {
	exponents += m_unit_list[i].unit.dimensions() * m_unit_list[i].exponent;
    }
    return exponents;
}

// Needed to satisfy "sort" method of SPList template - never called
int operator<(const DerivedUnitElement& e1, const DerivedUnitElement& e2)
{
    return e1.exponent < e2.exponent;
}

// Return a string containing enough information to recreate the unit
String DerivedUnit::encodeUnit() const
{
    static const String mark(';');
    static char buffer[32];
    sprintf(buffer, "%g", m_scale);
    return m_name + mark + m_label + mark + m_components + mark + String(buffer);
}

DerivedUnit operator*(const DerivedUnit& unit1, const DerivedUnit& unit2)
{
    DerivedUnit res;
    res.m_scale = unit1.m_scale * unit2.m_scale;
    res.m_name = unit1.m_name + String(" * ") + unit2.m_name;
    res.m_label = unit1.m_label + String(" ") + unit2.m_label;
    res.m_components = unit1.m_components + String(',') + unit2.m_components;
    res.setup();
    return res;
}
    
#if defined(__GNUG__)
template class SPListBase<DerivedUnitElement>;
template class SPList<DerivedUnitElement>;
#endif /* __GNUG__ */

#ifdef __DECCXX
#pragma define_template SPListBase<DerivedUnitElement>
#pragma define_template SPList<DerivedUnitElement>
#endif /* __DECCXX */
