/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "geniedefs.h"
// -----------------------------------------------------------------------------
// DEVICE_FUNCTIONS.C
//
// Description: C++ code to allow creation/deletion & clearing of a 
//		Genie object-graphics device window.
//
//
// *** includes for the picture list
#include "Pictlist.h"

// *** local includes
#include "Genie_functions.h"
#include "Pginclude.h"
#include "colourmap.h"
#include "Window.h"
#include "splash.h"
#include "assoc.h"

static const char* rcsid = "@(#)$Id: device_functions.cxx,v 1.20 1997/10/28 17:56:49 cmm Exp $";
USE_RCSID(rcsid);

static SPList<Window*> PTRlist;
static Assoc<ColourMap*> colour_maps;

// *** pointer must be static
static Window *PTR = 0;				// Current active device
static Window *prev_PTR = 0;			// Previous active device (used as quick hack for HARDCOPY command)

// Returns -1 on error, else device ID ()
    long open_device(const char *device_type,
		    const char *filename,
		    float width,
		    float height,
		    float ra,
		    const char *title, 
		    const char *icont,
		    funcPtrI exf,
		    funcPtr redrf,
		    const char *menutitle,
		    int n,
		    const char *menuitems,
		    funcPtrI menuf) {

	ColourMap* colour_map = 0;		// Current colour map
	ColourMap* prev_colour_map = 0;		// previous colour map

	// *** NEW a window object
	prev_PTR = PTR;
	if (prev_PTR != 0)
	    prev_colour_map = prev_PTR->m_map;

        PTR = new Window(device_type,filename,width,height,ra,
			title,icont,exf,
			redrf,menutitle,n,menuitems,menuf);
	PTRlist.push(PTR);

	// *** display the window
	if (PTR->create() != 0)
	    return -1;			// error creating device

	// *** get a colourmap
	SPString map_key(device_type);
	if (!colour_maps.isin(map_key))
	{
//	    gdbg << "Creating colourmap for device \"" << device_type << "\"" << endl;
	    colour_maps(map_key) = new ColourMap(prev_colour_map);
	}
	colour_map = colour_maps(map_key);

	// *** install colourmap
	PTR->installColourmap(colour_map);

	// *** start graphics buffering
	start_buffer();

//	PTR->m_map->print(gout);

	return PTRlist.count() - 1;		// position in list for [] call

}

void select_device(long device_id)
{
    prev_PTR = PTR;
    PTR = PTRlist[device_id];
    if (PTR != 0)
    {
        PTR->select();
    }
}

String current_device_type()
{
    if (PTR)
    {
	return PTR->deviceType();
    }
    else
    {
	return String("");
    }
}

    void close_device() {

	// *** stop graphics buffering
	close_buffer();

	// *** remove the window
	PTR->remove();

	// *** delete the object

        for(int i=0; i<PTRlist.count(); i++)
	{
	    if (PTRlist[i] == PTR)
		PTRlist[i] = 0;
  	}
	delete PTR;
	if (PTR != prev_PTR)
	    PTR = prev_PTR;
	else
	    PTR = 0;
	prev_PTR = 0;
}

    void clear_device() {

	// *** clear the screen or advance page
	if (PTR)
	{
	    PTR->clean();
	}
	else
	{
	    gerr << "No valid device open" << endl;
	}
}

// Return -1 on error
int define_named_colour(const String& name)
{
    String temp_name(name);
    temp_name.downcase();
    if (PTR != 0)
    {
        return PTR->m_map->newNamedColour(name);
    }
    else
    {
	*g_err << "COLOUR: No device currently open or selected" << endl;
	return -1;
    }
}

// Return -1 on error
int define_hls_colour(float hue, float lightness, float saturation)
{
    if (PTR != 0)
    {
        return PTR->m_map->newHLSColour(hue, lightness, saturation);
    }
    else
    {
	*g_err << "COLOUR: No device currently open or selected" << endl;
	return -1;
    }
}

// Return -1 on error
int query_rgb_colour(int index, g_real& r, g_real& g, g_real& b)
{
    if (PTR != 0)
    {
	return PTR->m_map->queryColour(index, r, g, b);
    }
    else
    {
	*g_err << "COLOUR: No device currently open or selected" << endl;
	return -1;
    }
}

// Return -1 on error
int define_rgb_colour(float r, float g, float b)
{
    if (PTR != 0)
    {
        return PTR->m_map->newRGBColour(r, g, b);
    }
    else
    {
	*g_err << "COLOUR: No device currently open or selected" << endl;
	return -1;
    }
}

g_integer* allocate_colourtable(g_real* ra, g_real* ga, g_real* ba, int len_ra,
	 int& ncol_wanted, g_real contrast, int nbands, g_real diffuse, 
	g_real shine, g_real polish)
{
    if (PTR != 0)
    	return PTR->m_map->newRGBColourTable(ra, ga, ba, len_ra, ncol_wanted, 
		contrast, nbands, diffuse, shine, polish);
    *g_err << "COLOUR: No device currently open or selected" << endl;
    return 0;
}

int deallocate_colourtable(g_integer* table, int ncol)
{
    if (PTR != 0)
        return PTR->m_map->freeColourTable(table, ncol);    
    *g_err << "COLOUR: No device currently open or selected" << endl;
    return 0;
}

// change colours for demo
// float border_r=255.0/255.0, border_g=228.0/255.0, border_b=196.0/255.0, fac=0.9/1.1;
// pgscr(12, border_r, border_g, border_b);
// pgscr(13, 1.0, 1.0, 240.0/255.0);
// pgscr(14, border_r/1.1, border_g/1.1, border_b/1.1);
// pgscr(15, border_r*fac, border_g*fac, border_b*fac);

// Instantiate templates
#if defined(__GNUG__)
template class SPList<Window*>;
template class SPListBase<Window*>;
template class Assoc<ColourMap*>;
template class Binar<ColourMap*>;
template class SPList<ColourMap*>;
template class SPListBase<ColourMap*>;
template class SPList< Binar<ColourMap*> >;
template class SPListBase< Binar<ColourMap*> >;
#endif /* __GNUG__ */

#ifdef __DECCXX
#pragma define_template SPList<Window*>
#pragma define_template SPListBase<Window*>
#pragma define_template Assoc<ColourMap*>
#pragma define_template Binar<ColourMap*>
#pragma define_template SPList<ColourMap*>
#pragma define_template SPListBase<ColourMap*>
#pragma define_template SPList< Binar<ColourMap*> >
#pragma define_template SPListBase< Binar<ColourMap*> >
#endif /* __DECCXX */
