/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

//
// $Id: dlfcn_vms.cxx,v 1.12 1997/10/28 17:03:57 cmm Exp $
//
// Module: 	dlfcn_vms
// Author:	Freddie Akeroyd, ISIS Facility, Rutherford Appleton Laboratory
// Purpose:	Emulate the dlopen() set of routines on VMS
//		using the native lib$find_image_symbol() function, including
//		searching of a comma separated list of directories in
//		logical name LD_LIBRARY_PATH. Pathnames to dlopen()
//		may be specified in either UNIX or VMS style
//
#include "geniedefs.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <gnustring.h>
#include <unixlib.h>
#include <lib$routines.h>
#include <ssdef>
#include <libdef>
#include <descrip.h>
#include <rmsdef>
#include "splash.h"

#include "dlfcn_vms.h"

static String error_message;

#define make_descriptor(name,string)    \
    struct dsc$descriptor_s name = { strlen(string), DSC$K_DTYPE_T, DSC$K_CLASS_S, \
					strcpy((char*)malloc(strlen(string)+1), string) }

#define make_descriptor_size(name,size)    \
    struct dsc$descriptor_s name = { size, DSC$K_DTYPE_T, DSC$K_CLASS_S, (char*)memset(malloc(size+1), 0, size+1) }

// Helper routine for decc$to_vms()
static const char* action_result = 0;
static int action_routine(char* unix_file, int type)
{
    action_result = unix_file;
    return 1;
}

static char* translatePath(char* path)
{
    int status;
    status = decc$to_vms(path, action_routine, 0, 0);
    if (status > 0 && action_result != 0)
    {
//	return strcpy((char*)malloc(1+strlen(action_result)), action_result);
	return action_result;
    }
    else
    {
	return path;
    }
}

// Locate file, return NULL on error
static const char* vms_find_file(const char* path, const char* logical_name)
{
    char* result;
    int i;
    make_descriptor(filespec, path);
    make_descriptor(default_filespec, logical_name);
    make_descriptor_size(resultant_filespec, 256);
    unsigned long status, context(0);
// switch order as only first argument can be a search list
    status = lib$find_file(&filespec, &resultant_filespec, &context, &default_filespec, 0, 0, 0); 
    lib$find_file_end(&context);
    if (status & 1)
    {
	result = (char*)malloc(1+resultant_filespec.dsc$w_length);
	strncpy(result, resultant_filespec.dsc$a_pointer, resultant_filespec.dsc$w_length);
        result[resultant_filespec.dsc$w_length] = '\0';
	i = resultant_filespec.dsc$w_length;
	while(i> 0 && result[--i] == ' ')
	    result[i] = '\0';		// remove training spaces (if any)
        return result;
    } 
    else
    {
	return 0;
    }
}

class SharedLibraryItem
{
    private:
    	int m_mode;
	String m_path;

    public:
    	SharedLibraryItem() : m_mode(0), m_path() {}
    	SharedLibraryItem(const char* path, int mode) : m_mode(mode), m_path(path) {}
	const char* getPath() const { return m_path; }
	int getMode() const { return m_mode; }
};

static SPList<SharedLibraryItem> library_list;

void* dlopen(const char* path, int mode)
{
    static int ld_library_path = (getenv("LD_LIBRARY_PATH") != 0 ? 1 : 0);
    int ret;
    if (path == 0)
    {
	error_message = "dlopen: NULL handle not implemented\n";
	return 0;
    }
    struct stat stat_buffer;
    ginf << "MODULE: trying \"" << path << "\" ..." << endl;
    ret = stat(path, &stat_buffer);
    if (ret != 0 && ld_library_path)
    {
	path = vms_find_file(path, "LD_LIBRARY_PATH");
	if (path != 0)
	{
	    ginf << "MODULE: trying \"" << path << "\" ..." << endl;
	    ret = stat(path, &stat_buffer);
	}
    }
    if (ret != 0)
    {
	error_message = "dlopen: unable to stat() - check file exists and is readable\n";
	return 0;
    }
    ginf << "MODULE: loaded library \"" << path << "\"" << endl;
    library_list.push(SharedLibraryItem(path, mode));
    return (void*)library_list.count();	// return 1 + position in list (can't return position as that could be "0")
}

// A version of lib$find_image_symbol with a condition handler - we
// need this to avoid getting a large traceback printed to the users' screen

static unsigned long find_image_symbol(void* filename, void* symbol, void* symbol_value, void* image_name)
{
    lib$establish(lib$sig_to_ret);
    return lib$find_image_symbol(filename, symbol, symbol_value, image_name);
}

void* dlsym(void* handle, const char* name)
{
    static char buffer[64], module_logical[64];
    unsigned long symbol_value(0), status;
    unsigned long pos = (unsigned long)handle;
    if ((pos == 0) || (pos > library_list.count()))
    {
	error_message = "dlsym: invalid handle\n";
	return 0;
    }
    if (name == 0)
    {
	error_message = "dlsym: invalid symbol\n";
	return 0;
    }
    const char* path = library_list[pos-1].getPath();
// For moment, path must contain no directory qualifier etc.
// filename must contain no directory or extension information
// image_name contains the other information for a PARSE
    sprintf(module_logical, "%s%ld", "GENIE_MODULE_IMAGE_", pos);
    make_descriptor(logical_name_d, module_logical);
    make_descriptor(logical_value_d, translatePath(path));
    status = lib$set_logical(&logical_name_d, &logical_value_d);
    make_descriptor(symbol_d, name);
    make_descriptor(image_name_d, "SYS$DISK:[].EXE");
// Catch the message about "compiler warnings" in the library and retry
    unsigned long eomwarn = LIB$_EOMWARN, keynotfou = LIB$_KEYNOTFOU;
    status = find_image_symbol(&logical_name_d, &symbol_d, &symbol_value, &image_name_d);
    if (lib$match_cond(&status, &eomwarn) > 0)
    {
	status = find_image_symbol(&logical_name_d, &symbol_d, &symbol_value, &image_name_d);
    }
    if (status & 1)
	return (void*)symbol_value;	// All OK
// Some other error
    sprintf(buffer, "%ld", status);
    error_message = String("dlsym: error reading symbol \"") + String(name) + 
		    String("\" from file \"") + String(path) + String("\" - VMS status code = ") + String(buffer);
    return 0;
}

// return 0 on success, 1 on error
int dlclose(void* handle)
{
    unsigned long pos = (unsigned long)handle;
    if ((pos == 0) || (pos > library_list.count()))
    {
	error_message = "dlclose: invalid handle\n";
	return 1;
    }
    return 0;		// Nothing else possible on VMS!
}

const char* dlerror(void)
{
    return (const char*)error_message;
}

int operator<(const SharedLibraryItem& a, const SharedLibraryItem& b)
{
    return 0;
}

#ifdef __DECCXX
#pragma define_template SPList<SharedLibraryItem>
#pragma define_template SPListBase<SharedLibraryItem>
#endif /* __DECCXX */

#if defined(__GNUG__) || defined(_MSC_VER)
template class SPList<SharedLibraryItem>;
template class SPListBase<SharedLibraryItem>;
#endif /* __GNUG__ */
