/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

// fileset.cxx: implementation of the CFileSet class.
//
//////////////////////////////////////////////////////////////////////

#include "geniedefs.h"

#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif /* HAVE_UNISTD_H */
#include <stdlib.h>
#include <fcntl.h>
#ifdef _WIN32
#	define STRICT 1
#   include <io.h>
#   include <winreg.h>
#   include <tchar.h>
#   include <BitsNBobs.h>
#endif /* _WIN32 */


#include "gnustring.h"
#include "fileset.h"

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

// Find any and all information about the file paths for this invocation
// of the application (GENIE).	Once this has constructed itself in the running
// environment, it does little else than provide a central point to get file
// information from.

// If file names are given explicitly in the constructor, they will be
// absolute and override all else.  If they are not given, symbols and
// default names will be used according to the vagaries of the code below.

// NT access to the registry for similar path names etc. should also be
// documented here.

// Create a single private instance of a CFileSet
CFileSet JustOneOfMe;					// Could add initial defaults here for UNIX/VMS (not PC)

GENIEEXPORT CFileSet* CFileSet::s_this = NULL;		// Set sentinel value for "this" class variable

CFileSet::CFileSet(
    const char* szGCLDir,
    const char* szImagePath,
    const char* szHistoryFile,
    const char* szInstrumentationFile,
    const char* szDebuggingOutput,
    const char* szGCLInitialisation,
    const char* szTCLInitialisation
    ) :
    m_gstrGCLDir(szGCLDir),
    m_gstrSmalltalkImage(szImagePath),
    m_gstrHistoryFile(szHistoryFile),
    m_gstrParserInstrumentation(szInstrumentationFile),
    m_gstrDebuggingOutput(szDebuggingOutput),
    m_gstrGCLInitialisation(szGCLInitialisation),
    m_gstrTCLInitialisation(szTCLInitialisation)
{
    char* cp;

    // Only expect one instance to be created so record the this member as a way of access
    // for access to functions from other places where "fileset.h" is included.
    if (s_this == NULL)
    {
	s_this = this;		// Make a reference available from the class declaration
    }
    else
    {
	m_gstrFileSetError = "Trying to create more than one instance of CFileSet is an error\n";
    }
    
    // The way these file names are found is slightly varied, mainly because of the
    // way they were originally done.  This code aims more to document and encapsulate
    // than make logical! (DIRECTORY separators are system dependent macros set in geniedefs.h)

    // (1) Source path will either be a translation of GENIE_DIR or will be taken to
    // be the current directory
    if ( m_gstrGCLDir == "" )
    {
	if ( (cp = getenv("GENIE_DIR")) != NULL )
	{
	    m_gstrGCLDir = String(cp) + DIRECTORY_SEPARATOR;
	}
	else
	{
	    m_gstrGCLDir = String(CURRENT_DIRECTORY_PATH);
	}
    }

    // (2) Image file will be located by environment variable or stored in current directory
    // as "genie.im"
    if ( m_gstrSmalltalkImage == "" )
    {
//DIAG	system("set > c:\\tmp.out");
	if ( (cp = getenv("GENIE_SMALLTALK_IMAGE")) != NULL )
	{
	    m_gstrSmalltalkImage = String(cp);
	}
	else
	{
	    m_gstrSmalltalkImage = String(CURRENT_DIRECTORY_PATH) + "genie.im";
	}
    }

    // (3) Start time GCL initialization file
    if ( m_gstrGCLInitialisation == "" )
    {
	if ( (cp = getenv("GENIE_GCL_INIT")) != NULL )
	{
	    m_gstrGCLInitialisation = String(cp);
	}
	else
	{
	    m_gstrGCLInitialisation = String("genieinit.gcl");
	}
    }

    // (4) Start time TCL initialization file
    if ( m_gstrTCLInitialisation == "" )
    {
	if ( (cp = getenv("GENIE_TCL_INIT")) != NULL )
	{
	    m_gstrTCLInitialisation = String(cp);
	}
	else
	{
	    m_gstrTCLInitialisation = String("genieinit.tcl");
	}
        
	// Check the file exists - we need to say if it's missing
	int fd;
	if ( (fd = open(m_gstrTCLInitialisation, O_RDONLY, 0)) != -1 )
	{
	    m_gstrFileSetError  = m_gstrFileSetError + "Loading TCL commands from \"" + m_gstrTCLInitialisation + "\"";
	    close(fd);
	}
	else
	{
	    m_gstrFileSetError  = m_gstrTCLInitialisation + "\" not present: was GENIE_TCL_INIT defined correctly?";
	    m_gstrTCLInitialisation = "";
	}

    }

    // (5) Put history file at user root on whatever system we are on
    if ( m_gstrHistoryFile == "" )
    {
#if defined(__VMS) && !defined(__VMS_POSIX)
	m_gstrHistoryFile = String(getenv("SYS$LOGIN")) + ".genie_history";
#elif defined(_WIN32)
	m_gstrHistoryFile = String(getenv("HOMEDRIVE")) + getenv("HOMEPATH");
	if ( m_gstrHistoryFile.lastchar() != '\\' )
	    m_gstrHistoryFile += "\\";			// Add the \ if necessary (HOMEPATH can be either!)
	m_gstrHistoryFile += ".genie_history";
#else
	m_gstrHistoryFile = String(getenv("HOME")) + "/" + ".genie_history";
#endif
    }

    // (6) Intermediate Smalltalk preload/debugging output file
    if ( m_gstrDebuggingOutput == "" )
    {
	m_gstrDebuggingOutput = String("boot.st");
    }

    // (7) File to dump parser instrumentation into.
    if ( m_gstrParserInstrumentation == "" )
    {
	m_gstrParserInstrumentation = String("procdefs.out");
    }
}


CFileSet::~CFileSet()
{
    s_this = NULL;
}


// Static definitions of parser boot sequences, all GCL files, the lists
// terminate with a null string.
const char *CFileSet::s_basefiles[] =
{
    // Basic GENIE system for running procedures etc.
    "stfixes.gcl",
    "stextensions.gcl",
    "external.gcl",

    // Predefined variables, constants
    "setup.gcl",

    // Load the PROCEDURE/FUNCTION running subsystem
    "function.gcl",

    // Genie & smalltalk types/extensions
    "maths.gcl",				// Basic mathematical functions (generic)
    "workspace.gcl",				// Workspace class
    "multidimensionalarray.gcl",		// needed by array.gcl
    "array.gcl",
    "integerarray.gcl",
    "realarray.gcl",
    "genietypes.gcl",				// GXreal, GXstring & GXinteger etc.

    // System primitive procedures
    "builtins.gcl",
    "objects.gcl",				// Define mechanisms for creating (workspace) classes
    "baseprimitives.gcl",
    "directory.gcl",

    "errors.gcl",				// GCL <-> C++ error handling functions

    // Neutron specific GENIE from now on
    "nexus.gcl",				// NeXus Definitions
    "workspace_user.gcl",			// Workspace add ons
    "asciifile.gcl",

    // Graphics
    "drivers.gcl",				// Auto generated from ../pgplot/makefile
    "io.gcl",
    "operations.gcl",
    "objectgraphics.gcl",
    "GraphicsGlobals.gcl",
    "GraphicalObjects.gcl",
    "colour.gcl",
    "Device.gcl",
    "DrawableObjects.gcl",
    "DrawTwoD.gcl",
    "TwoDContour.gcl",
    "TwoDCell.gcl",
    "DrawTwoDContour.gcl",
    "DrawTwoDContourArray.gcl",
    "DrawTwoDContourFunction.gcl",
    "DrawTwoDContourLabel.gcl",
    "DrawTwoDCell.gcl",
    "DrawTwoDCellArray.gcl",
    "DrawTwoDCellFunction.gcl",
    "DrawTwoDCellWedge.gcl",
    "Window.gcl",
    "ScaledWindow.gcl",
    "TwoDWindow.gcl",
    "Multiplot.gcl",
    "Picture.gcl",
    "GraphicalObjectsMisc.gcl",
    "DeviceMisc.gcl",
    "AddToList.gcl",
    "DrawOneD.gcl",
    "DrawOneDGraph.gcl",
    "DrawOneDLineGraph.gcl",
    "DrawMarkers.gcl",
    "isInAWindowII.gcl",
    "isOverlapped.gcl",
    "isOverlappedII.gcl",
    "isInAWindow.gcl",
    "FindOverLappingPoint.gcl",
    "makeNewWindow.gcl",
    "GetNewCoordinates.gcl",
    "GetCoordinates.gcl",
    "FindWindows.gcl",
    "zoom.gcl",
    "DrawPolyLine.gcl",
    "Drawtext.gcl",
    "Polygon.gcl",
    "DrawAnnotation.gcl",
    "DrawOneDAxes.gcl",
    "DrawOneDGraticule.gcl",
    "DrawOneDLabels.gcl",
    "DrawOneDTitle.gcl",
    "DrawVerrors.gcl",
    "DrawHErrors.gcl",
    "DrawOneDErrors.gcl",
    "DrawMultiplot.gcl",
    "Drawhistogram.gcl",
    "whichWindow.gcl",

    // 3d graphics routines
    "ThreeDObject.gcl",
    "ThreeDWindow.gcl",
    "Plane.gcl",
    "ThreeDText.gcl",
    "Ellipsoid.gcl",
    "ThreeDLine.gcl",
    "Rod.gcl",
    "Cone.gcl",
    "Ball.gcl",
    "ThreeDLattice.gcl",
    "IsoSurface.gcl",
    "LatticePlane.gcl",
    "TwoDLattice.gcl",
    "Surface.gcl",
    "LatticeSlice.gcl",
    "newGraphics.gcl",
    "MiscNewGraphics.gcl",
    "select.gcl",
    "old_graphics.gcl",
    "graphics_v2.gcl",
    "genie_v2_compat.gcl",
    "peak.gcl",
    "module.gcl",
    "tclMisc.gcl",			 
    "info.gcl",
    "help.gcl", 			    // procedures to implement /HELP qualifier on commands
    NULL				    // terminator
};


bool CFileSet::GetBaseFile(int& index, String& gstr)
{
    if (s_basefiles[index] != NULL)
    {
	gstr = m_gstrGCLDir + s_basefiles[index];
	return true;
    }
    else    
    {
	return false;
    }
}


#ifdef _WIN32

//
// Specify the Registry key for Open GENIE settings information
//
LPCTSTR CFileSet::s_szKey = _T("SOFTWARE\\CCLRC ISIS Facility\\Open GENIE\\tkgenie32\\Settings");

//
// Utility routines for getting/setting the default values
//
GENIEEXPORT BOOL CFileSet::GetRegistrySettings(HKEY hk)
{
    // Access the default settings key
    HKEY hkSettings;
    if ( ERROR_SUCCESS != RegOpenKeyEx(hk, s_szKey, 0, KEY_EXECUTE, &hkSettings) )
    {
	 return FALSE;	// Can't find key, give up now
    }

    // Directory paths
    m_gstrGCLDir =	GetRegString(hkSettings, _T("GCL Path"));
    m_gstrTCLDir =	GetRegString(hkSettings, _T("TCL Path"));
    m_gstrLibraryDir =  GetRegString(hkSettings, _T("Library Path"));
    m_gstrHelpDir =	GetRegString(hkSettings, _T("Help Path"));

    // Initialization files
    m_gstrTCLInitialisation =	GetRegString(hkSettings, _T("TCL Init"));
    m_gstrGCLInitialisation =	GetRegString(hkSettings, _T("GCL Init"));
    m_gstrSmalltalkImage =	GetRegString(hkSettings, _T("Smalltalk Image"));

    // Interface configuration settings
    m_bFileCheck = GetRegInt(hkSettings, _T("File Check"));
    m_bDontShow = GetRegInt(hkSettings, _T("Don't Show"));

    RegCloseKey(hkSettings);    // Finished registry access

    return TRUE;
}


// Reads Per user defaults from the registry.  Note that attempting to read them
GENIEEXPORT BOOL CFileSet::SetRegistrySettings(HKEY hk)
{
    // Access the default settings key
    HKEY hkSettings;
    if ( ERROR_SUCCESS != RegOpenKeyEx(hk, s_szKey, 0, KEY_SET_VALUE, &hkSettings) )
    {
	 return FALSE;	// Can't find key, give up now
    }

    BOOL bStatus = 0;

    // Directory paths
    bStatus |= SetRegString(hkSettings, _T("GCL Path"), m_gstrGCLDir);
    bStatus |= SetRegString(hkSettings, _T("TCL Path"), m_gstrTCLDir);
    bStatus |= SetRegString(hkSettings, _T("Library Path"), m_gstrLibraryDir);
    bStatus |= SetRegString(hkSettings, _T("Help Path"), m_gstrLibraryDir);

    // Initialization files
    bStatus |= SetRegString(hkSettings, _T("TCL Init"), m_gstrTCLInitialisation);
    bStatus |= SetRegString(hkSettings, _T("GCL Init"), m_gstrGCLInitialisation);
    bStatus |= SetRegString(hkSettings, _T("Smalltalk Image"), m_gstrSmalltalkImage);

    // Interface configuration settings
    bStatus |= SetRegInt(hkSettings, _T("File Check"), m_bFileCheck);
    bStatus |= SetRegInt(hkSettings, _T("Don't Show"), m_bDontShow);

    return bStatus;
}


// Reads a string out of the registry into a CString
String CFileSet::GetRegString(HKEY& rhk, LPTSTR pszValueName)
{
    TCHAR szRegString[_MAX_PATH];
    DWORD dwType, cbData = sizeof(szRegString);
    LONG status;

    status = RegQueryValueEx(rhk, pszValueName, 0, &dwType, (LPBYTE)szRegString, (LPDWORD)&cbData);
    if (status != ERROR_SUCCESS)
    {
	BitsNBobs::Message("GetRegString", status);
	return "";
    }

    szRegString[cbData] = 0;
    return szRegString;
}

// Writes a CString into the registry
int CFileSet::SetRegString(HKEY& rhk, LPTSTR pszValueName, LPCTSTR pszValue)
{
    LONG status;

    status = RegSetValueEx(rhk, pszValueName, 0, REG_SZ, (LPBYTE)pszValue, _tcslen(pszValue));
    if (status != ERROR_SUCCESS)
    {
	BitsNBobs::Message("SetRegString", status);
	return status;
    }

    return 0;
}

// Writes a CString into the registry
int CFileSet::SetRegInt(HKEY& rhk, LPTSTR pszValueName, DWORD dwValue)
{
    LONG status;

    status = RegSetValueEx(rhk, pszValueName, 0, REG_DWORD, (LPBYTE)&dwValue, sizeof(DWORD));
    if (status != ERROR_SUCCESS)
    {
	BitsNBobs::Message("SetRegInt", status);
	return status;
    }

    return 0;
}

// Simple read of registry integer value, (-1 on failure)
int CFileSet::GetRegInt(HKEY& rhk, LPTSTR pszValueName)
{
    DWORD dwRegWord, dwType, cbData = sizeof(DWORD);
    LONG status;

    status = RegQueryValueEx(rhk, pszValueName, 0, &dwType, (LPBYTE)&dwRegWord, (LPDWORD)&cbData);
    if (status != ERROR_SUCCESS)
    {
	BitsNBobs::Message("GetRegInt", status);
	return -1;
    }

    return dwRegWord;
}


GENIEEXPORT void CFileSet::SetVars()
{
    // Now we have the settings, pass on locations to the rest of GENIE via the
    // environment
    putenv("GENIE_DIR=" + m_gstrGCLDir);
    putenv("GENIE_TCL_DIR=" + m_gstrTCLDir);

    // On PC root all ancillary files in fixed directories below GENIE_LIBRARY_DIR
    putenv("GENIE_LIBRARY_DIR=" + m_gstrLibraryDir);
    putenv("GENIE_DOC_DIR=" + m_gstrHelpDir);
    putenv("GENIE_INCLUDE_DIR=" + m_gstrLibraryDir + "include");

    putenv("GENIE_TCL_INIT=" + m_gstrTCLInitialisation);
    putenv("GENIE_GCL_INIT=" + m_gstrGCLInitialisation);
    putenv("GENIE_SMALLTALK_IMAGE=" + m_gstrSmalltalkImage);
    putenv("PGPLOT_FONT=" + m_gstrGCLDir + "..\\..\\grfont.dat-nt-4.0");
    putenv("PGPLOT_DIR=" + m_gstrGCLDir + "..\\..");

    // Put variables required by the smalltalk interpreter
    putenv("SMALLTALK_KERNEL=" + m_gstrGCLDir + "..\\smalltalk");    // locate from GCL directory
    putenv("SMALLTALK_IMAGE=" + m_gstrSmalltalkImage);		    // (same as GENIE_SMALLTALK_IMAGE)

    // Set a base path for the TCL library
    putenv("TCL_LIBRARY=" + m_gstrLibraryDir + "tcl8.0");    // TCL Runtime (others locate from this)

    return;
}

#endif