/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "geniedefs.h"
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include "smalltalk.h"      // smalltalk
#include "parsedefs.h"      // parser
#include "parser.h"
#include "ghistory.h"       // genie
#include "fileset.h"
#include "genie.h"
#include "g_commandline.h"
#include "genie_commands.h"
// GNU Readline
extern "C"
{
#    include <readline/readline.h>
#    include <readline/history.h>
}

static void initialize_readline();
static void stripwhite(char* string);

int g_commandline(CGenie* pGenie)
{
    static char old_line[MAXSTRLEN] = "";
    static char *lp = old_line + 1;                     // Sets initial condition to be ready to read
    static bool end_of_line_to_send = FALSE;             // Keep a note of when an EOL is needed
    static char *lp_unexpanded = 0, *lp_expanded = 0;
    int expand;
    bool done;

    pGenie->PrintBanner(gout);

    initialize_readline();
    while(1)
    {
    done = FALSE;
    while(!done)
    {
         if (lp_expanded != 0)
         {
	     free(lp_expanded);
	     lp_expanded = 0;
	 }
         while( (lp_unexpanded = readline(">> ")) == 0 )
	 {
	    crlf();
	    rl_on_new_line();	// Trap NULL returns from e.g. ^D
         }
#if 0	/* I've turned off ! history subs as they cause problems */
         expand = history_expand(lp_unexpanded, &lp_expanded);
         free(lp_unexpanded);
#else
	 expand = 0;
	 lp_expanded = lp_unexpanded;
#endif
	 lp_unexpanded = 0;
         stripwhite(lp_expanded);                         
	 if (*lp_expanded)
	 {
	     switch(expand)
	     {
	         case 0:		        // no expansion too place, so no need to echo result
		    done = TRUE;
		    break;

	         case 1:		        // history expansion took place
	            done = TRUE;	// and now fall though to next statement ...
	         case 2:		        // line should be displayed, but not executed (e.g. it was a "!string:p")
	         case -1:		    // error in expanding, error message returned
	            fprintf(rl_outstream, "%s\n", lp_expanded);
	            rl_on_new_line();
		    break;

	         default:
	            fprintf(rl_outstream, "Unusual code returned from history_expand(): %i\n", expand);
	            rl_on_new_line();
		    break;
	     }
	}
    }
    pGenie->EvaluateGCLCommand(lp_expanded, 1);
    }
    return 0;
}

static void stripwhite(char* string)
{
    int i = 0;

    while(whitespace(string[i]))
    {
        i++;
    }

    if (i)
    {
      strcpy(string, string + i);
    }

    i = strlen(string) - 1;

    while(i > 0 && whitespace(string[i]))
    {
        i--;
    }

    string[++i] = '\0';
    return;
}

/* Generator function for command completion.  STATE lets us know whether
   to start from scratch; without any state (i.e. STATE == 0), then we
   start at the top of the list. */
static char* genie_command_generator(char* text, int state)
{
  static int list_index, len;
  const char *name;

  /* If this is a new word to complete, initialize now.  This includes
     saving the length of TEXT for efficiency, and initializing the index
     variable to 0. */
  if (!state)
    {
      list_index = 0;
      len = strlen (text);
    }

  /* Return the next name which partially matches from the command list. */
  while ((name = genie_commands[list_index][COMMAND]) != 0)
    {
      list_index++;

      if (strncasecmp (name, text, len) == 0)
        return strcpy((char*)malloc(1+strlen(name)), name);		/* the memory is free'd by readline */
    }

  /* If no names matched, then return NULL. */
  return ((char *)NULL);
}


// text should be in lower case
static char* genie_keyword_generator(char* text, int state)
{
  static int list_index, len;
  static String help_string;
  const char *help;
  const char *name;
  /* If this is a new word to complete, initialize now.  This includes
     saving the length of TEXT for efficiency, and initializing the index
     variable to 0. */
  if (!state)
    {
      list_index = 0;
      len = strlen (text);
    }
  /* Return the next name which partially matches from the command list. */
  while ((name = genie_commands[list_index][COMMAND]) != 0)
    {
      help = genie_commands[list_index][COMMENT];
      cat(name, " ", help_string);
      cat(help_string, help, help_string);
      help_string.downcase();
      list_index++;
      if (help_string.contains(text))
      {
//	cat(BLUE_ON, name, help_string);
//        cat(help_string, BLUE_OFF, help_string);
//        cat(help_string, " ", help_string);
	cat(name, " ", help_string);
        cat(help_string, help, help_string);
        return strcpy((char*)malloc(1+strlen(help_string)), help_string);	/* the memory is free'd by readline */
      }
    }

  /* If no names matched, then return NULL. */
  return ((char *)NULL);
}

static char* genie_help_generator(char* text, int state)
{
  static int list_index, len;
  const char *name;
  const char *help;

  /* If this is a new word to complete, initialize now.  This includes
     saving the length of TEXT for efficiency, and initializing the index
     variable to 0. */
  if (!state)
    {
      list_index = 0;
      len = strlen (text);
    }

  /* Return the next name which partially matches from the command list. */
  while ((name = genie_commands[list_index][COMMAND]) != 0)
    {
      help = genie_commands[list_index][COMMENT];
      list_index++;

      if (strncasecmp (name, text, len) == 0)
      {
        return strcpy((char*)malloc(1+strlen(help)), help);		/* the memory is free'd by readline */
      }
    }

  /* If no names matched, then return NULL. */
  return ((char *)NULL);
}

/* 
 * Attempt to complete on the contents of TEXT.  START and END show the
 * region of TEXT that contains the word to complete.  We can use the
 * entire line in case we want to do some simple parsing.  Return the
 * array of matches, or NULL if there aren't any. 
 */
static char** genie_command_completion(char* text, int start, int end)
{
    static char line[132];
    strncpy(line, text+start, end-start);
    line[end-start] = '\0';
    char** matches = 0;
    if (start == 0)  /* the word to complete is at the start of the line */
    {
	matches = completion_matches(line, genie_command_generator);
    }
    return matches;
}

static char** genie_help_completion(char* text, int start, int end)
{
    static char line[132];
    strncpy(line, text+start, end-start);
    line[end-start] = '\0';
    char** matches = 0;
    if (start == 0)  /* the word to complete is at the start of the line */
    {
	matches = completion_matches(line, genie_help_generator);
    }
    return matches;
}

static char** genie_keyword_completion(char* text, int start, int end)
{
    static char line[132];
    strncpy(line, text+start, end-start);
    line[end-start] = '\0';
    char** matches = 0;
    if (start == 0)  /* the word to complete is at the start of the line */
    {
        matches = completion_matches(line, genie_keyword_generator);
    }
    return matches;
}

static int genie_keyword_search(int dummy1, int dummy2)
{
    static int count;
    static char buffer[132];
    static char* prompt = "(command-keyword-search) ";
    int c, done=0;
    int rl_save_point = rl_point, rl_save_end = rl_end;
    rl_modifying(0, rl_end);
    rl_extend_line_buffer(132);
    strcpy(rl_line_buffer, prompt);
    rl_point = rl_end = strlen(prompt);
    rl_redisplay(); 
    count = 0;
    while(!done)
    {
//	c = rl_read_key();
	c = rl_getc(rl_instream);
	switch(c)
	{
	    case RETURN:
	    case NEWLINE:
		done=1;
		break;

	    default:
		buffer[count] = tolower(c);
		count++;
		break;
	}
	fputc(c, rl_outstream);
	fflush(rl_outstream);
    }
    buffer[count] = '\0';
    char** matches = genie_keyword_completion(buffer, 0, count);
    int i = 0;
    if (matches != 0)
    {
    	while(matches[i] != 0)
    	{
	    fprintf(rl_outstream, "%s", matches[i]);
	    free(matches[i]);
	    i++;
	    crlf();
    	}
	free(matches);
    }
    else
    {
	crlf();
    }
    fflush(rl_outstream);
// Remove junk left over from previous line
    for(i=rl_save_end; i<rl_end; i++)
    {
	rl_line_buffer[i] = ' ';
    }
    rl_do_undo();
    rl_on_new_line();
    rl_point = rl_save_point;
    rl_end = rl_save_end;
    rl_redisplay(); 
    return 0;
}

/*
 * Print a list of genie commands followd by the "comment" field
 * from Smalltalk onto standard output. This procedure is invoked by the
 * <Ctrl>-h key on the command line, which is bound in initialize_readline()
 */
static int genie_command_help(int dummy1, int dummy2)
{
    unsigned i, j, maxlen;
    int end = rl_point;
// Strip off any spaces at the end - we may have pressed ^H after completing the command
    while((end > 0) && whitespace(rl_line_buffer[end-1]))
	end--;
    
    char** matches_command = genie_command_completion(rl_line_buffer, 0, end);
    char** matches_help = genie_help_completion(rl_line_buffer, 0, end);
    if (matches_command != 0)
    {
	i = maxlen = 0;
// Calculate the max length of a completed command (used for "pretty printing")
	while(matches_command[i] != 0)
	{
	    if (strlen(matches_command[i]) > maxlen) 
	    { 
		maxlen = strlen(matches_command[i]); 
	    }
	    i++;
	}
// If we couldn't complete the command, first element of "matches" is the incomplete command - we don't want this printed
// However if we _did_ complete the command, the first element is what we want (and all we're given in fact!)
	if (i == 1)
	{
	    i = 0;
	}
	else
	{
	    i = 1;
	    free(matches_command[0]);		// The incomplete command we matched on
	    free(matches_help[0]);
	}
	maxlen += 4;		// add in number of blank spaces printed between command and help
	crlf();
	while(matches_command[i] != 0)
	{
	    fprintf(rl_outstream, "%s", matches_command[i]);
	    for(j=0; j<maxlen-strlen(matches_command[i]); j++)
	    {
		putc(' ', rl_outstream);
	    }
	    fprintf(rl_outstream, "%s", matches_help[i]);
	    free(matches_command[i]);
	    free(matches_help[i]);
	    crlf();
	    i++;
	}
	free(matches_command);
	free(matches_help);
	fflush(rl_outstream);
        rl_on_new_line();	// start a new prompt line
    }
    return 0;
}

// This is where we will set up options such as copmmand name completion
static void initialize_readline()
{
    /* Allow conditional parsing of the ~/.inputrc file. */
    rl_readline_name = (char*)"Genie";

    /* Tell the completer that we want a crack first. */
    rl_attempted_completion_function = (CPPFunction *)genie_command_completion;

    /* Register the <Ctrl>-h key as out help key */
    rl_add_defun("genie-command-help", (Function*)genie_command_help, CTRL('H'));

    /* Register the <Ctrl>-k key as out keyword search key */
    rl_add_defun("genie-keyword-search", (Function*)genie_keyword_search, CTRL('K'));

    return;
}
