/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "geniedefs.h"
#include <stdio.h>
#include <fcntl.h>
#include <sys/types.h>
#include <time.h>
#include "smalltalk.h"
#include "smalltalk_object.h"
#include "g_fileaccess.h"
#include "genievars.h"
#include "genie_tmpfile.h"
#include "iostream_list.h"
#include "read_ascii.h"
#include "genie_types.h"	/* this and g3io.h just for translatePath() */
#include "g3io.h"
#include "genie_tmpfile.h"
#ifdef HAVE_IO_H
#	include <io.h>
#endif
#ifdef __VMS
#include <rms.h>
#include <lib$routines.h>
#include <starlet.h>
#include <stsdef.h>
/* Modify from Severe error to error, otherwise program will exit in call to lib$signal() */
#define VMS_FATAL_TO_ERROR(__status) \
        if ( (__status & 07) == STS$K_SEVERE ) \
        { \
            __status &= ~STS$K_SEVERE;  \
            __status |= STS$K_ERROR;  \
        }
#endif /* __VMS */

static iostream_list open_streams;

static OOP asciiFileOpen(SmalltalkObject& asciiFile);
static OOP asciiFileClose(SmalltalkObject& asciiFile);
static OOP asciiFileSkip(SmalltalkObject& asciiFile, g_integer nskip);
static OOP asciiFileRewind(SmalltalkObject& asciiFile);
static OOP asciiFileLines(SmalltalkObject& asciiFile);
static OOP asciiFileReadFixedFormat(SmalltalkObject& asciiFile, String& fields, String& format, g_integer count);
static OOP asciiFileReadFreeFormat(SmalltalkObject& asciiFile, String& fields, String& separator, g_integer count);
static OOP asciiFileWrite(FileAccessCode code, SmalltalkObject& asciiFile, SmalltalkObject& aVariable, String& format_separator, String& user_comment);

static OOP asciiStringOpen(SmalltalkObject& asciiFile);

// Get "handle" and "fs" objects
// '\002' is used to indicate no comment character
#define GET_HANDLE_AND_STREAM \
    GenieInteger handle(asciiFile("handle")); \
    if ((g_integer)handle < 0 || (g_integer)handle >= open_streams.count()) \
    { \
	gerr << "ASCIIFILE: Invalid handle" << endl; \
	return nilOOP; \
    } \
    iostream* fs = open_streams[handle]; \
    if (fs == 0) \
    { \
	gerr << "ASCIIFILE: Stream number " << (g_integer)handle << " is invalid" << endl; \
	return nilOOP; \
    } \
    if (fs->fail()) \
    { \
	gerr << "ASCIIFILE: Stream number " << (g_integer)handle << " is showing an error" << endl; \
	return nilOOP; \
    } \
    GenieString comment(asciiFile("comment")); \
    GenieString delimiter(asciiFile("delimiter"));

OOP g_fileAccess(FileAccessCode code, OOP arguments)
{
    BEGINSAVEBLOCK

    OOP result(nilOOP);
    SmalltalkObject asciiFile(arguments, 1);

    switch(code)
    {
	case gxAsciiFileOpen:
	    result = asciiFileOpen(asciiFile);
	    break;

    	case gxAsciiFileRewind:
	    result = asciiFileRewind(asciiFile);
	    break;

    	case gxAsciiFileClose:
	    result = asciiFileClose(asciiFile);
	    break;

    	case gxAsciiFileSkip:
	{
	    GenieInteger nskip(arguments, 2);
	    result = asciiFileSkip(asciiFile, nskip);
	    break;
	}

    	case gxAsciiFileLines:
	    result = asciiFileLines(asciiFile);
	    break;

    	case gxAsciiFileReadFreeFormat:
  	{
	    GenieString fields(arguments, 2);
	    GenieString separator(arguments, 3);
	    GenieInteger count(arguments, 4);
	    result = asciiFileReadFreeFormat(asciiFile, fields, separator, count);
	    break;
	}

    	case gxAsciiFileReadFixedFormat:
  	{
	    GenieString fields(arguments, 2);
	    GenieString format(arguments, 3);
	    GenieInteger count(arguments, 4);
	    result = asciiFileReadFixedFormat(asciiFile, fields, format, count);
	    break;
	}

    	case gxAsciiFileWriteFreeFormat:
        {
	    SmalltalkObject aVariable(arguments, 2);	// Array
	    GenieString separator(arguments, 3);
	    GenieString user_comment(arguments, 4);
	    result = asciiFileWrite(code, asciiFile, aVariable, separator, user_comment);
	    break;
	}

    	case gxAsciiFileWriteFixedFormat:
        {
	    SmalltalkObject aVariable(arguments, 2);	// Array
	    GenieString format(arguments, 3);
	    GenieString user_comment(arguments, 4);
	    result = asciiFileWrite(code, asciiFile, aVariable, format, user_comment);
	    break;
	}

	case gxAsciiStringOpen:
	    result = asciiStringOpen(asciiFile);
	    break;

	default:
	    gerr << "ASCIIFILE: Invalid callout number = " << code << endl;
	    break;
    }
    ENDSAVEBLOCK
    return result;
}

static OOP asciiStringOpen(SmalltalkObject& asciiFile)
{
    GenieVariable var(asciiFile("file"));
    GenieString gs, delimiter(asciiFile("delimiter"));
    GenieStringArray* gsa;
    char* bytes;
    if (var.type() == GenieStringType)
    {
	gs = (GenieString)var;
	bytes = new char[gs.length()];
	strcpy(bytes, gs);
    }
    else if (var.type() == GenieStringArrayType)
    {
	gsa = (GenieStringArray*)var;
	bytes = gsa->asString(delimiter[0]);
    }
    else
    {
	*g_err << "asciiStringOpen: not a String or StringArray type" << endl;
	return nilOOP;
    }
    strstream *pss = new strstream(bytes, strlen(bytes), ios::in);
    GenieInteger handle;
    if (pss != 0)
    {
        open_streams.push(pss);
        handle = open_streams.count() - 1;
    }
    else
    {
	handle = -1;
    }
    asciiFile("handle", handle);
    return (OOP)asciiFile;
}

// return 1 if OK, 0 if error
// Return "file" if do nothing - important not to even clean up the
// filename as later code may think it is a temporary and remove it!
static const char* fix_file_attributes(const char* file)
{
#ifdef __VMS
    int rms_status;
    const char* trans_file = translatePath(file, 'F'); /* need native VMS style, not UNIX style */
    struct FAB fblock = cc$rms_fab;
    fblock.fab$l_fna = trans_file; /* need native VMS style */
    fblock.fab$b_fns = strlen(trans_file);
    fblock.fab$b_fac = FAB$V_GET;
    fblock.fab$b_shr = FAB$V_SHRPUT | FAB$V_SHRGET;
    rms_status = sys$open(&fblock);
    if (rms_status == RMS$_FNF)
    {
	return file;
    }
    if (rms_status != RMS$_NORMAL)
    {
	gerr << "Error returned from SYS$OPEN for \"" << trans_file << "\""  << endl;
	VMS_FATAL_TO_ERROR(rms_status);
	lib$signal(rms_status);
	return file;
    }
    if ( (fblock.fab$b_org == FAB$C_SEQ) && ((fblock.fab$b_rat & FAB$M_CR) != 0) && (fblock.fab$b_rfm == FAB$C_STMLF) )
    {
	sys$close(&fblock);
	return file;
    }
    rms_status = sys$close(&fblock);
    if (rms_status != RMS$_NORMAL)
    {
	gerr << "Error returned from SYS$CLOSE for \"" << trans_file << "\"" << endl;
	VMS_FATAL_TO_ERROR(rms_status);
	lib$signal(rms_status);
	return file;
    }
// now convert file
    const char* conv_file = genie_tmpfile("genie","tmp");
    ginf << "Converting \"" << file << "\" to STREAM_LF format (" << conv_file << ")" << endl;
    FILE *new_fp = fopen(conv_file, "w"), *old_fp = fopen(file, "r", "shr=get");
    static char conv_buffer[10240];
    int num_read;
    while(!feof(old_fp))
    {
	num_read = fread(conv_buffer, 1, sizeof(conv_buffer), old_fp);
	fwrite(conv_buffer, 1, num_read, new_fp);
    }
    if (ferror(old_fp) || ferror(new_fp))
    {
	gerr << "Error performing STREAM_LF format conversion" << endl;
    }
    fclose(old_fp);
    fclose(new_fp);
    return conv_file;
#else
    return file;
#endif /* __VMS */
}

static OOP asciiFileOpen(SmalltalkObject& asciiFile)
{
    GenieString file(asciiFile("file"));
    const char* conv_file = fix_file_attributes(file);
    fstream* fs = new fstream;
    int flags = ios::in | ios::out;
    int ret = open(conv_file, O_RDONLY, 0);
    if (ret == -1)
    {
	ginf << "ASCIIFILE: \"" << conv_file << "\" does not exist and will be created on write" << endl;
    }
    else
    {
	close(ret);
    }
// Try for read/write initially
    fs->open(conv_file, flags);
    if (!fs->good())
    {
	ginf << "ASCIIFILE: Cannot open \"" << conv_file << "\" for READ/WRITE - trying READONLY" << endl;
	flags = ios::in;
	fs->open(conv_file, flags);
    }
    GenieInteger handle;
    if (strcmp(file, conv_file))	// did we create a temporary file?
    {
	remove(conv_file);
    }
    if (fs->good())
    {
	open_streams.push(fs);
	handle = open_streams.count() - 1;
    }
    else
    {
	gerr << "ASCIIFILE: Error opening \"" << conv_file << "\"" << endl;
	gerr << "ASCIIFILE: " << strerror(errno) << endl;
	handle = -1;
    }
    asciiFile("handle", handle);
    return (OOP)asciiFile;
}

static OOP asciiFileClose(SmalltalkObject& asciiFile)
{
    GET_HANDLE_AND_STREAM;
// need to check type if want to do a close()
//    fs->close();
    delete fs;
    open_streams[handle] = 0; 
    asciiFile("handle", GenieInteger(-1));
    return asciiFile;
}

static OOP asciiFileRewind(SmalltalkObject& asciiFile)
{
    GET_HANDLE_AND_STREAM;
    fs->clear();		// set stream state to good
    fs->seekg(0, ios::beg);
    if (!fs->good())
	*g_err << "ASCIIFILE: Error rewinding stream" << endl;
    return asciiFile;
}

static OOP asciiFileLines(SmalltalkObject& asciiFile)
{
    GET_HANDLE_AND_STREAM;
    GenieInteger nlines;
    nlines = lines_in_file(*fs, comment, delimiter[0]);
    if (!fs->good())
	*g_err << "ASCIIFILE: Error computing lines in stream" << endl;
    return (OOP)nlines;
}

// Tries to skip "nskip" lines, returns actual number skipped
static OOP asciiFileSkip(SmalltalkObject& asciiFile, g_integer nskip)
{
    GET_HANDLE_AND_STREAM;
    int nlines(0);
    while(nskip > 0 && skip_comments(*fs, comment, delimiter[0]))
    {
	fs->ignore(INT_MAX, delimiter[0]);
	nlines++;
	nskip--;
    }
    if (!fs->good())
	*g_err << "ASCIIFILE: Error skipping stream" << endl;
    return (OOP)GenieInteger(nlines);
}


enum ItemType { UnknownItem, GenieIntegerItem, IntegerItem, 
		LongIntegerItem, GenieRealItem, FloatItem, 
		DoubleItem, GenieStringItem, StringItem };

static OOP asciiFileReadFixedFormat(SmalltalkObject& asciiFile, String& field_names,
					String& format, g_integer count)
{
#define MAXFIELDS	15	/* max items in an input line */
    GET_HANDLE_AND_STREAM;
    if (count == -1)
	count = lines_in_file(*fs, comment, delimiter[0]);
// "field_names" is a comma separated list of names to call the workspace fields 
    int i, j, nitems;
    String field_labels[MAXFIELDS];
    nitems = split(field_names, field_labels, MAXFIELDS, ',');
    ItemType field_type[MAXFIELDS];
    size_t field_size[MAXFIELDS];
    char buffer[SHORTCHARBUFF];
    void* result_array[MAXFIELDS];
    for(i=0; i<MAXFIELDS; i++)
    {
	result_array[i] = 0;
	field_type[i] = UnknownItem;
	field_size[i] = 0;
    }
// Now parse line to see what type of data we have
// work out field types
// first work out a reduced format string
    const char* fmt_pointer;
    i = 0;
    int in_format = 0;
    char* _format = new char[1+strlen(format)];
    for(fmt_pointer = format; *fmt_pointer != '\0'; fmt_pointer++)
    {
	if (!in_format && *fmt_pointer != '%')
	{
	    continue;
	}
	if (*fmt_pointer == '%' && *(fmt_pointer + 1) == '%')
	{
	    fmt_pointer++;
	    continue;
	}
	if (*fmt_pointer == '%' && *(fmt_pointer+1) != '*' && !in_format)
	{
	    in_format = 1;
	    _format[i++] = *fmt_pointer;
	    continue;
	}
	if (in_format && isalpha(*fmt_pointer))
	{
	    _format[i++] = *fmt_pointer;
	}
	if (strchr("diounxefgsc", *fmt_pointer) != 0)
	{
	    in_format = 0;
	}
    }
    _format[i] = '\0';
    String format_names[MAXFIELDS+1];
    int nfields = split(_format, format_names, MAXFIELDS+1, '%');
// as input string is e.g. %f%d we will get a split into
// three parts with the first part null; hence must decrement nfields
    nfields--;
    if (nfields != nitems)
    {
	*g_err << "ASCIIFILE: Inconsistency - nfields = " << nfields << ", nitems = " << nitems << endl;
        *g_err << "ASCIIFILE: field names = \"" << field_names << "\"" << endl;
        *g_err << "ASCIIFILE: item format = \"" << format << "\"" << endl;
        *g_err << "ASCIIFILE: reduced format = \"" << _format << "\"" << endl;
	ginf << "ASCIIFILE: Your format string must match up with all items on a line" << endl;
	ginf << "ASCIIFILE: If you wish to ignore something, use the %* notation e.g. %*f to ignore a float" << endl;
	return nilOOP;
    }
    delete[] _format;
    Regex integer_item("[diounx]", 1),  float_item("[efg]", 1), 
	string_item("[sc]", 1), long_integer_item("l[diounx]"), 
	double_item("l[efg]");
    for(i=0; i<nfields; i++)
    {
	if (format_names[i+1].matches(long_integer_item))
	{
	    field_type[i] = LongIntegerItem;
	    result_array[i] = new long[count];
	    field_size[i] = sizeof(long) / sizeof(char);
	    continue;
	}
	if (format_names[i+1].matches(integer_item))
	{
	    field_type[i] = IntegerItem;
	    result_array[i] = new int[count];
	    field_size[i] = sizeof(int) / sizeof(char);
	    continue;
	}
	if (format_names[i+1].matches(double_item))
	{
	    field_type[i] = DoubleItem;
	    result_array[i] = new double[count];
	    field_size[i] = sizeof(double) / sizeof(char);
	    continue;
	}
	if (format_names[i+1].matches(float_item))
	{
	    field_type[i] = FloatItem;
	    result_array[i] = new float[count];
	    field_size[i] = sizeof(float) / sizeof(char);
	    continue;
	}
	if (format_names[i+1].matches(string_item))
	{
	    field_type[i] = StringItem;
	    result_array[i] = new char[512*count];
	    field_size[i] = 512;
// This makes sure any string we write is NULL terminated - the %c in
// sscanf() does not ensure this
	    memset(result_array[i], '\0', 512*count);
	    continue;
	}
	*g_err << "ASCIIFILE: unknown format conversion \"" << format_names[i+1] << "\"" << endl;
	return nilOOP;
    }
// Now go back to beginning and read whole file
    int nfound(0);
    for(i=0; i<count; i++)
    {
        if (skip_comments(*fs, comment, delimiter[0]))
        {
            fs->getline(buffer, sizeof(buffer), delimiter[0]);
        }
        if (fs->good())
	{
// need to use all of result_array[MAXFIELDS] in call
	    if ((j = sscanf(buffer, format, 
		(char*)result_array[0] + nfound * field_size[0],
		(char*)result_array[1] + nfound * field_size[1],
		(char*)result_array[2] + nfound * field_size[2],
		(char*)result_array[3] + nfound * field_size[3],
		(char*)result_array[4] + nfound * field_size[4],
		(char*)result_array[5] + nfound * field_size[5],
		(char*)result_array[6] + nfound * field_size[6],
		(char*)result_array[7] + nfound * field_size[7],
		(char*)result_array[8] + nfound * field_size[8],
		(char*)result_array[9] + nfound * field_size[9],
		(char*)result_array[10] + nfound * field_size[10],
		(char*)result_array[11] + nfound * field_size[11],
		(char*)result_array[12] + nfound * field_size[12],
		(char*)result_array[13] + nfound * field_size[13],
		(char*)result_array[14] + nfound * field_size[14]
	    )) == nfields)
	    {
		nfound++;
	    }
//	    else
//	    {
//	 	ginf << "ASCIIFILE: sscanf() can only match " << j << " items on line (" << nfields << " expected)" << endl;
//		*g_err << "ASCIIFILE: please check format \"" << format << "\" is correct" << endl;
//	        return nilOOP;
//	    }
	}
	else
	{
	    *g_err << "ASCIIFILE: Error in input stream when reading data line" << i << endl;
	    ginf << "ASCIIFILE: Probably trying to read more lines than available in stream" << endl;
	    return nilOOP;
	}
    }
    if (nfound == 0)
    {
	ginf << "ASCIIFILE: no items found matching format string \"" << format << "\"" << endl;
        for(j=0; j<nitems; j++)
	{
	    delete []result_array[j];
	}
	return nilOOP;
    }
    else if (nfound != count)
    {
	ginf << "ASCIIFILE: format matched " << nfound << " lines with " << count-nfound << " discarded" <<  endl; 
    }
    GenieWorkspace result(nitems);
    float* fa;
    int* ia;
    long* la;
    double* da;
    for(j=0; j<nitems; j++)
    {
        switch(field_type[j])
        {
            case FloatItem:
		fa = (float*)result_array[j];
    		result(field_labels[j], GenieRealArray(fa, nfound));
		break;

            case DoubleItem:
		da = (double*)result_array[j];
    		result(field_labels[j], GenieRealArray(da, nfound));
		break;

            case IntegerItem:
		ia = (int*)result_array[j];
    		result(field_labels[j], GenieIntegerArray(ia, nfound));
		break;

            case LongIntegerItem:
		la = (long*)result_array[j];
    		result(field_labels[j], GenieIntegerArray(la, nfound));
		break;

	    case StringItem:
	  	{
		    GenieStringArray gsa(nfound);
		    for(i=0; i<nfound; i++)
		    {
			gsa(i, (char*)result_array[j] + i * field_size[j]);
		    }
    		    result(field_labels[j], gsa);
		}
	 	break;

	    default:
	        *g_err << "ASCIIFILE: unknown item type in input stream" << endl;
		*g_err << "ASCIIFILE: position  = " << j << ", label = \"" 
			<< field_labels[j] << "\"" << endl;
	        return nilOOP;
    	        break;
	}
	delete []result_array[j];
    }
//    if (nitems == 1)
//	return result(field_labels[0]);
//    else
//	return result;
    return result;
#undef MAXFIELDS
}

// Returns workspace unless only one item on line
static OOP asciiFileReadFreeFormat(SmalltalkObject& asciiFile, String& field_names,
					String& separator, g_integer count)
{
    GET_HANDLE_AND_STREAM;
    bool no_workspace(false);
    if (count == -1)
	count = lines_in_file(*fs, comment, delimiter[0]);
// "field_names" is a comma separated list of names to call the workspace fields 
#if defined(__DECCXX) || defined(_WIN32)
    const int MAXFIELDS(15);
#else
    const int MAXFIELDS(field_names.freq(',') + 1);
#endif
    String field_labels[MAXFIELDS];
// nfields and be 0 -> 1 item & don't return a workspace
    int nfields = split(field_names, field_labels, MAXFIELDS, ',');
    if (nfields == 0)
    {
	nfields = 1;
	no_workspace = true;
    }
    ItemType field_type[MAXFIELDS];
    char buffer[SHORTCHARBUFF];
    streampos mark = fs->tellg();
// Read first line
    if (skip_comments(*fs, comment, delimiter[0]))
    {
        fs->getline(buffer, sizeof(buffer), delimiter[0]);
    }
    if (!fs->good())
    {
        gerr << "ASCIIFILE: unable to read first line of stream" << endl;
        return nilOOP;
    }
// Now parse line to see what type of data we have
    int i, j, nitems;
    unsigned skip;
    Regex item_separator(separator, 1);
    Regex integer_item("[+-]?[0-9]+", 1);
    Regex real_item1("[+-]?[0-9]*\\.[0-9]*", 1);
    Regex real_item2("[+-]?[0-9]*[\\.]?[0-9]*[eE][+-]?[0-9]+", 1);
    String items[MAXFIELDS];
    g_real* gra;
    g_integer* gia;
    const char **gsa;
    void* result_array[MAXFIELDS];
// ignore any leading ' ' or '\t' characters in line
    for(skip = 0; (skip < sizeof(buffer)) && (buffer[skip] == ' ' || buffer[skip] == '\t'); skip++)
	;
    nitems = split(buffer+skip, items, MAXFIELDS, item_separator);
    if (nitems < nfields)
    {
	*g_err << "ASCIIFILE: Insufficient items (" << nitems 
		<< ") available on data line - " << nfields 
		<< " required" << endl;
	return nilOOP;
    }
    for(i=0; i<nfields; i++)
    {
	if (items[i].matches(real_item1) || items[i].matches(real_item2))
	{
	    field_type[i] = GenieRealItem;
	    gra = new g_real[count];
	    for(j=0; j<count; j++)
	    {
		gra[j] = UndefinedReal;
	    }
	    result_array[i] = gra;
 	}
	else if (items[i].matches(integer_item))
	{
	    field_type[i] = GenieIntegerItem;
	    gia = new g_integer[count];
	    for(j=0; j<count; j++)
	    {
		gia[j] = UndefinedInteger;
	    }
	    result_array[i] = gia;
	}
	else
	{
	    field_type[i] = GenieStringItem;
	    gsa = (const char**)new char*[count];
	    for(j=0; j<count; j++)
	    {
		gsa[j] = UndefinedString;
	    }
	    result_array[i] = gsa;
	}
    }
// Now go back to beginning and read whole file
    fs->seekg(mark);
    for(i=0; i<count; i++)
    {
        if (skip_comments(*fs, comment, delimiter[0]))
        {
            fs->getline(buffer, sizeof(buffer), delimiter[0]);
        }
        if (fs->good())
	{
// ignore any leading ' ' or '\t' characters in line
            for(skip = 0; (skip < sizeof(buffer)) && (buffer[skip] == ' ' || buffer[skip] == '\t'); skip++)
            	;
            nitems = split(buffer+skip, items, MAXFIELDS, item_separator);
	    if (nitems > nfields)
		nitems = nfields;	// discard extra data
            for(j=0; j<nfields; j++)
	    {
	        switch(field_type[j])
	        {
	            case GenieRealItem:
			gra = (g_real*)result_array[j];
		        gra[i] = (j < nitems ? atof(items[j]) : 0.0);
		        break;

	            case GenieIntegerItem:
			gia = (g_integer*)result_array[j];
		        gia[i] = (j < nitems ? atol(items[j]) : 0);
		        break;
		    case GenieStringItem:
			gsa = (const char**)result_array[j];
//		   	gsa[i] = (j < nitems ? strcpy((char*)malloc(1+strlen(items[j])), items[j]) : 0);
		   	gsa[i] = (j < nitems ? strcpy((char*)malloc(1+strlen(items[j])), items[j]) : strcpy((char*)malloc(1), ""));
			break;

		    default:
		        gerr << "ASCIIFILE: unknown item type in input stream at position " << j << endl;
			ginf << "ASCIIFILE: object is \"" << items[j] << "\"" << endl;
		        return nilOOP;
    		        break;
  	        }
	    }
        }
	else
	{
	    *g_err << "ASCIIFILE: error reading input stream at line " << i << endl;
	    ginf << "ASCIIFILE: Probably trying to read more lines than available in stream" << endl;
	    return nilOOP;
	}
    }
    GenieWorkspace result(nfields+1);
    for(j=0; j<nfields; j++)
    {
        switch(field_type[j])
        {
            case GenieRealItem:
		gra = (g_real*)result_array[j];
    		result(field_labels[j], GenieRealArray(gra, count));
		break;

            case GenieIntegerItem:
		gia = (g_integer*)result_array[j];
    		result(field_labels[j], GenieIntegerArray(gia, count));
		break;

	    case GenieStringItem:
		gsa = (const char**)result_array[j];
		result(field_labels[j], GenieStringArray(gsa, count));
		for(i=0; i<count; i++)
		{
		    if (gsa[i] != 0)
			free((void*)gsa[i]);
		}
		break;

	    default:
	        gerr << "ASCIIFILE: unknown item type in input stream" << endl;
	        return nilOOP;
    	        break;
	}
	delete []result_array[j];
    }
    if (no_workspace)
    {
	const GenieVariable& gv = result(field_labels[0]);
	return gv.value();
    }
    return result;
}

typedef long ARG_TYPE;		// alignment of argument list

static OOP asciiFileWrite(FileAccessCode code, SmalltalkObject& asciiFile, SmalltalkObject& aVariable, String& format_separator, String& user_comment)
{
    GET_HANDLE_AND_STREAM;
    int gv_len = aVariable.arrayLength();
    fs->flags(ios::showpoint);	// show decimal point
    int len, i, j, k, ncolumns(0), nlines(0);
// Determine number of lines - use longest array length found
    GenieVariable** gv_list = new GenieVariable*[gv_len+1];
    int* gv_list_len = new int[gv_len+1];
    union gv_pass_t { long l; double d; const char* s; ARG_TYPE a[sizeof(double)/sizeof(ARG_TYPE)]; };
    gv_pass_t gv_pass_tmp;
    int gv_pass_size;
#define GV_MAXARGS	20
    ARG_TYPE gv_pass[GV_MAXARGS];
    static char buffer[1024];
    GenieVariable* pgv;
    String format;
    for(j=1; j<=gv_len; j++)
    {
	if (aVariable.arrayAt(j) != nilOOP)
	{
	    ++ncolumns;
	    pgv = new GenieVariable(aVariable.arrayAt(j));
	    switch(pgv->type())
	    {
		case GenieIntegerArrayType:
		    len = ((GenieIntegerArray*)*pgv)->length();
		    gv_list_len[j] = len;
		    break;

		case GenieRealArrayType:
		    len = ((GenieRealArray*)*pgv)->length();
		    break;

		case GenieStringArrayType:
		    len = ((GenieStringArray*)*pgv)->length();
		    break;

		default:
		    len = 1;
		    break;
	    }
	    if (len > nlines)
		nlines = len;
	    gv_list[j] = pgv;
	    gv_list_len[j] = len;
	}
 	else
	{
	    gv_list[j] = 0;
	}
    }
    if (nlines == 0 || ncolumns == 0)
    {
	*g_err << "WRITEASCII: No lines" << endl;
	return nilOOP;
    }
    ginf << "ASCIIFILE: Number of lines written = " << nlines << endl;
    if (comment[0] != '\002')
    {  
        time_t clock;
	time(&clock);
	struct tm* tm_struct = localtime(&clock);
	if (strftime(buffer, sizeof(buffer), "%a %d %h at %H:%M:%S %Z", tm_struct) == 0)
        {
            strcpy(buffer, "<unknown>");
        }
	*fs << comment[0] << delimiter[0];
        *fs << comment[0] << " Data written by Open GENIE on " << buffer << delimiter[0];
	*fs << comment[0] << " " << nlines << " data line(s) and " << ncolumns << " column(s)" << delimiter[0];
	*fs << comment[0] << delimiter[0];
	if (strlen(user_comment) > 0)
	{
	    *fs << comment[0];
	    for(j=0; (size_t)j<strlen(user_comment); j++)
	    {
	        if (user_comment[j] == '\n')
	 	    *fs << delimiter[0] << comment[0];
	        else
	            fs->put(user_comment[j]);
	    }
	    *fs << delimiter[0] << comment[0] << delimiter[0];
	}
    }
// Now write out data
    if (code == gxAsciiFileWriteFreeFormat)
    {
        for(i=0; i<nlines; i++)
        {
	    k = 0;
	    for(j=1; j<=gv_len; j++)
	    {
	        pgv = gv_list[j];
	        len = gv_list_len[j];
	        if (pgv != 0)
	        {
		    k++;
		    {
	                switch(pgv->type())
	                {
		            case GenieIntegerType:
			        *fs << (GenieInteger)*pgv;
		                break;

		    	    case GenieRealType:
			        *fs << (GenieReal)*pgv;
		                break;

		            case GenieStringType:
			        *fs << (const char*)(GenieString)*pgv;
		                break;

		            case GenieIntegerArrayType:
			        *fs << (*(GenieIntegerArray*)*pgv)[i%len];
		                break;

		            case GenieRealArrayType:
			        *fs << (*(GenieRealArray*)*pgv)[i%len];
		                break;

		            case GenieStringArrayType:
			        *fs << (*(GenieStringArray*)*pgv)[i%len];
		                break;

		            default:
			        *g_err << "ASCIIFILE: Invalid type in free format write" << endl;
			        *fs << "Invalid";
		                break;	
	                }
		        if (k != ncolumns)
		            *fs << format_separator;
	            }
	        }
            }
	    *fs << delimiter[0];
	}
    } 
    else if (code == gxAsciiFileWriteFixedFormat)
    {
// Hack for Alpha - pass 6 zeros to fill up registers, as we
// don't know whether to use integer or float registers
	format = String("%n%n%n%n%n%n") + format_separator;
        for(i=0; i<nlines; i++)
        {
	    k = 0;
	    for(j=0; j<GV_MAXARGS; j++)
	    {
		gv_pass[j] = 0;
	    }
	    for(j=1; j<=gv_len; j++)
	    {
	        pgv = gv_list[j];
	        len = gv_list_len[j];
	        if (pgv != 0)
	        {
	            switch(pgv->type())
	            {
		        case GenieIntegerType:
			    gv_pass_tmp.l = (GenieInteger)*pgv;	
			    gv_pass_size = sizeof(long);
			    break;

		    	case GenieRealType:
			    gv_pass_tmp.d = (GenieReal)*pgv;	
			    gv_pass_size = sizeof(double);
		            break;

		        case GenieStringType:
			    gv_pass_tmp.s = (const char*)(GenieString)*pgv;	
			    gv_pass_size = sizeof(const char*);
		            break;

		        case GenieIntegerArrayType:
			    gv_pass_tmp.l = (*(GenieIntegerArray*)*pgv)[i%len];
			    gv_pass_size = sizeof(long);
		            break;

		        case GenieRealArrayType:
			    gv_pass_tmp.d = (*(GenieRealArray*)*pgv)[i%len];
			    gv_pass_size = sizeof(double);
		            break;

		        case GenieStringArrayType:
			    gv_pass_tmp.s = (*(GenieStringArray*)*pgv)[i%len];
			    gv_pass_size = sizeof(const char*);
		            break;

		        default:
			    *g_err << "ASCIIFILE: Invalid type in fixed format write" << endl;
			    gv_pass_tmp.l = 0;
			    gv_pass_size = sizeof(long);
		            break;	
		    }
		    if (gv_pass_size > sizeof(ARG_TYPE))
		    {
		        gv_pass[k++] = gv_pass_tmp.a[0];
		        gv_pass[k++] = gv_pass_tmp.a[1];
		    }
		    else
		    {
		        gv_pass[k++] = gv_pass_tmp.a[0];
		    }
		}
	    }
// Hack for Alpha - pass 6 zeros to fill up registers, as we
// don't know whether to use integer or float registers
       	    if (sprintf(buffer, format, &j, &j, &j, &j, &j, &j, 
			gv_pass[0], gv_pass[1], gv_pass[2], gv_pass[3], 
			gv_pass[4], gv_pass[5], gv_pass[6], gv_pass[7], 
			gv_pass[8], 0) > 0)
	    {
	        *fs << buffer << delimiter[0];
	    }
	    else
	    {
	        *g_err << "ASCIIFILE: Error writing fixed format line \"" << format_separator << "\"" << endl;
	    }
	}
    }
    for(j=1; j<=gv_len; j++)
    {
	delete gv_list[j];
    }
    delete[] gv_list;
    delete[] gv_list_len;
    return 0;
}
