/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

/*
 *      Module:         g_module
 *      Author:         Freddie Akeroyd, ISIS
 *      Purpose:        Callout for MODULE interface to calling user programs
 *
 */
#include "geniedefs.h"
#include <limits.h>

#include <iostream.h>
#ifdef _WIN32
#	include <io.h>
#	include "stream_nt.h"
#else
#	include <stream.h>
#endif /* _WIN32 */
#include <signal.h>
#include <setjmp.h>
#include "gsignals.h"
#include <gnustring.h>
#include "smalltalk.h"
#include "genievars.h"
#include "module.h"
#include "fortran_types.h"
#include "g_module.h"
#include "Genie_functions.h"
#include "callout_list.h"
#include "genie_cmodule.h"


//extern "C" 
//{
//    void for_rtl_init_(int*, char**);
//    int for_rtl_finish_();
//}
static const char* rcsid = "@(#)$Id: g_module.cxx,v 1.65 1998/10/16 13:17:58 faa Exp $";
USE_RCSID(rcsid);

typedef void (*module_func_t)(GenieWorkspace& pars, GenieWorkspace& res);
static char* current_module_function = 0;	// pointer to static string of name of executing function - used for error reporting
#ifdef _WIN32
typedef void (__stdcall *fortran_module_func_t)(GenieWorkspace& pars, GenieWorkspace& res);
typedef module_func_t c_module_func_t;
#else
typedef module_func_t fortran_module_func_t;
typedef module_func_t c_module_func_t;
#endif /* _WIN32 */

Module genie_modules;

static int execute_module(module_func_t funcptr, ModuleFunctionType func_type, GenieWorkspace& pars, GenieWorkspace& res);

extern "C" {
int map_stdio();
int unmap_stdio();
}

static void map_fortran_stdio();
static void unmap_fortran_stdio();
static void module_unmap_all();

// return nilOOP on error, arguments on success
OOP g_module(module_function_code function, OOP arguments)
{
    GenieCalloutList callout_list(arguments);
    OOP result(arguments);		// success by default

    BEGINSAVEBLOCK

    switch(function)
    {
	case Module_test:
	    break;

	case Module_load:
	{
	    GenieString file(callout_list, "FILE"), desc(callout_list, "DESC");
	    if (genie_modules.load(file, desc) == -1)
		result = nilOOP;
	}
	break;

	case Module_execute:
	{
	    ModuleFunctionType func_type;
	    GenieString func(callout_list, "FUNC"), type(callout_list, "TYPE");
	    GenieWorkspace pars(callout_list, "PARS"), res(callout_list, "RES");
	    if (type == "FORTRAN")
	    {
		func_type = FortranModuleFunction;
	    }
	    else
	    {
		func_type = CModuleFunction;
	    }
#if HAVE_DLFCN_H || defined(__VMS) || defined(__hpux) || defined(WIN32)
	    module_func_t funcptr = (module_func_t)genie_modules.getFunctionPointer(func, func_type);
#else
// The system does not support dynamic libraries, so we can only return this function
// Modules can be tested by changing their function name to "module_test_function" and linking them into genie
	    module_func_t funcptr = module_test_function;
#endif
	    if (funcptr != 0)
	    {
// No need to call map_fortran_stdio() and unmap_fortran_stdio() now as we are one process
//		int argc = 1;
//	 	char* argv[] = { "gmain", 0 };
//		if (func_type == FortranModuleFunction)
//	  	    for_rtl_init_(&argc, argv);
		if (execute_module(funcptr, func_type, pars, res) != 0)
	  	{
		    gerr << "MODULE: An error occurred while executing \"" << (const char*)func << "\"" << endl;
		    if (current_module_function != 0)
		    {
		        gerr << "MODULE: last executed function was \"" << current_module_function << "\"" << endl;
			current_module_function = 0;
		    }
		}
//		if (func_type == FortranModuleFunction)
//	  	    for_rtl_finish_();
	    }
	    else
	    {
		result = nilOOP;
	    }
	}
	break;

        case Module_list:
        {
            genie_modules.list(ginf);
//	    gout.flush();
        }
        break;

	case Module_unload:
	{
	    GenieString file(callout_list, "FILE");
	    if (genie_modules.unload(file) == -1)
		result = nilOOP;
	}
	break;

	default:
            PERROR("Module routines called with invalid function(g_module)");
            break;
    }

    ENDSAVEBLOCK

    return arguments;
}

static sigjmp_buf module_signal_context;

static void module_signal_handler(int sig)
{
    siglongjmp(module_signal_context, sig);
}

#ifdef _WIN32
	static int module_signals_to_handle[] = { SIGILL, SIGSEGV, SIGFPE }; // no SIGBUS
#else
	static int module_signals_to_handle[] = { SIGILL, SIGSEGV, SIGBUS, SIGFPE };
#endif

#define NMODULESIGNALS (sizeof(module_signals_to_handle) / sizeof(int))

#ifdef __VMS

#include <ssdef.h>
#include <stsdef.h>
#include <lib$routines.h>

static int vms_execute_module(module_func_t funcptr, ModuleFunctionType func_type, GenieWorkspace& pars, GenieWorkspace& res)
{
    if (getenv("GENIE_DEBUG") == NULL)	// If we are in the VMS debugger, we don't want a condition handler
    {
        lib$establish(lib$sig_to_ret);
    }
    (*funcptr)(pars, res);
    return SS$_NORMAL;
}

// 0 on success, otherwise signal causing problem
static int execute_module(module_func_t funcptr, ModuleFunctionType func_type, GenieWorkspace& pars, GenieWorkspace& res)
{
    volatile signal_handler old_sig_handler[32];
    int i;
    unsigned long status;
// Set signals to default action
    for(i=1; i<32; i++)
    {
	if ( (old_sig_handler[i] = genie_signal(i, SIG_DFL)) == SIG_ERR )
	    old_sig_handler[i] = SIG_DFL;
    }
    status = vms_execute_module(funcptr, func_type, pars, res);
// Free up memory allocated with cmodule_malloc()
    cmodule_freeall(NULL);	// free default heap
// Unmap arrays
    module_unmap_all();
    if ( (status & 07) != STS$K_SUCCESS )
    {
	*g_err << "MODULE: exception raised during module execution" << endl;
// Modify from Sever error to error, otherwise program will exit in call to lib$signal()
	if ( (status & 07) == STS$K_SEVERE )	
	{
	    status &= ~STS$K_SEVERE;	// clear severe error
	    status |= STS$K_ERROR;	// signal error instead
	}
	lib$signal(status);
    }
// Restore signals
    for(i=1; i<32; i++)
    {
	genie_signal(i, old_sig_handler[i]);
    }
    return status == SS$_NORMAL ? 0 : 1;
}

#else

// 0 on success, otherwise signal causing problem
static int execute_module(module_func_t funcptr, ModuleFunctionType func_type, GenieWorkspace& pars, GenieWorkspace& res)
{
    unsigned i;
    volatile int sigret = 0;
    volatile signal_handler old_sig_handler[NMODULESIGNALS];
// Attempt to catch likely signals
    if (getenv("GENIE_DEBUG") == NULL)
    {
        for(i=0; i<NMODULESIGNALS; i++)
        {
	    old_sig_handler[i] = genie_signal(module_signals_to_handle[i], module_signal_handler);
        }
        if ( (sigret = sigsetjmp(module_signal_context, 1)) == 0 )
        {
	    if (func_type == FortranModuleFunction)
	    {
    	        (*(fortran_module_func_t)funcptr)(pars, res);
	    }
	    else
	    {
    	        (*(c_module_func_t)funcptr)(pars, res);
	    }
        }
        else // signalled due to some error
        {
  	    gerr << "MODULE: Signal raised in user code during execution" << endl;
	    gerr << printsig(sigret) << endl;
        }
// restore signals
        for(i=0; i<NMODULESIGNALS; i++)
        {
	    genie_signal(module_signals_to_handle[i], old_sig_handler[i]);
        }
    }
    else
    {
	if (func_type == FortranModuleFunction)
	{
    	    (*(fortran_module_func_t)funcptr)(pars, res);
	}
	else
	{
    	    (*(c_module_func_t)funcptr)(pars, res);
	}
    }
// Free up memory allocated with cmodule_malloc()
    cmodule_freeall(NULL);	// free default heap
// Unmap arrays
    module_unmap_all();
    return sigret;
}

#endif /* __VMS */

void GENIEMECHANISM  g3_module_get_string(GenieWorkspace& pars, const STRING& name, STRING& val) 
{ 
    current_module_function = "MODULE_GET_STRING";
    String message;
    GenieString gs = pars(name.spot);
    if (gs.defined())
    {
        strcpy(val.spot, (const char*)gs);
    }
    else
    {
	cat("No string parameter ", name.spot, message);
	error_add("MODULE", message, "Fix code and reload module");
        strcpy(val.spot, "<Undefined>");
    }
    val.length = strlen(val.spot);
}

void GENIEMECHANISM  g3_module_put_string(GenieWorkspace& res, const STRING& name, const STRING& val) 
{ 
    current_module_function = "MODULE_PUT_STRING";
    res(name.spot, val.spot); 
}

// handle is typically of size fort_int[2] i.e. big enough to hold a pointer
void GENIEMECHANISM  g3_module_string_array_alloc(const fort_int& length, fort_int* handle)
{
    current_module_function = "MODULE_STRING_ARRAY_ALLOC";
    GenieStringArray* gsa = new GenieStringArray(length);
    memcpy(handle, &gsa, sizeof(GenieStringArray*));
}

void GENIEMECHANISM  g3_module_string_array_add(fort_int* handle, const fort_int& element, const STRING& val)
{
    current_module_function = "MODULE_STRING_ARRAY_ADD";
    GenieStringArray* gsa = 0;
    memcpy(&gsa, handle, sizeof(GenieStringArray*));
    (*gsa)(element - 1, val.spot);			// FORTRAN arrays start at 1, C arrays at 0
}

void GENIEMECHANISM  g3_module_string_array_free(GenieWorkspace& res, const STRING& name, fort_int* handle)
{
    current_module_function = "MODULE_STRING_ARRAY_FREE";
    GenieStringArray* gsa = 0;
    memcpy(&gsa, handle, sizeof(GenieStringArray*));
    res(name.spot, *gsa);
    delete gsa;
    memset(handle, 0, sizeof(GenieStringArray*));
}

void GENIEMECHANISM  g3_module_get_real(GenieWorkspace& pars, const STRING& name, fort_real& val) 
{ 
    current_module_function = "MODULE_GET_REAL";
    String message;
    GenieReal gr = pars(name.spot);
    if (gr.defined())
    {
	val = F(gr);
    }
    else
    {
	cat("No real parameter ", name.spot, message);
	error_add("MODULE", message, "Fix code and reload module");
	val = FLT_MAX;
    }
}

void GENIEMECHANISM  g3_module_get_double(GenieWorkspace& pars, const STRING& name, fort_double& val) 
{ 
    current_module_function = "MODULE_GET_DOUBLE";
    String message;
    GenieReal gr = pars(name.spot);
    if (gr.defined())
    {
	val = (g_real)gr;
    }
    else
    {
	cat("No real parameter ", name.spot, message);
	error_add("MODULE", message, "Fix code and reload module");
	val = DBL_MAX;
    }
}

void GENIEMECHANISM  g3_module_put_real(GenieWorkspace& res, const STRING& name, const fort_real& val) 
{ 
    current_module_function = "MODULE_PUT_REAL";
    res(name.spot, (g_real)val); 
}

void GENIEMECHANISM  g3_module_put_double(GenieWorkspace& res, const STRING& name, const fort_double& val) 
{ 
    current_module_function = "MODULE_PUT_DOUBLE";
    res(name.spot, (g_real)val); 
}

void GENIEMECHANISM  g3_module_get_int(GenieWorkspace& pars, const STRING& name, fort_int& val) 
{ 
    current_module_function = "MODULE_GET_INT";
    String message;
    GenieInteger gi = pars(name.spot);
    if (gi.defined())
    {
	val = (g_integer)gi;
    }
    else
    {
	cat("No integer parameter ", name.spot, message);
	error_add("MODULE", message, "Fix code and reload module");
	val = INT_MAX;
    }
}

void GENIEMECHANISM  g3_module_put_int(GenieWorkspace& res, const STRING& name, const fort_int& val) 
{ 
    current_module_function = "MODULE_PUT_INT";
    res(name.spot, (g_integer)val); 
}

void GENIEMECHANISM  g3_module_get_real_array(GenieWorkspace& pars, const STRING& name, fort_real* val, fort_int& len)
{
    current_module_function = "MODULE_GET_REAL_ARRAY";
    g3_module_get_real_array_alloc(pars, name, &val, len);
}


// input: len is length of input array. If *val == NULL, routine allocates
//        memory using malloc()
// output: len will be set to 0 for an error, else the length of the array 
//         that is trying to be read
void g3_module_get_real_array_alloc(GenieWorkspace& pars, const STRING& name, fort_real** val, fort_int& len)
{
    String message;
    GenieRealArray gra;
    int l_len = 0;
    gra.become(pars(name.spot));
    if (!gra.defined())
    {
	cat("No real array parameter ", name.spot, message);
	cat(message, " of length ", message);
	cat(message, dec(len), message);
	error_add("MODULE", message, "Fix code and reload module");
	len = 0;
	return;
    }
    if (*val == 0)
    {
	l_len = gra.length();
	*val = (fort_real*)cmodule_malloc(l_len*sizeof(fort_real), NULL);
    }
    else
    {
	if ( len == 0 )
	{
	    gerr << "Error: module_get_real_array called with len=0" << endl;
	    ginf << "Please specify an integer variable here, initialized to the correct length" << endl;
	    ginf << "	INTEGER nlength" << endl;
	    ginf << "	nlength = value	    (value is the number of array elements to get)" << endl;
	}
	else
	{
	    l_len = (len > gra.length() ? gra.length() : len);
	}
    }
    for(int i=0; i<l_len; i++)
    {
	(*val)[i] = F(gra[i]);
    }
    len = gra.length();
}

void GENIEMECHANISM  g3_module_get_double_array(GenieWorkspace& pars, const STRING& name, fort_double* val, fort_int& len)
{
    current_module_function = "MODULE_GET_DOUBLE_ARRAY";
    g3_module_get_double_array_alloc(pars, name, &val, len);
}
// input: len is length of input array. If *val == NULL, routine allocates
//        memory using malloc()
// output: len will be set to 0 for an error, else the length of the array 
//         that is trying to be read
void g3_module_get_double_array_alloc(GenieWorkspace& pars, const STRING& name, fort_double** val, fort_int& len)
{
    String message;
    int l_len = 0;
    GenieRealArray gra;
    gra.become(pars(name.spot));
    if (!gra.defined())
    {
	cat("No real array parameter ", name.spot, message);
	cat(message, " of length ", message);
	cat(message, dec(len), message);
	error_add("MODULE", message, "Fix code and reload module");
	len = 0;
	return;
    }
    if (*val == 0)
    {
	l_len = gra.length();
	*val = (fort_double*)cmodule_malloc(l_len*sizeof(fort_double), NULL);
    }
    else
    {
	if ( len == 0 )
	{
	    gerr << "Error: module_get_double_array called with len=0" << endl;
	    ginf << "Please specify an integer variable here, initialized to the correct length" << endl;
	    ginf << "	INTEGER nlength" << endl;
	    ginf << "	nlength = value	    (value is the number of array elements to get)" << endl;
	}
	else
	{
	    l_len = (len > gra.length() ? gra.length() : len);
	}
    }
    for(int i=0; i<l_len; i++)
    {
	(*val)[i] = gra[i];
    }
    len = gra.length();
}

void GENIEMECHANISM  g3_module_get_int_array(GenieWorkspace& pars, const STRING& name, fort_int* val, fort_int& len)
{
    current_module_function = "MODULE_GET_INT_ARRAY";
    g3_module_get_int_array_alloc(pars, name, &val, len);
}

// input: len is length of input array. If *val == NULL, routine allocates
//        memory using malloc()
// output: len will be set to 0 for an error, else the length of the array 
//         that is trying to be read
void g3_module_get_int_array_alloc(GenieWorkspace& pars, const STRING& name, fort_int** val, fort_int& len)
{
    String message;
    GenieIntegerArray gia;
    int l_len = 0;
    gia.become(pars(name.spot));
    if (!gia.defined())
    {
	cat("No integer array parameter ", name.spot, message);
	cat(message, " of length ", message);
	cat(message, dec(len), message);
	error_add("MODULE", message, "Fix code and reload module");
	len = 0;
	return;
    }
    if (*val == 0)
    {
	l_len = gia.length();
	*val = (fort_int*)cmodule_malloc(l_len*sizeof(fort_int), NULL);
    }
    else
    {
	if ( len == 0 )
	{
	    gerr << "Error: module_get_integer_array called with len=0" << endl;
	    ginf << "Please specify an integer variable here, initialized to the correct length" << endl;
	    ginf << "	INTEGER nlength" << endl;
	    ginf << "	nlength = value	    (value is the number of array elements to get)" << endl;
	}
	else
	{
	    l_len = (len > gia.length() ? gia.length() : len);
	}
    }
    for(int i=0; i<l_len; i++)
    {
        (*val)[i] = gia[i];
    }
    len = gia.length();
}

void GENIEMECHANISM  g3_module_put_double_array(GenieWorkspace& pars, const STRING& name, const fort_double* val, const fort_int& len)
{
    current_module_function = "MODULE_PUT_DOUBLE_ARRAY";
    GenieRealArray gra(val, len);
    pars(name.spot, gra);
}

void GENIEMECHANISM  g3_module_put_real_array(GenieWorkspace& pars, const STRING& name, const fort_real* val, const fort_int& len)
{
    current_module_function = "MODULE_PUT_REAL_ARRAY";
    GenieRealArray gra(val, len);
    pars(name.spot, gra);
}

fort_int GENIEMECHANISM g3_module_put_nd_real_array(GenieWorkspace& pars, const STRING& name, const fort_real* val, 
					const fort_int* dims_array, const fort_int& ndims)
{
    current_module_function = "MODULE_PUT_ND_REAL_ARRAY";
    GenieRealArray gra(val, dims_array, ndims);
    pars(name.spot, gra);
    return 0;
}

void GENIEMECHANISM  g3_module_put_int_array(GenieWorkspace& pars, const STRING& name, const fort_int* val, const fort_int& len)
{
    current_module_function = "MODULE_PUT_INT_ARRAY";
    GenieIntegerArray gia(val, len);
    pars(name.spot, gia);
}

// See genie_cmodule.h for options flag
void GENIEMECHANISM  g3_module_print(const STRING& s, const fort_int& option)
{
    current_module_function = "MODULE_PRINT";
    if (option == CMODULE_PRINT_NORMAL)
        gout.write(s.spot, strlen(s.spot)) << endl;
    else if (option == CMODULE_PRINT_INFORMATION)
        ginf.write(s.spot, strlen(s.spot)) << endl;
    else if (option == CMODULE_PRINT_ERROR)
        g_err->write(s.spot, strlen(s.spot)) << endl;
    return;
}

void GENIEMECHANISM g3_module_get_string_array_go(GenieWorkspace& pars, const STRING& iname, fort_int& length, fort_int* handle)
{
    current_module_function = "MODULE_GET_STRING_ARRAY_GO";
    String message;
    GenieStringArray* sa = new GenieStringArray(pars(iname.spot));
    if (sa->defined())
    {
        memcpy(handle, &sa, sizeof(GenieStringArray*));
	if ( length == 0 )
	{
	    gerr << "Error: module_get_string_array called with length=0" << endl;
	    ginf << "Please specify an integer variable here, initialized to the length of the array" << endl;
	}
	else
        {
	    length = (length > sa->length() ? sa->length() : length);
        }
    }
    else
    {
	cat("No string array parameter ", iname.spot, message);
	length = 0;
	error_add("MODULE", message, "Fix code and reload module");
	delete sa;
        memset(handle, 0, sizeof(GenieStringArray*));
    }
} 

void GENIEMECHANISM g3_module_get_string_array_i(fort_int* handle, fort_int& element, STRING& ivalue)
{
    current_module_function = "MODULE_GET_STRING_ARRAY_I";
    GenieStringArray* sa;
    memcpy(&sa, handle, sizeof(GenieStringArray*));
    if (sa == 0)
    {
	return;
    }  
    char* s = (*sa)[element-1];
    if (s != 0)
    {
        strncpy(ivalue.spot, s, 512);
        ivalue.length = strlen(s);
        free((void *)s);
    }
}

void GENIEMECHANISM g3_module_get_string_array_end(fort_int* handle)
{
    current_module_function = "MODULE_GET_STRING_ARRAY_END";
    GenieStringArray* sa;
    memcpy(&sa, handle, sizeof(GenieStringArray*));
    if (sa == 0)
    {
	return;
    }
    delete sa;
    memset(handle, 0, sizeof(GenieStringArray*));
}
    

// Get the major and minor version ID of the interface we are using
fort_int GENIEMECHANISM g3_module_get_version(fort_int& major, fort_int& minor)
{
    current_module_function = "MODULE_GET_VERSION";
    major = atoi(GENIE_INTERFACE_MAJOR_VERSION);
    minor = atoi(GENIE_INTERFACE_MINOR_VERSION);
    return 0;
}

// Check to see if "major" & "minor" are compatable with the interface we are currently using
// Return 1 if OK, 0 if not
fort_int GENIEMECHANISM g3_module_version_ok(const fort_int& major, const fort_int& minor)
{
    current_module_function = "MODULE_VERSION_OK";
    if ( (major == atoi(GENIE_INTERFACE_MAJOR_VERSION)) && (minor <= atoi(GENIE_INTERFACE_MINOR_VERSION)) )
    {
	return 1;		// All OK
    }
    else
    {
	*g_err  << "Interface version mismatch: this MODULE requires (at least) V" << major << "." << minor << endl
		<< "while GENIE is providing only V" << GENIE_INTERFACE_MAJOR_VERSION << "." << GENIE_INTERFACE_MINOR_VERSION << endl;
	ginf << "To use this MODULE in GENIE you could try to rebuild it by" << endl
             << "executing MODULE/COMPILE on the original program source;"  << endl
	     << "however, you may find you are missing some library functions" << endl;
	return 0;		// Not OK
    }
}
    
static void map_fortran_stdio()
{
    static const char* local_tty = OOPToString(typeNameToOOP("gXUUlocalUtty"));
    static fort_int len_local_tty = strlen(local_tty);
#ifndef HAVE_LIBF2C
    f_map_fortran_stdio(local_tty, len_local_tty);	// doesn't work for f2c
#endif
}

static void unmap_fortran_stdio()
{
#ifndef HAVE_LIBF2C
    f_unmap_fortran_stdio();				// doesn't work for f2c
#endif
}

// On input, tty_len is the number of chacters in "itty"
// On output, it should be set to the length of itty
void GENIEMECHANISM  g3_module_open_stdio(char* itty, fort_int& tty_len)
{
    static const char* local_tty = OOPToString(typeNameToOOP("gXUUlocalUtty"));
    static int len_local_tty = strlen(local_tty);
    strncpy(itty, local_tty, tty_len);		// provided tty_len is big enough, we get NULL termination
    tty_len = (len_local_tty + 1 > tty_len ? tty_len : len_local_tty);
    return;
}

void GENIEMECHANISM  g3_module_close_stdio(fort_int& stdio_unit)
{
    return;
}

/*
 * Re-assign I/O streams so FORTRAN can freely use units 5 and 6
 * Unfortunatelf, it doesn't work :-(
 */

static int old_stdin(-1), old_stdout(-1), old_stderr(-1); 	// initialise as an invalid descriptor

int map_stdio()
{
// Save current I/O channels
    if ((old_stdin = dup(0)) == -1)
	perror("map_stdio: dup()");
    if ((old_stdout = dup(1)) == -1)
	perror("map_stdio: dup()");
    if ((old_stderr = dup(2)) == -1)
	perror("map_stdio: dup()");
    return 0;
}

int unmap_stdio()
{
// Restore old I/O channels
    if (dup2(0, old_stdin) == -1)
        perror("unmap_stdio: dup2()");
    if (dup2(1, old_stdout) == -1)
        perror("unmap_stdio: dup2()");
    if (dup2(2, old_stderr) == -1)
        perror("unmap_stdio: dup2()");
// Close temporary channels so we don't run out of descriptors later on
    if (close(old_stdin) == -1)
	perror("unmap_stdio: close()");
    if (close(old_stdout) == -1)
	perror("unmap_stdio: close()");
    if (close(old_stderr) == -1)
	perror("unmap_stdio: close()");
    old_stdin = old_stdout = old_stderr = -1;		// Invalidate descriptors
    return 0;
}

struct memory_item
{
    void* address;
    size_t nbytes;
    memory_item(void* a = NULL, size_t n = 0) : address(a), nbytes(n) {}
};
typedef SPList<memory_item> MemoryItemList;

static MemoryItemList memory_list;	// default Heap
int operator<(const memory_item& a, const memory_item& b)
{
    return a.nbytes < b.nbytes;
}

typedef SPList<MemoryItemList*> HeapList;
static HeapList heap_list;	// For Additional Heaps

void GENIEMECHANISM g3_module_makeheap(fort_int* heap)
{
    current_module_function = "MODULE_MAKEHEAP";
    MemoryItemList* the_heap = new MemoryItemList;
    heap_list.push(the_heap);
    memcpy(heap, &the_heap, sizeof(MemoryItemList*));
}


void GENIEMECHANISM g3_module_malloc(const int& nelements, const int& element_size, const void* array, int address[], 
	long& offset, const fort_int* heap)
{
// int address[2]
    current_module_function = "MODULE_MALLOC";
    void* mem;
    MemoryItemList* the_heap;
    memcpy(&the_heap, heap, sizeof(MemoryItemList*));
    if (the_heap == NULL)
    {
	the_heap = &memory_list;
    }
    int extra = 16;		// extra space in case we have to shift for alignment
    size_t nbytes = nelements * element_size;
    offset = 0;
    memset(address, 0, sizeof(void*));
    mem = (void*)malloc(nbytes+extra);	//extra space in case we have to shift for alignment
    if (mem == NULL)
    {
	gerr << "g3_module_malloc: not enough memory to allocate " << nbytes << " bytes" << endl;
	return;
    }
// calculate offset such that array[offset] points to "address"
    offset = ( (const char*)mem - (const char*)array ) / element_size;
    if ( (offset * element_size + (const char*)array) != (const char*)mem )
    {
	gerr << "g3_module_malloc: alignment error " << (unsigned long)mem << " - " << (unsigned long)array << endl;
	return;
    }
    memcpy(address, &mem, sizeof(void*));
    memory_item item(mem, nbytes);
//    ginf << "malloc: allocating " << nbytes << " bytes at " << (void*)mem << endl;
    the_heap->push(item);
}

void GENIEMECHANISM g3_module_free(fort_int* address, const fort_int* heap)
{
    current_module_function = "MODULE_FREE";
    void* mem;
    MemoryItemList* the_heap;
    memcpy(&the_heap, heap, sizeof(MemoryItemList*));
    if (the_heap == NULL)
    {
	the_heap = &memory_list;
    }
    memcpy(&mem, address, sizeof(void*));
    if (mem == NULL)
    {
	gerr << "g3_module_free: attempting to free NULL pointer" << endl;
	return;
    }
    for(int i=0; i<the_heap->count() && mem != NULL; i++)
    {
	if ((*the_heap)[i].address == mem)
	{
	    free(mem);
//    	    ginf << "free: freeing " << (*the_heap)[i].nbytes << " bytes at " << (void*)mem << endl;
	    (*the_heap)[i].address = NULL;
	    mem = NULL;
	}
    }
// This is memory not allocated with a genie_malloc() - we shall assume it has been allocated
// with a malloc inside e.g. strdup() and free it anyway
    if (mem != NULL)
    {
	ginf << "g3_module_free: freeing memory at " << (void*)mem << " that was not allocated by genie_malloc()" << endl;
	ginf << "g3_module_free: this was probably allocated by malloc() from within e.g. strdup()" << endl;
	free(mem);
    }
    memset(address, 0, sizeof(void*));
}

void GENIEMECHANISM g3_module_freeall(const fort_int* heap)
{
    current_module_function = "MODULE_FREEALL";
    MemoryItemList* the_heap = NULL;
    if (heap != NULL)
    {
        memcpy(&the_heap, heap, sizeof(MemoryItemList*));
    }
    if (the_heap == NULL)
    {
	the_heap = &memory_list;
    }
    for(int i=0; i<the_heap->count(); i++)
    {
	if ((*the_heap)[i].address != NULL)
	{
//    	    ginf << "freeall: freeing " << (*the_heap)[i].nbytes << " bytes at " << (*the_heap)[i].address << endl;
	    free((*the_heap)[i].address);
	    (*the_heap)[i].address = NULL;
	}
    }
    the_heap->reset();
}

void GENIEMECHANISM g3_module_freeallheaps()
{
    current_module_function = "MODULE_FREEALLHEAPS";
    g3_module_freeall(0);	// free default heap
    for(int i=0; i<heap_list.count(); i++)
    {
	g3_module_freeall((fort_int*)heap_list[i]);
    }
    heap_list.reset();
}

typedef SPList<GenieVariableBase*> GenieVariableList;
static GenieVariableList g_var_list;

static void module_unmap_all()
{
    for(int i=0; i<g_var_list.count(); i++)
    {
	delete g_var_list.pop();
    }
    g_var_list.reset();
}

#if 0
void GENIEMECHANISM g3_module_map_real_array(GenieWorkspace& pars, const STRING& name, fort_real* val, long& offset, fort_int& len)
{
    fort_int dims_array[20], ndims = 20;
    g3_module_map_nd_real_array(pars, name, val, offset, dims_array, &ndims);
    len = 1;
    for(i=0; i<ndims; i++)
    {
	len *= dims_array[i];
    }
}
    
void GENIEMECHANISM g3_module_map_nd_real_array(GenieWorkspace& pars, const STRING& name, fort_real* val, long& offset, fort_int* dims_array, fort_int& ndims)
#endif

void GENIEMECHANISM g3_module_map_real_array(GenieWorkspace& pars, const STRING& name, fort_real* val, long& offset, fort_int& len)
{
    current_module_function = "MODULE_MAP_REAL_ARRAY";
    String message;
    GenieRealArray* gra;
    offset = 0;
    if (pars.getFieldOOP(name.spot) == nilOOP)
    {
// We need to allocate space
	if (len > 0)
	{
            gra = new GenieRealArray(len);
	    pars(name.spot, *gra);
	    gout << "array set" << endl;
	}
	else
	{
	    cat("Cannot create real array ", name.spot, message);
	    cat(message, " of length ", message);
	    cat(message, dec(len), message);
	    error_add("MODULE", message, "Fix code and reload module");
	    len = 0;
	    return;
	}
    }
    else
    {
	gra = new GenieRealArray;
        gra->become(pars(name.spot));
    }
    gra->setFlag(WriteBackInternalArrays);
    g_var_list.push(gra);
    len = gra->length();
    fort_real* vals = (fort_real*)(*gra);
    offset = (vals - val);		// so val[offset] = vals[0]
}

void GENIEMECHANISM g3_module_map_double_array(GenieWorkspace& pars, const STRING& name, fort_double* val, long& offset, fort_int& len)
{
    current_module_function = "MODULE_MAP_DOUBLE_ARRAY";
    String message;
    GenieRealArray* gra;
    offset = 0;
    if (pars.getFieldOOP(name.spot) == nilOOP)
    {
// We need to allocate space
	if (len > 0)
	{
            gra = new GenieRealArray(len);
	    pars(name.spot, *gra);
	}
	else
	{
	    cat("Cannot create double array ", name.spot, message);
	    cat(message, " of length ", message);
	    cat(message, dec(len), message);
	    error_add("MODULE", message, "Fix code and reload module");
	    len = 0;
	    return;
	}
    }
    else
    {
	gra = new GenieRealArray;
        gra->become(pars(name.spot));
    }
    gra->setFlag(WriteBackInternalArrays);
    g_var_list.push(gra);
    len = gra->length();
    fort_double* vals = (fort_double*)(*gra);
    offset = (vals - val);		// so val[offset] = vals[0]
}

// Instantiate templates
#if defined(__GNUG__)
template class SPList<memory_item>;
template class SPListBase<memory_item>;
template class SPList<MemoryItemList*>;
template class SPListBase<MemoryItemList*>;
template class SPList<GenieVariableBase*>;
template class  SPListBase<GenieVariableBase*>;
#endif /* __GNUG__ */

#ifdef __DECCXX
#pragma define_template SPList<memory_item>
#pragma define_template SPListBase<memory_item>
#pragma define_template SPList<MemoryItemList*>
#pragma define_template SPListBase<MemoryItemList*>
#pragma define_template SPList<GenieVariableBase*>
#pragma define_template SPListBase<GenieVariableBase*>
#endif /* __DECCXX */
