/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "geniedefs.h"
#include <string.h>
#include <math.h>
#include <gnustring.h>
#include "genie_types.h"
#include "genievars.h"
#include "g_objgraphics.h"
#include "Buffering.h"
#include "ranges.h"
#include "g_fortextern.h"
#include "g_module.h"
#include "callout_list.h"
#include "Genie_functions.h"

static const char* rcsid = "@(#)$Id: g_objgraphics.cxx,v 1.58 1998/01/30 18:41:58 faa Exp $";
USE_RCSID(rcsid);

/*
 *  BEWARE:
 *	precision may be lost in this process!
 */

void DoubleArrayToFloatArray( float *f_array, double *d_array, const long &len ) {

    for( long i=0; i <= len-1; i++ )
//	if (d_array[i] < FLT_MAX)
		f_array[i] = F(d_array[i]);
//	else
//		f_array[i] = FLT_MAX;
    return ;
}

// Convert an xlen length bin boundaries array into an xlen-1 bin centres array
static g_real* bins_to_points(g_real* xa, int xlen)
{
    int np = xlen - 1;
    g_real* res = new g_real[np];
    for(int i=0; i<np; i++)
    {
	res[i] = (xa[i] + xa[i+1]) / 2.0;
    }
    return res;
}

#if 0
/*
 *  BEWARE:
 *	precision may be lost in this process!
 */

void LongArrayToIntArray( int *f_array, long *d_array,const long &len ) {

    for( long i=0; i <= len-1; i++ )
	f_array[i] = d_array[i];

    return ;
}
#endif

// ---
// *** driver utility functions
//
void MyRedrawFunct() { 
	clear_device();
	redraw_picture(0, 0);
}
void MyExitFunct(int *i) { PERROR("MyExitFunct:DEVICE/CLOSE"); }
void MyHardcopyFunct(int *i) { PERROR("MyHardcopyFunct:HARDCOPY"); }

OOP g_objgraphics( graphics_function_code function, OOP parameters ) 
{
    static int debug_callout = (getenv("GENIE_DEBUG_CALLOUT") != 0 ? 1 : 0);

    GenieCalloutList gcl(parameters);

    BEGINSAVEBLOCK

    if (debug_callout)
	gdbg << "CALLOUT: calling g_objgraphics with function \"" << graphics_function_names[function] << "\"\n";

    switch(function){

/* *** GRAPHICS TEST CASE *** */
    case Gtest:

	break;

/* *** INIT DEFAULTS *** */
    case Init_defaults: {

		/* no parameters, just call it */
		init_defaults();
	}
	break;

/* *** CHECK THE ERROR STATUS *** */
    case Error_status: {
		GenieInteger retval(gcl, "STATUS");

		/* no parameters, integer return */
		retval = error_status();
	}
	break;

/* *** CLEAR THE ERRORS *** */
    case Clear_errors: {

		/* no parameters, no return */
		clear_errors();
	}
	break;

/* *** GET THE ERROR OBJECT *** */
    case Error_object: {
		/* get value to return */
	GenieString retval(gcl, "OBJECT");
		/* get error number */
	GenieInteger value(gcl, "OBJNUM");


		retval = error_object(value);
	}
	break;

/* *** GET THE ERROR FAILURE *** */
    case Error_failure: {
		/* get error number */
	GenieString retval(gcl, "FAILURE");
	GenieInteger value(gcl, "OBJNUM");

		/* get value to return */
		retval = error_failure(value);

	}
	break;

/* *** GET THE ERROR SOLUTION *** */
    case Error_solution: {
	GenieString retval(gcl, "SOLUTION");
		/* get error number */
	GenieInteger value(gcl, "OBJNUM");

		/* get value to return */
		retval = error_solution(value);

	}
	break;

/* *** GET THE ERROR MESSAGE *** */
    case Error_message: {
		/* get error number */
	GenieString retval(gcl, "MESSAGE");
	GenieInteger value(gcl, "OBJNUM");

		/* get value to return */
		retval = error_message(value);

	}
	break;
/* *** ADD AN ERROR MESSAGE *** */
    case Add_error: {
		GenieString obj(gcl, "OBJECT"), fail(gcl, "FAILURE"), soln(gcl, "SOLUTION");
		error_add(obj,fail,soln);
	}
	break;

/* *** OPEN THE DEVICE *** */
    case Open_device: {
	funcPtrI MEF = MyExitFunct;
	funcPtrI MHF = MyHardcopyFunct;
	funcPtr MRF = MyRedrawFunct;
	GenieString devname(gcl, "DEVICE"), filename(gcl, "FILENAME");
	GenieReal x(gcl, "X"), y(gcl, "Y"), ra(gcl, "SIZE");

		/* device name is first parameter */
		/* second parameter is filename */
		/* third parameter is X dimension of view surface */
		/* fourth parameter is Y dimension os view surface */
		/* fourth parameter is Y dimension os view surface */
		/* call the GENIE FUNCTION */
		open_device(devname,filename,F(x),F(y),F(ra),
		"Open GENIE","Genie",MEF,MRF,"Hardcopy",0,NULL,MHF);
	}
	break;

/* *** CLOSE THE DEVICE *** */
    case Close_device: {

		/* no parameters, just call the GENIE FUNCTION */
		close_device();
	}
	break;

/* *** CLEAR THE DEVICE *** */
   case Clear_device: {

		/* no parameters */
		clear_device();
	}
	break;

/* *** CREATE A PICTURE *** */
    case Create_picture: {

		/* no parameters */
		create_picture();
	}
	break;

/* *** DELETE ALL PICTURES *** */
    case Delete_pictures: {

		/* no parameters */
		delete_pictures();
	}
	break;

/* *** REDRAW A PICTURE *** */
    case Redraw_picture: {
		/* get picture number */
		/* get picture number */
	GenieInteger value(gcl, "PICTNUM"), value2(gcl, "RESET");

		/* call function */
		redraw_picture(value,value2);
	}
	break;

/* *** GET NUMBER OF CURRENT PICTURE *** */
    case Getpictnum: {
		/* get picture number variable */
	GenieInteger value(gcl, "PICTNUM");

		value = getpictnum();
	}
	break;
		
/* *** GET CURSOR IN A PICTURE *** */
    case Get_picture_cursor: {
	char getl[2] = { '\0', '\0' };
	GenieInteger wind(gcl, "WINDOW");
	GenieRealArray Warr(gcl, "WORLD"), Darr(gcl, "DEVICE");
	GenieString temp3(gcl, "LETTER");
	Warr.setFlag(WriteBackInternalArrays);
	Darr.setFlag(WriteBackInternalArrays);

		/* get window number */
		/* get the letter */
		/* get world array pointer */
		/* get device array pointer */

		/* call the function */
		get_picture_cursor(wind, Warr, Darr, &getl[0]);

		/* the letter must be put back            */
		temp3 = getl;

	}
	break;

/* *** UNDRAW A PRIMITIVE *** */
	case Undraw: {
	GenieInteger number(gcl, "PRIMNO");
	if (!number.defined()) { number = 1; }

		undraw(number);
	}
	break;

/* *** UNDRAW ALL PICTURES *** */
	case Undraw_all: {

		/* no parameters, complete clearout */
		undraw_all();
	}
	break;

/* *** DELETE A WINDOW AND ALL PRIMITIVES INSIDE IT *** */
	case Window_delete: {
		/* get window number */
		GenieInteger wind(gcl, "WINDOW");
		window_delete(wind);
	}
	break;

/* *** UNDO LAST UNDRAW *** */
	case Undo_undraw: {

		/* no parameters */
		undo_undraw();
	}
	break;

/* *** CREATE AN UNSCALED WINDOW *** */
	case Create_window_unscaled: {
		GenieReal xdl(gcl, "XMIN"),xdu(gcl, "XMAX"),ydl(gcl, "YMIN"),ydu(gcl, "YMAX");
		GenieInteger col(gcl, "COLOUR");
	  	GenieInteger id(gcl, "ID");

		/* get window device coordinates */
		/* get window colour */
		id = create_window_unscaled(xdl,ydl,xdu,ydu,col);
	}
	break;

/* *** CREATE A SCALED WINDOW *** */
	case Create_window_scaled: {
		GenieReal xdl(gcl, "XMIN"),xdu(gcl, "XMAX"),ydl(gcl, "YMIN"),ydu(gcl, "YMAX");
		GenieReal LXMIN(gcl, "LXMIN"),LXMAX(gcl, "LXMAX"),LYMIN(gcl, "LYMIN"),LYMAX(gcl, "LYMAX");
		GenieRealArray xa(gcl, "XARRAY");
		GenieRealArray ya(gcl, "YARRAY");
		GenieRealArray ea(gcl, "EARRAY");
		GenieInteger col(gcl, "COLOUR"),mins(gcl, "MINMODE"),maxs(gcl, "MAXMODE");
	  	GenieInteger id(gcl, "ID");
		int np;

		/* get window device coordinates */
		/* get any limits on the world co-ordinates */
		/* first find the array lengths (shortest) */
		/* get three arrays, find shortest	   */
		/* get window colour */
		/* get scaling factors */

		np = xa.length();
		if (ya.length() < np) 
			np = ya.length();
		if (ea.defined() && (ea.length() < np))
			np = ea.length();

		id = create_window_scaled(xdl,ydl,xdu,ydu,xa,ya,ea,np,col,
				mins,maxs,LXMIN,LXMAX,LYMIN,LYMAX);
	}
	break;

/* *** CREATE A MULTIPLOT WINDOW *** */
	case Create_window_multiplot: {
		GenieReal xdl(gcl, "XMIN"),xdu(gcl, "XMAX"),ydl(gcl, "YMIN"),ydu(gcl, "YMAX");
		GenieInteger mins(gcl, "MINMODE"),maxs(gcl, "MAXMODE");
		GenieInteger col(gcl, "COLOUR");
	  	GenieInteger id(gcl, "ID");

		/* get window device coordinates */
		/* get window colour */
		/* get scaling factors */

		id = create_window_multiplot(xdl,ydl,xdu,ydu,col,mins,maxs);
	}
	break;



/* *** GET CURSOR IN A WINDOW *** */

    case Get_window_cursor: {
	GenieRealArray Warr(gcl, "WORLD");
	GenieRealArray Darr(gcl, "DEVICE");
	Warr.setFlag(WriteBackInternalArrays);
	Darr.setFlag(WriteBackInternalArrays);
	char getl[2] = { '\0', '\0' };
	GenieString temp3(gcl, "LETTER");

		/* get the letter */
		/* get world array pointer */
		/* convert it to single precision */
		/* get device array pointer */

		/* call the function  */
		get_window_cursor(Warr, Darr, &getl[0]);

		/* the letter must be put back            */
		temp3 = getl;
	}
	break;


/* *** DRAW AXES *** */
	case Draw_axes: {
		GenieInteger col(gcl, "COLOUR"),qa(gcl, "QUALITY");
		GenieInteger xl(gcl, "XLOG"),yl(gcl, "YLOG");
		GenieReal lw(gcl, "LINEWIDTH"),th(gcl, "TEXTHT");
		GenieInteger xnum(gcl, "XNUM"), ynum(gcl, "YNUM");	// Whether to show axis numbering
		GenieInteger yhoriz(gcl, "YHORIZ");

		/* get line colour */
		/* get line width */
		/* get text quality */
		/* get text height */
		/* get log/linear parameter */

		/* MAKE THE CALLOUT */
		draw_axes(col, lw, th, qa, xl, yl, xnum, ynum, yhoriz,0);
	}
	break;

/* *** DRAW_TEXT *** */
	case Draw_text: {
		GenieReal xp(gcl, "X"),yp(gcl, "Y");
		GenieString ts(gcl, "STRING");
		GenieReal th(gcl, "TEXTHEIGHT"),an(gcl, "ANGLE"),lw(gcl, "LINEWIDTH");
		GenieInteger qa(gcl, "QUALITY"),col(gcl, "COLOUR");

		/* get the X/Y position */
		/* get the text string */
		/* get text height, angle and quality */
		/* get colour */
		/* get the linewidth */
		/* callout */

		draw_text(xp, yp, ts, th, an, col, qa, lw);
	}
	break;

/* *** DRAW_HISTOGRAM *** */
	case Draw_histogram: {
		GenieInteger col(gcl, "COLOUR"), bc(gcl, "BIN_CENTRE"), ty(gcl, "LINETYPE");
		GenieReal lw(gcl, "LINEWIDTH");
		GenieRealArray xa(gcl, "XARRAY"), ya(gcl, "YARRAY");
		int np;

		/* get data array pointers */
		/* get colour */
		/* get line type */
		/* get bin centre parameter */
		/* get line width */

		/* find shortest */
		np = xa.length();
		if (ya.length() < np) np = ya.length();

		/* callout */
		draw_histogram(np, xa, ya, col, bc, ty, lw);
	}
	break;

/* *** CREATE_MULTIPLOT *** */
	case Create_multiplot: {
		GenieRealArray xa(gcl, "XARRAY");
		long int npp;
		GenieInteger col(gcl, "COLOUR"),YP(gcl, "NUMHIST"),bc(gcl, "BINCENTRE"),ty(gcl, "LINETYPE");
		GenieInteger ci(gcl, "CHMIN"),ca(gcl, "CHMAX");
		GenieReal gap(gcl, "YGAP"), lt(gcl, "LINEWIDTH");

		/* get X and Y arrays, sizes array */
		/* get (NUMHIST) */
		/* get channel range to plot */
		/* get Y gap */
		/* get colour */
		/* get bin centre */
		/* get line type */
		/* get line thickness */

		npp = xa.length();
		if ((int)ci == -1) 
		    ci = 0;
		if ((int)ca == -1) 
		    ca = npp - 1;

		/* callout */
		create_multiplot(YP, xa, npp, ci, ca, gap, col,bc,ty,lt);

	}
	break;

/* *** ADD_SPECTRA *** */
	case Add_one_spectrum: 
	{
		GenieRealArray ya(gcl, "YARRAY");
		const g_real* ydata = ya.data();
		/* 
		 * yarray may be two dimensionsal 
		 */
		int nloop, step, ndims = ya.dimsSize();
		if (ndims == 1)
		{
		    nloop = 1;
		    step = 0;
		}
		else // 2 or more
		{
		    step = ya.dope(1);
		    nloop = ya.dims(1);
		}
		int YP = ya.dims(ndims);		// C arrays are row-wise, so size = last dimension
		for(int i=0; i<nloop; i++)
		{
		    add_spectra(ydata, YP);
		    ydata += step;
		}
	}
	break;

	case TotalInRange:
	{
	    GenieInteger res(gcl, "RES");
	    GenieString range(gcl, "RANGE");
	    RangeList* range_list = parse_range(range);
	    res = sum_range(range_list);
	    delete range_list;
	}
	break;
	
// NPOINTS is the number of raw points in a spectrum
// These will be grouped into bunches of NGROUP before being added to the multiplot
	case Add_spectra:
	{
	    int np;
	    GenieInteger npoints(gcl, "NPOINTS"), ngroup(gcl, "NGROUP");
	    int* ya = new int[(size_t)npoints+1];
	    GenieString file(gcl, "FILE");
	    GenieIntegerArray spectra_array(gcl, "SPECTRA_ARRAY");
	    for(int i=0; i<spectra_array.length(); i++)
	    {
		    if ((np = get_spectrum(file, spectra_array[i], ya, npoints, ngroup)) > 0)
		        add_spectra(ya, np);
	    }
	    delete[] ya;
	}
	break;

/* *** DRAW MULTIPLOT *** */
	case Draw_multiplot: {
		draw_multiplot();
	}
	break;

#if 0		/* COMMENTED OUT */
/* *** CREATE A 2D PLOT *** */
	case Create_2d_plot: {

		int Xdims[2];
		int Ydims[2];
		int Zdims[2];
		g_real *conts;
		g_integer *shades;
		int *colours;
		int nconts,i;
		int type;
		float width;
		g_variable temp1, temp2;
		
		/* get dimensions */
		Xdims[0]=getInteger(getElementByField(arguments,"X0"));
		Xdims[1]=getInteger(getElementByField(arguments,"X1"));

		Ydims[0]=getInteger(getElementByField(arguments,"Y0"));
		Ydims[1]=getInteger(getElementByField(arguments,"Y1"));

		Zdims[0]=getInteger(getElementByField(arguments,"Z0"));
		Zdims[1]=getInteger(getElementByField(arguments,"Z1"));

		/* get contour array */
		temp1 = getElementByField(arguments,"CONTOURS");
		conts = arrayGetAllReal(temp1);

		/* get number of contours */
		nconts = (int)ArrayLength(temp1);

		/* get colour array */
		temp2 = getElementByField(arguments,"COLOURS");
		shades = arrayGetAllInteger(temp2);

		/* create an integer array */
		colours = new int[nconts];
		for(i=0; i<nconts; i++) {
			colours[i] = shades[i]; }

		/* get linetype */
		type = getInteger(getElementByField(arguments,"LINETYPE"));

		/* get linewidth */
		width = getInteger(getElementByField(arguments,"LINEWIDTH"));

		/* call the constructor */
		create_2d_plot(Xdims, Ydims, Zdims, conts, nconts,
				colours, type, width);
		/* tidy up */
		delete [] colours;
	}
	break;

/* *** add a 2d spectra *** */
	case Add_2d_spectra: {
		g_real *x, *y,*z;
		g_variable temp1, temp2, temp3;
		int num;

		temp1 = getElementByField(arguments,"XARRAY");
		x = arrayGetAllReal(temp1);
		temp2 = getElementByField(arguments,"YARRAY");
		y = arrayGetAllReal(temp2);
		temp3 = getElementByField(arguments,"ZARRAY");
		z = arrayGetAllReal(temp3);

		num = getInteger(getElementByField(arguments,"NUMBER"));

		/* add the spectra */
		add_2d_spectra(x, y, z, num);
	}
	break;

/* *** DRAW 2D PLOT *** */
	case Draw_2d_plot: {

		draw_2d_plot();
	}
	break;

#endif		/* COMMENTED OUT */

/* *** DRAW_PLOT *** */
	case Draw_plot: {
		int np;
		g_real* xxa;
		GenieInteger col(gcl, "COLOUR"), ty(gcl, "LINETYPE");
		GenieReal lw(gcl, "LINEWIDTH");
		GenieRealArray xa(gcl, "XARRAY"), ya(gcl, "YARRAY");

		/* get data array pointers */
		/* get colour */
		/* get line type */
		/* get line width */

		int xlen = xa.length(), ylen = ya.length();

		/* find shortest */
		np = (xlen < ylen ? xlen : ylen);
	  	if ( (xlen - ylen) == 1 ) 		// Are we trying to draw a histogram?
		{
		    xxa = bins_to_points(xa, xlen);
		    draw_plot(np, xxa, ya, col, ty, lw);
		    delete[] xxa;
		}
		else
		{
		    draw_plot(np, xa, ya, col, ty, lw);
		}
	}
	break;

/* *** DRAW MARKERS *** */
	case Draw_markers: {
		g_real* xxa;
		GenieInteger col(gcl, "COLOUR"), sy(gcl, "SYMBOL");
		int np;
		GenieReal cs(gcl, "CHARSIZE"), lw(gcl, "LINEWIDTH");
		GenieRealArray xa(gcl, "XARRAY"), ya(gcl, "YARRAY");

		/* get data array pointers */
		/* get colour */
		/* get line type */
		/* get bin centre parameter */
		/* get line width */

		/* find shortest */
		int xlen = xa.length(), ylen = ya.length();

		/* find shortest */
		np = (xlen < ylen ? xlen : ylen);
	  	if ( (xlen - ylen) == 1 ) 		// Are we trying to draw a histogram?
		{
		    xxa = bins_to_points(xa, xlen);
		    draw_markers(np, xxa, ya, col, sy, cs, lw);
		    delete[] xxa;
		}
		else
		{
		    draw_markers(np, xa, ya, col, sy, cs, lw);
		}
	}
	break;

/* *** DRAW GRATICULE *** */
	case Draw_graticule: {
		GenieInteger col(gcl, "COLOUR"),xl(gcl, "XLOG"),yl(gcl, "YLOG");
		GenieReal lw(gcl, "LINEWIDTH");
		float ns; 
		ns = 1.0F;
		/* get colour */
		/* get xlog parameter */
		/* get ylog parameter */
		/* get line width */
		/* callout */

		draw_graticule(col, lw, xl,yl,ns,1,1,1,1,0);
	}
	break;

/* *** DRAW LABELS *** */
	case Draw_labels: {
		GenieString xl(gcl, "XLABEL"), yl(gcl, "YLABEL");
		GenieInteger col(gcl, "COLOUR"),qa(gcl, "QUALITY");
		GenieReal cs(gcl, "TEXTHT"), lw(gcl, "LINEWIDTH");
		GenieInteger xvert(gcl, "XVERTICAL"), yvert(gcl, "YVERTICAL");

		/* get colour */
		/* get quality */
		/* get character size*/
		/* get the two strings */
		/* callout */
		draw_labels(xl, yl, col, qa, cs, lw, xvert, yvert);
	}
	break;

/* *** DRAW TITLE *** */
	case Draw_title: {
		GenieString xl(gcl, "TITLE");
		GenieInteger col(gcl, "COLOUR"),qa(gcl, "QUALITY");
		GenieReal cs(gcl, "TEXTHT");

		/* get colour */
		/* get quality */
		/* get character size*/
		/* get the string */
		/* callout */
		draw_title(xl, col, qa, cs);
	}
	break;

/* *** DRAW_BOX *** */
	case Draw_box: {
		GenieReal xl(gcl, "XLOW"),yl(gcl, "YLOW"),xh(gcl, "XHIGH"),yh(gcl, "YHIGH");
		GenieInteger col(gcl, "COLOUR"),so(gcl, "SOLID");
		GenieReal lw(gcl, "LINEWIDTH");

		/* get dimensions */
		/* get colour */
		/* get solid parameter */
		/* get line width */
		/* callout */
		draw_box(xl, yl, xh, yh, col, lw,so);
	}
	break;

/* *** DRAW LINE *** */
	case Draw_line: {
		GenieReal xl(gcl, "XLOW"),yl(gcl, "YLOW"),xh(gcl, "XHIGH"),yh(gcl, "YHIGH");
		GenieInteger col(gcl, "COLOUR"),ty(gcl, "LINETYPE");
		GenieReal lw(gcl, "LINEWIDTH");

		/* get dimensions */
		/* get colour */
		/* get line type */
		/* get line width */
		/* callout */
		draw_line(xl, yl, xh, yh, col, ty, lw);
	}
	break;

/* *** DRAW CURVE *** */
	case Draw_curve: {

	}
	break;

/* *** DRAW ERRORS *** */
    case Draw_errors: {
		GenieInteger col(gcl, "COLOUR"), sy(gcl, "ANGLE");
		g_real* xxa;
		int np;
		GenieReal lw(gcl, "LINEWIDTH");
		GenieRealArray xa(gcl, "XARRAY");
		GenieRealArray ya(gcl, "YARRAY");
		GenieRealArray ea(gcl, "DYARRAY");

		/* get data array pointers */
		/* get colour */
		/* get error type */
		/* get line width */

		/* find shortest */
		int xlen = xa.length(), ylen = ya.length(), elen = ea.length();
		np = xlen;
		if (ylen < np) np = ylen;
		if (elen < np) np = elen;

	  	if ( ((xlen - ylen) == 1) && ((xlen - elen) == 1) ) 		// Are we trying to draw a histogram?
		{
		    xxa = bins_to_points(xa, xlen);
		    draw_errors(np, xxa, ya, ea, col, sy, lw);
		    delete[] xxa;
		}
		else
		{
		    draw_errors(np, xa, ya, ea, col, sy, lw);
		}
	}
	break;

/* *** CHECKDEV *** */
	case Checkdev: {
		GenieString retval(gcl, "DEVNAME");

		// check for Tektronix
		if (retval.contains("TE") !=0) retval = "TEK4010";
		if (retval.contains("te") !=0) retval = "TEK4010";
		if (retval.contains("Te") !=0) retval = "TEK4010";
		if (retval.contains("tE") !=0) retval = "TEK4010";

		// check for X windows
		if (retval.contains("XW") !=0) retval = "XW";
		if (retval.contains("xw") !=0) retval = "XW";
		if (retval.contains("Xw") !=0) retval = "XW";
		if (retval.contains("xW") !=0) retval = "XW";

	}
	break;

/* *** Set_display_update *** */
// 0 to disable buffering and flush, 1 to enable buffering, anything else is ignored (used to get current update value)
	case Set_display_update: 
	{
	    GenieInteger enable_buffering(gcl, "ENABLE");
	    enable_buffering =  set_buffer_update(enable_buffering);
	}
	break;

/* *** Flush_buffer *** */
	case Flush_buffer: 
	{
	    int old_buffering_value = set_buffer_update(0); 	// temporarily disable buffering
	    flush_buffer();					// flush graphics events
	    set_buffer_update(old_buffering_value);		// reset buffer status to old value
	}
	break;

/* *** DRAW_POLYGON *** */
	case Draw_polygon: 
	{
		GenieRealArray x(gcl, "X"), y(gcl, "Y");
		GenieInteger col(gcl, "COLOUR"), so(gcl, "SOLID"), numsides(gcl, "NUMSIDES");
		GenieReal lw(gcl, "LINEWIDTH");

		draw_polygon(x, y, numsides, col, lw, so);
	}
	break;

/* *** DEFINE_RGB_COLOUR *** */
	case Define_rgb_colour:
	{
	    GenieReal r(gcl, "R"), g(gcl, "G"), b(gcl, "B");
	    GenieInteger colour_index(gcl, "COLOUR_INDEX");

	    colour_index = define_rgb_colour(r, g, b);
	}
	break;

/* *** DEFINE_HLS_COLOUR *** */
	case Define_hls_colour:
	{
	    GenieReal hue(gcl, "H"), lightness(gcl, "L"), saturation(gcl, "S");
	    GenieInteger colour_index(gcl, "COLOUR_INDEX");

	    colour_index = define_hls_colour(hue, lightness, saturation);
	}
	break;

/* *** DEFINE_NAMED_COLOUR *** */
	case Define_named_colour:
	{
	    GenieString name(gcl, "NAME");	
	    GenieInteger colour_index(gcl, "COLOUR_INDEX");

	    colour_index = define_named_colour(name);
	}
	break;

	case Contour_plot:
	{
/*
 * IMODE = 0             DRAW/CONTOUR
 * IMODE = 1             DRAW/CONTOUR/ARRAY
 * IMODE = 2             DRAW/CONTOUR/FUNCTION
 *
 * ITYPE = 0		linear contours
 * ITYPE = 1 		LOG contours
 * ITYPE = 2 		SQRT contours
 */
	    g_real *cont, *tr_matrix;
	    GenieRealArray values(gcl, "VALUES"), wx(gcl, "WX"), wy(gcl, "WY"), 
				contours(gcl, "CONTOURS"), tr(gcl, "TR");
	    GenieString func(gcl, "FUNC");
	    GenieInteger ncont(gcl, "NCONT"), line_type(gcl, "LINETYPE"), colour(gcl, "COLOUR"), 
				imode(gcl, "IMODE"), itype(gcl, "ITYPE");
	    GenieReal line_width(gcl, "LINEWIDTH");
	    void* func_ptr;
// Work out contour levels if required
	    if (contours.defined())
	    {
		cont = contours;
		ncont = contours.length();
	    }
	    else
	    {
		cont = get_contour_levels(values, values.length(), ncont, itype);
 	    }
// Work out TR matrix if required
 	    if (tr.defined())
	    {
	 	tr_matrix = tr;
	    }
	    else
	    {
		tr_matrix = get_contour_tr_matrix(values.dims(1), values.dims(2));
	    }
	    
	    switch(imode)
	    {
		case 0:
		    draw_contour(values, values.dims(1), values.dims(2), tr_matrix,
				 cont, ncont, line_type, colour, (g_real)line_width);
		    break;

		case 1:
	    	    draw_contour_array(values, values.dims(1), values.dims(2), 
				       wx, wx.dimsArray(), wx.dimsSize(), wy, wy.dimsArray(), wy.dimsSize(),
				       cont, ncont, line_type, colour, (g_real)line_width);
		    break;

		case 2:
#if HAVE_DLFCN_H || defined(__VMS) || defined(__hpux)
		    func_ptr = genie_modules.getFunctionPointer(func);	// (genie_cont_func_t)
#else
		    func_ptr = sample_genie_coord_tran_func;	// in module_test_function.f
#endif /* HAVE_DLFCN_H */
		    if (func_ptr != 0)
		    {
		        draw_contour_function(values, values.dims(1), values.dims(2), func_ptr,
				 	  cont, ncont, line_type, colour, (g_real)line_width);
		    }
		    else
		    {
			error_add("draw_contour_function", "Function name unknown", "");
		    }
		    break;

		default:
		    error_add("Contour_plot", "Invalid mode", "");
		    break;
	    }
	}
	break;

	case Contour_label:
	{
	    GenieString label(gcl, "LABEL");
	    GenieInteger start(gcl, "START"), end(gcl, "END"), st(gcl, "ST"), colour(gcl, "COLOUR"), font(gcl, "FONT");
	    GenieInteger intval(gcl, "INTVAL"), minint(gcl, "MININT");
	    GenieReal height(gcl, "HEIGHT"), line_width(gcl, "LINEWIDTH");
	    label_contour(label, start, end, st, intval, minint, colour, font, (g_real)height, (g_real)line_width);
	}
	break;

// convert an HLS colour to RGB 
 	case HLS_to_RGB:
	{
	    GenieWorkspace rgb_colour(gcl, "RGB_COLOUR");			// "Colour" structure
	    GenieReal hue(gcl, "H"), lightness(gcl, "L"), saturation(gcl, "S");
	    pg_float red, green, blue;
	    grxrgb(hue, lightness, saturation, red, green, blue);
	    rgb_colour("R", red);
	    rgb_colour("G", green);
	    rgb_colour("B", blue);
	}
	break;

	case Allocate_predefined_colourtable:
	{
	    GenieString name(gcl, "NAME");
	    GenieWorkspace wk(gcl, "WK");
	    predefined_colourtable(name, wk);
	}
	break;

	case Allocate_colourtable:
	{
	    static int ic_col = 256;	// for IC1,IC2 emulation
	    GenieWorkspace table(gcl, "TABLE");
	    GenieRealArray r(gcl, "R"), g(gcl, "G"), b(gcl, "B");
 	    GenieInteger ncol(gcl, "NCOL"), nbands(gcl, "NBANDS");
	    int ncol_alloc = ncol;	// The number really allocated ( <= ncol on return )
	    GenieReal contrast(gcl, "CONTRAST"), diffuse(gcl, "DIFFUSE"),
		shine(gcl, "SHINE"), polish(gcl, "POLISH");
	    g_integer* ct = allocate_colourtable(r, g, b, r.length(), 
			ncol_alloc, contrast, nbands, diffuse, shine, polish);
	    GenieIntegerArray ci(ct, ncol_alloc);
	    table("CI", ci);
	    GenieRealArray ra(ncol_alloc), ga(ncol_alloc), ba(ncol_alloc);
	    for(int i=0; i<ncol_alloc; i++)
	    {
		query_rgb_colour(ct[i], ra[i], ga[i], ba[i]);
	    }
	    table("R", ra);
	    table("G", ga);
	    table("B", ba);
	    delete[] ct;
// Bit of a hack - the 3D routines assume contiguous colours between
// IC1 and IC2. As we don't do this in GENIE, we emulate it by returning
// A dummy range that just has to be unique. Later on we correct this
// Range by passing the integer array of tru colours and using SBCCRT
	    table("IC1", ic_col);
	    table("IC2", ic_col+ncol_alloc-1);
	    ic_col += ncol_alloc;
	}
	break;

	case Deallocate_colourtable:
	{
	    GenieWorkspace table(gcl, "TABLE");
	    GenieIntegerArray ci = table("CI");
	    deallocate_colourtable(ci, ci.length());
	}
	break;

	case Cellarray_plot:
	{
/*
 * IOPT = 0             DRAW/CELLARRAY 
 * IOPT = 1             DRAW/CELLARRAY/ARRAY
 * IOPT = 2             DRAW/CONTOUR/FUNCTION
 *
 * IMODE = 0            Same colour for all of cell
 * IMODE = 1            Interpolate colour across cell
 *
 * ITYPE = 0		linear levels
 * ITYPE = 1 		log levels
 * ITYPE = 2 		sqrt levels
 */
	    GenieRealArray values(gcl, "VALUES"), wx(gcl, "WX"), wy(gcl, "WY"), tr(gcl, "TR");
	    GenieInteger iopt(gcl, "IOPT"), imode(gcl, "IMODE"), itype(gcl, "ITYPE");
	    GenieReal valmax(gcl, "VALMAX"), valmin(gcl, "VALMIN");
	    GenieWorkspace colourtable(gcl, "COLOURTABLE");
	    GenieString func(gcl, "FUNC");
            GenieIntegerArray table = colourtable("CI");
	    void* func_ptr;
	    if (!valmax.defined())
		valmax = values.array_max(1);
	    if (!valmin.defined())
		valmin = values.array_min(1);
	    switch(iopt)
	    {
		case 0:
		case 1:
	            draw_cellarray_array(values, values.dims(1), values.dims(2), wx, wx.dimsArray(), wx.dimsSize(), 
				wy, wy.dimsArray(), wy.dimsSize(), valmax, valmin, itype, imode, table, table.length());
		    break;

		case 2:
#if HAVE_DLFCN_H || defined(__VMS) || defined(__hpux)
		    func_ptr = genie_modules.getFunctionPointer(func);	// (genie_cont_func_t)
#else
		    func_ptr = sample_genie_coord_tran_func;	// in module_test_function.f
#endif /* HAVE_DLFCN_H */
                    if (func_ptr != 0)
                    {
                        draw_cellarray_function(values, values.dims(1), values.dims(2), valmax, valmin, func_ptr,
                                          itype, imode, table, table.length());
                    }
                    else
                    {
                        error_add("draw_cellarray_function", "Function name unknown", "");
                    }
                    break;

		default:
		    error_add("Contour_plot", "Invalid option", "");
		    break;
	    }
	}
	break;

	case Create_window_2d:
	{
	    GenieReal dev_xmin(gcl, "DEV_XMIN"), dev_xmax(gcl, "DEV_XMAX"), dev_ymin(gcl, "DEV_YMIN"), dev_ymax(gcl, "DEV_YMAX");
	    GenieRealArray x(gcl, "X"), y(gcl, "Y");
	    GenieInteger colour(gcl, "COLOUR");
	    GenieInteger id(gcl, "ID");
	    id = create_window_two_d(dev_xmin, dev_xmax, dev_ymin, dev_ymax, x, x.length(), y, y.length(), colour);
	}
	break;

	case Cellarray_wedge:
	{
	    GenieInteger objnum(gcl, "OBJNUM"), mode(gcl, "MODE");
	    draw_cellarray_wedge(objnum, mode);
	}
	break;

// Find out information about an object on the display list
	case Query_object:
	{
	    GenieObjectID objid(gcl, "OBJID");		// position on list
	    GenieWorkspace res(gcl, "RES");		// returned object information into pre-allocated variable
	    query_object(res, objid);
	}
	break;

/* *** DEFAULT *** */
    default: {
	}
	break;
    }

    ENDSAVEBLOCK

    return parameters;
}

