/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "geniedefs.h"
#include <iostream.h>
#include <math.h>
#include <limits.h>

#include "g_types.h"
#include "smalltalk.h"
#include "g_operations.h"
#include "g_module.h"
#include "genievars.h"
#include "transform_utils.h"

static const char* rcsid = "@(#)$Id: g_operations.cxx,v 1.36 1998/06/16 14:49:20 faa Exp $";
USE_RCSID(rcsid);

// Constants
const int INST_VAR_OFFSET = 2;			    /* Offset to data in a GXArray (dims & dope are inst vars 1 & 2) */
const int PASS_NILS = 0;			    /* Pass nil values onto function */
const int SKIP_NILS = 1;			    /* don't pass nil values to function */
const double LN2LOG10 = log(10.0);

// Auxilliary functions to return the correct OOP for a single operation, these functions
// are the analogue of "aBlock" in the smalltalk collect:

// Unary real operations
static g_real g_negate(g_real val) { return -val; }
static g_real g_sin(g_real val) { return sin(val); }
static g_real g_cos(g_real val) { return cos(val); }
static g_real g_tan(g_real val) { return tan(val); }
static g_real g_arcSin(g_real val) { return -1.0 <= val && val <= 1.0 ? asin(val) : UndefinedReal; }
static g_real g_arcCos(g_real val) { return -1.0 <= val && val <= 1.0 ? acos(val) : UndefinedReal; }
static g_real g_arcTan(g_real val) { return atan(val); }
static g_real g_sqrt(g_real val) { return (val >= 0.0 ? (g_real)sqrt(val) : UndefinedReal); }
static g_real g_abs(g_real val) { return (val >= 0.0 ? val : -val); }
static g_real g_ln(g_real val) { return (val > 0.0 ? (g_real)log(val) : UndefinedReal); }
static g_real g_log10(g_real val) { return (val > 0.0 ? (g_real)log(val)/LN2LOG10 : UndefinedReal); }
static g_real g_exp(g_real val) { return exp(val); }

// Binary real operations
static g_real g_remainder(g_real val1, g_real val2) 
{ 
    if (val2 != 0.0) 
    { 
	return val1 - val2 * (g_real)(long)(val1 / val2);
    }   
    else 
    { 
 	return UndefinedReal; 
    } 
}
static g_real g_raised_to(g_real val1, g_real val2) { return pow(val1, val2); }
// Fix undefined values to "val2"
static g_real g_fix(g_real val1, g_real val2) { return(IS_UNDEFINED_REAL(val1) ? val2 : val1);  } 
// Undefine values that are equal to "val2"
// We do the comparison as "float" rather than "g_real" as values may be read in as "float" into
// genie (e.g. from FORTRAN) and then not be equal to the same value specified in a "double"
static g_real g_unfix(g_real val1, g_real val2) { return(F(val1) == F(val2) ? UndefinedReal : val1);  } 

// Unary integer operations (others done by coercing to real)
static g_integer g_negate(g_integer val) { return -val; }
static g_integer g_abs(g_integer val) { return (val >= 0 ? val : -val); }

// Binary integer operations (others done by coercing to real)
static g_integer g_raised_to(g_integer val1, g_integer val2) { return (g_integer)pow(val1, (double)val2); }
static g_integer g_remainder(g_integer x, g_integer y) 
{ 
    if (y != 0) 
    { 
 	return (y > 0 ? x % y : x-((x-(y-1))/y)*y); 
    }   
    else 
    { 
 	return UndefinedInteger; 
    } 
}
// Fix undefined values to "val2"
static g_integer g_fix(g_integer val1, g_integer val2) { return(IS_UNDEFINED_INTEGER(val1) ? val2 : val1);  } 
// Undefine values that are equal to "val2"
static g_integer g_unfix(g_integer val1, g_integer val2) { return(val1 == val2 ? UndefinedInteger : val1);  } 

// Callin from Smalltalk
// =====================

OOP g_operations( operation_function_code function, OOP arguments )
{
    static int debug_callout = (getenv("GENIE_DEBUG_CALLOUT") != 0 ? 1 : 0);

    OOP result(nilOOP);

    BEGINSAVEBLOCK
 
    if (debug_callout)
	gdbg << "CALLOUT: calling g_operations with function \"" << operation_function_names[function] << "\"\n";

    switch(function)				    /* NOTE: each part of the CASE is responsible for doing a return OOP; */
    {
        case Operation_test:
	{
	    cout << "Operations test called" << endl;
	    break;
	}

        case GXRealarray_extend:						// extend a single value to an array
	{
	    GenieRealArray self(arguments, 1);
	    GenieReal aReal(arguments, 2);
	    GenieRealArray extended_array(self);
	    result = (OOP)self;		// reference self to get round CXX bug - not really needed
	    extended_array.fill(aReal);
	    result = (OOP)extended_array;
	    break;
	}

        case GXIntegerarray_extend:						// extend a single value to an array
	{
	    GenieIntegerArray self(arguments, 1);
	    GenieInteger anInteger(arguments, 2);
	    GenieIntegerArray extended_array(self);
	    result = (OOP)self;		// reference self to get round CXX bug - not really needed
	    extended_array.fill(anInteger);
	    result = (OOP)extended_array;
	    break;
	}
	
        case GXRealarray_shallowCopy:
	{
	    GenieRealArray self(arguments, 1);
	    result = (OOP)self.shallowCopy();
	    break;
	}

        case GXIntegerarray_shallowCopy:
	{
	    GenieIntegerArray self(arguments, 1);
	    result = (OOP)self.shallowCopy();
	    break;
	}

        case GXIntegerarray_asFloat:						// float an integer
	{
	    GenieIntegerArray self(arguments, 1);
	    GenieRealArray aRealArray(self);
	    result = (OOP)aRealArray;
	    break;
 	}

        case GXRealarray_collect_hint:						// an optimized collect: ( hint: is what to do )
	{
	    GenieRealArray self(arguments, 1), aBlock(arguments, 2);
	    GenieChar hint(arguments, 3);
	    GenieRealArray result_array(self);
	    switch(hint)
	    {
		case 'n': result_array.collect(g_negate); break; 				// negated
		case 's': result_array.collect(g_sin); break; 					// sin
		case 'c': result_array.collect(g_cos); break; 					// cos
		case 't': result_array.collect(g_tan); break; 					// tan
		case 'S': result_array.collect(g_arcSin); break; 				// arcSin
		case 'C': result_array.collect(g_arcCos); break; 				// arcCos
		case 'T': result_array.collect(g_arcTan); break; 				// arcTan
		case 'l': result_array.collect(g_ln); break; 					// ln
		case 'e': result_array.collect(g_exp); break; 					// exp
		case 'L': result_array.collect(g_log10); break; 				// log10
		case 'q': result_array.collect(g_sqrt); break; 					// sqrt
		case 'a': result_array.collect(g_abs); break; 					// |a|
	        default:
		PERROR("Undefined unary real function");
		    break;
	    }
	    result = (OOP)result_array;
	    break;
	}

        case GXIntegerarray_collect_hint:				// an optimized collect: ( hint: is what to do )
	{
	    GenieIntegerArray self(arguments, 1), aBlock(arguments, 2);
	    GenieChar hint(arguments, 3);
	    GenieIntegerArray result_array(self);

	    switch(hint)
	    {
		case 'n': result_array.collect(g_negate); break; 					// negated
		case 'a': result_array.collect(g_abs); break; 						// |x|
	        default:
		    PERROR("Undefined unary integer function");
		    break;
	    }
	    result = (OOP)result_array;
	    break;
	}

        case GXRealarray_with_collect_hint:
	{
	    GenieRealArray self(arguments, 1), anArray(arguments, 2), aBlock(arguments, 3);
	    GenieChar hint(arguments, 4);
	    GenieRealArray result_array;

	    switch(hint)
	    {
		case '+' : result_array.become(self + anArray); break;
		case '-' : result_array.become(self - anArray); break;
		case '*' : result_array.become(self * anArray); break;
		case '/' : result_array.become(self / anArray); break;
	    }
	    if (!result_array.defined())
	    {
		result_array = self;
	        switch(hint)
		{
	            case '\\': result_array.withCollect(g_remainder, anArray, SKIP_NILS); break;	// remainder
	  	    case '^' : result_array.withCollect(g_raised_to, anArray, SKIP_NILS); break;
	  	    case 'f' : result_array.withCollect(g_fix, anArray, PASS_NILS); break;
	            default:
		        PERROR("Undefined binary real function");
		        break;
		}
	    }
	    result = (OOP)result_array;
	    break;
	}

        case GXIntegerarray_with_collect_hint:
	{
	    GenieIntegerArray self(arguments, 1), anArray(arguments, 2), aBlock(arguments, 3);
	    GenieChar hint(arguments, 4);
	    GenieIntegerArray result_array;

	    switch(hint)
	    {
		case '+' : result_array.become(self + anArray); break;
		case '-' : result_array.become(self - anArray); break;
		case '*' : result_array.become(self * anArray); break;
		case '/' : result_array.become(self / anArray); break;
	    }
	    if (!result_array.defined())
	    {
		result_array = self;
	        switch(hint)
		{
	            case '\\': result_array.withCollect(g_remainder, anArray, SKIP_NILS); break;	// remainder
	  	    case '^' : result_array.withCollect(g_raised_to, anArray, SKIP_NILS); break;
	  	    case 'f' : result_array.withCollect(g_fix, anArray, PASS_NILS); break;
	            default:
		        PERROR("Undefined binary integer function");
		        break;
		}
	    }
	    result = (OOP)result_array;
	    break;
	}

        case GXRealarray_equals:
	{
	    GenieRealArray self(arguments, 1), anArray(arguments, 2);
	    result = (self == anArray ? trueOOP : falseOOP);
	    break;
	}

        case GXIntegerarray_equals:
	{
	    GenieIntegerArray self(arguments, 1), anArray(arguments, 2);
            result = (self == anArray ? trueOOP : falseOOP);
	    break;
	}

        case GXRealarray_max:
	{
	    GenieRealArray self(arguments, 1);
	    result = GenieReal::toOOP(self.array_max());
	    break;
	}

        case GXIntegerarray_max:
	{
	    GenieIntegerArray self(arguments, 1);
	    result = GenieInteger::toOOP(self.array_max());
	    break;
	}

        case GXRealarray_min:
	{
	    GenieRealArray self(arguments, 1);
	    result = GenieReal::toOOP(self.array_min());
	    break;
	}

        case GXIntegerarray_min:
	{
	    GenieIntegerArray self(arguments, 1);
	    result = GenieInteger::toOOP(self.array_min());
	    break;
	}

        case GXRealarray_sum:
	{
	    GenieRealArray self(arguments, 1);
            result = GenieReal::toOOP(self.sum());
	    break;
	}

        case GXIntegerarray_sum:
	{
	    GenieIntegerArray self(arguments, 1);
	    result = GenieInteger::toOOP(self.sum());
	    break;
	}

    	case GXRealarray_at:
        {
            GenieRealArray   self(arguments, 1);
            GenieInteger        index(arguments, 2);
            result = GenieReal::toOOP(self[(g_integer)index-1]);
	    break;
        }

    	case GXRealarray_at_put:
        {
            GenieRealArray  self(arguments, 1);
            GenieInteger    index(arguments, 2);
	    GenieReal	    aReal(arguments,3);
            self[(g_integer)index-1] = (g_real)aReal;
            result = (OOP)aReal;
	    break;
        }


    	case GXIntegerarray_at:
        {
            GenieIntegerArray   self(arguments, 1);
            GenieInteger        index(arguments, 2);
            result = GenieInteger::toOOP(self[(g_integer)index-1]);
	    break;
        }

    	case GXIntegerarray_at_put:
        {
            GenieIntegerArray   self(arguments, 1);
            GenieInteger        index(arguments, 2), anInteger(arguments,3);
            self[(g_integer)index-1] = (g_integer)anInteger;
            result = (OOP)anInteger;
	    break;
        }

    	case GXRealarray_element_size:
            result = fromInt(sizeof(g_real));
	    break;

    	case GXIntegerarray_element_size:
            result = fromInt(sizeof(g_integer));
	    break;

	case GXRealarray_fix:
	{
	    GenieRealArray self(arguments, 1);
	    GenieReal aFloat(arguments, 2);
	    for(int i=0; i<self.length(); i++)
	    {
	   	self[i] = g_fix(self[i], aFloat);
	    }
	    result = (OOP)self;
	    break;
	}

	case GXRealarray_unfix:
	{
	    GenieRealArray self(arguments, 1);
	    GenieReal aFloat(arguments, 2);
	    for(int i=0; i<self.length(); i++)
	    {
	   	self[i] = g_unfix(self[i], aFloat);
	    }
	    result = (OOP)self;
	    break;
	}

	case GXIntegerarray_fix:
	{
	    GenieIntegerArray self(arguments, 1);
	    GenieInteger anInteger(arguments, 2);
	    for(int i=0; i<self.length(); i++)
	    {
	   	self[i] = g_fix(self[i], anInteger);
	    }
	    result = (OOP)self;
	    break;
	}

	case GXIntegerarray_unfix:
	{
	    GenieIntegerArray self(arguments, 1);
	    GenieInteger anInteger(arguments, 2);
	    for(int i=0; i<self.length(); i++)
	    {
	   	self[i] = g_unfix(self[i], anInteger);
	    }
	    result = (OOP)self;
	    break;
	}

// MODE: 0 = arithmetic, 1 = geometric
        case GXIntegerarray_extend_with_mode:						// extend a single value to an array
	{
	    GenieIntegerArray self(arguments, 1);
	    GenieInteger anInteger(arguments, 2), aStep(arguments, 3), aMode(arguments, 4);
	    GenieIntegerArray extended_array(self);
	    if (aStep.defined())
	    {
		extended_array.fill(anInteger, aStep, aMode);
	    }
	    else
	    {
	        extended_array.fill(anInteger);
	    }
	    result = (OOP)extended_array;
	    break;
	}

// MODE: 0 = arithmetic, 1 = geometric
        case GXRealarray_extend_with_mode:						// extend a single value to an array
	{
	    GenieRealArray self(arguments, 1);
	    GenieReal aReal(arguments, 2), aStep(arguments, 3);
	    GenieInteger aMode(arguments, 4);
	    GenieRealArray extended_array(self);
	    if (aStep.defined())
	    {
		extended_array.fill(aReal, aStep, aMode);
	    }
	    else
	    {
	        extended_array.fill(aReal);
	    }
	    result = (OOP)extended_array;
	    break;
	}

	case GXIntegerarrayCopyFromTo:
	{
	    GenieIntegerArray self(arguments, 1);
	    GenieInteger start(arguments, 2), stop(arguments, 3);
	    int size = (int)stop - (int)start + 1;
	    GenieIntegerArray result_array(size);
	    for(int i=0; i<size; i++)
	    {
		result_array[i] = self[i+(int)start - 1];	// -1 as smalltalk arrays start at 1
	    }
	    result = (OOP)result_array;
	    break;
	}

	case GXRealarrayCopyFromTo:
	{
	    GenieRealArray self(arguments, 1);
	    GenieInteger start(arguments, 2), stop(arguments, 3);
	    int size = (int)stop - (int)start + 1;
	    GenieRealArray result_array(size);
	    for(int i=0; i<size; i++)
	    {
		result_array[i] = self[i+(int)start - 1];	// -1 as smalltalk arrays start at 1
	    }
	    result = (OOP)result_array;
	    break;
	}

	case GXIntegerarrayReplaceFromToWithStartingAt:
	{
	    GenieIntegerArray self(arguments, 1), anArray(arguments, 4);
	    GenieInteger start(arguments, 2), stop(arguments, 3), repStart(arguments, 5);
	    int i_start = (int)start - 1, i_repStart = (int)repStart - 1;
	    int size = min((int)stop - i_start, (int)anArray.length() - i_repStart);
	    for(int i=0; i<size; i++)
	    {
		self[i_start + i] = anArray[i_repStart + i];
	    }
	    result = (OOP)self;
	    break;
	}

	case GXRealarrayReplaceFromToWithStartingAt:
	{
	    GenieRealArray self(arguments, 1), anArray(arguments, 4);
	    GenieInteger start(arguments, 2), stop(arguments, 3), repStart(arguments, 5);
	    int i_start = (int)start - 1, i_repStart = (int)repStart - 1;
	    int size = min((int)stop - i_start, (int)anArray.length() - i_repStart);
	    for(int i=0; i<size; i++)
	    {
		self[i_start + i] = anArray[i_repStart + i];
	    }
	    result = (OOP)self;
	    break;
	}
	
// Apply function of type   g_real (*)(g_real&)
	case GXRealarray_apply_function:
	{
	    GenieRealArray self(arguments, 1);
	    GenieString func(arguments, 2);
	    real_array_collect_ref_t func_ptr = (real_array_collect_ref_t)genie_modules.getFunctionPointer(func);
	    if (func_ptr != 0)
	    {
	        self.collect(func_ptr);
	    }
	    else
	    {
		gerr << "ApplyFunction - real function \"" << func << "\" not loaded" << endl;
	    }
	    result = (OOP)self;
	    break;
	}

// Apply function of type   g_real (*)(g_real&, int[], int&)
	case GXRealarray_apply_nd_function:
	{
	    GenieRealArray self(arguments, 1);
	    GenieString func(arguments, 2);
	    real_array_collect3_ref_t func_ptr = (real_array_collect3_ref_t)genie_modules.getFunctionPointer(func);
	    if (func_ptr != 0)
	    {
	        self.collect(func_ptr);
	    }
	    else
	    {
		gerr << "ApplyNDFunction - real function \"" << func << "\" not loaded" << endl;
	    }
	    result = (OOP)self;
	    break;
	}

// Apply function of type   g_integer (*)(g_integer&)
	case GXIntegerarray_apply_function:
	{
            GenieIntegerArray self(arguments, 1);
            GenieString func(arguments, 2);
            integer_array_collect_ref_t func_ptr = (integer_array_collect_ref_t)genie_modules.getFunctionPointer(func);
	    if (func_ptr != 0)
	    {
	        self.collect(func_ptr);
	    }
	    else
	    {
		gerr << "ApplyFunction - integer function \"" << func << "\" not loaded" << endl;
	    }
            result = (OOP)self;
	    break;
	}

// Apply function of type   g_integer (*)(g_integer&, int[], int&)
	case GXIntegerarray_apply_nd_function:
	{
            GenieIntegerArray self(arguments, 1);
            GenieString func(arguments, 2);
            integer_array_collect3_ref_t func_ptr = (integer_array_collect3_ref_t)genie_modules.getFunctionPointer(func);
	    if (func_ptr != 0)
	    {
	        self.collect(func_ptr);
	    }
	    else
	    {
		gerr << "ApplyNDFunction - integer function \"" << func << "\" not loaded" << endl;
	    }
            result = (OOP)self;
	    break;
	}

	case GXRealarray_uncentre:
	{
	    GenieRealArray self(arguments, 1);
	    GenieRealArray halfwidth(arguments, 2);
	    result = (OOP)self.pointsToBins(halfwidth);
	    break;
	}

	case GXRealarray_centre:
	{
	    GenieRealArray self(arguments, 1);
	    result = (OOP)self.binsToPoints();
	    break;
	}

        case GXRealarray_cut:
	{
	    GenieRealArray self(arguments, 1);
	    GenieInteger dimension(arguments, 2);
	    GenieReal value(arguments, 3);
	    result = (OOP)self.slice(dimension, value);
	}
	break;

        case GXRealarray_bracket:
	{
	    int i;
	    GenieRealArray self(arguments, 1);
	    GenieReal value(arguments, 2);
	    GenieReal res(nilOOP);
	    i = bracket((g_real)value, (g_real*)self, (g_integer)self.length());
	    if (i != -1)
	    {
		res = (g_real)(i + 1) + (value - self[i]) / (self[i+1] - self[i]);
	    }
	    result = (OOP)res;
	}
	break;
	
        case GXRealarray_rank:
	{
	    GenieRealArray self(arguments, 1);
	    GenieInteger direction(arguments, 2);
	    g_integer* ranks = self.rank(direction);
	    GenieIntegerArray res(ranks, self.length());
	    delete[] ranks;
	    result = (OOP)res;
	}
	break;

        case GXIntegerarray_rank:
	{
	    GenieIntegerArray self(arguments, 1);
	    GenieInteger direction(arguments, 2);
	    g_integer* ranks = self.rank(direction);
	    GenieIntegerArray res(ranks, self.length());
	    delete[] ranks;
	    result = (OOP)res;
	}
	break;

        case GXRealarray_sort:
	{
	    GenieRealArray self(arguments, 1);
	    GenieIntegerArray ranks(arguments, 2);
	    self.sort(ranks);
	    result = (OOP)self;
	}
	break;

        case GXIntegerarray_sort:
	{
	    GenieIntegerArray self(arguments, 1);
	    GenieIntegerArray ranks(arguments, 2);
	    self.sort(ranks);
	    result = (OOP)self;
	}
	break;

    default:
	PERROR("Undefined function");
	break;

    }
    ENDSAVEBLOCK
    return result;
}
