/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "geniedefs.h"
#include <stdlib.h>

#ifdef HAVE_UNISTD_H
#	include <unistd.h>
#endif

#ifdef _WIN32
#include <strstrea.h>
#else
#include <strstream.h>
#endif

#include <stdio.h>		// Only for sprintf()
#include <string.h>
// For getcputimes
//#include <sys/times.h>
//#include <unistd.h>
//#include <sys/resource.h>

#ifndef _WIN32
#	include <sys/wait.h>
#endif

#include <signal.h>
#include <setjmp.h>
#include <fcntl.h>
#if defined(__VMS) && !defined(__VMS_POSIX)
#    include <unixio.h>
#    include <descrip.h>
#    include <lib$routines.h>
#endif

#if defined(_WIN32)
#   include <io.h>
#	define WEXITSTATUS(__arg)	__arg
#	define WIFEXITED(__arg)	(__arg != 1)
static char* get_win32_error()
{
    static LPVOID lpMsgBuf = NULL;
    if (lpMsgBuf != NULL)
    {
	LocalFree(lpMsgBuf);
	lpMsgBuf = NULL;
    }
    FormatMessage(
            FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
            NULL,
            GetLastError(),
            MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
            (LPTSTR) &lpMsgBuf,    0,    NULL );
    return (char*)lpMsgBuf;
}
#endif /* _WIN32 */

#include "gsignals.h"

static const char* rcsid = "@(#)$Id: g_system.cxx,v 1.57 1998/06/19 09:48:34 cmm Exp $";
USE_RCSID(rcsid);

#undef MIN		// lose an unecessary extra definition of this !!

#include "smalltalk.h"
#include "genie_types.h"
#include "g_system.h"
#include "genievars.h"
#include "genielog.h"
#include "genie_tmpfile.h"

#include "callout_list.h"

static void print_parameters( g_variable arguments );
static void run_test( g_variable arguments );
static void clear_pipe();

// return -1 on error, else fd to keep for call to restore_stream()
static int save_stream(const char* file, int fd)
{
    int save_fd, new_fd;
    if ( (save_fd = dup(fd)) == -1 )
	return -1;
    if ( (new_fd = open(file, O_RDWR|O_CREAT|O_TRUNC|O_APPEND, 0600)) == -1 )
	return -1;
    if (close(fd) == -1)
	return -1;
    if (dup2(new_fd, fd) == -1)
	return -1;
    if (close(new_fd) == -1)
	return -1;
    return save_fd;
}
    
// return 0 on success, else failure
static int restore_stream(int save_fd, int fd)
{
    if (close(fd) == -1)
	return -1;
    if (dup2(save_fd, fd) == -1)
	return -1;
    if (close(save_fd) == -1)
	return -1;
    return 0;
}
    
static void output_stream(const char* file_fd, genie_ostream* g_stream)
{
    static char line_buffer[256];
    fstream fs;
    bool done = false;
    fs.open(file_fd, ios::in);
    if (!fs.good())
    {
	gerr << "output_stream: cannot open \"" << file_fd << "\"" << endl;
	return;
    }
    while(!done)
    {
        fs.getline(line_buffer, sizeof(line_buffer), '\n');
	if (fs.good())
	{
	    *g_stream << line_buffer << endl;
 	}
	else
	{
	    done = true;
	}
    }
    fs.close();
}

//ostream Gout;
//ostream Gerr;
//istream Gin;

extern char *symToCString( OOP oopSym );

bool gcTortureInCallout = false;
int gcTortureFacility = -1;

// Return oopParameters on success, nilOOP on failure
OOP g_system( system_function_code function, OOP oopParameters )
{
    static char buffer[LONGCHARBUFF];
    OOP result(oopParameters);		// default: success
    GenieCalloutList callout_list(oopParameters);

    BEGINSAVEBLOCK


    switch(function){

    case Test:
    {
        cout << "Called Test: entrypoint in g_system.cc" << endl;
        break;
    }

    case SystemCommand:	     //  Send a command to the system for execution
    {
// Set res to 0 on success, 1 on error
        int status(-1), fstatus(-1);
        GenieString command(callout_list, "command");
	GenieInteger res(callout_list, "res");

// -- the system() call doesn't like signals so restore to defaults
        Restore_signals();   
#ifdef __VMS
#define make_descriptor(name,string)    \
    struct dsc$descriptor_s name = {  strlen(string), DSC$K_DTYPE_T, DSC$K_CLASS_S, \
					strcpy((char*)malloc(strlen(string)+1), string) }
	const char *file_fd1 = genie_tmpfile("genie1", "tmp");
	make_descriptor(vms_in, "NL:");		// to avoid %DCL-E-TRMMBX
	make_descriptor(vms_out, file_fd1);
	make_descriptor(vms_command, command);
	unsigned sub_status = 1;
	status = lib$spawn(&vms_command, &vms_in, &vms_out, 0, 0, 0, &sub_status, 0, 0, 0, 0, 0, 0);
	output_stream(file_fd1, g_out);
	unlink(file_fd1);
	free(vms_in.dsc$a_pointer);
	free(vms_out.dsc$a_pointer);
	free(vms_command.dsc$a_pointer);
	if ( (status & 1) && (sub_status & 1) )
	{
	    res = 0;
	}
	else
	{
            *g_err << "SYSTEM: LIB$SPAWN or subprocess error" << endl;
	    *g_err << "SYSTEM: executed command was \"" << command << "\"" << endl;
	    result = nilOOP;
	    res = 1;
	}
#elif defined(_WIN32)
	STARTUPINFO si;
	PROCESS_INFORMATION pi;
	ZeroMemory(&si, sizeof(si));
	si.cb = sizeof(si);
	char* winCommand;
	const char *file_fd1 = genie_tmpfile("genie1", "tmp");
	command = command + " > " + file_fd1;
	winCommand = strdup((const char*)command);
	if (!CreateProcess(NULL, winCommand, NULL, NULL, TRUE, 0, NULL, NULL, 
	    &si, &pi))
	{
	    gerr << "SYSTEM: Error executing \"" << command << "\"" << endl;
	    gerr << "SYSTEM: Error = " << get_win32_error() << endl;
	}
	WaitForSingleObject(pi.hProcess, INFINITE);
	CloseHandle(pi.hProcess);
	CloseHandle(pi.hThread);
	free(winCommand);
	output_stream(file_fd1, g_out);
	unlink(file_fd1);
#else
	const char *file_fd0 = NULL_OUTPUT_DEVICE;
	const char *file_fd1 = genie_tmpfile("genie1", "tmp");
	const char *file_fd2 = genie_tmpfile("genie2", "tmp");
	int save_fd0 = save_stream(file_fd0, 0);
	int save_fd1 = save_stream(file_fd1, 1);
	int save_fd2 = save_stream(file_fd2, 2);
	if ( (save_fd0 == -1) || (save_fd1 == -1) || (save_fd2 == -1) )
	{
	    gerr << "Error saving file descriptors" << endl;
	}
        // System return wait(2) code
        status = system(command);
	if ( (restore_stream(save_fd0, 0) != 0) || (restore_stream(save_fd1, 1) != 0) || (restore_stream(save_fd2, 2) != 0) )
	{
	    gerr << "Error restoring streams" << endl;
	}
	output_stream(file_fd1, g_out);
//	output_stream(file_fd2, g_err);		// Just seems to be a copy of the other
	unlink(file_fd1);
	unlink(file_fd2);
	if (status == -1)
	{
	    PERROR("g_system(SystemCommand)");
	    result = nilOOP;
	}
	else if ( (status != 0) && WIFEXITED(status) )
	{
            *g_err << "SYSTEM: system() returned unusual status = " << WEXITSTATUS(status) << endl;
	    *g_err << "SYSTEM: executed command was \"" << command << "\"" << endl;
	    result = nilOOP;
	}
	res = (status != 0 ? 1 : 0);
#endif /* __VMS */
        Set_signals();	       // -- Return signal state to normal for GENIE
        break;
    }

    case GetEnv:	        // Read an environment variable
    {
        GenieString name(callout_list, "name");
        GenieString value(callout_list, "value");
        char* cp = 0;
	if ((const char*)name != NULL && (cp = getenv(name)) != NULL)
	{
	    value = cp;
	}
        break;
    }

    case GCTorture:
    {
	GenieInteger flag(callout_list, "flag"), facility(callout_list, "facility"); 
#if defined(GC_TORTURE)
	if ((int)flag == 0)
	{
	    gcTortureInCallout = false;
	    gcTortureFacility = -1;
	}
 	else if ((int)flag == 1)
	{
	    gcTortureInCallout = true;
	    gcTortureFacility = facility;
	}
#else
	gerr << "GCTORTURE unavaliable - Smalltalk was not configured with the \"--enable-gc-torture\" option" << endl;
	result = nilOOP;
#endif /* GC_TORTURE */
	break;
    }

    case GetGenieVersion:
    {
	    GenieString info(callout_list, "VERSION_INFO");
	    info = /*(const char*)*/g_version;
	    break;
    }

    case GenieLog:
    {
	    GenieString message(callout_list, "MESSAGE");
	    if (genielog(message) == -1)
	    {
	  	result = nilOOP;
		*g_err << "GENIELOG: error sending message" << endl;
	    }
	    break;
    }

    case MakeTempFile:
    {
	GenieString prefix(callout_list, "PREFIX"), directory(callout_list, "DIRECTORY"), suffix(callout_list, "SUFFIX"), tempfile(callout_list, "TEMPFILE");
	const char* tmpfile = genie_tmpfile(prefix, suffix, directory);
	if (tmpfile != NULL)
  	{
	    tempfile = tmpfile;
	}
	else
	{
	    *g_err << "TMPFILE: Unable to generate unique temporary filename" << endl;
	    result = nilOOP;
	}
	break;
    }

    case GetCPUTimes:
    {
#if 0
	GenieWorkspace structtms(callout_list, "STRUCTTMS");
	struct tms struct_tms;
	times(&struct_tms);
	structtms("user", struct_tms.tms_utime);
	structtms("system", struct_tms.tms_stime);
	structtms("user_child", struct_tms.tms_cutime);
	structtms("system_child", struct_tms.tms_cstime);
	struct rusage usage;
	if (getrusage(RUSAGE_SELF, &usage) == 0)
	{
	    structtms("maxrss", usage.ru_maxrss);
	    structtms("pagefaults", usage.ru_majflt);
	    structtms("pagereclaims", usage.ru_majflt);
	    structtms("nswap", usage.ru_nswap);
	}
	else
	{
	    perror("getrusage");
	    result = nilOOP;
	}
#endif
	break;
    }

    case RaiseSignal:
    {
	GenieInteger sig(callout_list, "SIG");
	if (raise(sig) != 0)
	{
	    *g_err << "RAISE: error raising signal " << (g_integer)sig << endl;
	    result = nilOOP;
	}
	break;
    }

    default:
	PERROR("unknown system callout");
	result = nilOOP;
        break;
    }
 
    ENDSAVEBLOCK

    return result;
}

