/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "geniedefs.h"
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/types.h>
#include <new.h>
#include "genievars.h"
#include "smalltalk_object.h"
#ifdef __VMS
#include <unixlib.h>
#endif /* __VMS */

#include "smalltalk.h"	    // smalltalk
#include "parsedefs.h"	    // parser
#include "parser.h"
#include "g_tktcl.h"	    // Tcl/Tk
#include "ghistory.h"	    // Access to genie objects
#include "fileset.h"
#include "genie.h"
#include "genielog.h"

#if HAVE_LIBTK

#include "g_tktcl.h"
#include "module_inquire.h"
extern "C" {
#	include <readline.h>
#	include <history.h>
}

/*
 * Module:		Tk/Tcl <-> Genie interface
 * Purpose:		Define commands to allow passing of data and commands from Genie <-> TCL
 * Author:		Freddie Akeroyd, ISIS
 * Date:		September 1996
 */

static const char* rcsid = "@(#)$Id: g_tktcl.cxx,v 1.68 1998/10/15 08:03:39 cmm Exp $";
USE_RCSID(rcsid);

GENIEEXPORT void GenieTclPanic(char* format, ... )
{
	static char buffer[1024];
	va_list ap;
	va_start(ap, format);
    vsprintf(buffer, format, ap);
	va_end(ap);
#ifdef _WIN32
	MessageBeep(MB_ICONEXCLAMATION);
	MessageBox(NULL, buffer, "Fatal Error in TkGenie",
		MB_ICONSTOP | MB_OK | MB_TASKMODAL | MB_SETFOREGROUND);
#else
	gerr << buffer << endl;
#endif /* _WIN32 */
}

static void tcl_trace(ClientData clientData, Tcl_Interp* interp, int level, 
    char* command, Tcl_CmdProc* cmdProc, ClientData cmdClientData, 
    int argc, char* argv[])
{
    int i;
    gprintif("*** Trace on %s at level %d ***\n", command, level);
    for(i=0; argv[i] != 0; i++)
		gprintif("%s ", argv[i]);
    gprintif("\n");
}

static int return_list(Tcl_Interp* interp)
{
    CGenie* pGenie = (CGenie*)g_pGenie; // Get the global pointer to the running genie
    assert(pGenie!=NULL);

    char *tmp = (char*)pGenie->Streams()->Out();
    int len = strlen(tmp);
    if (len == 0)
	return TCL_OK;
    if (tmp[len-1] == '\2')
	tmp[len-1] = '\0';	// remove trailing '\2' if present
    char* bit = strtok(tmp, "\2");
    if (bit == 0)
	return TCL_OK;
    Tcl_AppendElement(interp, bit);
    while ( (bit = strtok(0, "\2")) != 0 )
    {
	    Tcl_AppendElement(interp, bit);
    }
    return TCL_OK;
}

// sends a tcpip command to a particular port and server
static int tcpip(ClientData clientData, Tcl_Interp *interp, int argc, char* argv[])
{
    CGenie* pGenie = (CGenie*)g_pGenie; // Get the pointer to the running genie
    assert(pGenie!=NULL);
    int rep_len = 50000;
    char* reply = (char*)malloc(rep_len+1);
    if (reply == 0)
    {
	Tcl_AppendResult(interp, "g_tcpip: out of memory error", 0);
	return TCL_ERROR;
    }
    if (argc != 4)
    {
	Tcl_AppendResult(interp, "Usage: g_tcpip server port message", 0);
	return TCL_ERROR;
    }
    if (tcpip_get(argv[1], atoi(argv[2]), argv[3], strlen(argv[3]), reply, &rep_len) == 0)
    {
        reply[rep_len] = '\0';
        Tcl_ResetResult(interp);
        Tcl_SetResult(interp, reply, TCL_DYNAMIC);	// free() done by TCL
        return TCL_OK;
    }
    else
    {
	Tcl_AppendResult(interp, (char*)pGenie->Streams()->Err());
	return TCL_ERROR;
    }
}

// converts a single post field string into the correct format
static int httpPost(ClientData clientData, Tcl_Interp *interp, int argc, char* argv[])
{
    if (argc != 2)
    {
	    Tcl_AppendResult(interp, "Usage: g_httpPost string", 0);
	    return TCL_ERROR;
    }
    int c;
    char* result = (char*)malloc(strlen(argv[1])*3+1);
    char* input = argv[1];
    if (result == 0)
    {
	    Tcl_AppendResult(interp, "g_httpPost: out of memory error", 0);
	    return TCL_ERROR;
    }
    result[0] = '\0';
    for (c=0;(unsigned)c<strlen(input);c++) {
        if ((isalnum(input[c])) && ((unsigned char)input[c] < 127)) {
            strncat(result,input+c,1);
        } else {
            if (input[c] == ' ') {
                strncat(result, "+\0", 1);
            } else {
                sprintf(result, "%s%%%02X", result, (unsigned char)input[c]); 
            }
        }
    }
    Tcl_ResetResult(interp);
    Tcl_SetResult(interp, result, TCL_DYNAMIC);	// free() done by TCL
    return TCL_OK;
}

static int trace(ClientData clientData, Tcl_Interp *interp, int argc, char* argv[])
{
    static Tcl_Trace t = (Tcl_Trace)0;
    if (clientData == (ClientData)1) {
        if (t == (Tcl_Trace)0) {
            t = Tcl_CreateTrace(interp, 10, tcl_trace, 0);
        } else {
            Tcl_AppendResult(interp, "g_traceOn: trace already on", 0);
            return TCL_ERROR;
        }
    } else {
        if (t != (Tcl_Trace)0) {            
            Tcl_DeleteTrace(interp, t);
            t = (Tcl_Trace)0;
        } else {
            Tcl_AppendResult(interp, "g_traceOff: no trace active", 0);
            return TCL_ERROR;
        }
    }
    return TCL_OK;
}

// returns the previous gcl command typed at the command line
static int previous(ClientData clientData, Tcl_Interp *interp, int argc, char* argv[])
{
    CHistory* his = ((CGenie*)g_pGenie)->GetHistory();
    Tcl_ResetResult(interp);
    Tcl_SetResult(interp, (char*)his->GetPreviousCommand(), TCL_STATIC);
    return TCL_OK;
}

// returns the next gcl command typed at the command line
static int next(ClientData clientData, Tcl_Interp *interp, int argc, char* argv[])
{
    CHistory* his = ((CGenie*)g_pGenie)->GetHistory();
    Tcl_ResetResult(interp);
    Tcl_SetResult(interp, (char*)his->GetNextCommand(), TCL_STATIC);
    return TCL_OK;
}

// searches the command history for a string, and returns the full command
static int search(ClientData clientData, Tcl_Interp *interp, int argc, char* argv[])
{
    CHistory* his = ((CGenie*)g_pGenie)->GetHistory();
    if (!strcmp(argv[2],"new")) {
	    his->Search(NULL);	// reset search position index
    }    
    Tcl_ResetResult(interp);
    Tcl_SetResult(interp, (char*)his->Search(argv[1]), TCL_STATIC);
    return TCL_OK;
}

// pack strings for storing in files by swaping \n with \06
static int pack(ClientData clientData, Tcl_Interp *interp, int argc, char* argv[])
{
    int length,c;
    char oldChar = '\n';
    char newChar = '\6';
    printf("initial string = %s\n",argv[1]);
    if (clientData == (ClientData)1) {
	oldChar = '\6';
	newChar = '\n';
    }	 
    char* result;
    length = strlen(argv[1]);
    result = (char*)malloc(length + 1);
    for(c = 0; c <= length; c++)
    {
	if (argv[1][c] == oldChar) {
	    result[c] = newChar;
	    printf("translating\n");
	} else {
	    result[c] = argv[1][c];
	}
    }
    printf("result = %s\n",result);
    Tcl_ResetResult(interp);
    Tcl_SetResult(interp, result, TCL_DYNAMIC);    
    return TCL_OK;
}

// Execute a command line in Smalltalk
/* static */ int 
smalltalk(ClientData clientData, Tcl_Interp *interp, int argc, char* argv[])
{
    CGenie* pGenie = (CGenie*)g_pGenie; // Get the global pointer to the running genie
    assert(pGenie!=NULL);

    int result;
    if (argc < 2)
    {
	interp->result = "Usage: smalltalk command_string ?options?";
	return TCL_ERROR; 
    }
// command string to run is in argv[1]
    if ( pGenie->EvaluateSmalltalkCommand(argv[1]) != 0)
	result = TCL_ERROR;
    else
	result = TCL_OK;

    Tcl_ResetResult(interp);
    if (argc == 3 && !strcmp(argv[2], "-out"))
    {
	Tcl_SetResult(interp, (char*)pGenie->Streams()->Out(), TCL_VOLATILE);
// bin the rest
	pGenie->Streams()->ClearAll();
    }
// gout contains a set of list elements separated by "\2"
    else if (argc == 3 && !strcmp(argv[2], "-list"))
    {
	result = return_list(interp);
// bin the rest
	pGenie->Streams()->ClearAll();
    }
    else
    {
	Tcl_AppendElement(interp, (char*)pGenie->Streams()->Out());
	Tcl_AppendElement(interp, (char*)pGenie->Streams()->Inf());
	Tcl_AppendElement(interp, (char*)pGenie->Streams()->Err());
	Tcl_AppendElement(interp, (char*)pGenie->Streams()->Dbg());
    }
    return result;
}

static int 
g_setinstvars(ClientData clientData, Tcl_Interp *interp, int argc, char* argv[])
{
//    int result;
    if (argc != 3)
    {
	interp->result = "Usage: g_setinstvars oop_tag instvars_list";
	return TCL_ERROR; 
    }
    char* oop_tag = argv[1];
    char* instvars_list = argv[2];
    char** argv_list =0;
    int argc_list;
    if (Tcl_SplitList(interp, instvars_list, &argc_list, &argv_list) != TCL_OK)
    {
	Tcl_AppendResult(interp, "g_setinstvars: error splitting list", 0);
	return TCL_ERROR;
    }
    OOP oop = (OOP)(unsigned long)atol(oop_tag);
    SmalltalkObject obj(oop);
    for(int i=0; i<argc_list; i+=2)
    {
	obj.fromString(argv_list[i], argv_list[i+1]);
    }
    if (argv_list != 0)
	free((char*)argv_list);
    return TCL_OK;
}

// Execute a command line in GCL
static int 
gcl(ClientData clientData, Tcl_Interp *interp, int argc, char* argv[])
{
    CGenie* pGenie = (CGenie*)g_pGenie; // Get the global pointer to the running genie
    assert(pGenie!=NULL);

    int result, error;
    if (argc < 2)
    {
	interp->result = "Usage: gcl command_string ?options?";
	return TCL_ERROR; 
    }
// command string to run is in argv[1]
    if (argc == 3 && !strcmp(argv[2], "-history")) {
        error = pGenie->EvaluateGCLCommand(argv[1],1);
    } else {
        error = pGenie->EvaluateGCLCommand(argv[1],0);
    }
    if (error != 0) {
	    result = TCL_ERROR;
    } else {
	    result = TCL_OK;
    }
    Tcl_ResetResult(interp);
    if (argc == 3 && !strcmp(argv[2], "-out"))
    {
	Tcl_SetResult(interp, (char*)pGenie->Streams()->Out(), TCL_VOLATILE);
// bin the rest
	pGenie->Streams()->ClearAll();
    }
    else
    {
	Tcl_AppendElement(interp, (char*)pGenie->Streams()->Out());
	Tcl_AppendElement(interp, (char*)pGenie->Streams()->Inf());
	Tcl_AppendElement(interp, (char*)pGenie->Streams()->Err());
	Tcl_AppendElement(interp, (char*)pGenie->Streams()->Dbg());
    }
    return result;
}

// Global DATA
PGTkItem* PGTkItemList[MAX_PGTKITEMS];
Tcl_Interp* PGTkDefaultInterp = 0;

// create a PGPLOT window (basically a canvas)
// Only ever gets called in the creation command
//  pgwindow .pg1
// As the created object is a canvas, so all other processing done
// via canvas procedure
//
// Window names should be of the form .where.ever.pgwin1 etc., with
// pgwin1 accessed from PGPLOT by the device name /TK1
//

static Tcl_CmdDeleteProc *canvas_delete_proc = 0;

static void
pgwindow_delete_proc(ClientData clientData)
{
    int i;
    for(i=0; i<MAX_PGTKITEMS; i++)
    {
	if (PGTkItemList[i] != 0 && clientData == PGTkItemList[i]->cmdinfo.deleteData)
	    break;
    }
    if (i != MAX_PGTKITEMS)
    {
	PGTkItemList[i] = 0;
//	gprintif("Deleting window %d\n", i);
    }
    if (canvas_delete_proc)
	(*canvas_delete_proc)(clientData);
}

static char* 
pgwindow_trace(ClientData clientData, Tcl_Interp* interp, char* name1, char* name2, int flags)
{
// name2 will be NULL as we are not tracing an array
    PGTkItem* item = (PGTkItem*)clientData;
    if (flags & TCL_TRACE_WRITES)
	gprintif("WRITE to %s\n", name1);
    if (flags & TCL_TRACE_READS)
	gprintif("READ of %s\n", name1);
    if (flags & TCL_TRACE_UNSETS)
	gprintif("UNSET of %s\n", name1);
    return NULL;
}

static int 
pgwindow(ClientData clientData, Tcl_Interp *interp, int argc, char* argv[])
{
    static Tcl_CmdInfo* canvas_cmdinfo = 0;
    int result;
    Tcl_ResetResult(interp);
// Do we know how to create a new canvas?
    if (canvas_cmdinfo == 0)
    {
	for(int i=0; i<MAX_PGTKITEMS; i++)
	    PGTkItemList[i] = 0;
	canvas_cmdinfo = new Tcl_CmdInfo;
	if (!Tcl_GetCommandInfo(interp, "canvas", canvas_cmdinfo))
	{
	    Tcl_AppendResult(interp, "pgwindow: \"canvas\" command not found", 0);
	    return TCL_ERROR;
	}
	PGTkDefaultInterp = interp;
    } 
// overwrite the command name of "pgwindow" with "canvas"
//   argv[0] = Tk_GetUid("canvas");
// Create a new canvas
    result = (*(canvas_cmdinfo->proc))(canvas_cmdinfo->clientData, interp, argc, argv);
    if (result != TCL_OK)
	return result;
// Execute events generated by canvas command
    while(Tk_DoOneEvent(TK_ALL_EVENTS|TK_DONT_WAIT))
	;
// Parse window name - the last character should be a number
    char* pgwin_name = strdup(argv[1]);
    int pgwin_number = pgwin_name[strlen(pgwin_name)-1] - '1'; 
// Find an open device slot in the table
    if (pgwin_number < 0 || pgwin_number >= MAX_PGTKITEMS)
    {
	Tcl_AppendResult(interp, "pgwindow: invalid PG window number", 0);
	return TCL_ERROR;
    }
    if (PGTkItemList[pgwin_number] != 0)
    {
	Tcl_AppendResult(interp, "pgwindow: PG window already in use", 0);
	return TCL_ERROR;
    }
// Now store the address of the new Tk widget command
    PGTkItem* item = new PGTkItem;
    item->interp = interp;
    item->opened = 0;		/* No PGOPEN yet */
    item->cmd = Tk_GetUid(pgwin_name);
//    if (Tcl_TraceVar(interp, item->cmd, TCL_GLOBAL_ONLY|TCL_TRACE_READS|TCL_TRACE_WRITES|TCL_TRACE_UNSETS, pgwindow_trace, item) != TCL_OK)
//	return TCL_ERROR;
    item->tkwin = Tk_NameToWindow(interp, item->cmd, Tk_MainWindow(interp));
    if (item->tkwin == 0)
    {
	Tcl_AppendResult(interp, "pgwindow: cannot locate window", 0);
	return TCL_ERROR;
    }
    if (!Tcl_GetCommandInfo(interp, Tk_PathName(item->tkwin), &(item->cmdinfo)))
    {
	Tcl_AppendResult(interp, "pgwindow: cannot get window command info", 0);
	return TCL_ERROR;
    }
    canvas_delete_proc = item->cmdinfo.deleteProc;
    item->cmdinfo.deleteProc = pgwindow_delete_proc;
    if (!Tcl_SetCommandInfo(interp, Tk_PathName(item->tkwin), &(item->cmdinfo)))
    {
	Tcl_AppendResult(interp, "pgwindow: cannot set window command info", 0);
	return TCL_ERROR;
    }
    PGTkItemList[pgwin_number] = item; 
    return TCL_OK;
} 

// Files searched for in directory given by GENIE_TCL_DIR environment variable
// Last entry must be NULL
static const char* tcl_init_files[] = { "objecttags.tcl", "globals.tcl", "printOutput.tcl", 
	"alter.tcl", "html_library.tcl", "help.tcl", "module.tcl", 
	"commands.tcl", 0, "interface.tcl",
// rest commented out for moment
	"display.tcl", "graph.tcl", "st1.tcl",
	"miscDis.tcl", "oneDgraph.tcl", "oneDHistogram.tcl",
	"oneDErrors.tcl", "oneDMarkers.tcl", "axes.tcl", "label.tcl", "title.tcl",
	"text.tcl", "polygon.tcl", "polyline.tcl", 0 };

/*
 * Application init for TK/TCL, called by Tk_Main
 */

#ifdef _WIN32
extern "C" GENIEEXPORT int TkConsoleInit(Tcl_Interp*);
#endif /* _WIN32 */


// Not normally exported (original source in generic/tclEnv.c) but copied here to
// access our copy of environ.  Assumes that env is already set up and watched by
// the tcl version of this routine and simply re-stores the environment variables
// to it.
static void
TclSetupEnv(Tcl_Interp *interp)
{
    char *p, *p2;
    Tcl_DString ds;
    int i, sz;

    /*
     * Next, initialize the DString we are going to use for copying
     * the names of the environment variables.
     */
    Tcl_DStringInit(&ds);
    
    /*
     * Store the environment variable values into the interpreter's
     * "env" array.
     */
// We will only have added of modified, so we don't need to delete
// We also need to keep any extra ones set up that are not in environ
// e.g. DISPLAY on VMS
//    (void) Tcl_UnsetVar2(interp, "env", (char *) NULL, TCL_GLOBAL_ONLY);
    for (i = 0; ; i++) {
	p = environ[i];
	if (p == NULL) {
	    break;
	}
	for (p2 = p; *p2 != '='; p2++) {
	    if (*p2 == 0) {
		/*
		 * This condition doesn't seem like it should ever happen,
		 * but it does seem to happen occasionally under some
		 * versions of Solaris; ignore the entry.
		 */
		goto nextEntry;
	    }
	}
        sz = p2 - p;
        Tcl_DStringSetLength(&ds, 0);
        Tcl_DStringAppend(&ds, p, sz);
	(void) Tcl_SetVar2(interp, "env", Tcl_DStringValue(&ds),
                p2+1, TCL_GLOBAL_ONLY);
	nextEntry:
	continue;
    }

    /*
     * Finally clean up the DString.
     */
    Tcl_DStringFree(&ds);
}


GENIEEXPORT int Genie_Tcl_AppInit(Tcl_Interp* interp)
{
    // Re-read environment variables here
    TclSetupEnv(interp);

    if (Tcl_Init(interp) == TCL_ERROR)
    {
	GenieTclPanic("Error returned from Tcl_Init");
	return TCL_ERROR;
    }
    if (Tk_Init(interp) == TCL_ERROR)
    {
	GenieTclPanic("Error returned from Tk_Init");
	return TCL_ERROR;
    } 

#ifdef _WIN32
//    Tcl_StaticPackage(interp, "Tk", Tk_Init, Tk_SafeInit);
//	if (TkConsoleInit(interp) == TCL_ERROR)
//	{
//		GenieTclPanic("Error initialising WIN32 console");
//		return TCL_ERROR;
//	}
#endif /* _WIN32 */

#ifdef HAVE_LIBTIX
    if (Tix_Init(interp) == TCL_ERROR)
    {
	GenieTclPanic("Error returned from Tix_Init");
	return TCL_ERROR;
    } 
#else
    ginf << "Tix_Init not called - Tix is not installed" << endl;
#endif /* HAVE_LIBTIX */
/*
 * Now add the new commands
 */
    if (!Tcl_CreateCommand(interp, "gcl", gcl, (ClientData)0, (Tcl_CmdDeleteProc*)0))
    {
	GenieTclPanic("Error defining \"gcl\" command");
	return TCL_ERROR;
    }
    if (!Tcl_CreateCommand(interp, "g_packString", pack, (ClientData)0, (Tcl_CmdDeleteProc*)0))
    {
	GenieTclPanic("Error defining \"g_packString\" command");
	return TCL_ERROR;
    }
    if (!Tcl_CreateCommand(interp, "g_unpackString", pack, (ClientData)1, (Tcl_CmdDeleteProc*)0))
    {
	GenieTclPanic("Error defining \"g_unpackString\" command");
	return TCL_ERROR;
    }
    if (!Tcl_CreateCommand(interp, "g_previousCommand", previous, (ClientData)0, (Tcl_CmdDeleteProc*)0))
    {
	GenieTclPanic("Error defining \"g_previousCommand\" command");
	return TCL_ERROR;
	} 
    if (!Tcl_CreateCommand(interp, "g_nextCommand", next, (ClientData)0, (Tcl_CmdDeleteProc*)0))
    {
	GenieTclPanic("Error defining \"g_nextCommand\" command");
	return TCL_ERROR;
	}
    if (!Tcl_CreateCommand(interp, "g_searchHistory", search, (ClientData)0, (Tcl_CmdDeleteProc*)0))
    {
	GenieTclPanic("Error defining \"g_searchHistory\" command");
	return TCL_ERROR;
	}
    if (!Tcl_CreateCommand(interp, "g_tcpip", tcpip, (ClientData)0, (Tcl_CmdDeleteProc*)0))
    {
	GenieTclPanic("Error defining \"g_tcpip\" command");
	return TCL_ERROR;
	}
    if (!Tcl_CreateCommand(interp, "g_httpPost", httpPost, (ClientData)0, (Tcl_CmdDeleteProc*)0))
    {
	GenieTclPanic("Error defining \"g_httpPost\" command");
	return TCL_ERROR;
	}
    if (!Tcl_CreateCommand(interp, "g_traceOn", trace, (ClientData)1, (Tcl_CmdDeleteProc*)0))
    {
	GenieTclPanic("Error defining \"g_traceOn\" command");
	return TCL_ERROR;
	}
    if (!Tcl_CreateCommand(interp, "g_traceOff", trace, (ClientData)0, (Tcl_CmdDeleteProc*)0))
    {
	GenieTclPanic("Error defining \"g_traceOff\" command");
	return TCL_ERROR;
	}								       
    if (!Tcl_CreateCommand(interp, "smalltalk", smalltalk, (ClientData)0, (Tcl_CmdDeleteProc*)0))
    {
	GenieTclPanic("Error defining \"smalltalk\" command");
	return TCL_ERROR;
    }
    if (!Tcl_CreateCommand(interp, "pgwindow", pgwindow, (ClientData)0, (Tcl_CmdDeleteProc*)0))
    {
	GenieTclPanic("Error defining \"pgwindow\" command");
	return TCL_ERROR;
    }
    if (!Tcl_CreateCommand(interp, "g_setinstvars", g_setinstvars, (ClientData)0, (Tcl_CmdDeleteProc*)0))
    {
	GenieTclPanic("Error defining \"g_setinstvars\" command");
	return TCL_ERROR;
    }
/*
 * now load any stuff we need
 */
    const char* genie_tcl_dir = (getenv("GENIE_TCL_DIR") != 0 ? strdup(getenv("GENIE_TCL_DIR")) : CURRENT_DIRECTORY);
    char loadpath[256];
    for(int i = 0; tcl_init_files[i] != 0; i++)
    {
	sprintf(loadpath, "%s%s%s", genie_tcl_dir, DIRECTORY_SEPARATOR, tcl_init_files[i]);
	if (Tcl_EvalFile(interp, loadpath) != TCL_OK)
	{
	    GenieTclPanic("Error executing TCL startup commands from \"%s\"", loadpath);
	    return TCL_ERROR;
	}
    }
    if ( /*((CGenie*)g_pGenie)->TestOption(CGenie::GENDEBUG) &&*/ getenv("GENIE_TCL_DEBUG") != 0 ) {
    	Tcl_CreateTrace(interp, 10, tcl_trace, 0);
   	}
    return TCL_OK;
}

#endif /* HAVE_LIBTK */

void GENIEMECHANISM g3_module_tk_inquire_begin(const STRING& title, int* dialog)
{
    DialogueBox* d = new (dialog) DialogueBox(title.spot, PGTkDefaultInterp);
//    memcpy(dialog, d, sizeof(DialogueBox));
    return;
}

void GENIEMECHANISM g3_module_tk_inquire_int(int* dialog, const STRING& label, fort_int& val)
{
    ((DialogueBox*)dialog)->addElement(new IntegerDialogueElement(label.spot, &val));	 
    return;
}

void GENIEMECHANISM g3_module_tk_inquire_real(int* dialog, const STRING& label, fort_real& val)
{
    ((DialogueBox*)dialog)->addElement(new RealDialogueElement(label.spot, &val));    
    return;
}

void GENIEMECHANISM g3_module_tk_inquire_double(int* dialog, const STRING& label, fort_double& val)
{
    ((DialogueBox*)dialog)->addElement(new DoubleDialogueElement(label.spot, &val));	
    return;
}

void GENIEMECHANISM g3_module_tk_inquire_string(int* dialog, const STRING& label, char* val, fort_int len)
{
    ((DialogueBox*)dialog)->addElement(new StringDialogueElement(label.spot, val, len-1));    
    return;
}

void GENIEMECHANISM g3_module_tk_inquire_end(int* dialog)
{
    ((DialogueBox*)dialog)->display();
//    delete dialog;
    return;
}
