/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

/*
 *      Module:         g_transform
 *      Author:         Freddie Akeroyd, ISIS
 *      Purpose:        Callout to perform operations on internal data
 *
 */
#include "geniedefs.h"

#ifdef HAVE_UNISTD_H
#	include <unistd.h>
#endif

#include <stdio.h>
#include <math.h>
#include "genie_types.h"
#include "transform_utils.h"
#include "units_p.h"
#include "genievars.h"
#include "g_transform.h"
#include "Genie_functions.h"
#ifdef _WIN32
#	include <io.h>
#	include "stream_nt.h"
#else
#	include <stream.h>
#endif /* _WIN32 */
#include "callout_list.h"
#include "local_units.h"
#include "fortran_types.h"
#include "isis_raw_file.h"

#include <string.h>
#include <gnustring.h>

static int transformIntegrateSpectra(GenieRealArray& xmin, 
	GenieRealArray& xmax, GenieWorkspace& wksp, GenieWorkspace& res);
static int transformIntegrateFile(GenieRealArray& xmin, GenieRealArray& xmax, 
	GenieString& file, GenieIntegerArray& iarray, GenieWorkspace& res);

static const char* rcsid = "@(#)$Id: g_transform.cxx,v 1.38 1998/02/26 01:29:47 faa Exp $";
USE_RCSID(rcsid);

OOP g_transform( transform_function_code function, OOP parameters ) 
{
    static char buffer[LONGCHARBUFF];

    GenieCalloutList gcl(parameters);
    BEGINSAVEBLOCK

    switch(function)
    {

    	case TransformTest:
	{
            cerr << "g_transform(test function called)\n";
//	    printParams(parameters);
 	}
	break;

    	case TransformBinGroup:
	{
	    GenieRealArray xarray(gcl, "XARRAY"), yarray(gcl, "YARRAY"), earray(gcl, "EARRAY");
	    GenieInteger nbin(gcl, "NBIN");
	    g_integer ntc, nspec, i;

	    if (yarray.dimsSize() == 1)
	    {
		ntc = yarray.dims(1);
		nspec = 1;
	    }
	    else if (yarray.dimsSize() == 2)
	    {
		ntc = yarray.dims(2);
		nspec = yarray.dims(1);
	    }
	    else
	    {
		gerr << "BINGROUP: Invalid array dimensionality" << endl;
		ntc = 0;
		nspec = 0;
	    }
	    g_integer ntcnew = (ntc - 1) / nbin + 1;
	    g_real* xnew = new g_real[ntcnew+1];
	    g_real* ynew = new g_real[ntcnew*nspec];
	    g_real* enew;
	    if (earray.defined())
	    {
		enew = new g_real[ntcnew*nspec];
	    }
	    else
	    {
	 	enew = 0;
	    }
	    i = group_bins((g_real*)xarray, (g_real*)yarray, (g_real*)earray, ntc, nspec, xnew, ynew, enew, (g_integer)nbin);
	    if (i != ntcnew)
	    {
		gerr << "Error returned from group_bins of " << nspec << " spectra" << endl;
		gerr << "Calculated NTC (" << i << ") != Acutal (" << ntcnew << ")" << endl;
	    }
	    xarray(xnew, ntcnew+1);
	    if (nspec == 1)
	    {
		yarray(ynew, ntcnew);
		if (enew) { earray(enew, ntcnew); }
	    }
	    else
	    {
		yarray(ynew, nspec, ntcnew);
		if (enew) { earray(enew, nspec, ntcnew); }
	    }
	    delete []xnew;
	    delete []ynew;
	    delete []enew;
	}
	break;

    	case TransformBinBound:
	{
	    GenieInteger itype(gcl, "ITYPE"), isteps(gcl, "ISTEPS"), ibounds(gcl, "IBOUNDS");
	    GenieRealArray xbounds(gcl, "XBOUNDS"), xsteps(gcl, "XSTEPS");
	    GenieRealArray xarray(gcl, "XARRAY"), yarray(gcl, "YARRAY"), earray(gcl, "EARRAY");
	    g_integer ntc = yarray.length();
	    g_real ymax;
	    g_real* xnew = NULL;
	    int i = newAxis((g_real*)xbounds, (g_integer)ibounds, (g_real*)xsteps, (g_integer)isteps, xnew, 0, (g_integer)itype);
	    if (i == -1)
	    {
	 	error_add("REBIN", "NewAxis returned error value", "");
		break;
	    }
	    g_integer ntcnew = i - 1;
	    xnew = new g_real[ntcnew+1];
	    newAxis((g_real*)xbounds, (g_integer)ibounds, (g_real*)xsteps, (g_integer)isteps, xnew, 1, (g_integer)itype);
	    g_real* ynew = new g_real[ntcnew];
	    g_real* enew = new g_real[ntcnew];
	    rebin((g_real*)xarray, (g_real*)yarray, (g_real*)earray, ntc, xnew, ynew, enew, ntcnew, ymax);
	    xarray(xnew, ntcnew+1);
	    yarray(ynew, ntcnew);
	    earray(enew, ntcnew);
	    delete[] xnew;
	    delete[] ynew;
	    delete[] enew;
   	}
	break;

    	case TransformBinRange:
	{
	    GenieReal xmin(gcl, "XMIN"), xmax(gcl, "XMAX");
	    GenieRealArray xarray(gcl, "XARRAY"), yarray(gcl, "YARRAY"), earray(gcl, "EARRAY");
	    g_real ymax;
	    g_integer ntc = yarray.length();
 	    g_real* xnew = new g_real[ntc+5];
 	    g_real* ynew = new g_real[ntc+5];
 	    g_real* enew = new g_real[ntc+5];
	    rebinRange((g_real*)xarray, (g_real*)yarray, (g_real*)earray, (g_real*)xnew,
		    (g_real*)ynew, (g_real*)enew, ntc, (g_real)xmin, (g_real)xmax, ymax);
	    xarray(xnew, ntc+1);
	    yarray(ynew, ntc);
	    earray(enew, ntc);
	    delete[] xnew;
	    delete[] ynew;
	    delete[] enew;
	}
	break;

    	case TransformRebinAs:
	{
	     g_real ymax;
	     GenieRealArray xnew(gcl, "XNEW");
  	     GenieRealArray xold(gcl, "XOLD"), yold(gcl, "YOLD"), eold(gcl, "EOLD"); 
	     g_integer ntc = yold.length();
	     g_integer ntcnew = xnew.length() - 1;
 	     g_real* ynew = new g_real[ntcnew+5];
 	     g_real* enew = new g_real[ntcnew+5];
	     rebin((g_real*)xold, (g_real*)yold, (g_real*)eold, ntc, (g_real*)xnew, ynew, enew, ntcnew, ymax);
	     yold(ynew, ntcnew);
	     eold(enew, ntcnew);
	     delete[] ynew;
	     delete[] enew;
	}
	break;

	case TransformIntegrateSpectra:
	{
	    GenieRealArray xmin(gcl, "XMIN"), xmax(gcl, "XMAX");
	    GenieWorkspace res(gcl, "RES"), wksp(gcl, "WKSP");
	    transformIntegrateSpectra(xmin, xmax, wksp, res);
   	}
	break;

	case TransformIntegrateFile:
	{
	    GenieString file(gcl, "FILE");
	    GenieRealArray xmin(gcl, "XMIN"), xmax(gcl, "XMAX");
	    GenieWorkspace res(gcl, "RES");
	    GenieIntegerArray iarray(gcl, "IARRAY");
	    transformIntegrateFile(xmin, xmax, file, iarray, res);
	}
	break;

        case TransformUnits:
	{
	    String message;
	    GenieInteger iunit(gcl, "IUNIT");
	    GenieReal xmin(gcl, "XMIN"), xmax(gcl, "XMAX");
	    GenieWorkspace worksp(gcl, "T_WORKSP");
	    double ymax;
	    g_integer error_code, histogram_mode;
	    GenieRealArray xarray, yarray, earray;
	    xarray.become(worksp("X"));
	    yarray.become(worksp("Y"));
	    earray.become(worksp("E"));
	    histogram_mode = (xarray.length() != yarray.length());
	    g_real* xval = (g_real*)xarray;
	    g_real* yval = (g_real*)yarray;
	    g_real* eval = (g_real*)earray;
	    g_integer ntc = yarray.length();
	    g_real delta = (GenieReal)worksp("DELTA");
	    g_real l1 = (GenieReal)worksp("L1");
	    g_real l2 = (GenieReal)worksp("L2");
	    g_real TwoTheta = (GenieReal)worksp("TWOTHETA");
	    g_integer emode = (GenieInteger)worksp("EMODE");
	    g_real efixed = (GenieReal)worksp("EFIXED");
//	    g_integer xunit_in = worksp("X_UNIT_CODE");
//	    g_integer yunit_in = worksp("Y_UNIT_CODE");
	    String xunit_in = (GenieString)worksp("XLABEL");
	    String yunit_in = (GenieString)worksp("YLABEL");
	    const char *xunit_out, *yunit_out;
/* this may alter ntc */
 	    error_code = convert_units(xval, yval, eval, &ntc,
			   xunit_in, &xunit_out, yunit_in, &yunit_out,
			   iunit, l1, l2, emode, efixed,
			   TwoTheta, &ymax, xmin, xmax, delta, histogram_mode);
            if (error_code == 0)	// If non zero, we will have printed an error message
	    {
		worksp("X", GenieRealArray(xval, ntc+histogram_mode));
		worksp("Y", GenieRealArray(yval, ntc));
		worksp("E", GenieRealArray(eval, ntc));
	    	worksp("XLABEL", xunit_out);
	    	worksp("YLABEL", yunit_out);
	    	worksp("NTC", ntc);
	    }
	}
	break;

    	case TransformArrayBracket:
    	{
	    GenieReal xval(gcl, "XVAL");
	    GenieRealArray xarray(gcl, "XARRAY");
	    GenieInteger imin(gcl, "IMIN");
	    int i;
	    g_integer imax = xarray.length();
	    if ((i = bracket((g_real)xval, (g_real*)xarray, imax)) != -1)
	    {
/* +1 as c arrays start at 0, genie arrays at 1 */
	        imin = i +1;
	    }
	    else
	    {
	        imin = -1;
	        error_add("g_transform", "bracket", "");
	    }
	}
	break;

	case CreateNewUnit:
	{
	    GenieString name(gcl, "NAME"), label(gcl, "LABEL"), components(gcl, "COMPONENTS");
	    GenieReal scale(gcl, "SCALE");
	    GenieWorkspace encoded_unit(gcl, "ENCODED_UNIT");
	    encoded_unit = create_unit(name, label, components, scale);
	}
	break;

	case GetUnitDims:
	{
	    GenieWorkspace encoded_unit(gcl, "ENCODED_UNIT"), dims(gcl, "DIMS");
	    dims = get_unit_dimensions(encoded_unit);
	}
	break;

	case GetUnitProperties:
	{
	    GenieWorkspace encoded_unit(gcl, "ENCODED_UNIT");
	    GenieString properties(gcl, "PROPERTIES");
	    properties = get_unit_properties(encoded_unit);
	}
	break;

	case GetKnownUnits:
	{
	    GenieString list(gcl, "LIST");
	    list = get_known_units();
	}
	break;

	case MultiplyTwoUnits:
	{
	    GenieWorkspace unit1(gcl, "UNIT1"), unit2(gcl, "UNIT2"), res(gcl, "RES");	// encoded unit strings	
	    res = unit_multiply(unit1, unit2);
	}
	break;

	case ConvertUnit:	// Return multiplying factor to convert UNIT2 into UNIT1
	{
	    GenieWorkspace unit1(gcl, "UNIT1"), unit2(gcl, "UNIT2");	// encoded units
	    GenieReal conv(gcl, "CONV");
	    double conversion_factor = unit_convert(unit1, unit2);
	    if (conversion_factor != 0.0)
	    {
	        conv = conversion_factor;
	    }
	    else
	    {
	 	conv = UndefinedReal;
	    }
	}
	break;

    	default:
	{
            gerr << "g_transform(unknown callout number %d)\n";
	}
	break;
    }

    ENDSAVEBLOCK

    return parameters;
}

// Return 0 on success, 1 on error
static int transformIntegrateSpectra(GenieRealArray& xmin, 
	GenieRealArray& xmax, GenieWorkspace& wksp, GenieWorkspace& res)
{
	    GenieRealArray xarray, yarray, earray;
	    xarray.become(wksp("X"));
	    yarray.become(wksp("Y"));
	    earray.become(wksp("E"));
	    g_integer ntc(0), status(1), nspec(0), nmin, nmax;
	    if (!xmin.defined())
	    {
		xmin.alloc(1);
		xmin[0] = xarray[0];
	    }
	    if (!xmax.defined())
	    {
		xmax.alloc(1);
		xmax[0] = xarray[xarray.length()-1];
	    }
	    nmin = xmin.length();
	    nmax = xmax.length();
	    if (max(nmin, nmax) == 1)
	        ginf << "INTEGRATE: Integrating between " << xmin[0] << " and " << xmax[0] << endl;
	    if (yarray.dimsSize() == 1)
	    {
		ntc = yarray.dims(1);
		nspec = 1;
	    }
	    else if (yarray.dimsSize() == 2)
	    {
		ntc = yarray.dims(2);
		nspec = yarray.dims(1);
	    }
	    GenieRealArray sum_y(nspec), sum_e(nspec);
// I need all these casts to get the template to work :-(
	    status = integrateSpectra((g_real*)xarray, (g_real*)yarray, (g_real*)earray, 
				ntc, nspec, (g_real*)xmin, nmin, (g_real*)xmax, nmax, 
				(g_real*)sum_y, (g_real*)sum_e);
	    if (status == 0)
	    {
// Treat nspec = 1 as special (for backward compatability)
		if (nspec == 1)
		{
		    res("SUM", sum_y[0]);
		    res("ERROR", sum_e[0]);
		}
		else
		{
		    res("SUM", sum_y);
		    res("ERROR", sum_e);
		}
	    }
 	    else
	    {
		*g_err << "INTEGRATE: Error integrating spectra" << endl;
	    }
	    return status;
}

// 0 on success, something else on error
static int transformIntegrateFile(GenieRealArray& xmin, GenieRealArray& xmax, 
	GenieString& file, GenieIntegerArray& iarray, GenieWorkspace& res)
{
    ISISRawFile raw_file;
    if (raw_file.open(file) != 0)
	return -1;
    g_integer ntc = raw_file.numberOfTimeChannels();
    g_real* yarray = new g_real[ntc];
    g_real* earray = new g_real[ntc];
    g_integer nmin, nmax, nspec;
    nspec = iarray.length();
    int* dims_array = iarray.dimsArray();
    int ndims = iarray.dimsSize();
    GenieRealArray sum(dims_array, ndims), error(dims_array, ndims);
    int status(0), imin, imax;
    if (!xmin.defined())
    {
	xmin.alloc(1);
	xmin[0] = raw_file.timeChannels()[0];
    }
    if (!xmax.defined())
    {
	xmax.alloc(1);
	xmax[0] = raw_file.timeChannels()[ntc];
    }
    nmin = xmin.length();
    nmax = xmax.length();
    if (max(nmin, nmax) == 1)
	 ginf << "INTEGRATE: Integrating between " << xmin[0] << " and " << xmax[0] << endl;

    fort_int spec;
    for(int i=0; (i < nspec) && (status == 0); i++)
    {
	spec = iarray[i];
	if (raw_file.readSpectrum(yarray, earray, spec) == 0)
	{
	    imin = i % nmin;
	    imax = i % nmax;
	    status = integrateSpectrum(raw_file.timeChannels(), yarray, earray,
				ntc, xmin[imin], xmax[imax], sum[i], error[i]);
	}
	else
	{
	    status = 1;
	    *g_err << "INTEGRATE: Error reading spectrum " << spec << endl;
	}
    }
    delete[] yarray;
    delete[] earray;
    if (status == 0)
    {
// Treat nspec = 1 as special (for backward compatability)
	if (nspec == 1)
	{
	    res("SUM", sum[0]);
	    res("ERROR", error[0]);
	}
	else
	{
	    res("SUM", sum);
	    res("ERROR", error);
	}
    }
    else
    {
	*g_err << "INTEGRATE: Error integrating spectra" << endl;
    }
    return status;
}
