/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

//
//    Module:         genie_data_access
//    Author:         Chris Moreton-Smith, ISIS
//
//
//  Implements a class which provides access to the functionality
//  of Open GENIE.  Construction of CGenie runs up a full working
//  instance of GENIE but without any UI functionality.
//
//  Note that only one instance of Open GENIE can be created in process
//  at any one time, attempts to create any more instances will fail.

#include "geniedefs.h"  // Project include, keep this at the top always in all files

#include <time.h>
#include <signal.h>
#include <setjmp.h>
#include <limits.h>
#include <sys/stat.h>
#include <stdlib.h>

#if defined(_WIN32)
#	include "vms_uname.h"
#else
//#include <sys/types.h>
#	include <sys/utsname.h>
#endif

#include "smalltalk.h"
#include "parsedefs.h"
#include "parser.h"

#include "ghistory.h"
#include "fileset.h"
#include "genie.h"
#include "gsignals.h"
#include "genielog.h"

void* g_pGenie = NULL;				// Declare a global pointer the the genie object

unsigned CGenie::s_dwSessionID = 0;

GENIEEXPORT CGenie::CGenie(Options options, const char* default_format) :

// Member variable initialization (carried out in order of declaration - not necessarily order here)
    m_options(options),                     // Options which control the way GENIE is built and run (before useing TestOption())
    m_streams(TestOption(TERMINALSTREAMS),  // Use terminal streams if true
	      TestOption(INFORMATIONALMESSAGES), TestOption(DEBUGMESSAGES)),    // Select level of Output required
    m_pfileset(CFileSet::s_this),			    // Do this early as it initialises the locations for all file access
    m_history(m_pfileset->m_gstrHistoryFile), // Set GNU history class with default history file

    m_smalltalk(                        // Start a smalltalk
        m_pfileset->m_gstrSmalltalkImage,      // Path for compiled image to load from/save to
        TestOption(SAVEIMAGE) ),        // Starts the smalltalk, rebuild from new if the flag is set.

    m_parser(m_smalltalk,    // Start the parser using the smalltalk we just made.
        TestOption(SAVEIMAGE) || TestOption(GENDEBUG) ? (const char*)m_pfileset->m_gstrDebuggingOutput : "",   // intermediate smalltalk O/P
        TestOption(SAVEIMAGE) ? (const char*)m_pfileset->m_gstrParserInstrumentation : "",                     // parser instrumentation
        TestOption(NOSTOP), TestOption(YACCDEBUG))

// Main constructor body
{
    if (s_dwSessionID != 0)
    {
	    gerr << "Not allowed to create more than one genie object per process - aborting" << endl;
            exit(1);
    }
    else
    {
	s_dwSessionID++;
	
        // Ready to roll at this point - we can now use the GENIE object (make it available externally too)
	g_pGenie = (void*) this;

	Set_signals();					// Set up signal/exception handling

        Logit("Constructing CGENIE");                   // Internet Log this invocation of GENIE (also gets system info)

// comment out for moment - we don't see this with Tk/Tcl startup, and it leaves some
// runnish in the output stream that can confuse the first TCL startup comamnds
//        PrintBanner();                                  // Announce ourselves to the world!

        Stinit();           // smalltalk runtime initialization (for volatile things)
        DoOptionalGCLLoad();                            // Optionally load up GCL source code and save new image
        GCLinit();	    // GCL runtime initialisation

        // Load up a user genieinit.gcl always, but don't report if its not there (time 0 = load always)
        LoadFileByDate( m_pfileset->m_gstrGCLInitialisation, (time_t) 0, false);    

    }
}
	

// Destructor for the session (most gets destroyed automatically)
GENIEEXPORT CGenie::~CGenie()
{
    // Turn off the engine (don't forget to take the keys with you!)
    Unset_signals();			// best not to catch ourselves with this one !
    Logit("Stopping cleanly");		// Log the end of this session
    g_pGenie = NULL;			// No genie instance exists now

    // Do this last in case something bombs and we don't get here
    s_dwSessionID--;
}


// Functions CGenie::GCLnit() and CGenie::Stinit()
// execute any GCL or smalltalk code which must be re-done on each
// invocation of Open GENIE.  Remember, most globals etc are
// saved with the image but this may be an appropriate place to reinitialize
// some of them or set up volatile or hardware information.
void CGenie::Stinit()
{

#if defined(__VMS) && !defined(__VMS_POSIX)
#   define TTYNAME() ttyname()
#else
#ifdef _WIN32
#	define TTYNAME() "CON:"
#else
#   define TTYNAME() ttyname(Stdout)
#endif
#endif

    static char buffer[LONGCHARBUFF];

    sprintf(buffer, "Smalltalk at: #gXUUlocalUtty put: '%s'!", TTYNAME());
    m_smalltalk.EvaluateCommand(buffer);
    sprintf(buffer, "Smalltalk at: #gXUUplatformUdetails put: '%s'!", (const char*)m_gstrSystemInfo); // Version info found in Logit()
    m_smalltalk.EvaluateCommand(buffer);

    return;
}

void CGenie::GCLinit()
{
    m_parser.CompileString("__initprimitives");
    m_parser.CompileString("__initgraphics");
    m_parser.CompileString("__initio");

    return;
}


//
// Build a running genie up from all the appropriate bits of smalltalk or restore a
// saved image file instead.
//
int CGenie::LoadFileSetByImageDate(time_t date)
{
    int i, retval;
    String gstrFile;

    for ( i=0; m_pfileset->GetBaseFile(i, gstrFile); i++ )
    {
          retval = LoadFileByDate(gstrFile, date, true );
          if (retval > 0)		// errors on loading file
          {
              gerr << "Error loading file \"" << gstrFile << "\"" << endl;
          }
          if (retval == EOF)		// File not found
	    return EOF;
    }

    return 0;
}


// This takes a single command string in GENIE Command Language (GCL) and
// interprets it.  GENIE should already be in a fully loaded state before this
// routine is called.

int CGenie::EvaluateGCLCommand(const char* szCommand, int history)
{
    if (history == 1) {
        m_history.AddCommand(szCommand);        // Pass a copy to GNU history
    }
    return m_parser.CompileString(szCommand);     // Run through GCL (interactive mode)
}

// This takes a single command string in smalltalk and passes it on to be
// interpreted.  GENIE should already be in a fully loaded state before this
// routine is called.  Note, this should be used instead of CSmalltalk::EvaluateCommand
// being used directly.  Note the usual smalltalk "!" to end a command is required
// but not a \n

int CGenie::EvaluateSmalltalkCommand(const char* szCommand)
{
//    m_history.AddCommand(szCommand);			// Pass a copy to GNU history
    return m_smalltalk.EvaluateCommand(szCommand);       // Run through GCL (interactive mode)
}


// Returns 0 on a successful load, EOF, if a file was missing, or an error count
int CGenie::LoadFileByDate(const char filename[], const time_t date, bool bWarn)
{
    time_t filedate;
    int retval(0);

    // Load if the file is more recent than the date passed
    filedate = mdate(filename);
    if(filedate==0) filedate=INT_MAX;			// Go for very large date if file doesn't exist

    if ( (filedate > date) || (date == 0) )
    {
	ginf << "Loading file " << filename;
	retval = m_parser.CompileFile( filename );      // Call the parser to load the file
	if ( (retval == EOF) && bWarn )
	{
	    g_printef("\tError - Could not open file \"%s\"\n", filename );
	}
    }

    return retval;
}


void CGenie::Logit( const char callinfo[] )
{
    char szLogRecord[LONGCHARBUFF],              // Record logged to network
         szCompilerDetails[LONGCHARBUFF],        // Compiler details
         szOSDetails[LONGCHARBUFF],              // Operating system and version
         szReadableDetails[LONGCHARBUFF];        // OS compiler details for screen

    struct utsname details;                                     // details from uname()
    const char* current_command = m_history.GetLastCommand();   // last command executed (if any)

    // Format compiler information
#if defined(__DECCXX)
#   define MAJV 10000000
#   define MINV 100000
#   ifdef __ultrix
#	define __DECCXX_VER	10300000	/* Ultrix CXX doesn't define this - I use 1.3 so we'll assume that */
#   endif
    sprintf(szCompilerDetails, "DEC C++ V%.1f",
        __DECCXX_VER/MAJV + ((__DECCXX_VER-(__DECCXX_VER/MAJV)*MAJV)/MINV) /10.0);  // YYYYUUUCKKK
#elif defined(_MSC_VER)
    sprintf(szCompilerDetails, "MS Visual C++ V%s", _MSC_VER==1100?"5.0":"5.0 or greater");
#elif defined(__GNUC__)
    sprintf(szCompilerDetails, "GNU C++ V%d.%d", (int)__GNUC__, (int)__GNUC_MINOR__);
#else
    sprintf(szCompilerDetails, "Unknown C++ compiler");
#endif

    // record OS details
    uname(&details);			// Get OS and machine info
#if defined(__VMS)
    sprintf(szOSDetails, "%s:%s:%s", details.machine, details.sysname, details.version);
#else
    sprintf(szOSDetails, "%s:%s:%s", details.machine, details.sysname, details.release);
#endif

    // Format for log record
    sprintf(szLogRecord, "Event = %s\nCurrent command = %s\nEnvironment = #%s#%s#%s#%s#%s#%s\nVersion = %s",
            callinfo, current_command, details.nodename, details.machine, details.sysname,
            details.release, details.version, szCompilerDetails, g_version);

    // Now send the information
    genielog(szLogRecord);
    
    // Format for screen print and store
    sprintf(szReadableDetails, "%s\n(Running on %s - built with %s)",
        (const char*)&g_version[4], szOSDetails, szCompilerDetails);    // (Don't print the @(#) bit in g_version)
    m_gstrSystemInfo = szReadableDetails;                               // Save this string for later use
    return;
}


// Contain all legal notification of no warranty etc. also startup logging because it produces version info.
void CGenie::PrintBanner(ostream& os)
{
    // Report version & copyright notices to all and sundry
    os << m_gstrSystemInfo << endl;		        
    os << '\n';
    os << "   ******************************************************************" << '\n';
    os << "   *  Copyright (C) 1996,1997,1998 CCLRC Rutherford Appleton Lab.   *" << '\n';
    os << "   *  Open GENIE comes with ABSOLUTELY NO WARRANTY; for details     *" << '\n';
    os << "   *  type `warranty'.  This is free software, and you are welcome  *" << '\n';
    os << "   *  to redistribute it under certain conditions; type `licence'   *" << '\n'; 
    os << "   *  for details.                                                  *" << '\n';
    os << "   ******************************************************************" << '\n';
    os << endl;
}


// Either reload all GCL when SAVEIMAGE is specified, or check each GCL file against the current
// image date and load if newer.  After a full reload, save the new image file.
void CGenie::DoOptionalGCLLoad()
{
    time_t tImageDate;

    // the FILECHECK option specifies that we check the file dates against the image date
    if (TestOption(FILECHECK) || TestOption(SAVEIMAGE))
    {
        if (TestOption(SAVEIMAGE))   // Make some extra space for building images
        {
//            m_smalltalk.EvaluateCommand("Smalltalk enableGC: false!");              // disable garbage collecting
            m_smalltalk.EvaluateCommand("Smalltalk growTo: 20 * 1024 * 1024!");
	    tImageDate = 0;
        }
	else
	{
	    tImageDate = mdate(m_pfileset->m_gstrSmalltalkImage);                      // Get date of last image
	}

        if (LoadFileSetByImageDate(tImageDate) != 0)
        {
            gerr << "Essential system files not found, cannot start GENIE" << endl;
            ginf << "  Check definition of symbol GENIE_DIR" << endl;
            exit(1);
        }
    }

    // create a snapshot image file if we've been asked to load the whole lot
    if ( TestOption(SAVEIMAGE) )
    {
//        m_smalltalk.EvaluateCommand("Smalltalk enableGC: true!");		    // re-enable garbage collecting
        String command = "Smalltalk snapshot: '" + m_pfileset->m_gstrSmalltalkImage + "' !";
        m_smalltalk.EvaluateCommand(command);
        ginf << "Saved image file to \"" << m_pfileset->m_gstrSmalltalkImage
             << "\" and boot file to \"" << m_pfileset->m_gstrDebuggingOutput << "\"" << endl;
    	ginf << "Also saved procedure definitions to \"" << m_pfileset->m_gstrParserInstrumentation << "\"" << endl;
    }
}


// Called by GDAI, just makes debugging output available over the default
GENIEEXPORT int CGenie::gcl(const char* code)
{
    gdbg << code << endl;
    return EvaluateGCLCommand(code,0);
}


// Called by GDAI, just makes debugging output available over the default
GENIEEXPORT int CGenie::smalltalk(const char* code)
{
    gdbg << code << endl;
    EvaluateSmalltalkCommand(code);

    return 0;
}


// set the current default file O/P format for the gdai
GENIEEXPORT int CGenie::file_format(const char* file_format)
{
    String Fmt = file_format;
    Fmt.upcase();

    if ( Fmt == "GENIE" || Fmt == "" )
        Format = "";
    else if ( Fmt == "NEXUS" )
        Format = "/NEXUS";
    else if ( Fmt == "G2" )
        Format = "/G2";
    else if ( Fmt == "ASCII" )
        Format = "/ASCII";
    else
        gerr << "(file_format)*** Error: \"" << file_format << "\" is an unknown file format" << endl;

    return 0;
}


time_t mdate(const char *filename)
{
    static struct stat buf;
    if ( stat(filename, &buf) == -1 )
	return (time_t) 0;
    else
	return buf.st_mtime;
}



CStreams* CGenie::Streams()
{
    return &m_streams;
}
