/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#ifndef GENIE__
#define GENIE__

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000

//
//    Module:         CGenie
//    Author:         Chris Moreton-Smith, ISIS
//
//
//  Class which provides access to the functionality of Open GENIE
//  via a C++ class interface.
//
//  Currently this is just gives programmatic control suitable for a user
//  interface.  Note that more extensive functionality will be provided
//  by separate interfaces later.
//
//  This interface can be used in concert with direct access to Open GENIE
//  routines (e.g. the data access interface).  This method needs care
//  however and may not always be supported.
//
//  Note that only one instance of Open GENIE can be created in process
//  at any one time, attempts to create any more instances will fail.

#include "Streams.h"	// Added by ClassView
class CGenie
{

public:
    // Options mask for starting up Open GENIE engine
enum Options
    {
	YACCDEBUG=1,                // Output yacc parser debug information
	GENDEBUG=2,                 // Create Diagnostic Smalltalk output into file & via debug chan
	CRASHDEBUG=4,               // Remove exception handling to allow crash debugging
	SAVEIMAGE=8,                // Build and save a complete image
	FILECHECK=16,               // Load any newer GCL files from the base set
	NOSTOP=32,                  // Carry on parsing even after an error
	INFORMATIONALMESSAGES=64,   // Default state On for informational messages
	DEBUGMESSAGES=128,          // Default state On for debugging messages
        TERMINALSTREAMS=256,        // Create with terminal I/O streams, default is captured streams
	COMMANDLINE=512		    // purely command line (non Tk) genie
    };

private:
    // declare some private streams for subsequent use to trap I/O
    genie_ostream out, inf, err, dbg;

    // major classes used. Keep the order of declarations down to m_parser the same
    // as this is relied on by the member initializers on the constructor (i.e. the order
    // on the constructor is irrelevant but it's here that it matters).

    // Order sensitive
    Options m_options;              // Constructor options set up first
    CStreams m_streams;             // C++ view of the standard file streams, decide where I/O goes
    CFileSet* m_pfileset;            // Encapsulates all Open GENIE file paths
    CHistory m_history;             // History class implemented using GNU readline
    CSmalltalk m_smalltalk;         // Starts up smalltalk
    CParser m_parser;               // The GCL to Smalltalk parser
    // End Order sensitive

    // handy internal functions
public:
    bool TestOption( Options options ) { return (options & m_options) != 0; }
    void PrintBanner(ostream& os);
private:
    void DoOptionalGCLLoad();               // Load GCL source if required
    int  CGenie::LoadFileSetByImageDate(time_t date);  // Load and test dates of all GCL source files against a date
    void Stinit();                          // Initialization in smalltalk (for GENIE specifically)
    void GCLinit();                         // Initialization in GCL

    // other private members
    static unsigned s_dwSessionID;      // Prevent re-entrancy
    String m_gstrSystemInfo;            // Version, OS and compiler information

public:
    
    int EvaluateSmalltalkCommand(const char* szCommand);
    int EvaluateGCLCommand(const char* szCommand, int history);
    CHistory* GetHistory() {return &m_history;}
    // functions where access may be required elsewhere
    int  LoadFileByDate(const char filename[], const time_t date = 0, bool bWarn = true);   // Access required for LOAD command
    void Logit( const char callinfo[] );

    // Constructors
    GENIEIMPORT CGenie(Options options, const char* default_format);

    // Destructors
    GENIEIMPORT ~CGenie();

public:
   // Public accessor functions
   CStreams* Streams();                       // Access to smalltalk
   
    

// Definitions tied up with the GDAI (may move out of class soon)
// Send smalltalk or GCL commands to the Open GDAI
public:
    GENIEIMPORT int gcl(const char* code);
    GENIEIMPORT int smalltalk(const char* code);
    GENIEIMPORT int file_format(const char* file_format);
    GENIEIMPORT String file_format(void) const { return Format; }

private:
    // Definitions tied up with the GDAI (may change)
    String Format;              // Default file format for this session

};

// Global helpers
time_t mdate(const char *filename);



#endif // GENIE__
