/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

/*
 *      Module:         GenieArrayBase
 *      Author:         Freddie Akeroyd, ISIS
 *      Purpose:        Base definitions common to all genie arrays
 *
 */
#include "geniedefs.h"
#include <fstream.h>
#include <string.h>
#include <gnustring.h>
#include <ctype.h>
#include "Genie_functions.h"

#include "smalltalk.h"

static const char* rcsid = "@(#)$Id: genie_array_base.cxx,v 1.24 1998/10/22 07:26:36 cmm Exp $";
USE_RCSID(rcsid);

int associateGenieVariables(OOP oopParameters, ... );

char* Strdup(const char* old_string);
char* Strdup(const char* old_string, int length);
char* Strdup(const unsigned char* old_string);
char* Strdup(const unsigned char* old_string, int length);

class GenieVariable;	// forward declaration --- needed for workspace

#include "genievars.h"

const int INST_VAR_OFFSET = 3;

//void GenieArrayBase::reverse()
//{
//    for(int i=1; i<=m_length/2; i++)
//    {
//	swap(i, m_length + 1 - i);
//    }
//    return;
//}

GenieArrayBase::GenieArrayBase(OOP oop) : GenieVariableBase(oop), m_dope(0), m_ndims(0)
{
    init(oop);
    updateDimsAndDope();
}

ostream& GenieArrayBase::print(ostream& os, const char* note) const
{
    GenieVariableBase::print(os, note);
//    os << "Array length = " << length() << ", Dims size = " << dimsSize() << endl;
//    if (isOOP(dims()))
//    {
//	os << "Array DIMS constructor" << endl;
//	printOOPConstructor(dims());
//    }
//    else
//    {
//	os << "\nDIMS is not an OOP!" << endl;
//    }
//    if (isOOP(dope()))
//    {
//	os << "\nArray DOPE constructor" << endl;
//	printOOPConstructor(dope());
//    }
//    else
//    {
//	os << "\nDOPE is not an OOP!" << endl;
//    }
//    os << "\nPrinting DOPE and DIMS vector" << endl;
//    for(i=1; i<=dimsSize(); i++)
//    {
//	os << dope(i) << " = " << dims(i) << endl;
//    }
    return os;
}

void GenieArrayBase::setDimsAndDope(int* dims_array, int ndims)
{
    OOP dims_oop = makeDimsOOP(dims_array, ndims);
    instVarAtPut(m_value, MY_DIMS, dims_oop);
    instVarAtPut(m_value, MY_DOPE, makeDopeOOP(dims_oop)); 
    updateDimsAndDope();
    return;
}

void GenieArrayBase::setDimsAndDope(int dim1, int dim2)
{
    instVarAtPut(m_value, MY_DIMS, makeDimsOOP(dim1, dim2));
    instVarAtPut(m_value, MY_DOPE, makeDopeOOP(dim1, dim2));
    updateDimsAndDope();
    return;
}

void GenieArrayBase::copyDims(const GenieArrayBase& ga)
{
  BEGINSAVEBLOCK
    int m_dimsSize = ga.dimsSize();
    OOP m_dimsOOP = saveOOP(arrayNew(m_dimsSize));
    OOP m_dopeVectorOOP = saveOOP(arrayNew(m_dimsSize));
    for(int i=1; i<=m_dimsSize; i++)
    {
        arrayAtPut(m_dimsOOP, i, fromInt(ga.dims(i)));
        arrayAtPut(m_dopeVectorOOP, i, fromInt(ga.dope(i)));
    }
    instVarAtPut(m_value, MY_DIMS, m_dimsOOP);
    instVarAtPut(m_value, MY_DOPE, m_dopeVectorOOP);
  ENDSAVEBLOCK
    updateDimsAndDope();
    return;
}

GenieArrayBase& GenieArrayBase::operator=(const GenieArrayBase& gab)
{
    if (this != &gab)
    {
	setOOP(gab.shallowCopy());
    }
    return *this;
}

OOP GenieArrayBase::deepCopy(const GenieArrayBase& gab)
{
    gerr << "GenieArrayBase::deepCopy(),called ... error\n";
    return nilOOP;
}

OOP GenieArrayBase::deepCopyOOP(OOP oop)
{
    gerr << "GenieArrayBase::deepCopyOOP(),called ... error\n";
    return nilOOP;
}

OOP GenieArrayBase::shallowCopy(const GenieArrayBase& gab)
{
    if (this != &gab)
    {
        GenieVariableBase::shallowCopy(gab);
    }
    return m_value;
}

ostream& operator<<(ostream& os, const GenieArrayBase& gab)
{
    return gab.print(os);
}

int GenieArrayBase::arraySize(OOP a_dimsOOP)
{
    int size = 1;
    for(int i=1; i<=numIndexableFields(a_dimsOOP); i++)
    {
	size *= toInt(arrayAt(a_dimsOOP, i));
    }
    return size;
}

OOP GenieArrayBase::makeDopeOOP(const int* dims_array, int ndim)
{
    int dope = 1;
    OOP a_dopeOOP = saveOOP(arrayNew(ndim));
    for(int i=ndim; i>=1; i--)
    {
	arrayAtPut(a_dopeOOP, i, fromInt(dope));
	dope *= dims_array[i];
    }
    return a_dopeOOP;
}

OOP GenieArrayBase::makeDopeOOP(OOP a_dimsOOP)
{
    if (a_dimsOOP != nilOOP)
    {
        int dope = 1, size = numIndexableFields(a_dimsOOP);
        OOP a_dopeOOP = saveOOP(arrayNew(size));
        for(int i=size; i>=1; i--)
        {
	    arrayAtPut(a_dopeOOP, i, fromInt(dope));
	    dope *= toInt(arrayAt(a_dimsOOP, i));
        }
        return a_dopeOOP;
    }
    else 
    {
	return nilOOP;
    }
}

OOP GenieArrayBase::makeDopeOOP(int dim1, int dim2)
{
    OOP a_dopeOOP;
    if (dim2 == 0)              // 1-D array
    {
        a_dopeOOP = saveOOP(arrayNew(1));
        arrayAtPut(a_dopeOOP, 1, fromInt(1));
    }
    else                        // 2-D array
    {
        a_dopeOOP = saveOOP(arrayNew(2));
        arrayAtPut(a_dopeOOP, 1, fromInt(dim2));
        arrayAtPut(a_dopeOOP, 2, fromInt(1));
    }
    return a_dopeOOP;
}

OOP GenieArrayBase::makeDimsOOP(const int* dims_array, int ndim)
{
    OOP a_dimsOOP = saveOOP(arrayNew(ndim));
    for(int i=1; i<=ndim; i++)
    {
        arrayAtPut(a_dimsOOP, i, fromInt(dims_array[i-1]));
    }
    return a_dimsOOP;
}

// Return a new genie Array oop containing the given dimensions information
OOP GenieArrayBase::makeDimsOOP(int dim1, int dim2)
{
    OOP a_dimsOOP;
    if (dim2 == 0)              // 1-D array
    {
        a_dimsOOP = saveOOP(arrayNew(1));
        arrayAtPut(a_dimsOOP, 1, fromInt(dim1));
    }
    else                        // 2-D array
    {
        a_dimsOOP = saveOOP(arrayNew(2));
        arrayAtPut(a_dimsOOP, 1, fromInt(dim1));
        arrayAtPut(a_dimsOOP, 2, fromInt(dim2));
    }
    return a_dimsOOP;
}

int* GenieArrayBase::dimsArray() const
{
    int* result = new int[dimsSize()];
    for(int i=1; i<=dimsSize(); i++)
    {
	result[i-1] = dims(i); 
    }
    return result;
}

// Update the internal DIMS and DOPE values with those stored in the OOP
void GenieArrayBase::updateDimsAndDope()
{
    if (m_value != nilOOP)
    {
        m_ndims = numIndexableFields(dimsOOP(m_value));
        delete[] m_dope;
        m_dope = new int[m_ndims];
        OOP dope_oop = dopeOOP(m_value);
        for(int i=0; i<m_ndims; i++)
        {
	    m_dope[i] = toInt(arrayAt(dope_oop, i+1));
        }
    }
    return;
}

int GenieArrayBase::writeAsciiDims(fstream& fs, int indent)
{
    fs << setindent(indent) << dimsSize() << "\n"; 

    fs << setindent(indent);
    for(int i=1; i<=dimsSize(); i++)
    {
	fs << dims(i) << " ";
    } 
    fs << "\n";

    return 0;
}

// return a dimsOOP for use in an alloc()
OOP GenieArrayBase::readAsciiDims(fstream& fs)
{
    int ndims(-1);
    fs >> ndims;
    if (!fs.good() || (ndims <= 0))
    {
        error_add("GenieArrayBase::readAsciiDims", "Invalid number of dimensions", "");
        return nilOOP;
    }
    int* dims_array = new int[ndims];
    for(int i=0; i<ndims; i++)
    {
        fs >> dims_array[i];
    }
    OOP a_dimsOOP = makeDimsOOP(dims_array, ndims);
    delete[] dims_array;
    return a_dimsOOP;
}

OOP GenieArrayBase::become(OOP oop)
{
    GenieVariableBase::become(oop);
    updateDimsAndDope();
    return oop;
}

GenieArrayBase::GenieArrayBase(OOP params, int location) : GenieVariableBase(params, location), m_dope(0), m_ndims(0) 
{
    updateDimsAndDope();
}

void GenieArrayBase::init(OOP oop, OOP params, int location)
{
    GenieVariableBase::init(oop, params, location);
    updateDimsAndDope();
}

// dims_array pre allocated to dimensionality of array
// This function convers a 1D array reference into an ND one
// i.e. for A[10][10] converts A[12] into A[1][2]
void GenieArrayBase::arrayLocation(int index, int* dims_array)
{
    for(int i=0; i<m_ndims; i++)
    {
	dims_array[i] = index / m_dope[i];
	index %= m_dope[i];
    }
}
