/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#ifndef GENIE_BASE
#define GENIE_BASE

/*
 * this is the superclass for all genie variables
 */

extern OOP getClassOOP(OOP oop);

#include "genie_exception.h"

class GenieVariableBase;
class GenieCalloutList;
class IntermediateFile;

class GenieVariableBase
{
protected:
    void throwException(GenieExceptionType an_exception) { siglongjmp(genie_exception_context, an_exception); }

public:
/*
 * m_old_class and m_old_value are used to check for variables being 
 * returned to smalltalk with an illegal change of class
 * they are only set if that vaiable is "associated"
 */
    OOP m_old_class;    // The class at the time of the first call to "init" or "associate"
    OOP m_old_value;    // The value at the time of the first call to "init" or "associate"
                        // the above are used to check for illegal class changes

    OOP m_value;        // Current value of the variable
    OOP m_params; 	// Parameter list from smalltalk callout of which this variable is a part
    int m_location;     // Position of this variable in the m_params list - where result is written back to

public:
// associate genie variable with a named smalltalk variable in a callout parameter list
    virtual int associate(OOP parameters, const char* name);
//
// can either use:    associateGenieVariables(parameters, "name1", &var1, "name2", &var2, NULL)
//
// or         var1.associate(parameters, "name1")
//            var2.associate(parameters, "name2")
//
// This associate for anonymous parameters (g_operations.cc)
    virtual int associate(OOP parameters, int location);
// New CalloutList interface
    virtual int associate(const GenieCalloutList& gcl, const char* name);

    virtual ostream& print(ostream& os, const char* note=NULL) const;
    virtual void init(OOP oop = nilOOP, OOP params = nilOOP, int location = 0);	// called by Associate to set up variable
    virtual int defined() const { return !isNil(m_value); }
    virtual ~GenieVariableBase();
    GenieVariableBase(const GenieVariableBase& gvb) { init(gvb.deepCopy()); }
    GenieVariableBase(OOP value_oop = nilOOP, OOP subclass_oop = undefinedObjectClass) :
	m_old_class(nilOOP), m_old_value(nilOOP), m_value(nilOOP), m_params(nilOOP), m_location(0) { init(value_oop); }
// Constructor for anonymous callout, as in g_operations
    GenieVariableBase(OOP parameters, int location) { init(arrayAt(parameters, location), nilOOP, 0); }
// Constructor for new CalloutList interface
    GenieVariableBase(const GenieCalloutList& gcl, const char* name) { associate(gcl, name); }
    virtual GenieVariableBase& operator=(const GenieVariableBase& gvb);
    virtual GenieVariableType write(IntermediateFile* gif);
    virtual GenieVariableType read(IntermediateFile* gif);
    virtual GenieVariableType skip(IntermediateFile* gif);
    virtual GenieVariableType readAscii(fstream& fs);
    virtual GenieVariableType writeAscii(fstream& fs, int indent = 0);
    int indentAscii(fstream& fs, int indent);
    OOP value() const { return m_value; }
    operator OOP() const { return value(); }
    OOP operator<<(const GenieVariableBase& gvb) { return become(gvb); }
    int location() const { return m_location; }
    OOP params() const { return m_params; }
    OOP at(OOP pos) const;
    OOP atOrNil(OOP pos) const;
    OOP atPut(OOP pos, OOP val);
    OOP become(OOP oop) { return(m_value = oop); }
    OOP become(const GenieVariableBase& gvb) { return become(gvb.value()); }
    OOP shallowCopy() const { return shallowCopyOOP(m_value); }
    OOP deepCopy() const { return deepCopyOOP(m_value); }
    OOP setOOP(OOP oop) { return(m_value = oop); }
    void shallowCopy(const GenieVariableBase& gvb);
    void deepCopy(const GenieVariableBase& gvb);
    static OOP shallowCopyOOP(OOP oop);
    static OOP deepCopyOOP(OOP oop);
    static OOP becomeOOP(OOP oop) { return oop; }
    OOP getClass() const { return getClassOOP(m_value); }
    OOP inspect();		// simulate a smalltalk "inspect"
    static OOP inspectOOP(OOP oop);
    GenieVariableType type() const;
};

ostream& operator<<(ostream& os, const GenieVariableBase& gv);

#define BEGINSAVEBLOCK 	IncPtr incPtr = incSavePointer();

void saveGenieVariable(const GenieVariableBase& gvb);

OOP saveOOP(OOP oop);

#define ENDSAVEBLOCK	incRestorePointer(incPtr);

#endif	/* GENIE_BASE */
