/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

//
//    Module:         genie_data_access
//    Author:         Chris Moreton-Smith, ISIS
//
//

#include "geniedefs.h"
#include "smalltalk.h"
#include "parsedefs.h"
#include "parser.h"
#include "ghistory.h"
#include "fileset.h"
#include "genie.h"
#include "genievars.h"

#include "genie_data_access.h"

static const char* rcsid = "@(#)$Id: genie_data_access.cxx,v 1.45 1998/11/02 17:45:44 cmm Exp $";
USE_RCSID(rcsid);

CGenie *g_pgenie;	// The kept genie (once it's been created)

// Routines to allow simple but powerful data access to Freddie's
// wonder intermediate files and NeXus and ISIS raw files and ...
// who knows what else ?

// All these routines are based on a very simple assumption that it
// can all be done in Open GENIE anyway. Unfortunately Open GENIE looks so
// much like an interative thingy that no one will believe the power of the
// interface it makes possible. So we write a simple but powerful API and
// the point is made !!


// Constructor for an Open GDAI session

// Activate a closet Open GENIE session to do the work
int GX_one = 1, GX_zero = 0;
GENIEEXPORT int GENIEMECHANISM GX_activate_session(GDAI_STRING default_format = "GENIE", int inf = GX_one, int debug = GX_zero)
{
    int options = CGenie::TERMINALSTREAMS;	// Always enabled

    options |= inf ? CGenie::INFORMATIONALMESSAGES : 0;
    options |= debug ? CGenie::DEBUGMESSAGES : 0;

    if (!g_pgenie)
	g_pgenie = new CGenie((CGenie::Options)options, default_format);  // Constructs a session
    else
    {
	gerr << "GX_activate_session: error, one Open GENIE is already active" << endl;
	ginf << "  (in a module, you do not need to activate)" << endl;
    }    
    return g_pgenie == 0;   // assume memory allocation failure from failed new
}        
    

// Deactivate the session
GENIEEXPORT int GENIEMECHANISM GX_deactivate_session( void ) 
{
    delete g_pgenie;  // Closes the session

    return 0;
}        


// for file association and contents inspection
GENIEEXPORT int GENIEMECHANISM GX_select_source( GDAI_STRING filename )
{
    String Command = "set/file/input ";
    return g_pgenie->gcl(Command + "\"" + filename + "\"");
}


// for file association and contents inspection
GENIEEXPORT int GENIEMECHANISM GX_select_destination( GDAI_STRING filename, GDAI_STRING file_format = NULL )
{
    if (file_format)
    g_pgenie->file_format(file_format);                       // Make this a checking member

    String Command = "set/file/output ";
    return g_pgenie->gcl(Command + "\"" + filename + "\"");
}


GENIEEXPORT int GENIEMECHANISM GX_directory( GDAI_STRING filename )
{
    String Command = "list ";
    return g_pgenie->gcl(Command + "\"" + filename + "\"");
}


// Data reading/writing
GENIEEXPORT int GENIEMECHANISM GX_get( GDAI_STRING handle, GDAI_STRING tag, int object_id )
{
    char locator[SHORTCHARBUFF];
    String Handle = handle;
    ostrstream Locator(locator,SHORTCHARBUFF);

    if (tag && strlen(tag) > 0)
        Locator << "\"" << tag << "\"" << ends;
    else
        Locator << object_id << ends;
    
    return g_pgenie->gcl(Handle + "=get(" + locator + ")");
}


GENIEEXPORT int GENIEMECHANISM GX_put( GDAI_STRING handle, GDAI_STRING tag, int object_id, GDAI_STRING flag, GDAI_STRING comment )
{
    // Can be called for "New" or "Overwrite"
    
    String Command = "put ";
    String Flag = flag;

    if (capitalize(Flag) == "Overwrite")
        *g_err << "***(GX_put) Overwrite flag not yet implemented ***" << endl;
    else if (capitalize(Flag) == "New")
        Command = Command + "/New ";
    else if (capitalize(Flag) == "Append")
        Command = Command + "/Append ";

    Command = Command + g_pgenie->file_format() + " ";

    return g_pgenie->gcl(Command + handle + " \"" + upcase(String(tag)) + "\" " + "\"" + comment + "\"");
}


// Routines for manipulating handles
GENIEEXPORT int GENIEMECHANISM GX_assign_handle( GDAI_STRING handle_lval, GDAI_STRING handle_rval )
{
    String Lval = handle_lval;
    String Rval = handle_rval;
	if (Rval == "")
		return g_pgenie->gcl(Lval);					// Allow a single command
	else
		return g_pgenie->gcl(Lval + "=" + Rval);
}

GENIEEXPORT int GENIEMECHANISM GX_release_handle( GDAI_STRING handle )
{
    String Command = "free ";
    return g_pgenie->gcl(Command + "\"" + handle + "\"");
}

// Old name for GX_transfer - kept for backward compatability of VMS SYMBOL_VECTOR (but marked as PRIVATE_PROCEDURE)
extern "C" int GX_associate_handle( GDAI_STRING handle, GDAI_STRING direction, GDAI_STRING type, void* the_data, int ndims, int dims[] );
int GX_associate_handle(GDAI_STRING handle, GDAI_STRING direction, GDAI_STRING type, void* the_data, int ndims, int dims[] )
{
    gerr << "Obsolete procedure GX_ASSOCIATE_HANDLE called, use GX_TRANSFER instead";
    return 1;	// failure
}

GENIEEXPORT int GENIEMECHANISM GX_transfer( GDAI_STRING handle, GDAI_STRING direction, GDAI_STRING data_type, void* the_data, int ndims, int dims[] )
{
    // This allows data to go both ways, in or out of GENIE

    static OOP gXUUth = registerOOP(internString("gXUUth"));

    BEGINSAVEBLOCK;

    int status = 0;	// Set status OK to start with
    int i;		// loop counter
    
    size_t cTotalElements = dims[0];	// will hold number of elements in an array

    String Handle = handle;
    String Direction = direction;
    String Type = data_type;
    Type.upcase();

    
    // Check for an array going either way and calculate the size of
    // data we need to move
    if (ndims > 20 || ndims < 0)    // should be set to 0 if not used so check for silly values
    {
	gerr << "(GX_transfer)*** Error: too many or few dimensions for the array" << endl;
	ginf << "(ndims was given as " << ndims << ")" << endl;
	return 1;
    }
    else if (ndims !=0)		    // we have an array
    {
	for (i=0; i < ndims; i++)
	{
	    if (dims[i] < 1)
	    {
		gerr << "(GX_transfer)*** Error: 0 or -ve array dimension supplied" << endl;
		ginf << "(dimension " << i+1 << " was given as " << dims[i] << ")" << endl;
		return 1;
	    }
	}
    
	// Pre calculate the size of the array given dims & ndims
	for(i=1; i < ndims; i++)
	{
	    cTotalElements *= dims[i];
	}
    }    

    // Get the SystemDictionary
    GenieWorkspace Smalltalk(typeNameToOOP("Smalltalk"));
    
    // Create a dummy handle to use for all the IO and which we know the name of
    // this means that we know what to do with a non identifier hanldle (eg a[4,5]) which
    // We could not parse sufficiently to refer to in smalltalk.
    g_pgenie->gcl("__th=_");

    if ( Direction == "<--" )
    {
        if (Type == "STRING")
        {
            GenieString Data((g_string)the_data);
            Smalltalk.atPut(gXUUth, (OOP)Data);
        }
        else if (Type == "INT64") 
        {
            GenieInteger Data(*(g_integer*)the_data);
            Smalltalk.atPut(gXUUth, (OOP)Data);
        }
        else if (Type == "INT32")
        {
            GenieInteger Data(*(int*)the_data);
            Smalltalk.atPut(gXUUth, (OOP)Data);
        }
        else if (Type == "FLOAT64")
        {
            GenieReal Data(*(g_real*)the_data);
            Smalltalk.atPut(gXUUth, (OOP)Data);
        }
        else if (Type == "FLOAT32")
        {
            GenieReal Data(*(float*)the_data);
            Smalltalk.atPut(gXUUth, (OOP)Data);
        }
        else if (Type == "STRING[]") 
        {
            gerr << "(GX_transfer)*** Error: \"STRING*\"not yet implemented" << endl;
        }
        else if (Type == "INT64[]") 
        {
            GenieIntegerArray Data((g_integer*)the_data, cTotalElements);
            Smalltalk.atPut(gXUUth, (OOP)Data);
	    Data.setDimsAndDope(dims, ndims);
        }
        else if (Type == "INT32[]")
        {
            GenieIntegerArray Data((int*)the_data, cTotalElements);
            Smalltalk.atPut(gXUUth, (OOP)Data);
	    Data.setDimsAndDope(dims, ndims);
        }
        else if (Type == "FLOAT64[]")
        {
            GenieRealArray Data((g_real*)the_data, cTotalElements);
            Smalltalk.atPut(gXUUth, (OOP)Data);
	    Data.setDimsAndDope(dims, ndims);
        }
        else if (Type == "FLOAT32[]")
        {
            GenieRealArray Data((float*)the_data, cTotalElements);
            Smalltalk.atPut(gXUUth, (OOP)Data);
	    Data.setDimsAndDope(dims, ndims);
        }
        else
            *g_err << "(GX_transfer)*** Error: \"" << Type
	    << "\" is not a recognized data type" << endl
	    << " (Use INT32, INT64, FLOAT32, FLOAT64, STRING or" << endl
	    << "  INT32[], INT64[], FLOAT32[], FLOAT64[], STRING[] for arrays)" << endl;

        // now assign the value back to the given handle (without a copy)
	status = g_pgenie->gcl(Handle + " <~ __th");
    
    }
    else if ( Direction == "-->" )
    {
        // assign the value out of GENIE (without a copy)
        String Command = "__th";
	status = g_pgenie->gcl(Command + " <~ " + Handle);

        if (Type == "STRING")
        {
            GenieString Data(Smalltalk.at(gXUUth));
            strncpy((char*)the_data, (c_g_string)Data, dims[0]);
            ((char *)the_data)[dims[0]-1] = '\0';
        }
        else if (Type == "INT64") 
        {
            GenieInteger Data(Smalltalk.at(gXUUth));
            *(long*)the_data = (g_integer)Data;
        }

        else if (Type == "INT32")
        {
            GenieInteger Data(Smalltalk.at(gXUUth));
            *(int*)the_data = (int)Data;
        }
        else if (Type == "FLOAT64")
        {
            GenieReal Data(Smalltalk.at(gXUUth));
            *(g_real*)the_data = (g_real)Data;
        }
        else if (Type == "FLOAT32")
        {
            GenieReal Data(Smalltalk.at(gXUUth));
            *(float*)the_data = F(Data);
        }

        else if (Type == "STRING[]") 
        {
            gerr << "(GX_transfer)*** Error: \"STRING*\"not yet implemented" << endl;
        }
        else if (Type == "INT64[]") 
        {
            GenieIntegerArray Data(Smalltalk.at(gXUUth));
            memcpy(the_data, (long*)Data, cTotalElements * sizeof(long));
        }   
        else if (Type == "INT32[]")
        {
            GenieIntegerArray Data(Smalltalk.at(gXUUth));
            memcpy(the_data, (int*)Data, cTotalElements * sizeof(int));
        }
        else if (Type == "FLOAT64[]")
        {
            GenieRealArray Data(Smalltalk.at(gXUUth));
            memcpy(the_data, (double*)Data, cTotalElements * sizeof(double));
        }
        else if (Type == "FLOAT32[]")
        {
            GenieRealArray Data(Smalltalk.at(gXUUth));
            memcpy(the_data, (float*)Data, cTotalElements * sizeof(float));
        }
        else
            *g_err << "(GX_transfer)*** Error: \"" << Type << "\" is not a recognized the_data type" << endl;

    }
    else
    {
        *g_err << "(GX_transfer)*** Error: Direction must be specified as \"<--\" or \"-->\"" << endl
	       << " to move data out of or into your code respectively" << endl;
        status = 1;
    }

    ENDSAVEBLOCK;

    return status;
}
