/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

/*
 *	Module:		intermediate_file
 *	Author:		Freddie Akeroyd, ISIS
 *  	Purpose:	Definitions for Open Genie intermediate files
 *
 */
#include "geniedefs.h"
//#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <gnustring.h>
#ifdef _WIN32
#	include <io.h>
#	include "stream_nt.h"
#else
#	include <stream.h>
#endif /* _WIN32 */
#include "genievars.h"
#include "genie_intermediate_file.h"
#include "Genie_functions.h"

static const char* rcsid = "@(#)$Id: genie_intermediate_file.cxx,v 1.14 1998/01/30 18:41:57 faa Exp $";
USE_RCSID(rcsid);

#if defined(__VMS) && __VMS_VER < 70000000
/* #include "vms_password_funcs.h" */
#endif

#if SIZEOF_LONG == 4		/* not 64 bit architecture */

// This routine reads/writes 64bit integers on 32bit machines, simulating
// the xdr_hyper() routine available with Digital Unix
// If the number is too big to represent, a warning message is printed,
// but an error value is not returned
static bool_t xdr_hyper(XDR* xdrs, long* hp)
{
    long zero(0);
    bool_t res = xdr_long(xdrs, &zero) && xdr_long(xdrs, hp);		// 	XDR is big endian
    if (zero != 0)
    {
	gerr << "xdr_hyper: read number too big to store on this architecture - setting it to " << LONG_MAX << endl;
	*hp = LONG_MAX;
    }
    return res;
}

#endif /* SIZEOF_LONG == 4 */

// #define DEBUG
/*
 *  Written by Freddie Akeroyd, RAL, 1995
 *
 * keep version numbers only one charactrer in length and do not change G3_FILE_HEADER_BASE
 * this will ensure "start_of_g3_file" is always correct whatever file version
 */

//
// Version control examples: 
//
//     FileVersionNumber         GenieVersionNumber	Result
//	   1.3				1.2		Reads OK, but prints warning message suggesting obtain an updated genie
//	   1.3				2.1		Reads OK
//	   2.1				1.4		Error - file has too high major version to read
//
// e.g. If you add a new data type (but leave the format of existing type unchanged) or just do
//      some bug fixes, bump up GENIE_IO_MINOR_VERSION. All items (except the new type) are
//	readable by older genies
//
//	If you make a major change to file format, bump up GENIE_IO_MAJOR_VERSION - this will stop
//	older genies attempting to read the file
//

// Intermediate file version numbers written by this Genie (i.e. if we create a new file)

static const int g3io_major = atoi( GENIE_IO_MAJOR_VERSION );
static const int g3io_minor = atoi( GENIE_IO_MINOR_VERSION );

#define G3IO_VERSION	GENIE_IO_MAJOR_VERSION "." GENIE_IO_MINOR_VERSION

#define SETPOS(__pos)	( xdr_setpos(m_xdrs_decode, (__pos)) && xdr_setpos(m_xdrs_encode, (__pos)) )

/* 
 * G3IO version history:
 *     1.0    Initial release (using XDR)	(December 1995) 	FAA
 */

#define G3_FILE_HEADER_BASE 	"Open Genie Intermediate file V"	/* present in all file versions */
static const char* g3_file_header = G3_FILE_HEADER_BASE G3IO_VERSION ;
static const u_int start_of_g3_file = strlen(g3_file_header) * sizeof(char);

int GenieIntermediateFile::numberOfBlocks() const
{
    return m_total_blocks;
}

int GenieIntermediateFile::open(const char* file_name, ModeFlags mode)
{
    m_file = file_name;
    m_mode = mode;
    const char* flags;
    String message;
    if (mode & Output)
    {
	flags = "wb+";
    }
    else if (mode & Append)
    {
	flags = "rb+";
    }
    else	// Must be "Input" then
    {
	flags = "rb";
    }
#ifdef __VMS
    if (strcmp(flags, "rb") == 0)
    {
        m_file_ptr = ::fopen(file_name, flags, "shr=put", "ctx=stm");
    }
    else
    {
        m_file_ptr = ::fopen(file_name, flags);
    }
#else
    m_file_ptr = ::fopen(file_name, flags);
#endif
    if (m_file_ptr == 0)
    {
	cat("Cannot open \"", file_name, message);
	cat(message, "\" - ", message);
	cat(message, strerror(errno), message);
	error_add("GenieIntermediateFile", message, "");
	return 0;
    }
    m_xdrs_encode = new XDR;
    m_xdrs_decode = new XDR;
    if (m_xdrs_encode == 0 || m_xdrs_decode == 0)
    {
	error_add("GenieIntermediateFile", "Not enough memory to create XDR streams", "");
	return 0;
    }
    xdrstdio_create(m_xdrs_encode, m_file_ptr, XDR_ENCODE);
    xdrstdio_create(m_xdrs_decode, m_file_ptr, XDR_DECODE);
// Now read or write a header and position us ready to read/write the first block
    if (mode & Output)
    {
	m_total_blocks = 0;
	m_major_version = g3io_major;
	m_minor_version = g3io_minor;
	if (!writeHeader(false))
	    return 0;
    }
    else
    {
	if (!readHeader(false))
	    return 0;
    }
    return buildBlockList();
}

char* GenieIntermediateFile::header()
{
    static char head[128];
    sprintf(head, "Genie intermediate file version %ld.%ld", (long)m_major_version, (long)m_minor_version);
    return head;
}

int GenieIntermediateFile::readBlock(GenieVariable& gv)
{
    return skipLabel() && readItem(gv);
}

int GenieIntermediateFile::readBlock(char*& label, GenieVariable& gv)
{
    return readLabel(label) && readItem(gv);
}

int GenieIntermediateFile::appendBlock(GenieVariable& gv, const char* label, const char* comment)
{
    return seek() && writeBlock(gv, label, comment);
}

int GenieIntermediateFile::appendBlock(GenieVariableBase& gvb, const char* label, const char* comment)
{
    return seek() && writeBlock(gvb, label, comment);
}

int GenieIntermediateFile::writeBlock(GenieVariable& gv, const char* label, const char* comment)
{
    return updateBlockList(label, gv.type()) && writeLabel(label) && writeItem(gv, comment) && writeHeader(true);
}

int GenieIntermediateFile::writeBlock(GenieVariableBase& gvb, const char* label, const char* comment)
{
    return updateBlockList(label, gvb.type()) && writeLabel(label) && writeItem(gvb, comment) && writeHeader(true);
}

int GenieIntermediateFile::skipLabel()
{
    return skipString();
}

int GenieIntermediateFile::readLabel(char*& label)
{
    return readString(label);    
}

int GenieIntermediateFile::writeLabel(const char* label)
{
    return writeString(label);    
}

int GenieIntermediateFile::seek(const char* label)
{
    int j, k;
    for(k=0; k<m_total_blocks; k++)
    {
	j = strcmp(label, block_list[k].label);
	if (j == 0)	// We've found it
	{
	    if (seek(k))
		return k;
	    else
	        return -1;
	}
    }
    return -1;
}

int GenieIntermediateFile::seek(int block)
{
    if (block == -1)
    {
	m_block = m_total_blocks;
    	if (fseek(m_file_ptr, 0, SEEK_END) != -1)
	{
	    return SETPOS(ftell(m_file_ptr));
	}
	else
	{
	    return 0;
	}
    }
    if (block >= m_total_blocks)
    {
	gerr << "IntermediateFile: seeking too far" << endl;
	return 0;
    }
    else
    {
        m_block = block;
        return SETPOS(block_list[block].offset);
    }
}

int GenieIntermediateFile::skipHeader()
{
    if (SETPOS(start_of_g3_file))
    {
    	return skipInteger() && skipInteger() && skipInteger() && skipInteger();
    }
    else
    {
	error_add("IntermediateFile::skipHeader()", "Failed to position at file start", "");
	return 0;
    }
}

int GenieIntermediateFile::readOffset(u_int& offset)
{
    g3io_int_t temp;
    if (readInteger(temp))
    {
        offset = temp;
	return 1;
    }
    else
    {
	offset = 0;
	return 0;
    }
}

// "saved_offset" is the position of the "offset" field, obtained from a prior call to
//  saveOffset(). The current position is calculated and written to the "saved_offset"
//  to allow later fast seeks
int GenieIntermediateFile::writeOffset(u_int saved_offset)
{
    u_int mark = getFileOffset();
    unsigned res = SETPOS(saved_offset);
    res &= writeInteger(mark);
    res &= SETPOS(mark);
    return res;
}

int GenieIntermediateFile::saveOffset(u_int& saved_offset)	
{
    saved_offset = getFileOffset();
    return writeInteger(0);
}

u_int GenieIntermediateFile::setFileOffset(u_int offset)
{
    return SETPOS(offset);
}

u_int GenieIntermediateFile::getFileOffset() 
{ 
    return xdr_getpos(m_xdrs_encode); 
}

int GenieIntermediateFile::skipOffset() 
{ 
    return skipInteger(); 
}

GenieVariableType GenieIntermediateFile::skipBlock()
{
    return (skipLabel() != 0 ? skipItem() : GenieTypeError);
}

// Assumes we have already read the "info" sections
GenieVariableType GenieIntermediateFile::skipItem()
{
    return (skipInfo() != 0 ? GenieVariable::skip(this) : GenieTypeError);
}

GenieVariableType GenieIntermediateFile::skipItem(char*& comment, char*& user, char*& date)
{
    return readInfo(comment, user, date) != 0 ? GenieVariable::skip(this) : GenieTypeError;
}

int GenieIntermediateFile::readItem(GenieVariable& gv, char*& comment, char*& user, char*& date)
{
    return readInfo(comment, user, date) && gv.read(this);
}

int GenieIntermediateFile::readItem(GenieVariable& gv)
{
    return (skipInfo() && gv.read(this));
}

int GenieIntermediateFile::writeItem(GenieVariableBase& gvb, const char* comment)
{
    GenieVariable gv(gvb);
    return writeItem(gv, comment);
}

int GenieIntermediateFile::writeItem(GenieVariable& gv, const char* comment)
{
    return writeInfo(comment) && gv.write(this);
}

int GenieIntermediateFile::writeHeader(bool restore_position)
{
//    char version[10];
    u_int mark(0);
    if (restore_position) 
    {
	mark = xdr_getpos(m_xdrs_encode);
    }
    if (!SETPOS(0))
    {
	error_add("GenieIntermediateFile::writeHeader()", "Failed to position at start of file", "");
	return 0;
    }
    fprintf(m_file_ptr, "%s%1x.%1x", G3_FILE_HEADER_BASE, (int)m_major_version, (int)m_minor_version);
    if(!writeInteger(m_major_version) || !writeInteger(m_minor_version) || 
	!writeInteger(m_total_blocks) || !writeInteger(m_bytes))
    {
	error_add("GenieIntermediateFile::writeHeader()", "Failed to write header to file", "");
	return 0;
    }
    if (restore_position)
    {
	if (!SETPOS(mark))
	{
	    error_add("GenieIntermediateFile::writeHeader()", "Failed to re-position stream", "");
	    return 0;
	}
    }
    else
    {
	m_block = 0;
    }
    return 1;
}

int GenieIntermediateFile::readHeader(bool restore_position)
{
    u_int mark(0);
    if (restore_position) 
    {
	mark = xdr_getpos(m_xdrs_decode);
    }
    if (!SETPOS(0))
    {
	error_add("GenieIntermediateFile::readHeader()", "Failed to position at start of file", "");
	return 0;
    }
    if (!SETPOS(start_of_g3_file))
    {
	error_add("GenieIntermediateFile::readHeader()", "Failed to seek", "");
	return 0;
    }
    if (!readInteger(m_major_version) || !readInteger(m_minor_version))
    {
	error_add("GenieIntermediateFile::readHeader()", "Failed to read header version", "");
	return 0;
    }
    if (m_major_version > g3io_major)
    {
	gerr << "This version of Genie is unable to read V" << m_major_version << "." 
	     << m_minor_version << " intermediate files" << endl;
	gerr << "as it only supports V" G3IO_VERSION " and earlier." << endl;
	gerr << "Please obtain an updated copy of Genie to read this file" << endl;
	return 0;
    }
    if ((m_major_version == g3io_major) && (m_minor_version > g3io_minor))
    {
	gerr << "The version number of this intermediate file (" << m_major_version << "." << m_minor_version << ") "
	     << "is more recent than that supported by Genie (" G3IO_VERSION ")" << endl;
 	gerr << "While it will probably still be read correctly, you should obtain an updated version of Genie" << endl;
    }
// If we write files more recent than the current one, update the header to say so
    if ((g3io_major >= m_major_version) && (g3io_minor > m_minor_version))
    {
	m_major_version = g3io_major;
	m_minor_version = g3io_minor;
    }
    if(!readInteger(m_total_blocks) || !readInteger(m_bytes))
    {
	error_add("GenieIntermediateFile::readHeader()", "Faled to read header block count", "");
	return 0;
    }
    if (restore_position)
    {
	if (!SETPOS(mark))
	{
	    error_add("GenieIntermediateFile::readHeader()", "Failed to re-position stream", "");
	    return 0;
	}
    }
    else
    {
	m_block = 0;
    }
    return 1;
}

int GenieIntermediateFile::close()
{
    if ( (m_mode & Output) || (m_mode & Append) )
    {
    	writeHeader(false);
    }
    if (m_xdrs_encode != 0)
    {
    	xdr_destroy(m_xdrs_encode);
    	delete m_xdrs_encode;
        m_xdrs_encode = 0;
    }
    if (m_xdrs_decode != 0)
    {
    	xdr_destroy(m_xdrs_decode);
    	delete m_xdrs_decode;
    	m_xdrs_decode = 0;
    }
    if (m_file_ptr != 0)
    {
    	::fclose(m_file_ptr);
	m_file_ptr = 0;
    }
    m_block = -1;
    freeBlockList();
    return 1;
}

int GenieIntermediateFile::readInfo(char*& comment, char*& user, char*& date)
{
    return readString(comment) && readString(user) && readString(date);
}

int GenieIntermediateFile::skipInfo()
{
    return skipString() && skipString() && skipString();
}

int GenieIntermediateFile::writeInfo(const char* comment)
{
// User supplied comment
    if (!writeString(comment))
	return 0;
// user who wrote file
    if (!writeString(creator()))
	return 0;
// Write out the date and time of the update
    if (!writeString(dateAndTime()))
	return 0;
    return 1;
}

int GenieIntermediateFile::skipString()
{
    char* dummy;
    return readString(dummy) && freeString(dummy);
}

int GenieIntermediateFile::freeString(char*& s)
{
    xdr_free((xdrproc_t)xdr_string, (char*)&s);
    s = 0;
    return 1;
}

int GenieIntermediateFile::readReal(g3io_real_t& r)
{
    if (xdr_double(m_xdrs_decode, &r))
    {
	return 1;
    }
    else
    {
        error_add("GenieIntermediateFile", "Failed to read real", "");
        return 0;
    }
}

int GenieIntermediateFile::readInteger(g3io_int_t& i)
{
    if (xdr_hyper(m_xdrs_decode, &i))
    {
	return 1;
    }
    else
    {
        error_add("GenieIntermediateFile", "Failed to read integer", "");
        return 0;
    }
}

int GenieIntermediateFile::skipInteger()
{
    g3io_int_t dummy;
    return readInteger(dummy);
}

int GenieIntermediateFile::skipReal()
{
    g3io_real_t dummy;
    return readReal(dummy);
}

int GenieIntermediateFile::writeReal(g3io_real_t r)
{
    if (xdr_double(m_xdrs_encode, &r))
    {
	return 1;
    }
    else
    {
        error_add("GenieIntermediateFile", "Failed to write real", "");
        return 0;
    }
}

int GenieIntermediateFile::writeInteger(g3io_int_t i)
{
    if (xdr_hyper(m_xdrs_encode, &i))
    {
	return 1;
    }
    else
    {
        error_add("GenieIntermediateFile", "Failed to write integer", "");
        return 0;
    }
}

int GenieIntermediateFile::readString(char*& s)
{
    s = 0;	// This will ask the routine to allocate space for us. Use freeString() later
    if (xdr_string(m_xdrs_decode, &s, UINT_MAX))
    {
	return 1;
    }
    else
    {
	error_add("GenieIntermediateFile", "Failed to read string", "");
	s = 0;
	return 0;
    }
}

int GenieIntermediateFile::writeString(const char* s)
{
    if (xdr_string(m_xdrs_encode, (char**)&s, strlen(s)))
    {
	return 1;
    }
    else
    {
	error_add("GenieIntermediateFile", "Failed to write string", "");
	return 0;
    }
}

int GenieIntermediateFile::freeBlockList()
{
    for(int i=0; i<block_list.count(); i++)
    {
	freeString(block_list[i].label);
    }
    block_list.reset();
    return 1;
}

int GenieIntermediateFile::buildBlockList()
{
    String message;
    block_item item;
    u_int mark = xdr_getpos(m_xdrs_decode);
    if (!SETPOS(0))
    {
	error_add("GenieIntermediateFile::buildBlockList()", "Failed to position at start of file", "");
	return 0;
    }
    freeBlockList();
    if (!skipHeader())
    {
	error_add("GenieIntermediateFile::buildBlockList()", "Failed to skip header", "");
	return 0;
    }
    for(int i=0; i<m_total_blocks; i++)
    {
        item.offset = xdr_getpos(m_xdrs_decode);
	readLabel(item.label);
	item.type = skipItem();
	block_list.push(item);
    }
// Add in marker for end (non-used block)
//    item.offset = xdr_getpos(m_xdrs_decode);
//    item.label = 0;
//    block_list.push(item);
// put us back where we were
    if (!SETPOS(mark))
    {
        error_add("GenieIntermediateFile", "Failed to re-position stream", "");
    	return 0;
    }
    if (block_list.count() != m_total_blocks)
    {
	cat("Block total errror: ", dec(block_list.count()), message);
	cat(message, "!=", message);
  	cat(message, dec((long)m_total_blocks), message);
	error_add("GenieIntermediateFile", message, "");
    	return 0;
    }
    return 1;
}

ostream& GenieIntermediateFile::print(ostream& os) const
{
    block_item item;
    u_int mark;
    GenieVariableType type;
    const char* label;
    os << "Total number of blocks in \"" << m_file << "\" = " << m_total_blocks << endl;
    os << "Current block = " << m_block << endl;
    for(int i=0; i<block_list.count(); i++)
    {
	item = block_list[i];
	mark = item.offset;
	type = item.type;
	label = (item.label != 0 ? item.label : "no_label");
	os << i+1 << " = " << label << " (type = " << GXVariableNames[type] << ", offset = " << mark << ")" << endl;
    }
    return os;
}

int GenieIntermediateFile::updateBlockList(const char* label, GenieVariableType type)
{
    block_item item;
    u_int mark = xdr_getpos(m_xdrs_encode);
    item.offset = mark;
    item.label = strcpy(new char[strlen(label)+1], label);
    item.type = type;
    if (m_block < block_list.count())
    {
	block_list[m_block] = item;
    }	
    else
    {
	block_list.push(item);
	m_total_blocks = block_list.count();
    }
    return 1;
}

int GenieIntermediateFile::skipItemType()
{
    return skipInteger();
}

GenieVariableType GenieIntermediateFile::readItemType()
{
    g3io_int_t type;
    return readInteger(type) ? (GenieVariableType)type : GenieUndefinedType;
}

int GenieIntermediateFile::skipItemSubclassType()
{
    return skipString();
}

int GenieIntermediateFile::readItemSubclassType(char*& subclass_name)
{
    return readString(subclass_name);
}

int GenieIntermediateFile::writeItemSubclassType(const char* subclass_name)
{
    return writeString(subclass_name);
}

int GenieIntermediateFile::writeItemType(GenieVariableType gvt)
{
    return writeInteger(gvt);		// Superclass type used for IO
}

int GenieIntermediateFile::readRealArray(GenieRealArray& gra)
{
    unsigned res = readArrayDims(gra);
    u_int length = gra.length();
    double* array = gra;
    res &= xdr_array(m_xdrs_decode, (char**)&array, &length, length, sizeof(double), (xdrproc_t)xdr_double);
    return res; 
}

int GenieIntermediateFile::readIntegerArray(GenieIntegerArray& gia)
{
    unsigned res = readArrayDims(gia);
    u_int length = gia.length();
    long* array = gia;
    res &= xdr_array(m_xdrs_decode, (char**)&array, &length, length, sizeof(long), (xdrproc_t)xdr_hyper);
    return res; 
}

int GenieIntermediateFile::writeRealArray(GenieRealArray& gra)
{
    unsigned res = writeArrayDims(gra);
    u_int length = gra.length();
    double* array = gra;
    res &= xdr_array(m_xdrs_encode, (char**)&array, &length, length, sizeof(double), (xdrproc_t)xdr_double);
    return res; 
}

int GenieIntermediateFile::writeIntegerArray(GenieIntegerArray& gia)
{
    unsigned res = writeArrayDims(gia);
    u_int length = gia.length();
    long *array = gia;
    res &= xdr_array(m_xdrs_encode, (char**)&array, &length, length, sizeof(long), (xdrproc_t)xdr_hyper);
    return res; 
}

// returns 0 on failure, 1 on success
int GenieIntermediateFile::writeArrayDims(const GenieArrayBase& gab)
{
    unsigned res(1);
    res &= writeInteger(gab.dimsSize());
    for(int i=1; i<=gab.dimsSize(); i++)
    {
	res &= writeInteger(gab.dims(i));
    }
    return res;
}

// return 0 on failure, array length on success
int GenieIntermediateFile::readArrayDims(GenieArrayBase& gab)
{
    unsigned res(1);
    g3io_int_t temp;
    res &= readInteger(temp);    
    int ndims = temp;
    int *dims_array = new int[ndims];
    for(int i=0; i<ndims; i++)
    {
	res &= readInteger(temp);
	dims_array[i] = temp;
    }
    gab.alloc(dims_array, ndims);
    delete[] dims_array;
    return res;
}

int GenieIntermediateFile::writeStringArray(GenieStringArray& gsa)
{
    const char* temp_string;
    unsigned res = writeArrayDims(gsa);
    for(int i=0; i<gsa.length(); i++)
    {
        temp_string = gsa[i];
        if (temp_string == 0)           // is the element undefined ?
        {
            res &= writeString("");
        }
        else
        {
            res &= writeString(temp_string);
        }
    }
    return res;
}

int GenieIntermediateFile::readStringArray(GenieStringArray& gsa)
{
    char* temp_string;
    unsigned res = readArrayDims(gsa);
    for(int i=0; i<gsa.length(); i++)
    {
        res &= readString(temp_string);
        gsa(i, temp_string);
        res &= freeString(temp_string);
    }
    return res; 
}

int GenieIntermediateFile::skipStringArray()
{
    GenieStringArray gsa;
    return readStringArray(gsa);
}

int GenieIntermediateFile::skipIntegerArray()
{ 
    GenieIntegerArray gia;
    return readIntegerArray(gia);
}

int GenieIntermediateFile::skipRealArray()
{ 
    GenieRealArray gra;
    return readRealArray(gra);
}

#if defined(__GNUG__)
template class SPList<block_item>;
template class SPListBase<block_item>;
#endif /* __GNUG__ */

#ifdef __DECCXX
#pragma define_template SPList<block_item>
#pragma define_template SPListBase<block_item>
#endif /* __DECCXX */

