/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "geniedefs.h"
#include <gnustring.h>

#ifdef _WIN32
#	include <windows.h>
#	include <strstrea.h>
#else
#	include <strstream.h>
#endif

#include "genie_ostream.h"


//  example usage
//
// main()
// {
//	 genie_ostream gout(cout, "(", ")"), d;
//	 d = cerr;
//
//	 d << "WOW! " << 3 << endl;
//	 gout << "Hello " << 456.9 << endl;
//
//	return 0;
// }

#define GENIEOSTREAM_BUFFER_LENGTH	10240		/* max characters to output */
#define GENIEOSTREAM_BUFFER_OFFSET	32		/* extra space for an error message */

// normal constructor (redefine type of start and stop on Win 32).
GENIEEXPORT genie_ostream::genie_ostream(ostream& s, int c, int screen_output, int log) :
	col(c), buffer_length(GENIEOSTREAM_BUFFER_LENGTH),
	buffer(new char[GENIEOSTREAM_BUFFER_LENGTH+GENIEOSTREAM_BUFFER_OFFSET]), sofar(buffer, buffer_length, ios::out), 
	logging(log), screen(screen_output)
{
    init(s.rdbuf());
}

// Allows assignment of ordinary iostreams if desired
GENIEEXPORT genie_ostream& genie_ostream::operator=(ostream& s)
{
    if (&s != this)
    {
	    init(s.rdbuf());
    }
    return *this;
}

GENIEEXPORT genie_ostream& genie_ostream::flush()
{
    if (logging)
    {
	//        this_genie->logit(str());
    }
    if (screen)
    {
	(void)str();			// bin ostrstream output
        ostream::flush();
    }
    return *this;
}


GENIEEXPORT const char* genie_ostream::str()
{
    sofar << ends;
// We may have a problem if the buffer overflowed - it will have
// stopped accepting characters, and have a bad stream status.
// Thus we must make sure we have a '\0' and set it to good again
    if (!sofar.good())
    {
	strcpy(buffer + buffer_length, "\n*** TRUNCATED ***\n");	// into space left by GENIEOSTREAM_BUFFER_OFFSET
        sofar.clear();			// reset stream state
    }
    sofar.seekp(ios::beg);
    buffer[GENIEOSTREAM_BUFFER_LENGTH+GENIEOSTREAM_BUFFER_OFFSET-1] = '\0';	// make sure we are always terminated
    return buffer;
}

GENIEEXPORT genie_ostream& genie_ostream::operator<<(ostream &(*a)(ostream &))
{ 
    if (screen) 
    	ostream::operator<<(a); 
    sofar << a; 
    return *this;
}

GENIEEXPORT genie_ostream& genie_ostream::operator<<(ios &(*a)(ios &))
{ 
    if (screen) 
	ostream::operator<<(a);
    sofar << a;   
    return *this; 
}

// This function is defined differently on WIN32 to other systems as the NT console
// does not translate ANSI sequences in the same way as Unix or VMS.
#ifdef _WIN32

	ostream& sc( ostream& os, int col )
	{
		
		static CONSOLE_SCREEN_BUFFER_INFO csbi;
		WORD wKeptBackground, wForeground = FOREGROUND_INTENSITY ; // Use intense colours always

		HANDLE hStdOut = GetStdHandle(STD_OUTPUT_HANDLE);

		// Find out and maintain the previous background colour.
		GetConsoleScreenBufferInfo(hStdOut, &csbi);
		wKeptBackground = csbi.wAttributes &
			(WORD)(BACKGROUND_RED|BACKGROUND_BLUE|BACKGROUND_GREEN|BACKGROUND_INTENSITY);

		// Set up the foreground colour
		switch (col)
		{
		case BLUE:
			wForeground |= FOREGROUND_BLUE;
			break;
		case RED:
			wForeground |= FOREGROUND_RED;
			break;
		case PINK:
			wForeground |= FOREGROUND_BLUE | FOREGROUND_RED;
			break;
		default: // NEUTRAL, NEUTRAL_OFF, BLUE_OFF, RED_OFF, PINK_OFF:
			wForeground |= 0;
		}

		os << flush;	// Ensure everything up to here is output in the correct colour
		SetConsoleTextAttribute(hStdOut, wKeptBackground | wForeground );
		
		return os;
	}

#else // Anything UNIX or VMS

	ostream& sc( ostream& os, int col )
	{

		switch (col)
		{
		case BLUE:
			os << "\033[34m\033[1m";
			break;
		case RED:
			os << "\033[31m\033[1m";
			break;
		case PINK:
			os << "\033[35m\033[1m";
			break;
		default: // NEUTRAL, NEUTRAL_OFF, BLUE_OFF, RED_OFF, PINK_OFF:
			os << "\033[0m";
		}

		return os;
	}

#endif

GENIEEXPORT OMANIP(int) setcolour( int col )
{
	return OMANIP(int) ( sc, col );
}
