/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

//
// $Id: genie_tmpfile.cxx,v 1.5 1998/05/05 10:32:32 faa Exp $
//
// genie_tmpfile - create a unique temporary file for GENIE
//
// Freddie Akeroyd, ISIS, 23/8/97
//
#include "geniedefs.h"
#include <stdio.h>	// for P_tmpdir
#include <stdlib.h>
#include <sys/stat.h>
#include <fcntl.h>

#ifdef HAVE_UNISTD_H
#	include <unistd.h>
#endif

#ifdef HAVE_IO_H
#	include <io.h>
#	include <process.h>	// for getpid on NT/WIN95
#ifdef _WIN32
//#	define S_ISDIR(m) ( (m)&(_S_IFMT) == (_S_IFDIR) )
#	define S_ISDIR(m) ( (m)&(_S_IFDIR) != 0 )
#endif
#endif

#include "genievars.h"
#include "genie_tmpfile.h"

#ifdef __VMS
#include <unixlib.h>
#endif /* __VMS */

static int is_directory(const char* file);

const char* genie_tmpfile(const char* prefix, const char* suffix, const char* directory)
{
    GenieString p(prefix), s(suffix), d(directory);
    return genie_tmpfile(p, s, d);
}

// Returns a pointer to static storage, NULL on error
const char* genie_tmpfile(const GenieString& prefix, 
		const GenieString& suffix, const GenieString& directory)
{
    static String file;
    String dir;
    char mix[20];
    static int file_count = 0;
    struct stat stat_buffer;
// Environment variable/logical name for temporary file directory
#ifdef __VMS
    const char* tmpdir = "SYS$SCRATCH";
#elif defined(_WIN32)
    const char* tmpdir = "TEMP";
#else
    const char* tmpdir = "TMPDIR";
#endif /* __VMS */
    if (directory.defined())
    {
	dir = directory;
    }
    else
    {
	if (getenv(tmpdir) != 0)
	{
	    dir = getenv(tmpdir);
	} 
#ifdef P_tmpdir
	else if (strlen(P_tmpdir) > 0)
	{
	    dir = P_tmpdir;
	}
#endif /* P_tmpdir */
	else if (getenv("HOME") != 0)
	{
	    dir = getenv("HOME");
	}
    }
    if (!is_directory(dir))
    {
	*g_err << "TMPFILE: \"" << dir << "\" is not a valid directory" << endl;
    }
// Generate a unique name - stat will return an error if the file does not exist
    do
    {
        sprintf(mix, "%ld%d", (long)getpid(), file_count);
        if (suffix.defined())
            file = dir + DIRECTORY_SEPARATOR + prefix + mix + "." + suffix;
        else
            file = dir + DIRECTORY_SEPARATOR + prefix + mix;
        file_count++;
    }
    while (stat(file, &stat_buffer) == 0);
// Check we can create files
    int fd = creat(file, 0700);
    if (fd == -1)
    {
	*g_err << "TMPFILE: Unable to write to directory \"" << dir << "\"" << endl;
	return 0;
    }
    else
    {
	close(fd);
	remove(file);
    }
    return file;
}

static const char* vms_file_name;
static int vms_file_type;

static int vms_to_action(char* file, int type)
{
    vms_file_type = type;
    return 1;
}

static int vms_from_action(char* file)
{
    vms_file_name = file;
    return 1;
}

// Return 1 for a valid directory, 0 on error or invalid specification
static int is_directory(const char* file)
{
#ifdef __VMS
// We usually get 1 returned from "from_vms", but if we get a 0 it could mean that
// we have been given a UNIX style name, so just pass it through
    if (decc$from_vms(file, vms_from_action, 0) == 0)
    {
	vms_file_name = file;
    }
    if (decc$to_vms(vms_file_name, vms_to_action, 0, 2) == 0)
    {
	*g_err << "TMPFILE: no match for \"" << vms_file_name << "\"" << endl;
	return 0;
    }
    if (vms_file_type != DECC$K_DIRECTORY)
    {
	return 0;
    }
    return 1;
#else
// Check we really have a directory i.e. st_mode & S_IFMT == S_IFDIR
    struct stat stat_buffer;
    int status = stat(file, &stat_buffer);
    if (status != 0)
    {
	if (status == -1)
	{
	    *g_err << "TMPFILE: stat: " << strerror(errno) << endl;
	}
	else if (status == -2)		// VMS only?
	{
	    *g_err << "TMPFILE: stat: privilege violation" << endl;
	}
	return 0;
    }
    if (!S_ISDIR(stat_buffer.st_mode))
    {
	gprintef("TMPFILE: \"%s\" has st_mode = 0%o\n", (const char*)file, stat_buffer.st_mode);
	return 0;
    }
    return 1;
#endif /* __VMS */
}
