/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "geniedefs.h"
#include <stdlib.h>
#include <stdio.h>
#include <iostream.h>
#include <string.h>
#include "genie_types.h"
#include "parse_funcs.h"


struct memalloc
{
    void*   loc;
    size_t  length;
}			alloc_list[400];

int	alloc_count = 0;
size_t	alloc_total_size = 0;


void *M_alloc( size_t len )
{
    static char buffer[SHORTCHARBUFF];
    void *vp;

    if (len == 0) len = 1;

    if ( (vp = malloc(len)) )
    {
	alloc_list[alloc_count].loc = vp;
	alloc_list[alloc_count].length = len;
	alloc_count++;
	if (alloc_count == 400)
	{
	    cout << "EXPLODED" << endl;
	    exit(1);
	}
    }
    else
	gerr << ":Malloc failed while attempting to allocate" << len << " bytes!!" << endl;

    alloc_total_size += len;
    return vp;
}



size_t M_freeup()
{
    size_t i, temp;

    while ( alloc_count > 0 )
    {
	alloc_count--;
	for ( i=0; i < alloc_list[alloc_count].length; i++ )
	    ( (char *) alloc_list[alloc_count].loc ) [i] = '$';
	free( alloc_list[alloc_count].loc );
    }

    temp = alloc_total_size;
    alloc_total_size = 0;

    return temp;
}



/*
 * Construct an integer value
 */
g_variable newGinteger(long val){
        g_variable temp;
        temp.its_type = _Integer;
        temp.it.g_i = val;
        return temp;
}


/*
 * Construct a real value
 */
g_variable newGreal(double val){
        g_variable temp;
        temp.its_type = _Real;
        temp.it.g_r = val;
        return temp;
}


/*
 * Allocate store for a string
 */
g_variable newGstring(const char* val, int len){
        g_variable temp;
        temp.its_type = _String;		    /*   v - for the terminating '\0' */
	if ( ( temp.it.g_s= (g_string)M_alloc(len+1) ) == 0) PERROR("newGstring(M_alloc failed)");
        strncpy(temp.it.g_s, val, len);
	temp.it.g_s[len] = '\0';		/* Don't forget the terminating null */
        return temp;
}


/*
 * Allocate store for a workspace
 */
g_variable newGworkspace(int num_elems){	/* M_allocte one more element than asked for to hold a terminating Bel */

        g_variable temp;
	int i;

        temp.its_type = _Workspace;
	if ( ( temp.it.g_w = (g_workspace)M_alloc(sizeof(struct g_workspace_struct)) ) == 0)
	    PERROR("newGworkspace(M_alloc failed)");
	if ( ( temp.it.g_w->elems = (work_element *)M_alloc((num_elems)*sizeof(work_element)) ) == 0)
	    PERROR("newGworkspace(M_alloc failed)");
	temp.it.g_w->len = num_elems;
	for(i=0; i <= num_elems-1; i++)
	    temp.it.g_w->elems[i].name[0] = '\007';	/* Mark all of the elements as free (just for high water marking) */
        return temp;
}


/*
 * Allocate store for an array (Real, Integer, String or Workspace)
 */
g_variable newGarrayReal(int num_elems){
	g_variable temp;
	temp.its_type = _RealArray;
	if ( ( temp.it.g_r_a= (g_arr_real) M_alloc(sizeof(struct g_arr_real_struct)) ) == 0)
	    PERROR("newGarrayReal(M_alloc failed)");	
	if ( ( temp.it.g_r_a->values= (g_real *) M_alloc(sizeof(g_real)*num_elems) ) == 0)
	    PERROR("newGarrayReal(M_alloc failed)");	
	return temp;
}

g_variable newGarrayInteger(int num_elems){
	g_variable temp;
	temp.its_type = _IntegerArray;
	if ( ( temp.it.g_i_a= (g_arr_integer) M_alloc(sizeof(struct g_arr_integer_struct)) ) == 0)
	    PERROR("newGarrayInteger(M_alloc failed)");	
	if ( ( temp.it.g_i_a->values= (g_integer *) M_alloc(sizeof(g_integer)*num_elems) ) == 0)
	    PERROR("newGarrayInteger(M_alloc failed)");	
	return temp;
}

g_variable newGarrayString(int num_elems){
	g_variable temp;
	temp.its_type = _StringArray;
	if ( ( temp.it.g_s_a= (g_arr_string) M_alloc(sizeof(struct g_arr_string_struct)) ) == 0)
	    PERROR("newGarrayString(M_alloc failed)");	
	if ( ( temp.it.g_s_a->values= (g_string *) M_alloc(sizeof(g_string)*num_elems) ) == 0)
	    PERROR("newGarrayString(M_alloc failed)");	
	return temp;
}

g_variable newGarrayWorkspace(int num_elems){
	g_variable temp;
	temp.its_type = _WorkspaceArray;
	if ( ( temp.it.g_w_a= (g_arr_workspace) M_alloc(sizeof(struct g_arr_workspace_struct)) ) == 0)
	    PERROR("newGarrayWorkspace(M_alloc failed)");	
	if ( ( temp.it.g_w_a->values= (g_workspace *) M_alloc(sizeof(g_workspace)*num_elems) ) == 0)
	    PERROR("newGarrayWorkspace(M_alloc failed)");	
	return temp;
}



/*
 * find the number of elements currently used in the workspace, slightly analogous to strlen.
 */
int wrklenp(g_workspace workspace) {						    /* Length of a workspace */
        int i;
        for( i=0; (workspace->elems[i].name[0] != '\007') && (i <= workspace->len-1); i++);
        return i;
}



/*
 * Add one field and element pair to a workspace
 */
int addToWorkspace(g_variable workspace, const char *field, g_variable value){	    /* Value is passed by reference */

	int len;

	if ( workspace.its_type == _Workspace ) {
	    len = wrklen(workspace);						    /* Same sort of use as strlen() */
	    if ( len != workspaceLength(workspace) ){
		if ( findElementIndex(workspace,field,-1) == -1 ) {
		    strcpy(workspace.it.g_w->elems[len].name, field);
		    workspace.it.g_w->elems[len].value = value;
		    return 0;							/* Sucess return */
		} else
		    PERROR("addToWorkspace(field name duplicated)");
	    } else
		PERROR("addToWorkspace(workspace is full)");
	} else
	    PERROR("addToWorkspace(argument not a workspace)");
	return -1;								/* Failure return */
}



/*
 * Look up the index of an element by searching for the field name
 */
int findElementIndex( g_variable workspace, const char *field, int failflag ){
    int i;
    for( i=0; i <= wrklen(workspace)-1; i++ )
	if ( GStrcasecmp(field, workspace.it.g_w->elems[i].name) == 0 )
	    return i;

    if ( failflag != -1 ) PERROR("findElementIndex(Element not found)");
    return failflag;
}



/*
 * Counted string duplication, allocating storage for the new string (caller frees store)
 */
char *strndup(const char *string, int length) {

    char *cp;

    cp = (char*)M_alloc(length+1);
    strncpy(cp, string, length);
    cp[length] = '\0';
    return cp;
}



/*
 * Return a duplicate of the input workspace (note that this operates on the workspace pointer, not the whole g_variable)
 * By setting len short, the resultant workspace can be shortened or visa-versa.
 */
g_workspace wrkndup( g_workspace work, int len ) {

    int i;
    g_variable temp_work1, temp_work2;

    putWorkspace(temp_work1, work);
    temp_work2 = newGworkspace(len);
    for( i=0; i <= len-1; i++ )
	
	addToWorkspace(temp_work2,
	    getFieldByIndex(temp_work1, i),
	    getElementByIndex(temp_work1, i)
	    );

    return getWorkspace(temp_work2);

}



/*
 * Returns the number of elements in a genie array
 */

long ArrayLength( g_variable array ) {

    long i, len=1;

    switch ( variableType(array) ) {

    case _IntegerArray:
    case _RealArray:
    case _StringArray:
    case _WorkspaceArray:
	for( i=0; arrayGetDims(array)[i] != 0; i++ )
	    len=arrayGetDims(array)[i]*len;
	return len;
    default:
	PERROR("ArrayLength");
	return -1L;
    }
}



/*
 * This one is not running as the most beautiful example of 'C' coding ever beheld by
 * mankind!  However, it does produce a reliable representation of any form of (valid)
 * genie 'C' Variable you care to throw at it.  To make it flexible it uses buffer
 * space which is given to it and will return a result pointer to it.
 *
 *	    char plenty_buffer[LOAD_OF_BYTES];
 *	    printf("My genie variable:\n%s", printGvar(plenty_buffer, Avar) );
 *
 * yes, yes, I know the macros should be nicely coded into procedures, will do sometime!
 */


/*** MACROS just reduce the mess a little ***/

#define PRINTDIMS \
/*** print dims vector (and calculate total elements) ***/ \
sprintf(buff1, "%sdims[0..n]= [", padding); \
strcat(outbuff, buff1); \
len=1; \
for( i=0; arrayGetDims(var)[i] != 0; i++ ) { \
    sprintf(buff1, " %ld ", arrayGetDims(var)[i]); \
    strcat(outbuff, buff1); \
    len=arrayGetDims(var)[i]*len; \
} \
sprintf(buff1, "]\n"); \
strcat(outbuff, buff1)


#define PRINTDOPE \
/*** print dope vector ***/ \
sprintf(buff1, "%sdope[0..n]= [",padding); \
strcat(outbuff, buff1); \
for( i=0; arrayGetDope(var)[i] != 0; i++ ) { \
    sprintf(buff1, " %ld ", arrayGetDope(var)[i]); \
    strcat(outbuff, buff1); \
} \
sprintf(buff1, "]\n"); \
strcat(outbuff, buff1)

/*** for indenting to make output a little more readable ***/
#define PADUP	(strcat(padding, "    "))
#define PADDOWN	(padding[strlen(padding)-4]='\0')

char* printGvar(char outbuff[], g_variable var) {

	char buff1[LONGCHARBUFF], buff2[LONGCHARBUFF];
	static char padding[SHORTCHARBUFF] = "";
	int i, len;
	g_variable tgv;
	long ltmp;
	double dtmp;

	switch(variableType(var)) {

	case _Integer:
	    ltmp = getInteger(var);
	    if ( ltmp == LONG_MAX )
		sprintf(outbuff, "(undefined)");
	    else
		sprintf(outbuff, "%ld", ltmp);
	    return outbuff;
	case _Real:
	    dtmp = getReal(var);
	    if ( dtmp == DBL_MAX )
		sprintf(outbuff, "(undefined)");
	    else
		sprintf(outbuff, "%f", dtmp);
	    return outbuff;
	case _String:
	    sprintf(outbuff, "%s", getString(var) );
	    return outbuff;
	case _Workspace:
	    PADUP;
	    sprintf(outbuff, "\n%sWorkspace\n%s=========", padding, padding);
	    PADUP;
	    for( i=0; i <= wrklen(var)-1; i++ ) {
		sprintf(buff2, "\n%s(%s)\t= (%s)",
			padding, getFieldByIndex(var, i), printGvar(buff1, getElementByIndex(var, i)) );
		strcat(outbuff, buff2);
		}
	    PADDOWN;
	    PADDOWN;
	    return outbuff;
	case _IntegerArray:
	    PADUP;
	    sprintf(outbuff, "\n%sInteger array\n%s=============\n", padding, padding);
	    PRINTDIMS;
	    PRINTDOPE;
	    PADUP;
	    strcat(outbuff, padding);
	    /*** print out a sample of the array contents ***/
	    for( i=0; i <= MIN(len-1, 5); i++ ) {
		ltmp = arrayGetInteger(var, i);
		if ( ltmp == LONG_MAX )
		    sprintf(buff1, "(undefined)\t");
		else
		    sprintf(buff1, "%ld\t", ltmp);
		strcat(outbuff, buff1);
	    }
	    sprintf(buff1, " ...(%d elements)  ", len);
	    PADDOWN;
	    PADDOWN;
	    strcat(outbuff, buff1);
	    return outbuff;
	case _RealArray:
	    PADUP;
	    sprintf(outbuff, "\n%sReal array\n%s===========\n", padding, padding);
	    PRINTDIMS;
	    PRINTDOPE;
	    PADUP;
	    strcat(outbuff, padding);
	    /*** print out a sample of the array contents ***/
	    for( i=0; i <= MIN(len-1, 5); i++ ) {
		dtmp = arrayGetReal(var, i);
		if ( dtmp == DBL_MAX )
		    sprintf(buff1, "(undefined)\t");
		else
		    sprintf(buff1, "%f\t", dtmp);
		strcat(outbuff, buff1);
	    }
	    sprintf(buff1, " ...(%d elements)  ", len);
	    PADDOWN;
	    PADDOWN;
	    strcat(outbuff, buff1);
	    return outbuff;
	case _StringArray:
	    PADUP;
	    sprintf(outbuff, "\n%sString array\n%s============\n", padding, padding);
	    PRINTDIMS;
	    PRINTDOPE;
	    PADUP;
	    strcat(outbuff, padding);
	    /*** print out a sample of the array contents ***/
	    for( i=0; i <= MIN(len-1, 5); i++ ) {
		sprintf(buff1, "\"%s\"\t", arrayGetString(var, i));
		strcat(outbuff, buff1);
	    }
	    sprintf(buff1, " ...(%d elements)  ", len);
	    PADDOWN;
	    PADDOWN;
	    strcat(outbuff, buff1);
	    return outbuff;
	case _WorkspaceArray:
	    PADUP;
	    sprintf(outbuff, "\n%sWorkspace array\n%s===============\n", padding, padding);
	    PRINTDIMS;
	    PRINTDOPE;
	    /*** print out a sample of the array contents ***/
	    for( i=0; i <= MIN(len-1, 5); i++ ) {
		putWorkspace(tgv,arrayGetWorkspace(var, i));
		printGvar(buff1, tgv);
		strcat(outbuff, buff1);
	    }
	    sprintf(buff1, "%s ...(%d elements)  ",padding, len);
	    PADDOWN;
	    strcat(outbuff, buff1);
	    return outbuff;
	case _Undefined:
	    sprintf(outbuff, "<Undefined>");
	    return outbuff;
	default:
	    PERROR("printGvar(Unprintable type)");
	    strcpy(outbuff,"");
	    return outbuff;
	}
}
