/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

/*
 *      Module:         genie_var
 *      Author:         Freddie Akeroyd, ISIS
 *
 */
#include "geniedefs.h"
#include "genievars.h"
#include "g3io_types.h"
#include "intermediate_file.h"

static const char* rcsid = "@(#)$Id: genie_var.cxx,v 1.31 1997/10/22 09:14:20 cmm Exp $";
USE_RCSID(rcsid);

OOP GenieVariable::value() const 
{
    return pointerTypeC() != NULL ? pointerTypeC()->value() : nilOOP;
}

int GenieVariable::defined() const 
{ 
    return pointerTypeC() != NULL ? pointerTypeC()->defined() : 0;
}

GenieVariable::GenieVariable(const GenieCalloutList& gcl, const char* name) :
	m_params(nilOOP), m_location(0), m_type(GenieUndefinedType) 
{
    gu = 0;
    associate(gcl, name);
}

GenieVariable::GenieVariable(GenieVariableType gvt) : m_params(nilOOP), m_location(0)
{ 
    gu = 0;
    newType(gvt); 
}

GenieVariableType GenieVariable::type() const
{
    return m_type;
}

void GenieVariable::freeType()
{
    delete pointerType();
    gu = NULL;
    m_type = GenieUndefinedType;
    return;
}

void GenieVariable::newType(const char* subclass_type, OOP value_oop)
{
    OOP subclass_oop = typeNameToOOP(subclass_type);
    newType(getGenieVariableTypeFromClass(subclass_oop), value_oop, subclass_oop);
}

void GenieVariable::newType(GenieVariableType type, OOP value_oop, OOP subclass_oop)
{
    switch(type)
    {
	case GenieIntegerType:
	    gi = new GenieInteger(value_oop);
	    break;

	case GenieRealType:
	    gr = new GenieReal(value_oop);
            break;

	case GenieStringType:
	    gs = new GenieString(value_oop);
	    break;

	case GenieWorkspaceType:
	    gw = new GenieWorkspace(value_oop, subclass_oop);
	    break;

	case GenieRealArrayType:
	    gra = new GenieRealArray(value_oop);
	    break;

	case GenieIntegerArrayType:
	    gia = new GenieIntegerArray(value_oop);
	    break;

	case GenieStringArrayType:
	    gsa = new GenieStringArray(value_oop);
	    break;

	case GenieWorkspaceArrayType:
	    gwa = new GenieWorkspaceArray(value_oop);
//	    gwa = new GenieWorkspaceArray(value_oop, subclass_oop);
	    break;

	case GenieUndefinedType:
	    gu = NULL;
	    break;

	case GenieUndefinedArrayType:
	    gu = NULL;
	    break;

	case GenieBaseType:
	    gvb = new GenieVariableBase(value_oop);
	    break;

	default:
	    gerr << "newType: setting invalid type = " << (int)type << endl;
	    type = GenieUndefinedType;
	    gu = NULL;
	    break;
    }
    m_type = type;
// Now restore any previous associations with Smalltalk variables
    if (gu != NULL) 
    { 
	pointerType()->init(value(), m_params, m_location);
    }
    return;
}

void GenieVariable::checkType(GenieVariableType type) const
{
    if (m_type != type)
    {
	    gerr << "checkType: " << GXVariableNames[m_type] << " not same as " << GXVariableNames[type] << endl;
    }
    return;
}

void GenieVariable::checkType(GenieVariableType type)
{
    if (m_type != type)
    {
	if (gu != NULL && pointerTypeC()->value() != nilOOP)
	{
	    gerr << "checkType: changing from " << GXVariableNames[m_type] << " to " << GXVariableNames[type] << endl;
	}
	freeType();
	newType(type);
    }
    return;
}

GenieVariable::operator GenieString&()
{
    checkType(GenieStringType);
    return *gs;
}

GenieVariable::operator GenieWorkspace&()
{
    checkType(GenieWorkspaceType);
    return *gw;
}

GenieVariable::operator GenieWorkspaceArray&()
{
    checkType(GenieWorkspaceArrayType);
    return *gwa;
}

GenieVariable::operator GenieStringArray*()
{
    checkType(GenieStringArrayType);
    return gsa;
}

GenieVariable::operator GenieStringArray&()
{
    checkType(GenieStringArrayType);
    return *gsa;
}

GenieVariable::operator GenieInteger&()
{
    checkType(GenieIntegerType);
    return *gi;
}

GenieVariable::operator GenieIntegerArray&()
{
    checkType(GenieIntegerArrayType);
    return *gia;
}

GenieVariable::operator GenieIntegerArray*()
{
    checkType(GenieIntegerArrayType);
    return gia;
}

GenieVariable::operator GenieReal&()
{
    checkType(GenieRealType);
    return *gr;
}

GenieVariable::operator GenieRealArray&()
{
    checkType(GenieRealArrayType);
    return *gra;
}

GenieVariable::operator GenieRealArray*()
{
    checkType(GenieRealArrayType);
    return gra;
}

///*GenieVariable::operator const GenieString&() const
//{
//    checkType(GenieStringType);
//    return *gs;
//}*/
//
///*GenieVariable::operator const GenieWorkspace&() const
//{
//    checkType(GenieWorkspaceType);
//    return *gw;
//}*/
//
///*GenieVariable::operator const GenieWorkspaceArray&() const
//{
//    checkType(GenieWorkspaceArrayType);
//    return *gwa;
//}*/
//
///*ieVariable::operator const GenieStringArray*() const
//{
//    checkType(GenieStringArrayType);
//    return gsa;
//}*/
//
///*GenieVariable::operator const GenieStringArray&() const
//{
//    checkType(GenieStringArrayType);
//    return *gsa;
//}*/
//
///*GenieVariable::operator const GenieInteger&() const
//{
//    checkType(GenieIntegerType);
//    return *gi;
//}*/
//
///*GenieVariable::operator const GenieIntegerArray&() const
//{
//    checkType(GenieIntegerArrayType);
//    return *gia;
//}*/
//
///*GenieVariable::operator const GenieIntegerArray*() const
//{
//    checkType(GenieIntegerArrayType);
//    return gia;
//}*/
//
///*GenieVariable::operator const GenieReal&() const
//{
//    checkType(GenieRealType);
//    return *gr;
//}
//
//GenieVariable::operator const GenieRealArray&() const
//{
//    checkType(GenieRealArrayType);
//    return *gra;
//}
//
//GenieVariable::operator const GenieRealArray*() const
//{
//    checkType(GenieRealArrayType);
//    return gra;
//}*/

//GenieVariable::operator g_integer()
//{
//    checkType(GenieIntegerType);
//    return (g_integer)(*gi);
//}

//GenieVariable::operator g_real()
//{
//    checkType(GenieRealType);
//    return (g_real)(*gr);
//}

//GenieVariable::operator c_g_string()
//{
//    checkType(GenieStringType);
//    return (c_g_string)(*gs);
//}

GenieVariableBase* GenieVariable::pointerType()
{
    switch(m_type)
    {
	case GenieIntegerType:
	    return gi;
	    break;

	case GenieIntegerArrayType:
	    return gia;
	    break;

	case GenieRealType:
	    return gr;
            break;

	case GenieRealArrayType:
	    return gra;
            break;

	case GenieWorkspaceType:
	    return gw;
	    break;

	case GenieWorkspaceArrayType:
	    return gwa;
            break;

	case GenieStringType:
	    return gs;
            break;

	case GenieStringArrayType:
	    return gsa;
            break;

	case GenieUndefinedType:
	    break;

	case GenieUndefinedArrayType:
	    break;

	case GenieBaseType:
	    return gvb;

	default:
	    gerr << "pointerType: returning invalid type = " << (int)m_type << endl;
	    break;
    }
    return NULL;
}

const GenieVariableBase* GenieVariable::pointerTypeC() const
{
    switch(m_type)
    {
	case GenieIntegerType:
	    return gi;
	    break;

	case GenieIntegerArrayType:
	    return gia;
	    break;

	case GenieRealType:
	    return gr;
            break;

	case GenieRealArrayType:
	    return gra;
            break;

	case GenieStringType:
	    return gs;
            break;

	case GenieStringArrayType:
	    return gsa;
            break;

	case GenieWorkspaceType:
	    return gw;
	    break;

	case GenieWorkspaceArrayType:
	    return gwa;
            break;

	case GenieUndefinedType:
	    break;

	case GenieUndefinedArrayType:
	    break;

	case GenieBaseType:
	    return gvb;

	default:
	    gerr << "pointerType: returning invalid type = " << (int)m_type << endl;
	    break;
    }
    return NULL;
}

OOP GenieVariable::become(OOP oop)
{
    freeType();
    if (oop == nilOOP)
    {
	newType(GenieUndefinedType);
    }
    else
    {
    	newType(getGenieVariableType(oop), oop);
    }
    return oop;
}

GenieVariable::GenieVariable(OOP oop)
{
    m_location = 0;
    m_params = nilOOP;
    m_type = GenieUndefinedType;
    gu = 0;
    become(oop);
}

GenieVariable::~GenieVariable()
{
    delete pointerType();
}

ostream& operator<<(ostream& os, const GenieVariable& gv)
{
    return gv.print(os);
}

ostream& GenieVariable::print(ostream& os, const char* note) const
{
    if (note != NULL)
    {
        os << "Generic Genie Variable containing a " << GXVariableNames[m_type] << endl;
    }
    if (m_type == GenieUndefinedType || m_type == GenieUndefinedArrayType)
    {
	return os << "<Undefined>";
    }
    else
    {
        return pointerTypeC()->print(os, note);
    }
}
        
OOP GenieVariable::inspect()
{
    return gu != 0 ? pointerType()->inspect() : nilOOP;
}

GenieVariable::GenieVariable(const GenieVariableBase& gvb)
{
    newType(getGenieVariableType(gvb.value()), gvb.value());    
}

GenieVariable::GenieVariable(const GenieVariable& gv)
{
    newType(gv.m_type, gv.value());    
}

GenieVariable& GenieVariable::operator=(const GenieVariable& gv)
{
    if (this->gu != gv.gu)	// I can't just use (this != &gv) as i've overloaded the & operator ...
    {
	freeType();
    	newType(gv.m_type, gv.value());    
    }
    return *this;    
}

GenieVariable& GenieVariable::operator=(const GenieVariableBase& gvb)
{
    OOP val(gvb.value());
    freeType();
    newType(getGenieVariableType(val), val);    
    return *this;    
}

OOP GenieVariable::deepCopyOOP(OOP oop)
{
    GenieVariableType type = getGenieVariableType(oop);
    switch(type)
    {
	case GenieIntegerType:
	    return GenieInteger::deepCopyOOP(oop);

 	default:
	    gerr << "GenieVariable: no deep copy defined" << endl;
	    return nilOOP;
    }
}

int GenieVariable::associate(const GenieCalloutList& gcl, const char* name)
{
    freeType();
    newType(GenieBaseType, nilOOP);
    pointerType()->associate(gcl, name);
// now keep values in case we change type
    m_location = pointerType()->location();
    m_params = pointerType()->params();
    OOP val = value();
    freeType();
    newType(getGenieVariableType(val), val);
    return 0;
}

int GenieVariable::associate(OOP parameters, const char* name)
{
    freeType();
    newType(GenieBaseType, nilOOP);
    pointerType()->associate(parameters, name);
// now keep values in case we change type
    m_location = pointerType()->location();
    m_params = pointerType()->params();
    OOP val = value();
    freeType();
    newType(getGenieVariableType(val), val);    
    return 0;
}

// Reading and writing handled by functions in the contained variable
// All we have to do is write out the type field
GenieVariableType GenieVariable::read(IntermediateFile* gif)
{
    char* subclass_name;
    freeType();
    gif->skipItemType();	// We'll use the proper subclass type instead, though we should end up being this type
    unsigned res = gif->readItemSubclassType(subclass_name);
    newType(subclass_name);
    res &= gif->freeString(subclass_name);
    res &= gif->skipOffset();
// Don't use defined() in test as we are currently nilOOP!
    GenieVariableType item_type = (gu != 0 ? pointerType()->read(gif) : GenieUndefinedType);
    res &= (item_type != 0 ? 1 : 0);
    return res != 0 ? item_type : GenieTypeError;
}

GenieVariableType GenieVariable::write(IntermediateFile* gif)
{
    u_int mark;
    GenieVariableType item_type;
    unsigned res = gif->writeItemType(m_type);
    res &= gif->writeItemSubclassType(getClassString(value()));
    res &= gif->saveOffset(mark);
    item_type = (defined() ? pointerType()->write(gif) : GenieUndefinedType);
    res &= (item_type != 0 ? 1 : 0);
    res &= gif->writeOffset(mark);
    return res != 0 ? item_type : GenieTypeError;
}

GenieVariableType GenieVariable::skip(IntermediateFile* gif)
{
    u_int mark;
    GenieVariableType item_type = gif->readItemType();
    unsigned res = gif->skipItemSubclassType();
    res &= gif->readOffset(mark);
    res &= gif->setFileOffset(mark);
    return item_type;
}

GenieVariableType GenieVariable::writeAscii(fstream& fs, int indent)
{
    unsigned res;
//    fs << m_type << " " << getClassString(value()) << "\n";
    fs << setindent(indent) << getClassString(value()) << "\n";
    GenieVariableType item_type = (defined() ? pointerType()->writeAscii(fs, indent) : GenieUndefinedType);
    res = (item_type != 0 ? 1 : 0);
    return res != 0 ? item_type : GenieTypeError;
}

GenieVariableType GenieVariable::readAscii(fstream& fs)
{
    unsigned res;
    freeType();
//    int type;
//    fs >> type;
    String subclass_name;
    fs >> subclass_name;
    newType(subclass_name);
    GenieVariableType item_type = (gu != 0 ? pointerType()->readAscii(fs) : GenieUndefinedType);
    res = (item_type != 0 ? 1 : 0);
    return res != 0 ? item_type : GenieTypeError;
}

//GenieVariable::operator OOP() const
//{
//	return value();
//}

GenieVariable::GenieVariable(OOP parameters, int location)
{
    m_location = 0;
    m_params = nilOOP;
    m_type = GenieUndefinedType;
    gu = 0;
    become(arrayAt(parameters, location));
}
