/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "geniedefs.h"
#include <iostream.h>

#ifdef _WIN32
#	include <time.h>
#	include <float.h>	/* for _fpreset() */
#	include <strstrea.h>
#else
#	include <strstream.h>
#endif

#ifdef HAVE_UNISTD_H
#	include <unistd.h>
#endif

#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include "parsedefs.h"

#ifndef SIGCHLD
#define SIGCHLD		20	/* doesn't exist of VMS prior to 7.0 */
#endif

#include <setjmp.h>
#include "gsignals.h"

/*** MODULE GLOBAL ***/
sigjmp_buf gsignals_context;			    /* Saved context for longjmps ( not used until set other than 0 ) */
volatile int jumpsigs = 0;		    /* Flag set to TRUE if longjumping signals are used */

//
//  Minor differences between VMS/POSIX and ULTRIX.
//
#if !defined(_POSIX_SOURCE) && !defined(_WIN32)
#   undef SIG_IGN
#   undef SIG_ERR
#   undef SIG_DFL
#   define SIG_ERR         ((void (*)(int))(-1))
#   define SIG_DFL         ((void (*)(int))( 0))
#   define SIG_IGN         ((void (*)(int))( 1))
    typedef char *  caddr_t;	        	    // somehow avoids getting defined !!
#endif

volatile signal_handler	oldCtrlCaction = SIG_IGN;		    /* Used to store the original Ctrl-C handler from Smalltalk */

/* Routines for internal use only */
static void	Enable_signal(int sig, signal_handler handler);
static void	Disable_signal(int sig);
static void	Default_signal(int sig);

static void FpeInt(SIGTYPE) { Unset_signals(); if (jumpsigs) siglongjmp(gsignals_context, SIGFPE);  else sigfpe_action(); return; }
static void CtrlCInt(SIGTYPE) { Unset_signals(); if (jumpsigs) siglongjmp(gsignals_context, SIGINT);  else sigint_action(); return; }
static void SegvInt(SIGTYPE)  { Unset_signals(); if (jumpsigs) siglongjmp(gsignals_context, SIGSEGV); else sigother_action(SIGSEGV); return; }

#ifndef _WIN32

	static void BusInt(SIGTYPE)   { Unset_signals(); if (jumpsigs) siglongjmp(gsignals_context, SIGBUS);  else sigother_action(SIGBUS);  return; }
	static void HupInt(SIGTYPE)   { Unset_signals(); if (jumpsigs) siglongjmp(gsignals_context, SIGHUP);  else sigother_action(SIGHUP);	 return; }

#	ifndef __linux__
		static void SysInt(SIGTYPE)   { Unset_signals(); if (jumpsigs) siglongjmp(gsignals_context, SIGSYS);  else sigother_action(SIGSYS);  return; }
#	endif /* __linux__ */

	static void ChildInt(SIGTYPE) { Unset_signals(); if (jumpsigs) siglongjmp(gsignals_context, SIGCHLD); else sigother_action(SIGCHLD); return; }

#endif /* _WIN32 */

static void TermInt(SIGTYPE)  { Unset_signals(); if (jumpsigs) siglongjmp(gsignals_context, SIGTERM); else sigterm_action();	 return; }

#ifdef __ultrix
extern "C" int siginterrupt(int, int);
#endif

static const char* rcsid = "@(#)$Id: gsignals.cxx,v 1.39 1998/10/16 13:18:49 faa Exp $";
USE_RCSID(rcsid);


void Set_signals()
{
  if (getenv("GENIE_DEBUG") == NULL)
  {
    Enable_signal( SIGINT, CtrlCInt );
    Enable_signal( SIGSEGV, SegvInt );
    Enable_signal( SIGFPE, FpeInt );

#ifndef _WIN32

    Enable_signal( SIGBUS, BusInt);
    Enable_signal( SIGHUP, HupInt);

#   ifndef __linux__
	Enable_signal( SIGSYS, SysInt);
#   endif

    Disable_signal( SIGCHLD );		// We are now one process - only PGPLOT will create this

#endif /* _WIN32 */

//    Enable_signal( SIGTERM, TermInt);
  }

    return;
}


void Unset_signals()
{
  if (getenv("GENIE_DEBUG") == NULL)
  {
    Disable_signal( SIGINT );
    Disable_signal( SIGFPE );

#ifndef _WIN32

#   ifndef __sgi
// IRIX will go into an infinite loop if these are ignored and raised
        Disable_signal( SIGSEGV );		
        Disable_signal( SIGBUS );
#   endif

    Disable_signal( SIGHUP );

#   ifndef __linux__
        Disable_signal( SIGSYS );
#   endif

    Disable_signal( SIGCHLD );

#endif /* _WIN32 */

//    Disable_signal( SIGTERM);
  }

    return;
}


void Restore_signals()
{
  if (getenv("GENIE_DEBUG") == NULL)
  {
    Default_signal( SIGINT );
    Default_signal( SIGFPE );
    Default_signal( SIGSEGV );

#ifndef _WIN32

    Default_signal( SIGBUS );
    Default_signal( SIGHUP );

#   ifndef __linux__
	Default_signal( SIGSYS );
#   endif

    Default_signal( SIGCHLD );

#endif /* _WIN32 */

//    Default_signal( SIGTERM);
  }

    return;
}



/*
 * This routine tidies up after an exception of some sort, some
 * are more `expected' than others but all are dealt with !
 */
volatile void Handle_signal(int sig)
{
    switch (sig)
    {	
    case 0:			                /* Only called when setting up the longjmp */	
	    break;
    case SIGINT:
	    sigint_action();
	    break;
    case SIGTERM:
	    sigterm_action();
	    break;
    default:
	    sigother_action(sig);	    /* Pass the signal to allow the action to use it */
	    break;
    }

    Set_signals();	                /*** re-enable signals ***/
    return;
}


static void Enable_signal(int sig, signal_handler handler)
{
    if ( genie_signal(sig, handler) == (signal_handler) -1 ) PERROR("Enable_signal");
    return;
}



static void Disable_signal(int sig)
{
    if ( genie_signal(sig, (signal_handler) SIG_IGN ) == (signal_handler) -1 ) PERROR("Disable_signal");
    return;
}



static void Default_signal(int sig)
{
    if ( genie_signal(sig, (signal_handler) SIG_DFL ) == (signal_handler) -1 ) PERROR("Default_signal");
    return;
}


char *printsig(int sig)
{
    strstream buffer;

    switch (sig)
    {
    case SIGINT:
    	buffer << "Control-C";
    	break;
    case SIGSEGV:
	buffer << "Unexpected (SEGV) internal error within Open GENIE, please mail the command" << endl;
	buffer << "you typed along with any other error messages to \"genie@isise.rl.ac.uk\"" << endl;
	buffer << "this will help us fix this problem in future - thank you" << endl;
    	break;

#ifndef _WIN32
	case SIGBUS:
	buffer << "Unexpected (BUS) internal error within Open GENIE, please mail the command" << endl;
	buffer << "you typed along with any other error messages to \"genie@isise.rl.ac.uk\"" << endl;
	buffer << "this will help us fix this problem in future - thank you" << endl;
//    	buffer << "Bus error";
    	break;
    case SIGHUP:
	buffer << "Unexpected (HUP) internal error within Open GENIE, please mail the command" << endl;
	buffer << "you typed along with any other error messages to \"genie@isise.rl.ac.uk\"" << endl;
	buffer << "this will help us fix this problem in future - thank you" << endl;
//	    buffer << "Hangup signal";
	    break;

	#ifndef __linux__
		case SIGSYS:
		buffer << "Unexpected (SYS) internal error within Open GENIE, please mail the command" << endl;
		buffer << "you typed along with any other error messages to \"genie@isise.rl.ac.uk\"" << endl;
		buffer << "this will help us fix this problem in future - thank you" << endl;
//    		buffer << "System call error";
    		break;
	#endif
		case SIGCHLD:
			buffer << "Unexpected (CHLD) internal error within Open GENIE, please mail the command" << endl;
			buffer << "you typed along with any other error messages to \"genie@isise.rl.ac.uk\"" << endl;
			buffer << "this will help us fix this problem in future - thank you" << endl;
//		buffer << "Child process died or was killed";
    	break;

#endif /* _WIN32 */

    case SIGTERM:
    	buffer << "Process was terminated as the result of STOP/ID or Unix Kill";
    	break;
    case SIGFPE:
	buffer << "Unexpected (FPE) internal error within Open GENIE, please mail the command" << endl;
	buffer << "you typed along with any other error messages to \"genie@isise.rl.ac.uk\"" << endl;
	buffer << "this will help us fix this problem in future - thank you" << endl;
//        buffer << "Floating point exception (division by zero or number too big)";
        break;
    default:
	buffer << "Unexpected (SIG=" << sig << ") internal error within Open GENIE, please mail the command" << endl;
	buffer << "you typed along with any other error messages to \"genie@isise.rl.ac.uk\"" << endl;
	buffer << "this will help us fix this problem in future - thank you" << endl;
//    	buffer << "Signal number was " << sig;
	break;
    }
    buffer << ends;
    return buffer.str();
}

#undef HAVE_SIGACTION
#undef SA_RESTART

GENIEEXPORT volatile signal_handler genie_signal(int sig, volatile signal_handler func)
{
    signal_handler ret;
#if HAVE_SIGACTION

    struct sigaction act, old_act;
#	ifdef __ultrix
	    act.sa_handler = (void(*)())func;
#	else
	    act.sa_handler = func;
#	endif /* __ultrix */
    act.sa_flags = 0;
#	ifdef SA_RESTART
	    act.sa_flags |= SA_RESTART;			/* make system calls restartable */
#	endif /* SA_RESTART */
    sigemptyset(&act.sa_mask);
    sigemptyset(&old_act.sa_mask);
    sigaction(sig, &act, &old_act);
    ret = (signal_handler)old_act.sa_handler;
#else
    ret = signal(sig, func);
#endif /* HAVE_SIGACTION */
// Unfortunately, Linux has SA_RESTART but it doesn't do the same thing as on other machines
// We can't just use configure to look for siginterrupt() as some susyem will have both a working 
// SA_RESTART and siginterrupt
#if defined(__linux__) || defined(__ultrix)
    siginterrupt(sig, 0);			/* make system calls restartable */
#endif /* __linux__ || __ultrix */
    return ret;
}


// This is called from parser_main() once smalltalk is activated
// It modifies the genie signal handling to include the Smalltalk "stopExecuting()" call
void initializeSmalltalkSignals()
{
//    jumpsigs = 1; /* TRUE */			        // -- Signals here will default to a longjmp

    oldCtrlCaction = genie_signal(SIGINT, SIG_IGN);	// -- Keep what Smalltalk already does (for later use)

    return;
}

volatile void sigterm_action()
{
    gprintif("GENIE was killed/stopped by another process\n");
    exit(1);
}

volatile void sigfpe_action()
{
    *g_err << "* Floating Point Exception (overflow or division by zero) *" << endl;
#ifdef _WIN32
    _fpreset();		/* reset floating point state */
#endif /* _WIN32 */
    sigint_action();
}

volatile void sigint_action()
{
    *g_err << "* Ctrl-C *" << endl;
//    gdbg << (jumpsigs ? "In-parser" : "In-smalltalk") << endl;

    Set_signals();				// re-enable all signals

    // Only call the smalltalk handler if we are in evalCode() somewhere
    if (jumpsigs == 0)
    {
        if (oldCtrlCaction == SIG_DFL)
	        sigother_action(0);                     // Treat as an unexpected signal
        else if (oldCtrlCaction != SIG_IGN)
            ( *(void(*)(int))oldCtrlCaction )(0);	// Cast back to call the smalltalk routine correctly
    }
    return;
}

volatile void sigother_action(int sig)
{
    gerr << "Unexpected signal called - Exiting GENIE" << endl;
    ginf << printsig(sig) << endl;
//    CGenie::Logit(printsig(sig));
#ifdef _WIN32
    _sleep(2);
#else
    sleep(2);
#endif    
    if (getenv("GENIE_DUMP_CORE_ON_SIGNAL") == 0)
    {
	exit(1);                        // Definately broke here
    }
    else
    {
	abort();			// Gimme a core dump
    }
    _exit(sig);				// shouldn't be reached
}
