/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "geniedefs.h"
#include "hdf_utils.h"
#include "hdf_file.h"

//int HDFFile::addObject(HDFObject& object)
//{
//    return object.addToFile(*this);
//}

int HDFFile::open(const char* filename, int32 file_access_mode, int32 n_dds)
{
    int status;
    m_file_id = Hopen(filename, file_access_mode, n_dds);
    HDF_ERROR("HDFFile::open", m_file_id, -1);
    m_filename = filename;
    m_access_mode = file_access_mode;
    getFileVersion();
// Initalise Vgroup interface
    status = Vstart(m_file_id);
    HDF_ERROR("HDFFile::open", status, -1);
// Initialise SDS interface
//    status = SDstart(filename, file_access_mode);
    return 0;
}

void HDFFile::getFileVersion()
{
    Hgetfileversion(m_file_id, &m_major_v, &m_minor_v, &m_release, m_description);
}

int HDFFile::close()
{
    int32 status;
    if (m_file_id != INVALID_FID)
    {
	status = Vend(m_file_id);
	HDF_ERROR("HDFFile::close", status, -1);
	status = Hclose(m_file_id);
	m_file_id = INVALID_FID;
	HDF_ERROR("HDFFile::close", status, -1);
    }
    return 0;
}

int HDFFile::numberOfObjects(int32 tag) const
{
    int num = Hnumber(m_file_id, tag);
    HDF_ERROR("HDFFile::numberOfObjects", num, -1);
    return num;
}

ostream& HDFFile::print(ostream& os) const
{
    int i;
    if (m_file_id == INVALID_FID)
    {
	os << "HDF File not open" << endl;
        return os;
    }
    os << "HDF File \"" << m_filename << "\"" << endl;
    os << "Written by " << m_description << endl;
    os << "The file contains " << numberOfObjects() << " objects" << endl;
    os << "*** File Labels ***" << endl;
    for(i=0; i<numberOfFileLabels(); i++)
	os << readFileLabel(i) << endl;
    os << "*** File Descriptions ***" << endl;
    for(i=0; i<numberOfFileDescriptions(); i++)
	os << readFileDescription(i) << endl;
    return os;
}

int HDFFile::findNext(uint16 search_tag, uint16 search_ref, intn direction)
{
    uint16 find_tag, find_ref;                  // set both to 0 for start of search
    int32 find_offset, find_length;
    int status = Hfind(m_file_id, search_tag, search_ref, &find_tag, &find_ref, &find_offset, &find_length, direction);
    HDF_ERROR("HDFFile::find", status, -1);
    return 0;
}

int HDFFile::addFileAnnotation(const char* annotation, ann_type type)
{
    int32 status, an_id = ANstart(m_file_id);
    HDF_ERROR("HDFFile::addFileAnnotation", an_id, -1);
    int32 ann_id = ANcreatef(an_id, type);
    HDF_ERROR("HDFFile::addFileAnnotation", ann_id, -1);
    status = ANwriteann(ann_id, annotation, HDstrlen(annotation));
    HDF_ERROR("HDFFile::addFileAnnotation", status, -1);
    status = ANendaccess(ann_id);
    HDF_ERROR("HDFFile::addFileAnnotation", status, -1);
    status = ANend(an_id);
    HDF_ERROR("HDFFile::addFileAnnotation", status, -1);
    return 0;
}

int HDFFile::numberOfAnnotations(ann_type type) const
{
    int32 n_file_label, n_file_desc, n_data_label, n_data_desc, ret(-1);
    int32 status, an_id = ANstart(m_file_id);
    HDF_ERROR("HDFFile::numberOfAnnotations", an_id, -1);
    status = ANfileinfo(an_id, &n_file_label, &n_file_desc, &n_data_label, &n_data_desc);
    HDF_ERROR("HDFFile::numberOfAnnotations", status, -1);
    switch(type)
    {
	case AN_FILE_LABEL:
	    ret = n_file_label;
	    break;

	case AN_FILE_DESC:
	    ret = n_file_desc;
	    break;

	case AN_DATA_LABEL:
	    ret = n_data_label;
	    break;

	case AN_DATA_DESC:
	    ret = n_data_desc;
	    break;

	case AN_UNDEF:
	    ret = 0;
	    break;
    }
    status = ANend(an_id);
    HDF_ERROR("HDFFile::numberOfAnnotations", status, -1);
    return ret;
}

char* HDFFile::readFileAnnotation(int index, ann_type type) const
{
    int32 status, an_id = ANstart(m_file_id);
    HDF_ERROR("HDFFile::readFileAnnotation", an_id, 0);
    int32 ann_id = ANselect(an_id, index, type);
    HDF_ERROR("HDFFile::readFileAnnotation", ann_id, 0);
    int len = ANannlen(ann_id);
    char* buffer = new char[len+2];
    status = ANreadann(ann_id, buffer, len+1);
    HDF_ERROR("HDFFile::readFileAnnotation", status, 0);
    status = ANendaccess(ann_id);
    HDF_ERROR("HDFFile::readFileAnnotation", status, 0);
    status = ANend(an_id);
    HDF_ERROR("HDFFile::readFileAnnotation", status, 0);
    return buffer;
}
