/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "geniedefs.h"
#include "genievars.h"
#include "hdf_intermediate_file.h"
#include "hdf_utils.h"
#include "hdf_vgroup.h"
#include "hdf_dfsd.h"

// n_dds is the number of data descriptors for a new file
int HDFIntermediateFile::open(const char* filename, IntermediateFile::ModeFlags mode)
{
    int32 access_mode, n_dds(10);
// Map Genie to HDF file access code
    switch(mode)
    {
	case Input:
	    access_mode = DFACC_READ;
	    break;

	case Append:
	    access_mode = DFACC_WRITE;
	    break;

	case Input|Append:
	    access_mode = DFACC_RDWR;
	    break;

	case Output:
	case Input|Output:
	    access_mode = DFACC_CREATE;
	    break;

	default:
	    gerr << "Invalid mode " << mode << endl;
	    return 0;
    }
    m_file.open(filename, access_mode, n_dds);
    if (access_mode == DFACC_CREATE)
    {
        m_file.addFileLabel("NEXUS data");
        m_file.addFileDescription("Open GENIE data in NEXUS format");
    }
//    gout << m_file << endl;
    return 1;
}
    
// Return 0 on success, -1 on failure
int HDFIntermediateFile::close()
{
    return m_file.close();
}


// Total number of objects of a given tag in a file
// If get tag DFTAG_WILDCARD, gives total number of objects
// Returns -1 on error
int HDFIntermediateFile::numberOfBlocks() const
{
    int n;
    VgroupList* list = (VgroupList*)Vgroup::loneVgroups(m_file);
    if (list->count() > 0)
    {
	Vgroup* top = new Vgroup(m_file, ((*list)[0]).ref());		// already one in file
	n = top->numberOfMembers();
        delete top;
    }
    else
    {
	n = 0;
    }
    delete list;
    return n;
}

HDFIntermediateFile::HDFIntermediateFile()
{
}

int HDFIntermediateFile::appendBlock(GenieVariableBase& gvb, const char* label, const char* comment)
{
    GenieVariable gv(gvb);
    return appendBlock(gv, label, comment);
}

int HDFIntermediateFile::appendBlock(GenieVariable& gv, const char* label, const char* comment)
{
    static char entry[64];
// Find top level group in file - we'll assume it's the first lone one for simplicity
    Vgroup* top;
    VgroupList* list = (VgroupList*)Vgroup::loneVgroups(m_file);
    if (list->count() > 0)
    {
//	gout << "Connecting to exising top level vgroup" << endl;
	top = new Vgroup(m_file, ((*list)[0]).ref());		// already one in file
    }
    else
    {
//	gout << "Creating new top level vgroup" << endl;
	top = new Vgroup(m_file);				// create a new top level vgroup
	top->setClass("NEXUS data file");
	top->setName("Open GENIE output");
	String desc("Demonstration of creating a NEXUS file from Open GENIE data\n"
		    "\t\t\tOpen GENIE ---> HDF routines by Freddie Akeroyd, ISIS computer group, 10/10/96\n"
		    "\t\t\tThis file written by ");
	desc += creator();
	desc += " on ";
	desc += dateAndTime();
	top->addDescription(desc);
    }
// Create a new Vblock for the entry
    int nblocks = top->numberOfMembers();
    sprintf(entry, "Entry_%d", ++nblocks);		// a default label
    Vgroup new_block(m_file);
    new_block.addDescription(comment);
// Now create the object to add
    if (label && strcmp(label, "no_label"))
    {
        addGenieVariable(gv, new_block, label);
    }
    else
    {
	addGenieVariable(gv, new_block, entry);
    }
    new_block.addLabel(entry);
// Add outselves to the tree
    new_block.insertInVgroup(*top);
//    gout << *top << endl;
    delete top;
    delete list;
    return 0;
}

// Note - this routine will be called recursively
int HDFIntermediateFile::addGenieVariable(GenieVariable& gv, Vgroup& vg, const char* name, int set_class)
{
    if (set_class)
    {
        vg.setClass(getClassString(gv.value()));
        vg.setName(name);
    }
//    gout << "Adding " << vg << endl;
    switch(gv.type())
    {
	case GenieWorkspaceArrayType:
	{
	    char* field_name = new char[strlen(name)+10];	// enough space for some big numbers
	    GenieWorkspaceArray& work = *(gv.gwa);
	    GenieVariable v;
	    for(int i=0; i<work.length(); i++)
	    {
	        Vgroup new_vg(m_file);
	        v.become(work[i]);
		sprintf(field_name, "%s_%d", name, i+1);
		addGenieVariable(v, new_vg, field_name);
	        new_vg.insertInVgroup(vg);
	    }
	    delete[] field_name;
	    break;
	}

	case GenieWorkspaceType:
	{
	    char* field_name;
	    GenieWorkspace& work = *(gv.gw);
	    for(int i=1; i<=work.length(); i++)
	    {
	 	field_name = OOPToString(indexOOP(work.value(), i));
		if (field_name)
		{
	    	    GenieVariable v;
		    gout << "Field \"" << field_name << "\" item " << i << endl;
		    Vgroup new_vg(m_file);
		    v.become(work.getOOP(i));
		    addGenieVariable(v, new_vg, field_name);
		    new_vg.insertInVgroup(vg);
		    free(field_name);
		}
	    }
	    break;
	}

	case GenieRealArrayType:
	{
	    GenieRealArray& array = *(gv.gra);
	    int32* dims_array = (int32*)array.dimsArray();
    	    DFSD item(m_file, DFNT_FLOAT64, array.dimsSize(), dims_array, (float64*)array);
    	    item.writeToFile();
	    vg.insertObject(item);
	    break;
	}

	case GenieRealType:
	{
	    float64 val = *(gv.gr);
	    int32 dims_array[] = { 1 };
    	    DFSD item(m_file, DFNT_FLOAT64, 1, dims_array, &val);
    	    item.writeToFile();
	    vg.insertObject(item);
	    break;
	}

	case GenieIntegerArrayType:
	{
	    GenieIntegerArray& array = *(gv.gia);
	    int32* dims_array = (int32*)array.dimsArray();
    	    DFSD item(m_file, DFNT_INT32, array.dimsSize(), dims_array, (int32*)array);
    	    item.writeToFile();
	    vg.insertObject(item);
	    break;
	}

	case GenieIntegerType:
	{
	    int32 val = *(gv.gi);
	    int32 dims_array[] = { 1 };
    	    DFSD item(m_file, DFNT_INT32, 1, dims_array, &val);
    	    item.writeToFile();
	    vg.insertObject(item);
	    break;
	}
 	    
	case GenieStringArrayType:
	{
	    GenieStringArray& gsa = *(gv.gsa);
	    for(int i=0; i<gsa.length(); i++)
	    {
	    	GenieVariable v;
	    	GenieString gs = gsa(i);
		v.become(gs);
		addGenieVariable(v, vg, name, 0);
	    }
	    break;
	}

	case GenieStringType:
	{
	    c_g_string val = *(gv.gs);
	    int32 dims_array[] = { 1 };
    	    DFSD item(m_file, DFNT_UCHAR8, 1, dims_array, (uchar8*)val);
    	    item.writeToFile();
	    vg.insertObject(item);
	    break;
	}

	case GenieUndefinedType:
	{
// Need to write something - this is just a place holder
	    break;
	}
	
	default:
	    gerr << "HDFIntermediateFile::addGenieVariable - Invalid type = " << gv.type() << endl;
	    break;
    }
    return 0;
}
