/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "geniedefs.h"
#include "hdf_utils.h"
#include "hdf_sds.h"

// Open access to a file, close any existing file
// Return 0 on success, -1 on error
int SDS::open()
{
    if ((m_sd_id != -1) && close())
	return -1;
    m_sd_id = SDstart(m_filename, m_access_mode);
    HDF_ERROR("SDS::open", m_sd_id, -1);
    return 0;
}

// Open access to the Nth SDS in a file (0 based index)
// Return 0 on success, -1 on failure
int SDS::seek(int32 sds_index)
{
    if (flush())		// close access to current dataset
        return -1;
    m_sds_id = SDselect(m_sd_id, sds_index);
    HDF_ERROR("SDS::seek", m_sds_id, -1); 
    return 0;
}

int SDS::seek(const char* sds_name)
{
    int32 sds_index = SDnametoindex(m_sd_id, (char*)sds_name);
    HDF_ERROR("SDS::seek", sds_index, -1);
    return seek(sds_index);
}

int SDS::refToIndex(int32 ref)
{
    int32 index = SDreftoindex(m_sd_id, ref);
    HDF_ERROR("SDS::refToIndex", index, -1);
    return index;
}

int SDS::idToRef(int32 sds_id)
{
    int32 ref = SDidtoref(sds_id);
    HDF_ERROR("SDS::indexToRef", ref, -1);
    return ref;
}

int SDS::close()
{
    int status(0);		// Set to "success"
    if (flush())
	return -1;
    if (m_sd_id >= 0)
    {
        status = SDend(m_sd_id);
	m_sd_id = -1;
    }
    HDF_ERROR("SDS::close", status, -1);
    return 0;
}

int SDS::write(const char* sds_name, const char* comment, int32 data_type, int32 ndims, int32* dims_array, void* data)
{
    int32 status;
    if (flush())		// Clear out previous writes
	return -1;
    m_sds_id = SDcreate(m_sd_id, (char*)sds_name, data_type, ndims, dims_array);
    HDF_ERROR("SDS::write", m_sds_id, -1);
// The write supports "slabs", so we have to give it a starting point ...
    int32* start = new int32[ndims];
    for(int i=0; i<ndims; i++)
    {
	start[i] = 0;
    }
    status = SDwritedata(m_sds_id, start, 0, dims_array, data);
    delete []start;
    HDF_ERROR("SDS::write", status, -1);
// set creator attribute
    const char* creator = "Open Genie";
    status = SDsetattr(m_sds_id, "creator", DFNT_CHAR8, strlen(creator), (VOIDP)creator);
    HDF_ERROR("SDS::write", status, -1);
// Add comment filed
    status = SDsetattr(m_sds_id, "comment", DFNT_CHAR8, strlen(comment), (VOIDP)comment);
    HDF_ERROR("SDS::write", status, -1);
    return 0; 
} 

// Read the current data block. Correct values for "dims_array", "ndims" and the
// size of the pre-allocated array "data" can be obtained by calling blockInfo()
int SDS::read(int32 ndims, int32* dims_array, void* data) 
{ 
    int32 status;
    int32* start = new int32[ndims];
    for(int i=0; i<ndims; i++)
    {
	start[i] = 0;
    }
    status = SDreaddata(m_sds_id, start, 0, dims_array, data);
    delete []start;
    HDF_ERROR("SDS::read", status, -1);
    return 0;
}

// Retuns a null pointer on error
const char* SDS::blockName()
{
    static char sds_name[MAX_NC_NAME];
    int32 dims_array[MAX_VAR_DIMS];
    int32 ndims, data_type, nattrs, status;
    status = SDgetinfo(m_sds_id, sds_name, &ndims, dims_array, &data_type, &nattrs);
    HDF_ERROR("SDS::blockName", status, 0);
    return sds_name;
}

// Returns dims_array on heap
int32* SDS::blockInfo(int32& ndims, int32& data_type)
{
    char sds_name[MAX_NC_NAME];
    int32* dims_array = new int32[MAX_VAR_DIMS];
    int32 nattrs, status;
    status = SDgetinfo(m_sds_id, sds_name, &ndims, dims_array, &data_type, &nattrs);
    HDF_ERROR("SDS::blockInfo", status, 0);
    return dims_array;
}

#if 0
// return number of data sets, -1 onm error
int SDS::number()
{
    int32 n_datasets;		// number of datasets
    int32 n_file_attr;		// number of global attributes
    int status = SDfileinfo(m_sd_id, &n_datasets, &n_file_attr);
    HDF_ERROR("SDS::number", status, -1);
// now loop through and ignore dimension scales, which look like an SDS but aren't really ...
    int n(0);
    for(int i=0; i<n_datasets; i++)
    {
	seek(i);
	if (FLASE == SDiscoordvar(m_sds_id))
	    n++;
    }
    return n;
}
#endif

// Return 0 on success, 1 on error
int SDS::flush()
{
    int status;
    if (m_sds_id >= 0)
    {
        status = SDendaccess(m_sds_id);          // terminate a previous access
        m_sds_id = -1;
        HDF_ERROR("SDS::flush", status, -1);
    }
    return 0;
}

// Return data from an attribute (on heap), NULL on error
void* SDS::readAttribute(const char* attr_name, int32& data_type, int32& count)
{
    void* data;
    int32 status;
    char attr_name_buffer[MAX_NC_NAME];
    int32 attr_index = SDfindattr(m_sds_id, (char*)attr_name);
    HDF_ERROR("SDS::readAttribute", attr_index, 0);
    status = SDattrinfo(m_sds_id, attr_index, attr_name_buffer, &data_type, &count);
    HDF_ERROR("SDS::readAttribute", status, 0);
    data = new char[count * DFKNTsize(data_type)];
    status = SDreadattr(m_sds_id, attr_index, data);
    HDF_ERROR("SDS::readAttribute", status, 0);
    return data;
}
