/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "geniedefs.h"
#include "hdf_utils.h"
#include "hdf_vgroup.h"

int Vgroup::addToFile(HDFFile& hdf)
{
    return 1;
}

Vgroup& Vgroup::operator=(const Vgroup& vg)
{
    if (this != &vg)
    {
	cleanup();
        m_vgroup_id_ref_count = vg.m_vgroup_id_ref_count;
        m_vgroup_id = vg.m_vgroup_id;
        ++*m_vgroup_id_ref_count;
        setup(vg.fileName(), vg.fileId(), vg.ref(), vg.accessMode());
    }
    return *this;
}

Vgroup::Vgroup(const Vgroup& vg)
{
    m_vgroup_id_ref_count = vg.m_vgroup_id_ref_count;
    m_vgroup_id = vg.m_vgroup_id;
    ++*m_vgroup_id_ref_count;
    setup(vg.fileName(), vg.fileId(), vg.ref(), vg.accessMode());
}

Vgroup::Vgroup(const HDFFile& hdf, int32 vgroup_ref, const char* vgroup_access_mode) : m_vgroup_id(-1), 
	m_vgroup_id_ref_count(new int)
{
    *m_vgroup_id_ref_count = 1; 
    m_vgroup_id = Vattach(hdf.id(), vgroup_ref, vgroup_access_mode);
    HDF_ERROR_MESSAGE("Vgroup::Vgroup", m_vgroup_id);
    setup(hdf.name(), hdf.id(), vgroup_ref, vgroup_access_mode);
}

int Vgroup::setup(const char* filename, int32 file_id, int32 vgroup_ref, const char* vgroup_access_mode)
{
    setTag(DFTAG_VG);
    setFile(filename, file_id);
    int32 ref = VQueryref(m_vgroup_id);
    HDF_ERROR_MESSAGE("Vgroup::Vgroup", ref);
    setRef(ref);
    return 0;
}

int Vgroup::insertInVgroup(Vgroup& vg)
{
    int status = Vinsert(vg.vid(), m_vgroup_id);
    HDF_ERROR("Vgroup::insertInVgroup", status, -1);
    return 0;
}

Vgroup::~Vgroup()
{
    cleanup();
}

int Vgroup::cleanup()
{
    if (!--*m_vgroup_id_ref_count && (m_vgroup_id != -1))
    {
	int status = Vdetach(m_vgroup_id);
	m_vgroup_id = -1;
	HDF_ERROR("Vgroup::cleanup", status, -1);
    }
    return 0;
}

int Vgroup::insertObject(HDFObject& object)
{
    return object.insertInVgroup(*this);
}

int Vgroup::setName(const char* vgroup_name)
{
    int status = Vsetname(m_vgroup_id, vgroup_name);
    HDF_ERROR("Vgroup::setName", status, -1);
    return 0;
}

int Vgroup::setClass(const char* vgroup_class)
{
    int status = Vsetclass(m_vgroup_id, vgroup_class);
    HDF_ERROR("Vgroup::setClass", status, -1);
    return 0;
}

const char* Vgroup::getName() const
{
    static char vgroup_name[VGNAMELENMAX];
    int status = Vgetname(m_vgroup_id, vgroup_name);
    HDF_ERROR("Vgroup::getName", status, 0);
    return vgroup_name;
}

const char* Vgroup::getClass() const
{
    static char vgroup_class[VGNAMELENMAX];
    int status = Vgetclass(m_vgroup_id, vgroup_class);
    HDF_ERROR("Vgroup::getClass", status, 0);
    return vgroup_class;
}

void* Vgroup::allVgroups(HDFFile& hdf)		// Really returns VgroupList*
{
    VgroupList* list = new VgroupList;
    if (list == 0)
	return 0;
    int32 vgroup_ref, last_vgroup_ref(-1);
    while( (vgroup_ref = Vgetid(hdf.id(), last_vgroup_ref)) != -1 )
    {
	list->push(Vgroup(hdf, vgroup_ref));
    }
    return list;
}
	
void* Vgroup::loneVgroups(const HDFFile& hdf)		// Really returns VgroupList*
{
    VgroupList* list = new VgroupList;
    if (list == 0)
	return 0;
    int32 status, *ref_array(0);
    int maxsize = Vlone(hdf.id(), ref_array, 0);
    HDF_ERROR("Vgroup::loneVgroups", maxsize, 0);
    if (maxsize > 0)
    {
        ref_array = new int32[maxsize];
	status = Vlone(hdf.id(), ref_array, maxsize);
        HDF_ERROR("Vgroup::loneVgroups", status, 0);
	for(int i=0; i<maxsize; i++)
	{
	    list->push(Vgroup(hdf, ref_array[i]));
	}
	delete []ref_array;
    }
    return list;
}

int Vgroup::numberOfMembers() const
{
    int n = Vntagrefs(m_vgroup_id);
    HDF_ERROR("Vgroup::numberOfMembers", n, -1);
    return n;
}

HDFObject Vgroup::member(int index) const
{
    int32 tag(INVALID_TAG), ref(INVALID_REF), status;
    status = Vgettagref(m_vgroup_id, index, &tag, &ref);
    HDF_ERROR_MESSAGE("Vgroup::member", status);
    return HDFObject(m_filename, m_file_id, tag, ref);
}

int Vgroup::isMember(HDFObject& object)
{
    return Vinqtagref(m_vgroup_id, object.tag(), object.ref());
}

ostream& Vgroup::print(ostream& os) const
{
    int nmem = numberOfMembers();
    os << "Vgroup \"" << getName() << "\" (" << getClass() << ") with " << nmem << " members" << endl;
    if (nmem > 0)
    {
        os << "*** Vgroup members ***" << endl;
	for(int i=0; i<nmem; i++)
	{
	    os << member(i) << endl;
	}
    }
    return os;
}

#ifdef __DECCXX
#pragma define_template SPList<Vgroup>
#pragma define_template SPListBase<Vgroup>
#endif /* __DECCXX */

#if defined(__GNUG__) || defined(_MSC_VER)
template class SPList<Vgroup>;
template class SPListBase<Vgroup>;
#endif /* __GNUG__ */
