/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "geniedefs.h"
//+++
//
// CLASS: Multiplot
//
//
//---
// ----
// define String data type
//
#include <gnustring.h>
 
#include "Drawable_objects.h"
#include "Plot.h"
#include "Multiplot.h"
#include "Pginclude.h"
#include "Errors.h"
#include "g_types.h"
#include "SinglePrecision.h"

static const char* rcsid = "@(#)$Id: multiplot.cxx,v 1.13 1997/10/21 14:36:43 cmm Exp $";
USE_RCSID(rcsid);

// *** constructor
Multiplot::Multiplot(int totchan,
		     double *xdata,
		     int numhist,
		     int minch,
		     int maxch,
		     float bias,
		     int centre,
		     int colour,
		     int thickness,
		     int type) {

	    int c;

	    // *** set visibility
	    invisible = 0;


	    // ***get min/max channel numbers to plot
	    min_X_chan = minch + 1;	// fortran arrays start at 1
	    max_X_chan = maxch;

	    // *** get total number of channels in histograms (check for zero)
	    total_chan = totchan;
	    if (total_chan <= 0) {
			Errors syserr;
			syserr.add_error ("Multiplot ",
			    "too few data points ",
			    "make number_of_points greater than 0");
			return;
			}

	    // *** get number of histograms (check for zero)
	    if (numhist <= 0) {
			Errors syserr;
			syserr.add_error ("Multiplot ",
			    "too few histograms ",
			    "make number_of_histograms greater than 0");
			return;
			}
	    nhist = numhist;

	    // *** reserve space for arrays
	    X_array = new float[total_chan];
	    Y_array = new float[total_chan];
	    // *** fill the X array
	    if (X_array == 0) {
			Errors syserr;
			syserr.add_error ("Multiplot ",
			    "Insufficient virtual memory for X array ",
			    "call procedure DELETE_PICTURES");
	            return;
			}

	    XDL = F(xdata[0]);
	    XDU = F(xdata[0]);
	    DoubleArrayToFloatArray(X_array,  xdata, total_chan);
	    for (c = 0; c < total_chan; c++) {
		    if (XDL > xdata[c]) XDL = F(xdata[c]);
		    if (XDU < xdata[c]) XDU = F(xdata[c]);
	    }

	    // *** reserve some space for the 2d data
	    //	    (note that it is stored in a 1d array)
	    c = numhist * total_chan;
	    Y_2d_array = new float[c];
	    if (Y_2d_array == 0) {
			Errors syserr;
			syserr.add_error ("Multiplot ",
			    "Insufficient virtual memory for Y arrays ",
			    "call procedure DELETE_PICTURES");
	            return;
			}
	    counter = 0;
	    for (c=0;c<(numhist*total_chan);c++) {
		    Y_2d_array[c] = 0.0F;
		    }

	    // *** get offset
	    gap = bias;

	    // *** get bin centre
	    if (centre == -1)
		bin_centre = BIN_CENTRE;
	    else
		bin_centre = centre;

	    // *** get slant
	    slant = SLANT;

	    // *** get colour
	    if (colour == -1)
		line_colour = LINE_COLOUR;
	    else
		line_colour = colour;

	    // *** get line thickness
	    if (thickness == 0)
		line_thickness = LINE_THICKNESS;
	    else
		line_thickness = thickness;

	    // *** get line type
	    if (type == 0)
		line_type = LINE_TYPE;
	    else
		line_type = type;

	    // *** set Y limits
	    YDL = 0.0F;
	    YDU = 0.0F;
	    spectra_added = 0.0F;
	    }
		  

// *** add_spectra
void Multiplot::add_spectra(const int* spectrum,int length)
{
	    int i;
	    double speccy;

	    // *** check for exceeding array
	    if ((counter + total_chan) > (total_chan * nhist)) {
			Errors syserr;
			syserr.add_error ("Multiplot ",
			    "exceeded range of Y data points ",
			    "create a new multiplot");
			return;
		    }

	    // *** add channels to array (padding if required)
	    for(i=0; i<total_chan; i++) {
		if (i < length) {
		       speccy = (double)spectrum[i] / (X_array[i+1] - X_array[i]);
		       if (speccy < FLT_MAX)
			    Y_2d_array[counter+i] = F(speccy);
		       else
		            Y_2d_array[counter+i] = FLT_MAX;
		}
		else {
		    Y_2d_array[counter+i] =  Y_2d_array[counter + length - 1];
		}

		if (YDL > (Y_2d_array[counter+i] + (spectra_added * gap)))
			 YDL = Y_2d_array[counter+i] + (spectra_added * gap);
		if (YDU < (Y_2d_array[counter+i] + (spectra_added * gap)))
			 YDU = Y_2d_array[counter+i] + (spectra_added * gap);
	        }
	    spectra_added = spectra_added + 1.0F;
	    counter = counter + total_chan;

}

void Multiplot::add_spectra(const double *spectrum,int length) {
	    int i;

	    // *** check for exceeding array
	    if ((counter + total_chan) > (total_chan * nhist)) {
			Errors syserr;
			syserr.add_error ("Multiplot ",
			    "exceeded range of Y data points ",
			    "create a new multiplot");
			return;
		    }

	    // *** add channels to array (padding if required)
	    for(i=0; i<total_chan; i++) {
		if (i < length) {
		       if (spectrum[i] < FLT_MAX)
			    Y_2d_array[counter+i] = F(spectrum[i]);
		       else
		            Y_2d_array[counter+i] = FLT_MAX;
		}
		else {
		    Y_2d_array[counter+i] =  Y_2d_array[counter + length - 1];
		}

		if (YDL > (Y_2d_array[counter+i] + (spectra_added * gap)))
			 YDL = Y_2d_array[counter+i] + (spectra_added * gap);
		if (YDU < (Y_2d_array[counter+i] + (spectra_added * gap)))
			 YDU = Y_2d_array[counter+i] + (spectra_added * gap);
	        }
	    spectra_added = spectra_added + 1.0F;
	    counter = counter + total_chan;
	    }

// *** destructor
Multiplot::~Multiplot() {

	// *** delete the arrays
	delete [] Y_2d_array;
	delete [] X_array;
	delete [] Y_array;
	}

// *** get size of plot area
void Multiplot::get_size(float& XXM,
			 float& XXN,
			 float& YYM,
			 float& YYN) {

	// *** set the return values
	XXM = XDL;
	XXN = XDU;
	YYM = YDL;
	YYN = YDU;
	}

// *** display
void Multiplot::display() {

	// *** note that ALL histograms are plotted but only
	//	the channels requested are plotted.

	// *** set line thickness
	pgslw(line_thickness);

	// *** set colour
	pgsci(line_colour);

	// *** set the line type
	pgsls(line_type);

	// *** draw the multiplot
	pghi2d(Y_2d_array,total_chan,nhist,min_X_chan,max_X_chan,1,nhist,X_array,slant,gap,bin_centre,Y_array);

	}

