/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "geniedefs.h"
#include "genievars.h"
#include "nexus_intermediate_file.h"

// Return 1 on success, 0 on failure
int NeXusIntermediateFile::open(const char* filename, IntermediateFile::ModeFlags mode)
{
// Map Genie to NeXus file access code
    NXaccess open_mode;
    NXhandle fileid;
    switch(mode)
    {
	case Input:
	    m_access_mode = NXACC_READ;
	    open_mode = NXACC_READ;
	    break;

	case Append:
	case Input|Append:
	    m_access_mode = NXACC_RDWR;
	    open_mode = NXACC_RDWR;
	    break;

	case Output:
	case Input|Output:
	    m_access_mode = NXACC_RDWR;
	    open_mode = NXACC_CREATE;
	    break;

	default:
	    gerr << "NEXUS: Invalid mode " << mode << endl;
	    return 0;
    }
    m_filename = filename;
    if (NXopen((const char*)m_filename, open_mode, &fileid) == NX_OK)
    {
	NXclose(&fileid);
	return 1;
    }
    else
    {
	return 0;
    }
}
    
// Return 0 on success, -1 on failure
int NeXusIntermediateFile::close()
{
    return 0;
}

// Return number of entries in file
int NeXusIntermediateFile::numberOfBlocks() const
{
    NXhandle fileid;
    if (NXopen((const char*)m_filename, m_access_mode, &fileid) != NX_OK)
	return -1;
    int nblocks = countEntries(fileid);
    if (NXclose(&fileid) != NX_OK)
	return -1;
    return nblocks;
}

int NeXusIntermediateFile::NeXusToGenieVariable(GenieWorkspace& gw, 
		const char* name, void* data, 
		int* dims_array, int ndims, int type, bool one_element_arrays)
{
    int i, length(1);
    char* tname = 0;
    GenieVariableBase* gvb;
    OOP oop(nilOOP); 
    for(i=0; i<ndims; i++)
    {
	length *= dims_array[i];
    }
    switch(type)
    {
	case DFNT_FLOAT32:
	    if (length == 1 && !one_element_arrays)
	    {
	        gvb = new GenieReal(((float*)data)[0]);
	    }
	    else
	    {
		gvb = new GenieRealArray((float*)data, dims_array, ndims);
	    }
	    break;

	case DFNT_FLOAT64:
	    if (length == 1 && !one_element_arrays)
	    {
	        gvb = new GenieReal(((double*)data)[0]);
	    }
	    else
	    {
		gvb = new GenieRealArray((double*)data, dims_array, ndims);
	    }
	    break;

	case DFNT_INT32:
	case DFNT_UINT32:
	    if (length == 1 && !one_element_arrays)
	    {
	        gvb = new GenieInteger(((int*)data)[0]);
	    }
	    else
	    {
		gvb = new GenieIntegerArray((int*)data, dims_array, ndims);
	    }
	    break;

	case DFNT_INT8:
	case DFNT_UINT8:
	case DFNT_CHAR8:
	case DFNT_UCHAR8:
//	    gvb = new GenieString((const char*)data, length);
	    tname = (char*)data; 
	    if (tname[length-1] != '\0')
	    {
		tname[length] = '\0';	// NULL terminate - the calling routine should have given us enough extra space
	    }
	    gvb = new GenieString(tname);
	    break;

	case DFNT_INT16: // Not used by NeXus, so we'll use for undefined
	    gvb = new GenieReal(nilOOP);
	    break;

	default:
	    gerr << "NEXUS: Reading invalid type " << type << endl;
	    printHDFTag(*g_err, type);
	    return -1;
	    break;
    }
    tname = strcpy((char*)malloc(1+strlen(name)), name);
    cleanFilename(tname);
    gw(tname, *gvb);
    free(tname);
    return 0;
}

// Read attributes into a Genie Workspace
int NeXusIntermediateFile::readAttributes(GenieWorkspace& gw, NXhandle fileid)
{
    char name[VGNAMELENMAX];
    int length, type;
    void* data;
    while(NXgetnextattr(fileid, name, &length, &type) != NX_EOD)
    {
	length++;	// allow space for possible '\0' on string
        data = malloc(length*typeSize(type));
        NXgetattr(fileid, name, (char*)data, &length, &type);
	NeXusToGenieVariable(gw, name, data, &length, 1, type, 0);
	free(data);
    }
    return 0;
}

// Return size of a NeXus type in bytes
int NeXusIntermediateFile::typeSize(int type)
{
    switch(type)
    {
	case DFNT_FLOAT32:
	case DFNT_UINT32:
	case DFNT_INT32:
	    return 4;

	case DFNT_FLOAT64:
	    return 8;

	case DFNT_UCHAR8:
	case DFNT_CHAR8:
	case DFNT_UINT8:
	case DFNT_INT8:
	    return 1;

	case DFNT_INT16:	// Used to represent <Undefined>
	    return 2;
    }
    gerr << "NEXUS: invalid type " << type << endl;
    printHDFTag(*g_err, type);
    return 0;
}

// Read an SDS and associated attributes into a workspace
int NeXusIntermediateFile::readSDS(GenieWorkspace& gw, const char* name, 
					NXhandle fileid)
{
    NXstatus stat;
    int i, length(1);
    void* data;
    int ndims, dims_array[MAX_VAR_DIMS], type;
    stat = NXopendata(fileid, name);
    if (stat != NX_OK)
    {
	gerr << "NEXUS: Cannot read SDS \"" << name << "\"" << endl;
	return -1;
    }
    stat = NXgetinfo(fileid, &ndims, dims_array, &type);
    if (stat == NX_OK)
    {
        for(i=0; i<ndims; i++)
        {
	    length *= dims_array[i];
        }
	length++;	// Add one byte to terminate a string
        data = malloc(length*typeSize(type));
        stat = NXgetdata(fileid, data);
	if (stat == NX_OK)
	{
            NeXusToGenieVariable(gw, name, data, dims_array, ndims, type, 1);
            readAttributes(gw, fileid);
	}
        free(data);
    }
    stat = NXclosedata(fileid);
    return 0;
}

// block 0 is global atributes
int NeXusIntermediateFile::readBlock(GenieVariable& gv)
{
    GenieWorkspace gw;
    NXhandle fileid;
    if (NXopen((const char*)m_filename, m_access_mode, &fileid) != NX_OK)
	return -1;
    readAttributes(gw, fileid);		// global attributes
    int nblocks = readEntries(gw, fileid);
    if (NXclose(&fileid) != NX_OK)
	return -1;
    gv.become(gw);
    return nblocks;
}

int NeXusIntermediateFile::readEntries(GenieWorkspace& gw, NXhandle fileid)
{
    int i, status, nblocks(0), vg_type;
    char vg_name[VGNAMELENMAX],vg_class[VGNAMELENMAX];
    char vg_gclass[VGNAMELENMAX+2];  // +2 for GX
    while( (status = NXgetnextentry(fileid, vg_name, vg_class, &vg_type)) != NX_EOD )
    {
	if (vg_type == DFTAG_VG)
	{
	    if (!strcmp("UNKNOWN", vg_class) || !strcmp("Dim0.0", vg_class) || 
				!strcmp("CDF0.0", vg_class))
	    {
		;
	    }
	    else if (NXopengroup(fileid, vg_name, vg_class) == NX_OK)
	    {
		sprintf(vg_gclass, "GX%s", vg_class);
		vg_gclass[2] = toupper(vg_gclass[2]);
		for(i=3; i<strlen(vg_gclass); i++)
		{
		    vg_gclass[i] = tolower(vg_gclass[i]);
		}
    		GenieWorkspace gww(vg_gclass);
	        nblocks += readEntries(gww, fileid);
	 	cleanFilename(vg_name);
		gw(vg_name, gww);
		NXclosegroup(fileid);
	    }
	    else
	    {
		gerr << "NEXUS: Error opening group named \"" << vg_name << "\" of class \"" << vg_class << "\"" << endl;
	    }
	}
	else if (!strcmp("SDS", vg_class))
	{
	    if (vg_type == DFTAG_LINKED)
	    {
	        ginf << "NEXUS: \"" << vg_name << "\" is a link" << endl;
	    }
	    readSDS(gw, vg_name, fileid);
	    nblocks++;
	}
	else
	{
	    gerr << "NEXUS: unknow, class=\"" << (vg_class != 0 ? vg_class : "null") << "\", name=\""
                                     << (vg_name != 0 ? vg_name : "null") << "\"" << endl;
	    printHDFTag(*g_err, vg_type);
	}
    }
    return nblocks;
}

// Iterate over Vgroups in file, return number found
int NeXusIntermediateFile::countEntries(NXhandle fileid)
{
    int status, nblocks(0), vg_type;
    char vg_name[VGNAMELENMAX],vg_class[VGNAMELENMAX];
    while( (status = NXgetnextentry(fileid, vg_name, vg_class, &vg_type)) != NX_EOD )
    {
	if (vg_type == DFTAG_VG)
	{
	    if (!strcmp("UNKNOWN", vg_class) || !strcmp("Dim0.0", vg_class) || 
				!strcmp("CDF0.0", vg_class))
	    {
		;
	    }
	    else if (NXopengroup(fileid, vg_name, vg_class) == NX_OK)
	    {
	        nblocks += countEntries(fileid);
		NXclosegroup(fileid);
	    }
	    else
	    {
		gerr << "NEXUS: Error opening group named \"" << vg_name << "\" of class \"" << vg_class << "\"" << endl;
	    }
	}
	else if (!strcmp("SDS", vg_class))
	{
	    nblocks++;
	}
    }
    return nblocks;
}
    
NeXusIntermediateFile::NeXusIntermediateFile() : m_sds_open(false)
{
}

NeXusIntermediateFile::~NeXusIntermediateFile()
{
}

int NeXusIntermediateFile::appendBlock(GenieVariableBase& gvb, const char* label, const char* comment)
{
    GenieVariable gv(gvb);
    return appendBlock(gv, label, comment);
}

int NeXusIntermediateFile::appendBlock(GenieVariable& gv, const char* label, const char* comment)
{
    NXhandle fileid;
    if (NXopen((const char*)m_filename, m_access_mode, &fileid) != NX_OK)
	return -1;
    addGenieVariable(gv, label, fileid);
    return closeFile(&fileid);
}

// 0 on success, -1 on error
int NeXusIntermediateFile::closeFile(NXhandle* pFileid)
{
    closeOpenSDS(*pFileid);
    if (NXclose(pFileid) != NX_OK)
	return -1;
    return 0;
}
    
int NeXusIntermediateFile::closeOpenSDS(NXhandle fileid)
{
    if (m_sds_open)
    {
	NXclosedata(fileid);
	m_sds_open = false;
    }
    return 0;
}

inline OOP OOPCLASS(OOP oop)
{
    return isInt(oop) ? integerClass : oopClass(oop);
}

// Note - this routine will be called recursively
int NeXusIntermediateFile::addGenieVariable(GenieVariable& gv, 
		const char* name, NXhandle fileid)
{
// We don't want to close an SDS immediately after writing the
// array, as then any attributes we write will become global rather
// than attached to the SDS. Thus we use the m_sds_open variable to
// keep track ot this
    int i;
    if (name == 0)
	name = "unknown";

    switch(gv.type())
    {
	case GenieWorkspaceArrayType:
	{
	    char* field_name = new char[strlen(name)+10];	// enough space for some big numbers
	    GenieWorkspaceArray& work = *(gv.gwa);
	    closeOpenSDS(fileid);
//	    NXmakegroup(fileid, name, getClassString(gv.value()));
//	    NXopengroup(fileid, name, getClassString(gv.value()));
	    NXmakegroup(fileid, name, "GXWorkspace");
	    NXopengroup(fileid, name, "GXWorkspace");
	    for(i=0; i<work.length(); i++)
	    {
	        GenieVariable v;
	        v.become(work[i]);
		sprintf(field_name, "%s_%d", name, i+1);
		addGenieVariable(v, field_name, fileid);
	    }
	    delete[] field_name;
	    closeOpenSDS(fileid);
	    NXclosegroup(fileid);
	    break;
	}

	case GenieWorkspaceType:
	{
	    char* field_name;
	    char* class_string = strdup(getClassString(gv.value()));
	    closeOpenSDS(fileid);
// Fix up Nx (genie) -> NX (NeXus)
	    if (class_string[2] == 'N' && class_string[3] == 'x')
	    {
		class_string[3] = 'X';
	    }
// Use class_string+2 to skip GX prefix
	    NXmakegroup(fileid, name, class_string+2);
	    NXopengroup(fileid, name, class_string+2);
	    GenieWorkspace& work = *(gv.gw);
// We locate an array and write that out first so that
// any other values become attributes of that array
	    int do_first = 0;
	    for(i=1; i<=work.length(); i++)
	    {
		if ( indexOOP(work.value(), i) != nilOOP && 
			isAKindOf(OOPCLASS(work.getOOP(i)), arrayClass) )
		{
		    do_first = i;
		}
	    }
	    if (do_first != 0)
	    {
	 	field_name = OOPToString(indexOOP(work.value(), do_first));
		DeMangle(field_name);
	    	GenieVariable vv;
		vv.become(work.getOOP(do_first));
		addGenieVariable(vv, field_name, fileid);
		free(field_name);
	    }
	    for(i=1; i<=work.length(); i++)
	    {
	 	field_name = OOPToString(indexOOP(work.value(), i));
		if (field_name && i != do_first)
		{
	    	    GenieVariable v;
		    v.become(work.getOOP(i));
		    DeMangle(field_name);
		    addGenieVariable(v, field_name, fileid);
		    free(field_name);
		}
	    }
	    closeOpenSDS(fileid);
	    NXclosegroup(fileid);
	    free(class_string);
	    break;
	}

	case GenieRealArrayType:
	{
	    GenieRealArray& array = *(gv.gra);
	    int* dims_array = array.dimsArray();
	    closeOpenSDS(fileid);
	    NXmakedata(fileid, name, DFNT_FLOAT64, array.dimsSize(), dims_array);
	    NXopendata(fileid, name);
	    NXputdata(fileid, (g_real*)array);
	    m_sds_open = true;
	    break;
	}

	case GenieIntegerArrayType:
	{
	    GenieIntegerArray& array = *(gv.gia);
	    int* dims_array = array.dimsArray();
	    closeOpenSDS(fileid);
	    NXmakedata(fileid, name, DFNT_INT32, array.dimsSize(), dims_array);
	    NXopendata(fileid, name);
	    NXputdata(fileid, (int*)array);
	    m_sds_open = true;
	    break;
	}

	case GenieRealType:
	{
	    float64 val = *(gv.gr);
	    NXputattr(fileid, name, &val, 1, DFNT_FLOAT64);
	    break;
	}

	case GenieIntegerType:
	{
	    int32 val = *(gv.gi);
	    NXputattr(fileid, name, &val, 1, DFNT_INT32);
	    break;
	}

	case GenieStringType:
	{
	    const char* val = *(gv.gs);
	    if ( (val != 0) && (strlen(val) > 0) )
	    {
	        NXputattr(fileid, name, (void *)val, strlen(val), DFNT_CHAR8);
	    }
	    else // doesn't seem to cope with 0 length strings
	    {
	        NXputattr(fileid, name, " ", 1, DFNT_CHAR8);
	    }
	    break;
	}

	case GenieStringArrayType:
	    gerr << "NEXUS: Sorry, String Array is not yet implemented" << endl;
	case GenieUndefinedType:
	{
	    int val(0);	// DFNT_INT16 not used by NeXus, so we'll used as undefined
	    NXputattr(fileid, name, &val, 1, DFNT_INT16);
	    break;
	}
	
	default:
	    gerr << "NEXUS: Invalid GenieVariable type = " << gv.type() << endl;
	    break;
    }
    return 0;
}

ostream& NeXusIntermediateFile::print(ostream& os) const
{
    os << m_filename << endl;
    return os;
}

int NeXusIntermediateFile::seek(const char* label)
{
    return 1;
}

int NeXusIntermediateFile::seek(int block)
{
    return 1;
}

int NeXusIntermediateFile::readBlock(char*& label, GenieVariable& gv)
{
    return 1;
}

int NeXusIntermediateFile::readLabel(char*& label)
{
    return 1;
}

// rename illegal characters
int NeXusIntermediateFile::cleanFilename(char* name)
{
    for(int i=0; i<(int)strlen(name); i++)
    {
	switch(name[i])
	{
	    case '-':
		name[i] = '_';
	 	break;

	    default:
		break;
	}
    }
    return 0;
}

#include <hkit.h>

int NeXusIntermediateFile::printHDFTag(ostream& os, int tag)
{
    int i;
            for(i=0; i<sizeof(tag_descriptions)/sizeof(tag_descript_t); i++)
            {
                if (tag_descriptions[i].tag == tag)
                {
                    os << "Found HDF tag " << tag_descriptions[i].desc << " " << tag_descriptions[i].name << endl;
                }
            }
            for(i=0; i<sizeof(nt_descriptions)/sizeof(nt_descript_t); i++)
            {
                if (nt_descriptions[i].nt == tag)
                {
                    os << "Found HDF type " << nt_descriptions[i].desc << " " << nt_descriptions[i].name << endl;
                }
            }
    return 0;
}
