/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

// parser.cxx : implementation file
//

#include "geniedefs.h"

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <gnustring.h>
#include <assert.h>

#include "smalltalk.h"
#include "parsedefs.h"
#include "parser.h"

/////////////////////////////////////////////////////////////////////////////
// CParser
//
// This class encapsulates the Open GENIE command parser, i.e. the bit which
// takes in GCL and turns it into executeable smalltalk.  The result of
// compiling GCL input in either a file or buffer supplied to this class is
// to produce a smalltalk string suitable for submission to the Smalltalk's
// own parser.
//
// The GCL is parsed, first into an intermediate tree structure and then
// ultimately back out into an ASCII string of smalltak
//

// Global pointer to current parser declared only in the implementation
// to allow access to the class from the C called friend function getc_parser().
CParser *g_pCurrentParser = NULL;

unsigned CParser::s_dwSessionID = 0;

CParser::CParser(CSmalltalk& smalltalk,
                 const char* szDebugFile, const char* szInstrumentFile,
                 bool bErrorContinue, bool bYaccDebugging) :
    m_pSmalltalk(&smalltalk),
    m_gstrDebugFile(szDebugFile),
    m_gstrInstrumentFile(szInstrumentFile),
    m_bErrorContinue(bErrorContinue),
    m_bYaccDebugging(bYaccDebugging)
{
    if (s_dwSessionID != 0)
    {
	    gerr << "Not allowed to create more than one parser object per process- aborting" << endl;
            exit(1);
    }
    else
    {
	s_dwSessionID++;

        // Track all parser output in a new debug files if one is supplied (rename old one to .old)
        m_fpDebugOut = 0;
        if (m_gstrDebugFile != "" )
        {
            gprintif("Writing Parser output to \"%s\"\n", (const char*)m_gstrDebugFile);
            m_fpDebugOut = fopen(m_gstrDebugFile,"w");
        }
   
        // Parser instrumentation is written by YACC parser code if this FILE* is non-null
        g_fpTypeFile = 0;
        if (m_gstrInstrumentFile != "")
        {
            g_fpTypeFile = fopen(m_gstrInstrumentFile, "w");
        }
    
        // make this object available to getc_parser()
        g_pCurrentParser = this;
    }
}

CParser::~CParser()
{
    g_pCurrentParser = NULL;    // Publish our obituary
    
    if ( m_fpDebugOut )         // Close debugging file if open
    {
        fclose(m_fpDebugOut);
    }

    if (g_fpTypeFile)           // Close instrumenting file
    {
        fclose(g_fpTypeFile);
    }

    // Do this last in case something bombs in the kill
    s_dwSessionID--;
}


// Convert the parse tree created by compilation into an ASCII
// string and send it to the interpreter.
int CParser::SendToSmalltalk(node *message)
{
    char *buffer;
    int retval;

    buffer = (char *)malloc(nlength(message)+1);
    buffer[0] = '\0';

    printNode(buffer, message);                         // Decode the parse tree into buffer

    retval = m_pSmalltalk->EvaluateCommand(buffer);	// send the generated code to smalltalk

    if ( m_fpDebugOut )
    {
        fprintf(m_fpDebugOut, "%s", buffer); fflush(m_fpDebugOut);
    }

    free(buffer);
    return retval;
}


// Runs parser on either file or buffer input until we
// get an error or run out of input
int CParser::RunParser( bool bInteractive, bool bPrintDiagnostics, FILE* fpInput)
{

    int errors = 0, lines, status = 1;
    node *pCode;

    fill_buffers();				    /* Fill parser buffers with sentinel values */		        
    initialize_parser_stats();			    /* Set line and error counts to 0 */			

    while ( status )
    {
    	initialize_parser(!bInteractive, m_bYaccDebugging);     /* Set up the parser */

	m_fpInputFile = fpInput;				/* ensure we take input from the right file */
	status = yyparse();				        /* Start the parser */
        pCode = interrogate_parser(&errors, &lines);		/* Get the results */

    	if (status)
    	{
//            assert(pCode->valid);				// Hmm, code not valid after a parse error
    	    if ( nlength(pCode) && !bInteractive )
            {
                gprintif(".");
            }
    	    if (SendToSmalltalk(pCode))			    	/* call back with each chunk of code */
	    {
		errors++;
		if (!m_bErrorContinue)
		    status = 0;
	    }
    	}
        if (pCode->valid);
            deleteNode(pCode);
    }

    if (bPrintDiagnostics)
    {
        gprintif("\n");
        PrintBufferDiagnostics(lines);      /* Check no errors occurred and print stats */
    }

    return errors;
}


// Compile units of input from the source (file or internal buffer) until
// there is no more input or an error occurs.
//
// The default unit of compilation is one complete lexical construct
// from the YACC specified grammar, usually a single line or for non-interactive
// constructs, a single PROCEDURE or multi-line statement.
//
// These units are compiled and sent straight to the interpreter.

// Interprets a single string
int CParser::CompileString(const char* command, bool bInteractive)
{
    m_pszCommand = (char *)command;             // get pointer to string to parse

    return RunParser( bInteractive, FALSE, 0 ); // don't report lines or files compiled to the output
}


// Interprets a complete file from disk
int CParser::CompileFile(const char* szFilename)
{
    int nErrorCount = 0;
    FILE* fpInputFile;	    // Keep a copy of the pointer on the stack for recursion
    
    if ( (fpInputFile = fopen(szFilename, "r")) == 0 )
    {
        return EOF;
    }

    nErrorCount = RunParser( false, true, fpInputFile );     // non-interactive, print informational confirmation

    fclose(fpInputFile);

    return nErrorCount;
}



void CParser::PrintBufferDiagnostics(int lines)
{
    int bspace;

    gprintif("\t%d lines scanned/compiled", lines);
    if( (bspace = check_buffers()) > 0 )
    {
        gprintdf("\tbuffer space remaining = %d, ", lines, bspace);
        p_stats();
        gprintif("\n");
    }
    else
    {
    	gprintef("Buffer space overflowed on compilation - code not loaded properly\n");
        p_stats();
    	gprintif("\tOne procedure is probably too long or complex,\n\ttry splitting it into smaller procedures\n");
    }
}



///////////////////////////////////////////////////////////////////////////////////////////////
// External C based routines to link with LEX and YACC
///////////////////////////////////////////////////////////////////////////////////////////////


//CALLBACKS

// Called externally by YYLEX (in C) to get hold of the next character for the lexical analyser
// (note this is a friend of CParser).  m_fpInputFile contains the current input file pointer.
int getc_parser()
{
    if (g_pCurrentParser->m_fpInputFile)		// if this is NULL, we are parsing from a string
    {
        return getc(g_pCurrentParser->m_fpInputFile);
    }
    //
    // special cases when simulating getc from a buffer
    //
    else if (g_pCurrentParser->m_pszCommand == 0)	// Called again after returning '\n', so return EOF
    {
	return EOF;
    }
    else if (*g_pCurrentParser->m_pszCommand)		// in middle of string, return next character
    {
	return *g_pCurrentParser->m_pszCommand++;
    }
    else 				// End of string, so put back the stripped '\n'
    {
	g_pCurrentParser->m_pszCommand = 0;
	return '\n';
    }
}
