/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "geniedefs.h"
//+++
//
// CLASS: Picture
//
// CLASS DESCRIPTION: 
//
//---
// *** include string headers
#include <gnustring.h>

#include "Drawable_objects.h"
#include "Viewport.h"
#include "Drawable_objectsPTR_List.h"
#include "Picture.h"
#include "Pginclude.h"
#include "Errors.h"

static const char* rcsid = "@(#)$Id: picture.cxx,v 1.10 1997/03/26 16:26:38 cmm Exp $";
USE_RCSID(rcsid);

//---
// RE_DISPLAY function that calls
// the xxx.display() routine
//
void redisplay(Drawable_objectsPTR& item) {
	static int drawit;

	if (item == NULL) {
	    return;
	}

	// *** switch on drawing if viewport is visible
	if (item->line_colour == -1 && item->invisible == 0)
		drawit = 1;

	// *** switch off drawing if viewport is invisible
	if (item->line_colour == -1 && item->invisible > 0)
		drawit = 0;

	// *** if both item and its viewport are visible then draw it.
	if (item->invisible == 0 && drawit == 1)
		item->display();
	}


//---
// Constructor
//
Picture::Picture() {
	}

//---
// Destructor
//
Picture::~Picture() {
	}

//---
// Add a drawable object to the display list
// note that the list is in REVERSE order
//
int Picture::disp_add(Drawable_objectsPTR newobj) {

	return disp_list.push(newobj);
	}

//---
// Remove a drawable object from the list specified by
// number
// if (objnum > 0), it is relative to current object 
// if (objnum <= 0), it is an absolute item index on the visible list
//
void Picture::disp_remove(int objnum) {
	int i=0,j;
	Drawable_objectsPTR del;

	if (objnum <= 0) // absolute index
	{
  	    int n_visible(0);
	    for(j=0; j<disp_list.length();j++) 
	    {
	        del = disp_list.get(j);
	        if (del->invisible == 0) 
		    n_visible++;
	    }
	    objnum = n_visible + objnum;	
 	}

	// *** check that list is long enough
	if ( (objnum > disp_list.length()) || (objnum < 1) )
	{
		Errors syserr;
		syserr.add_error("Picture ",
		    "display list too short",
		    "primitive number was too high");
		return;
	}

	// go down the list, mark things for no-undelete
	for(j=0; j<disp_list.length();j++) {

		del = disp_list.get(j);
		if (del->invisible > 0) 
		{
		    del->invisible++;
		}
	}


	 // go UP the list to find the item to delete
	 //
	 for(j=disp_list.length()-1; j>=0; j--) {

		del = disp_list.get(j);

		if (del->invisible == 0) 
		    i++;		

		if (i == objnum) {
		    if (del->line_colour == -1){
			Errors syserr;
			syserr.add_error ("Picture ",
			    "Cannot delete window using UNDRAW ",
			    "use WINDOW/DELETE instead.");
		    }
		    else {
				del->invisible = 1;
		    }
		 }
	}
	return;
}

//---
// re display the whole list
// NB: the list order must be reversed
//
void Picture::re_display() {
	Drawable_objectsPTR del;
	int j;

	 for(j=0;j<disp_list.length();j++) {
		del = disp_list.get(j);
		redisplay(del);
	    }
	return;
	}

//---
// Remove a viewport from the list specified by
// number
//
void Picture::remove_View(int viewnum) {
	int count = 0,i;
	Drawable_objectsPTR del;

	// *** when a viewport is removed, all primitives
	//	plotted in that window will also be removed.

	// *** get name of viewport to be deleted
	for(i=disp_list.length()-1; i>=0; i--) { 
		del = disp_list.get(i);

		// mark it for no-undelete
		if (del->invisible > 0) {
			del->invisible++;
		}

		// *** is it a viewport?
		else if (del->line_colour == -1) {
		    count--;
		    if (count == viewnum) del->invisible = 1;
		}

	}
	Errors syserr;
	if (viewnum < count) {
			syserr.add_error ("Picture ",
			    "Viewport not on display list ",
			    "viewport-to-delete number too great.");
	}

// If we have a window ID (a +ve number) the above will not have found anything
	if (viewnum >= 0)
	{
	    if (viewnum < disp_list.length())
	    {
	        del = disp_list.get(viewnum);
	        if (del->line_colour == -1 && del->invisible == 0)
		{
		    del->invisible = 1;
		}
		else
		{
		    syserr.add_error("Picture", "Item is not a window or has already been deleted", "");
		}
	    }
	    else
	    {
		syserr.add_error("Picture", "Invalid window ID", "");
	    }
	}
		    
	return;
}

// ---
// Undo the last delete
//
void Picture::undo_remove() {
	Drawable_objectsPTR del;
	int i;

	// *** check for errors
	if (disp_list.length() == 0) {
		Errors syserr;
		syserr.add_error ("Picture ",
			"No items in picture list",
			"return to the previous picture.");
		}

	// *** step through display list
	//	reduce only last undo
	for(i=0;i<disp_list.length();i++) {
		del = disp_list.get(i);

		// *** reduce all deletion counts be one
		if (del->invisible > 0) del->invisible = del->invisible-1;
	}

}

// ---
// get cursor anywhere on picture
// safer than using Viewport::cursor alone
//
void Picture::cursor(int viewnumber,
			float* world,
			float* device,
			String& letter) {

	int counter = 0,i;
	Drawable_objectsPTR del;

	for(i=0;i<disp_list.length();i++) {

	    del = disp_list.get(i);

	    // *** if a viewport
	    if (del->line_colour == -1) {
		counter++;

		// *** display all views after required
		if (counter >= viewnumber) {
		    del->display();
		 }

		// *** get cursor
		if (counter == viewnumber) {
			del->cursor(world,device,letter);
		}
	    }
	}

    }
 
// Return an absolout object index
int Picture::objectNumber(Drawable_objectsPTR PTR)
{
    return disp_list.find(PTR);
}


int Picture::length()
{
    return disp_list.length();
}

ostream& Picture::identity(ostream& os, int objnum)
{
    return disp_list.identity(os, objnum);
}

int Picture::query(GenieWorkspace& info, int objnum)
{
    Drawable_objectsPTR del = objectPointer(objnum);
    if (del == 0)
	return -1;
    del->query(info);
    return 0;
}

// Convert an object ID (+ or -) info a display list object
// return NULL on error
Drawable_objectsPTR Picture::objectPointer(int objnum)
{
    Errors syserr;
    Drawable_objectsPTR del;
    int nvisible(0), j;
    if (objnum >= 0)		// absolute index
    {
	if (objnum < disp_list.length())
	{
	    return disp_list.get(objnum);
	}
	else
	{
	    syserr.add_error("Picture ","display list too short","primitive number was too high");
	}
    }
    else // relative index
    {
	for(j=disp_list.length()-1; j>=0; j--)
	{
	    del = disp_list.get(j);

	    if (del->invisible == 0)
                nvisible--; 

	    if (nvisible == objnum)
	 	return del;
	}
	syserr.add_error("Picture ","display list too short","primitive number was too high");
    }
    return 0;
}
