/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "geniedefs.h"
// -----------------------------------------------------------------------------
// PICTURE_FUNCTIONS.C
//
// Description: C++ code to allow creation of a Genie object-graphics Picture.
//		A pointer to the picture must remain in scope at all times
//		since it is the only link to the display list.
//
//

// *** local include only
#include "Pictlist.h"
#include "Genie_functions.h"
#include "string.h"

static const char* rcsid = "@(#)$Id: picture_functions.cxx,v 1.12 1997/07/06 14:13:28 faa Exp $";
USE_RCSID(rcsid);

// ID of last picture created i.e. the end of the current list
// Valid indices are 0 .. Last_Picture
static int Last_Picture = -1;

// return 0 on success, -1 if pictnum is invalid
int select_picture(int pictnum) 
{
    if (pictnum < 0 || pictnum  >= GENIE_MAXPICT)
    {
	*g_err << "PICTURE: Picture index " << pictnum << " is out of range" << endl;
	ginf << "PICTURE: Index must be in the range " << 0 << " to " << GENIE_MAXPICT-1 << endl;
 	return -1;
    }
    if (PicturePTR[pictnum] == 0)
    {
	*g_err << "PICTURE: Index " << pictnum << " does not point to a valid picture" << endl;
	return -1;
    }
    Current_Picture = pictnum;
    return 0;
}
    

// ---
// *** create a single new picture, remove an old picture
//	if required.
// Returns -1 on error
int create_picture() {

	// *** increment current picture number
	Last_Picture++;
	if (Last_Picture == GENIE_MAXPICT) Last_Picture = 0;

	// *** DELETE any picture to overwrite
	if (PicturePTR[Last_Picture] != 0) {
		delete PicturePTR[Last_Picture];
	}

	// *** NEW a picture on heap
	PicturePTR[Last_Picture] = new Picture;

	if (PicturePTR[Last_Picture] != 0)
	{
	    Current_Picture = Last_Picture;
	    return Last_Picture;
	}
	else
	    return -1;		// new failed, not enough memory for picture
}	

// ---
// *** destruct all the objects in all pictures
//
void delete_pictures() {

	int loop;

	for (loop = 0; loop<GENIE_MAXPICT; loop++) {

	    // *** check that picture is defined
	    if (PicturePTR[loop] != 0) {
		delete PicturePTR[loop];
		PicturePTR[loop] = 0;
		}
	 }
	}

// ---
// *** get a cursor in a viewport number contained within the picture
//
void get_picture_cursor(int view_num,
			float *world,
			float *device,
			char *letter) {

	// *** create a string
	String reply = "  ";

	// ** update the display
        flush_buffer();

	// *** get the cursor in the picture
	PicturePTR[Current_Picture]->cursor(view_num, world,device,reply);
	
	// *** convert the string to a character pointer
	strcpy(letter,reply);

}

// ---
// *** redraw a picture, this also sets it as the current one
//	if picture number 0 is specified then the current 
//	picture is redrawn, the number counts backwards.
//
void redraw_picture(int pictnum, int reset) {

	 // *** get picture to redraw
	 pictnum = Current_Picture + pictnum;

	 // *** if picture is out of bounds then wrap
	 if (pictnum < 0) pictnum = pictnum + GENIE_MAXPICT;
	 if (pictnum >= GENIE_MAXPICT) pictnum = pictnum - GENIE_MAXPICT;

	 // *** check for an undefined picture
	 if (PicturePTR[pictnum] == 0) {
	    error_add("Redraw picture", "Invalid picture", " ");
	    return;
	    }

	 // *** do the redraw
	 PicturePTR[pictnum]->re_display();

	// *** set this to be current
	if (reset == 1)	Current_Picture = pictnum;

	// ** update the display
	flush_buffer();

    }

// ---
// *** Remove a primitive from the display list
//	the last one to be drawn is removed, it may// ---
// *** redraw the current picture after a window resize
//
//	not be a viewport.
//
void undraw(int num) {

	PicturePTR[Current_Picture]->disp_remove(num);

	}

// ---
// *** undraw all starts a new picture
//
void undraw_all() {
	void clear_device();

	// *** clear the screen
	clear_device();

	// ** update the display
	flush_buffer();

	// *** start a new picture
	create_picture();

    }

// ---
// *** removes a window and all primitives drawn in it
//	from the picture
//
void window_delete(int viewnum) {

	PicturePTR[Current_Picture]->remove_View(viewnum);
	}

// ---
// *** undo_undraw, this function undoes the last delete,
//	be it a window or primitive.
//
void undo_undraw() {

	// *** check for empty picture
	if (PicturePTR[Current_Picture] != 0) {
	    PicturePTR[Current_Picture]->undo_remove();

	    // ** update the display
	    flush_buffer();

	    }
	}

// ---
// *** getpictnum, returns the number of the current picture
//
int getpictnum() {
	return Current_Picture;
	}


// Find out internal information about an object on the display list
int query_object(GenieWorkspace& info, g_objectid objid)
{
    PicturePTR[objid[ObjectPicture]]->query(info, objid[ObjectPosition]);
    return 0;
}
