/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "geniedefs.h"

#include <stdlib.h>
#include <iostream.h>
#include <fcntl.h>
#include <setjmp.h>
#include <signal.h>
#ifdef HAVE_UNISTD_H
#   include <unistd.h>
#else
#   include "getopt.h"
#endif

#ifdef __osf__
extern "C" {
void for_rtl_init_ (int *, char **);
int for_rtl_finish_( void );
	}
#endif /* __osf__ */

#ifdef _WIN32
#   include <string.h>
#   include <io.h>      // and some WIN 32 specific local helpers
    static void make_argv( LPSTR lpCmdLine, int& argc, char**& argv);
    static void fix_handle(FILE* stream, DWORD handle, const char* opts);
    extern "C" /*_declspec(dllimport)*/ void TkConsoleCreate(void);
#endif /* _WIN32 */

#include "smalltalk.h"	    // smalltalk
#include "parsedefs.h"	    // parser
#include "parser.h"
#include "ghistory.h"	    // genie
#include "fileset.h"
#include "genie.h"
#include "g_tktcl.h"	    // Tcl/Tk
#include "gsignals.h"
#include "genie_cmodule.h"  // for cmodule_freeallheaps()
#include "g_commandline.h"

static CGenie::Options parse_options(int argc, char** argv);   // Retrieve command line options
static void help_options();

static const char* rcsid = "@(#)$Id: tkgenie.cxx,v 1.30 1998/11/02 17:46:17 cmm Exp $";
USE_RCSID(rcsid);


//
// Set up local signal handler for escape from TK loop
//
static sigjmp_buf ljContext;
static void term_signal(int) { siglongjmp(ljContext, 1); }


//
// Main program is BinMain on NT but remains main on Unix & VMS
// (BinMain is called from TheApp::Run() in MFC with a command line
//  instead of argc/argv)
//
#if defined(_WIN32)

// WinMain on NT & WIN9x
int BinMain(LPSTR lpCmdLine)
{
    char **argv;
    int argc;

    make_argv(lpCmdLine, argc, argv);	    // construct an argv and argc

#else

// Main on Unix & VMS
int main(int argc, char* argv[], char *arge[])
{

#endif /* defined(_WIN32) */


    // Initialise FORTRAN RTL 
#ifdef __osf__
    int for_argc = 1;
    char* for_argv[] = { "gmain", 0 };
    for_rtl_init_(&for_argc, for_argv);
#endif /* __osf__ */

    // Set up any user configurable options for running the genie
    CGenie::Options options = parse_options( argc, argv );

    // Place to do any customising of file paths (add parameters to fs constructor)
    CGenie genie(options, "GENIE");
    CFileSet *pfs = CFileSet::s_this;		// get copy of fileset pointer
    cout << pfs->m_gstrFileSetError << endl;	// Report any errors or messages


    // Take startup commands form file pointed to by GENIE_TCL_INIT
    char* tk_argv[3];
    tk_argv[0] = argv[0];
    tk_argv[1] = (char *)(const char *)pfs->m_gstrTCLInitialisation;
    tk_argv[2] = 0;
    volatile int tk_argc = (tk_argv[1] != 0 ? 2 : 1);

#if defined(_WIN32)
    Tcl_SetPanicProc(GenieTclPanic);
    TkConsoleCreate();
#endif /* _WIN32 */

    if (getenv("GENIE_DEBUG") == NULL)
    {
        genie_signal(SIGABRT, term_signal);
        genie_signal(SIGTERM, term_signal);
        if (!sigsetjmp(ljContext, 1))		// Tk_Main exits with a user(ish) signal
        {    
	    if (genie.TestOption(CGenie::COMMANDLINE))
	    {
		g_commandline(&genie);
	    }
	    else
	    {
		Tk_Main(tk_argc, tk_argv, Genie_Tcl_AppInit);
	    }
        }
    }
    else
    {
	if (genie.TestOption(CGenie::COMMANDLINE))
	{
	    g_commandline(&genie);
	}
	else
	{
	    Tk_Main(tk_argc, tk_argv, Genie_Tcl_AppInit);
	}
    }
#ifdef __osf__
    for_rtl_finish_();
#endif /* __osf__ */
    cmodule_freeallheaps();
    return 0;
}

////// END OF MAIN PROGRAM //////


//////////////////////////////////////////////////////////////////////////
// Local routine to parse argv and set the options enum correctly from the
// command line parameters.
static CGenie::Options parse_options( int argc, char** argv)
{
    int options = CGenie::INFORMATIONALMESSAGES;	// Always enabled

    int ch;
    while ( (ch = getopt(argc, argv, "gscydhnlL")) != -1 )
       switch ( ch )
       {
        case 'g':
	    options |= (CGenie::GENDEBUG | CGenie::TERMINALSTREAMS);	// Log intermediate smalltalk code to boot.st
	    break;
        case 's':
	    options |= (CGenie::SAVEIMAGE | CGenie::TERMINALSTREAMS);	// Build a new image
	    break;
        case 'c':
	    options |= (CGenie::DEBUGMESSAGES | CGenie::FILECHECK | CGenie::TERMINALSTREAMS);   // For developers, reloads new files
	    break;
        case 'y':
	    options |= CGenie::YACCDEBUG;   // Yacc parser debugging - if you can interpret it!
	    break;
        case 'l':
        case 'L':
	    options |= (CGenie::COMMANDLINE | CGenie::TERMINALSTREAMS);   // comamnd line only genie
	    break;
        case 'd':
	    options |= (CGenie::DEBUGMESSAGES | CGenie::CRASHDEBUG);   // For developers, disables signal handling
	    break;
        case 'h':
	    help_options();    // for users
	    exit(0);
	    break;
        case 'n':
	    options |= (CGenie::DEBUGMESSAGES | CGenie::NOSTOP);	    // For developers, ignores smallatlk errors
	    break;
	}

    ch = optind;
    String temp;

//
// Look for args after flags:   genie [flags] [gcl_init_file] [tcl_init_file] [smalltalk_image]
// Trap zero length strings as wel get these from VMS due to coming via
// a command file
//
    if (argv[ch] != 0)
    {
	if (strlen(argv[ch]) > 0)
	{
	    cat("GENIE_GCL_INIT=", argv[ch], temp);
	    putenv(strdup(temp));
	}
	ch++;
    }
    if (argv[ch] != 0)
    {
	if (strlen(argv[ch]) > 0)
	{
	    cat("GENIE_TCL_INIT=", argv[ch], temp);
	    putenv(strdup(temp));
	}
	ch++;
    }
    if (argv[ch] != 0)
    {
	if (strlen(argv[ch]) > 0)
	{
	    cat("GENIE_SMALLTALK_IMAGE=", argv[ch], temp);
	    putenv(strdup(temp));
	}
	ch++;
    }

    return (CGenie::Options)options;	// Convert back to enum

}


// Just print out the options on a -h
static void help_options()
{
    // Have to cheat here because we have no streams at the time this would be useful!
#undef printf
    printf("Options understood by GENIE are:\n");
    printf("\t-g\tDebug code generation\n");
    printf("\t-s\tSave an execution image\n");
    printf("\t-c\tLoad any newer GCL files\n");
    printf("\t-y\tDebug yacc parser\n\n");
    printf("\t-l\tCommand line GENIE\n\n");
    printf("\t-d\tDebug disable exception handling\n");
    printf("\t-n\tDebug nostop - keep going even on interpreter errors\n");
    printf("\t-h\tThis message\n");
    printf("Symbols translated:\n");
    printf("\tGENIE_GCL_INIT\t\tUser initialization file ( default ./genieinit.gcl )\n");
    printf("\tGENIE_SMALLTALK_IMAGE\tSmalltalk saved image file ( default ./genie.im )\n");
    printf("\tGENIE_TK_DIR\t\tWhere tk files live ( default for ./tkinit.tcl )\n");

    return;
}

#if HAVE_LIBF2C
// For shared version of libf2c on linux - needed to remove undefined symbol
extern "C" void MAIN__()
{
    printf(__FILE__ ": MAIN__ called\n");
    exit(1);
}
#endif /* HAVE_LIBF2C */

//////////////////////////////////////////////////////////////////////
// Helpers for WIN32 only below here

#ifdef _WIN32
	    
static void fix_handle(FILE* stream, DWORD handle, const char* opts)
{
    int hCrt;
    FILE* hf;
    hCrt = _open_osfhandle((long)GetStdHandle(handle), _O_TEXT);
    hf = _fdopen(hCrt, opts);
    *stream = *hf;
    setvbuf(stream, NULL, _IONBF, 0);
}


void make_argv( LPSTR lpCmdLine, int& argc, char**& argv)
{
    static char buffer[SHORTCHARBUFF], buffer2[SHORTCHARBUFF];
    int i, max_size = 2;	// estimate of argv size
    char *p, *pszExecName, *pszFileName;

    char *vb[3] = {0,0,0}, *lp, *szCmdCopy = strdup(lpCmdLine);

    // The first section here sorts out a mix of parameter variations, typical examples being
    //
    //     1)   "The Executable path" "First file name" -c -s	    (WNT .exe name + initial file + args)
    //	   2)	DOSEXE~1.EXE "First file name" -c -s		    (DOS compatible name + initial file + args.)
    //	   3)   "The Executable path" "First file name" -c -s	    (initial filename + arguments)
    //	   4)   "The Executable path" -c -s			    (no initial filename)
    //
    // The initial document name if given MUST be double quoted or it will just be treated as another paramter
    // surplus quotes in later paramters will also break the parsing.

    // Parse out .EXE name, either double quoted with spaces or not quoted without spaces
    lp = szCmdCopy;
    if (lp[0] == '"')
    {
	pszExecName = strtok(lp, "\"");
    }
    else    
    {
	pszExecName = strtok(lp, " \t");
    }

    // Tokenize the rest on double quotes
    for ( i=0; vb[i] = strtok(NULL, "\""); i++);

    // use a heuristics to work out where the arguments are in the tokenized strings
    // (this assumes only one extra file in "" or this will break for sure)
    if (vb[0] == NULL)
    {
	pszFileName = "";
	lp = 0;
    }
    else if (vb[1] == NULL)
    {
	if (lpCmdLine[strlen(lpCmdLine)-1] != '"')	// Look for the " at the end to decide
	{
	    pszFileName = "";
	    lp = vb[0];
	}
	else
	{
	    pszFileName = vb[0];
	    lp = 0;	    
	}
    }
    else if (vb[2] == NULL)
    {
	if (strspn(vb[0], " \t") == strlen(vb[0]))	// All white space in vb[0]
	{
	    pszFileName = vb[1];
	    lp = vb[2];
	}
	else
	{
	    pszFileName = vb[0];
	    lp = vb[1];
	}
    }
    else
    {
	pszFileName = vb[1];
	lp = vb[2];
    }
    
    // from now on, we simply build an arvg and argc normally with what is left
    for(p = lp; p != 0 && *p != '\0'; p++)
    {
	if (isspace(*p))
	    max_size++;
    }
    max_size++;	// space for trailing NULL

//    char debuff[200];
//    sprintf(debuff, "echo %s > tmp.out", lpCmdLine);
//    system(debuff);
//    sprintf(debuff, "echo File = %s >> tmp.out", pszFileName);
//    system(debuff);
//    sprintf(debuff, "echo args = %s >> tmp.out", lp);
//    system(debuff);
//    sprintf(debuff, "echo i = %d >> tmp.out", i);
//    system(debuff);
//    sprintf(debuff, "echo strspn(vb[0] = %d, strlen(vb[0] = %d) >> tmp.out", strspn(vb[0], " \t"), strlen(vb[0]));
//    system(debuff);
//    sprintf(debuff, "echo vb[0] = (%s) >> tmp.out", vb[0]);
//    system(debuff);
//    sprintf(debuff, "echo vb[1] = (%s) >> tmp.out", vb[1]);
//    system(debuff);
//    sprintf(debuff, "echo vb[2] = (%s) >> tmp.out", vb[2]);
//    system(debuff);
    
    argv = (char**)malloc(max_size*sizeof(char*));

    // Get the program name and path (probably tkgenie)
    strcpy(buffer, pszExecName);
    for(p = buffer; *p != '\0'; p++)
    {
	if (*p == '\\')
	    *p = '/';
    }

    // split into words starting on whitespace
    argv[0] = buffer;
    for (i = 1, p = lp; p != 0 && *p != '\0'; i++)
    {
	while (isspace(*p)) 
	{
	    p++;
	}
	if (*p == '\0')
	{
	    break;
	}
	argv[i] = p;
	while (*p != '\0' && !isspace(*p)) 
	{
	    p++;
	}
	if (*p != '\0') 
	{
	    *p = '\0';
	    p++;
	}
    }
    argv[i] = NULL;
    argc = i;

    // Pass on the initial file to be looked at by TK interface
    strcpy(buffer2, "GENIE_FIRST_FILE=");
    strcat(buffer2, pszFileName);
    putenv(buffer2);

//    sprintf(debuff, "echo Argc = %d >> tmp.out", argc);
//    system(debuff);

//    free(szCmdCopy);  // buffer done with
//    exit(0);
}

#endif /* _WIN32*/
