/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "geniedefs.h"
#include <math.h>
#include "Pginclude.h"
#include "g_types.h"
#include "two_d_coord_trans_func.h"
#include "transform_utils.h"

static const char* rcsid = "@(#)$Id: two_d_coord_trans_func.cxx,v 1.21 1998/02/02 16:50:42 faa Exp $";
USE_RCSID(rcsid);

// for "val = data(i)", find i for a given val (i is C style i.e. 0 based)
// Returns -1.0 for undefined values
double TwoDimensionalCoordinateTransformFunction::reverseInterpolate1(DrawableValues& data, double val)
{
    if (val == UNDEFINED_PG_FLOAT)
    {
	return -1.0;
    }
    if (data.ndims() != 1)
    {
	gerr << "TwoDimensionalCoordinateTransformFunction - cannot do 2D interpolation" << endl;
	return -1.0;
    }
    int low = bracket(F(val), data.values(), data.npoints());
    if (low == -1)
    {
        return -1.0;
    }
    double frac;
    if (data(low) == data(low+1))
    {
	frac = 0.5;
    }
    else
    {
        frac = (val - data(low)) / (data(low+1) - data(low));
    }
    return (double)low + frac;
}

// Simple linear interpolation along an axis - for "val = data(i)", find val a given i
// index i is 0 based
// linear extrapolate off end of data if needed
double TwoDimensionalCoordinateTransformFunction::interpolate(DrawableValues& data, double i)
{
//    int i0 = (int)i, i1 = (int)(i+1.0); 
    int n, i0 = (int)floor(i), i1 = (int)ceil(i); 
    double alpha = i - (double)i0;
// Are we being asked for something out of range?
// linear extrapolate off end
    if (i0 < 0)
    {
	return data(0) + (data(1) - data(0)) * (double)i;
    }
    if (i1 >= data.npoints())
    {
	n = data.npoints() - 2;
	return data(n) + (data(n+1) - data(n)) * (double)(i - n);
    }
    if ((data(i0) == UNDEFINED_PG_FLOAT) || (data(i1) == UNDEFINED_PG_FLOAT))
    {
	return UNDEFINED_PG_FLOAT;
    }
    else
    {
        return data(i0) + alpha * (data(i1) - data(i0));
    }
}

// Simple 2D linear interpolation along an axis
// Cell formed by array elments (i0,j0) and (i1,j1), with (i,j) inside
// alpha and beta are fractional distances along i and j respectively
double TwoDimensionalCoordinateTransformFunction::interpolate(DrawableValues& data, double i, double j)
{
    int i0 = (int)floor(i), i1=(int)ceil(i);
    double alpha = i - (double)i0;
    int j0 = (int)floor(j), j1=(int)ceil(j);
    double beta = j - (double)j0;
    if ((data(i0,j0) == UNDEFINED_PG_FLOAT) || (data(i1,j0) == UNDEFINED_PG_FLOAT) || 
	(data(i0,j1) == UNDEFINED_PG_FLOAT) || (data(i1,j1) == UNDEFINED_PG_FLOAT))
    {
	return UNDEFINED_PG_FLOAT;
    }
    else
    {
        return data(i0,j0) + alpha * (data(i1,j0) - data(i0,j0))
		       + beta * (data(i0,j1) - data(i0,j0))
	  	       + alpha * beta * (data(i0,j0) + data(i1,j1) - data(i0,j1) - data(i1,j0));
    }
}


TwoDimensionalCoordinateTransformFunction::TwoDimensionalCoordinateTransformFunction(coord_tran_func_t forward_func, 
			coord_tran_func_t reverse_func) : CoordinateTransformFunction(2, forward_func, reverse_func),
	m_x_mode(PointMode), m_y_mode(PointMode)
{
}

TwoDimensionalCoordinateTransformFunction::TwoDimensionalCoordinateTransformFunction(g_real* matrix) :
	CoordinateTransformFunction(2, matrix), m_x_mode(PointMode), m_y_mode(PointMode)
{
}

TwoDimensionalCoordinateTransformFunction::TwoDimensionalCoordinateTransformFunction(g_real* wx, int* wx_dims, int wx_ndims,
	CoordinateTransformMode x_mode, g_real* wy, int* wy_dims, int wy_ndims, CoordinateTransformMode y_mode)
 : CoordinateTransformFunction(), m_x_data(wx, wx_dims, wx_ndims), m_y_data(wy, wy_dims, wy_ndims), m_x_mode(x_mode),
			m_y_mode(y_mode)
{
}

int TwoDimensionalCoordinateTransformFunction::forwardFunction(int* flags, const g_real* val_in, g_real* val_out)
{
    double x_offset, y_offset;
    x_offset = (m_x_mode == BinMode ? 0.5 : 0.0);
    y_offset = (m_y_mode == BinMode ? 0.5 : 0.0);
    if ((m_x_data.values() != 0) && (m_y_data.values() != 0))
    {
	if (m_x_data.ndims() == 1)
	    val_out[0] = interpolate(m_x_data, val_in[0] + x_offset);
	else
	    val_out[0] = interpolate(m_x_data, val_in[0] + x_offset, val_in[1] + y_offset);

	if (m_y_data.ndims() == 1)
	    val_out[1] = interpolate(m_y_data, val_in[1] + y_offset);
	else
	    val_out[1] = interpolate(m_y_data, val_in[0] + x_offset, val_in[1] + y_offset);

	return 0;
    }
    else
    {
	return CoordinateTransformFunction::forwardFunction(flags, val_in, val_out);
    }
}

// prototype should agree with "coord_tran_func_t" typedef
int TwoDimensionalCoordinateTransformFunction::reverseFunction(int* flags, const g_real* val_in, g_real* val_out)
{
    if ((m_x_data.values() != 0) && (m_y_data.values() != 0))
    {
// Assumes m_x_data and m_y_data are one dimensional
        double x_offset = (m_x_mode == BinMode ? 0.5 : 0.0);
        double y_offset = (m_y_mode == BinMode ? 0.5 : 0.0);
	val_out[0] = reverseInterpolate1(m_x_data, val_in[0]) - x_offset;
	val_out[1] = reverseInterpolate1(m_y_data, val_in[1]) - y_offset;
	return 0;
    }
    else if (m_reverse_matrix != 0)
    {
// This is an optimisation hook - CoordinateTransformFunction::reverseFunction() could also cope with this
	return matrixFunction(m_reverse_matrix, val_in, val_out);
    }
    else
    {
	return CoordinateTransformFunction::reverseFunction(flags, val_in, val_out);
    }
}
