/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

/*
 *      Module:         units
 *      Author:         Freddie Akeroyd, ISIS
 *      Purpose:        Routines for changing the units of spectra
 *
 */
#include "geniedefs.h"
#ifdef HAVE_UNISTD_H
#	include <unistd.h>
#endif
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "genie_types.h"
#include "units.h"
#include "units_p.h"

#define	MAX(__a, __b)	((__a) < (__b) ? (__b) : (__a))

#define DEFINED(__a)	((__a) != DBL_MAX)	/* to handle undefined real genie variables */

/*
 * error codes:
 *-1 = points truncated in array
 * 0 = success
 * 1 = input unit code invalid
 * 2 = output unit code invalid
 * 3 = inelastic, not specified emode/efixed properly
 * 4 = input unit is channel number --- cannot convert
 * 5 = invalid l1, l2 or TwoTheta
 */

static const char* rcsid = "@(#)$Id: units.cxx,v 1.14 1998/10/14 09:04:54 faa Exp $";
USE_RCSID(rcsid);

static char* convert_units_errmsg(void);

static g_integer truncate(g_real[], g_real[], g_real[], g_integer, 
				g_real, g_real,g_integer);

static int flip_order(g_real xarray[], g_real yarray[], g_real earray[], 
			g_integer ntc, g_integer histogram_mode);

int convert_units(g_real xarray[], g_real yarray[], g_real earray[], 
		  g_integer *ntcp, const char* xunit_in, const char** xunit_out, 
		  const char* yunit_in, const char** yunit_out, g_integer unit_convert,
		  g_real l1, g_real l2,
		  g_integer emode, g_real efixed, g_real TwoTheta, g_real *ymax, g_real emin, 
		  g_real emax, g_real delta, g_integer histogram_mode)
{
    g_integer ntc;
    int i, error_code = 0;
    g_real SinTheta, ltot;
    unit_code_struct xunit_code_in, xunit_code_out, yunit_code_in,
		yunit_code_out;
    ntc = *ntcp;
/* only y can be dimensionless */
    if ( (xunit_code_in = (unit_code_struct) unit_code(xunit_in, 'x')) <= 0)
    {
	error_code = 1;
	gerr << "UNITS Error: invalid xunit_in = " << xunit_in << endl;
	return(error_code);		
    }
    if ( xunit_code_in == 1 )
    {
	error_code = 4;
	gerr << "UNITS Error: cannot change units from channel number" << endl;
	return(error_code);		
    }
    if ( (yunit_code_in = (unit_code_struct) unit_code(yunit_in, 'y')) == -1)
    {
	error_code = 1;
	gerr << "UNITS Error: invalid yunit_in = " << yunit_in << endl;
	return(error_code);		/* invalid unit_in */
    }
    if ( unit_convert < 1 || unit_convert >= NUNIT )
    {
	error_code = 2;
	gerr << "UNITS Error: invalid output unit code = " << unit_convert << endl;
	return(error_code);
    }
    xunit_code_out = (unit_code_struct) unit_convert;
    yunit_code_out = ((yunit_code_in == 0) ? (unit_code_struct)0 : xunit_code_out);
/* 
 * deal with some easy options 
 */
    if ( xunit_code_in == xunit_code_out )
    {
	*xunit_out = xunit_in;
	*yunit_out = yunit_in;
	sprintf(error_message_buffer, "\n");
	return(error_code);
    }
    if (( xunit_code_in >= 9 || xunit_code_out >= 9 ) && ( emode <= 0 || emode >= 3 || efixed <= 0.0 ))
    {
	error_code = 3;
	gerr << "UNITS Error: either emode (" << emode << ") or efixed (" << efixed << ") wrongly specified" << endl;
	return(error_code);
    }

    ltot = l1 + l2;
    if ( ltot == 0.0 )
    {
	error_code = 5;
	gerr << "UNITS Error: total flight path L = 0" << endl;
	return(error_code);
    }
    SinTheta = sin(fabs(TwoTheta) * DEG2RAD / 2.0);
    if (histogram_mode)
    {
	error_code = convert_units_binmode(xarray, yarray, earray, 
		  ntc, xunit_in, xunit_code_in, xunit_code_out, 
		  yunit_code_in, yunit_code_out, 
		  l1, l2,
		  emode, efixed, SinTheta, delta);
    }
    else
    {
	error_code = convert_units_pointmode(xarray, yarray, earray, 
		  ntc, xunit_in, xunit_code_in, xunit_code_out, 
		  yunit_code_in, yunit_code_out, 
		  l1, l2,
		  emode, efixed, SinTheta, delta);
    }

    if (xarray[0] > xarray[ntc-1])
    {
	flip_order(xarray, yarray, earray, ntc, histogram_mode);
    }
    if ( xunit_code_out == DeltaE )
    {
	ntc = truncate(xarray, yarray, earray, ntc, emin, emax, histogram_mode);
    }
    else
    {
	ntc = truncate(xarray, yarray, earray, ntc, -BIGVAL*0.99, BIGVAL*0.99, histogram_mode);
    }

    *ymax = 0.0;
    for(i=0; i<ntc; i++)
    {
	if ( DEFINED(yarray[i]) && (yarray[i] > *ymax) )
    	{
	    *ymax = yarray[i];
    	}
    }

    *xunit_out = unit_name( (int)xunit_code_out, 'x');
    *yunit_out = unit_name( (int)yunit_code_out, 'y');

    if ( (xunit_out == NULL) || (yunit_out == NULL) )
    {
	gerr << "UNITS Error: xunit_out or yuit_out is NULL" << endl;
	error_code = 2;
    }
    *ntcp = ntc;
    return(error_code);
}


/* returns -1 on error */
int unit_code(const char* cunit, char axis)
{
    int i;
    if ( axis == 'x' )
    {
    	for(i=0; i<NUNIT; i++)
    	{
	    if (strcmp(cunit, xunit_names[i]) == 0)
	    {
	        break;
	    }
    	}
    }
    else if ( axis == 'y' )
    {
    	for(i=0; i<NUNIT; i++)
    	{
	    if (strcmp(cunit, yunit_names[i]) == 0)
	    {
	        break;
	    }
    	}
    }
    else
    {
	i = NUNIT;
    }


    if ( i == NUNIT  )
    {
	return(-1);		/* invalid unit string */
    }
    else
    {
	return(i);
    }
}


/* returns NULL pointer on error */
const char* unit_name(int iunit, char axis)
{
    static const char* invalid_unit = "Invalid Unit";
    if ( (iunit < 0) || (iunit >= NUNIT) )
    {
	return(invalid_unit);		/* invalid unit code */
    }
    else if ( axis == 'x' )
    {
	return(xunit_names[iunit]);
    }
    else if ( axis == 'y' )
    {
	return(yunit_names[iunit]);
    }
    else
    {
	return(invalid_unit);
    }

}

static char* convert_units_errmsg(void)
{
    return(error_message_buffer);
}

/* remove all points outside range */

static g_integer truncate(g_real xarray[], g_real yarray[], g_real earray[], g_integer ntc, g_real xmin, g_real xmax, g_integer histogram_mode)
{
    g_integer i,j=0;
    for(i=0; i<ntc; i++)
    {
	if (xarray[i] >= xmin && xarray[i] <= xmax)
	{
	    xarray[j] = xarray[i];
	    yarray[j] = yarray[i];
	    earray[j] = earray[i];
	    j++;
	}
    }
    if (histogram_mode)
    {
	if (xarray[ntc] >= xmin && xarray[ntc] <= xmax)
	{
	    xarray[j] = xarray[ntc];
	    j++;
	}
	return j - 1;	/* new ntc value */
    }
    else
    {
        return j;	/* new ntc value */
    }
}

static int flip_order(g_real xarray[], g_real yarray[], g_real earray[], 
			g_integer ntc, g_integer histogram_mode)
{
        int i,j;
	g_real factor;
        for(i=0; i < ntc/2; i++)
        {
            j = ntc - i - 1;

            factor = yarray[i];
            yarray[i] = yarray[j];
            yarray[j] = factor;

            factor = earray[i];
            earray[i] = earray[j];
            earray[j] = factor;
        }
        for(i=0; i < (ntc+histogram_mode)/2; i++)
        {
            j = ntc + histogram_mode - i -1;
            factor = xarray[i];
            xarray[i] = xarray[j];
            xarray[j] = factor;
        }
    	return 0;
}
