/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

/*
 *      Module:         units
 *      Author:         Freddie Akeroyd, ISIS
 *      Purpose:        Routines for changing the units of spectra
 *
 */
#include "geniedefs.h"
#ifdef HAVE_UNISTD_H
#	include <unistd.h>
#endif
#include <stdio.h>
#include <math.h>
#include <string.h>
#include "genie_types.h"
#include "units.h"
#include "units_p.h"

#define	MAX(__a, __b)	((__a) < (__b) ? (__b) : (__a))

#define DEFINED(__a)	((__a) != DBL_MAX)	/* to handle undefined real genie variables */

/*
 * error codes:
 *-1 = points truncated in array
 * 0 = success
 * 1 = input unit code invalid
 * 2 = output unit code invalid
 * 3 = inelastic, not specified emode/efixed properly
 * 4 = input unit is channel number --- cannot convert
 * 5 = invalid l1, l2 or TwoTheta
 */

static const char* rcsid = "@(#)$Id: units_pointmode.cxx,v 1.3 1997/10/20 10:14:54 cmm Exp $";
USE_RCSID(rcsid);

static int TOFToEnergy(g_real scale, g_integer emode, g_real efixed, 
	g_real xarray[], g_real yarray[], g_real earray[],
	g_integer ntc, g_real l1, g_real l2);

static int energyToTOF(g_real scale, g_integer emode, g_real efixed, 
	g_real xarray[], g_real yarray[], g_real earray[],
	g_integer ntc, g_real l1, g_real l2);

// compute derivative at element i of xarray[ntc]
inline double gradient(const g_real* xarray, g_integer ntc, g_integer i)
{
    if (i <= 0)
    {
	return xarray[1] - xarray[0];
    }
    else if (i >= (ntc - 1))
    {
	return xarray[ntc-1] - xarray[ntc-2];
    }
    else
    {
	return (xarray[i+1] - xarray[i-1]) / 2.0;
    }
}

int convert_units_pointmode(g_real xarray[], g_real yarray[], g_real earray[],
                  g_integer ntc, const char* xunit_in,
                  unit_code_struct xunit_code_in,
                  unit_code_struct xunit_code_out,
                  unit_code_struct yunit_code_in,
                  unit_code_struct yunit_code_out,
                  g_real l1, g_real l2,
                  g_integer emode, g_real efixed, g_real SinTheta, g_real delta)
{
    int i, /*j,*/ error_code = 0;
    g_real factor, fac2, FourPi, ltot, /*step,*/ t1, e2, t2;
    ltot = l1 + l2;
    FourPi = 4.0 * PI;
    if ( ltot == 0.0 )
    {
	error_code = 5;
	gerr << "UNITS Error: total flight path L = 0" << endl;
	return(error_code);
    }
    if (( xunit_code_in >= 9 || xunit_code_out >= 9 ) && ( emode <= 0 || emode >= 3 || efixed <= 0.0 ))
    {
	error_code = 3;
	gerr << "UNITS Error: either emode (" << emode << ") or efixed (" << efixed << ") wrongly specified" << endl;
	return(error_code);
    }

/* 
 * now convert units to a common base of TOF in microseconds 
 * C1 and C2 defined in units.h
 */

    switch(xunit_code_in) {

    case Dimensionless:
	error_code = 1;
	gerr << "UNITS Error: invalid xunit_in = " << xunit_in << endl;
	return(error_code);		
	break;

    case ChannelNo:
	error_code = 4;
	gerr << "UNITS Error: cannot change units from channel number" << endl;
	return(error_code);
	break;

    case TOF:
	for(i=0; i<ntc; i++)
	{
	    xarray[i] = xarray[i] - delta;
	}
	break;

    case Wavelength:
	factor = ltot * C1;
	for(i=0; i<ntc; i++)
	{
	    xarray[i] = xarray[i] * factor;
	    yarray[i] = yarray[i] / factor;
	    earray[i] = earray[i] / factor;
	}
	break;

    case dSpacing:
	factor = 2.0 * ltot * SinTheta * C1;
	for(i=0; i<ntc; i++)
	{
	    xarray[i] = xarray[i] * factor;
	    yarray[i] = yarray[i] / factor;
	    earray[i] = earray[i] / factor;
	}
	break;

    case Energy:
	factor = C1 * ltot * sqrt(C2);
	for(i=0; i<ntc; i++)
	{
	    xarray[i] = factor / sqrt(xarray[i]);
	    fac2 = 2.0 * factor * factor / (xarray[i] * xarray[i] * xarray[i]);
	    yarray[i] = yarray[i] * fac2;
	    earray[i] = earray[i] * fac2;
	}
	break;

    case MomentumTransfer:
	factor = FourPi * C1 * ltot * SinTheta;
	for(i=0; i<ntc; i++)
	{
	    xarray[i] = factor / xarray[i];
	    fac2 = factor / (xarray[i] * xarray[i]);
	    yarray[i] = yarray[i] * fac2;
	    earray[i] = earray[i] * fac2;
	}
	break;

    case Qsquared:
	factor = FourPi * C1 * ltot * SinTheta;
	for(i=0; i<ntc; i++)
	{
	    xarray[i] = factor / sqrt(xarray[i]);
	    fac2 = 2.0 * factor * factor / (xarray[i] * xarray[i] * xarray[i]);
	    yarray[i] = yarray[i] * fac2;
	    earray[i] = earray[i] * fac2;
	}
	break;

    case ReciprocalTOF:
	factor = ltot;
	for(i=0; i<ntc; i++)
	{
	    xarray[i] = xarray[i] * factor;
	    yarray[i] = yarray[i] / factor;
	    earray[i] = earray[i] / factor;
	}
	break;

    case DeltaE:
	energyToTOF(1.0, emode, efixed, xarray, yarray, earray, ntc, l1, l2);
	break;

    case DeltaE_inWaveNumber:
	energyToTOF(MEV2WAVENUMBER, emode, efixed, xarray, yarray, earray, ntc, l1, l2);
	break;

    case PrimaryWavelength:
	if ( emode == 2 )
	{
	    e2 = efixed;
	    t2 = C1 * l2 * sqrt(C2/e2);
	    factor = l1 * C1;
	    for(i=0; i<ntc; i++)
	    {
	    	t1 = factor * xarray[i];
		xarray[i] = t1 + t2;
		yarray[i] = yarray[i] / factor;
		earray[i] = earray[i] / factor;
	    }
	}
	break;

    default:
	error_code = 1;
	gerr << "UNITS Error: invalid xunit_in = " << xunit_in << endl;
	return(error_code);		
	break;
    }

// Now checked in individual commands - caused problems on instruments with "funny" hold off values
#if 0
/*
 * TOF cannot be -ve
 */
    for(i=0; i<ntc+1; i++)
    {
	if (xarray[i] <= 0.0)
	{
	    error_code = 1;
	    gerr << "UNITS Error: zero or negative time-of-flight with x[" << i+1 << "] = " << xarray[i] << endl;
	    gerr << "             check that the \"hold off\" (DELTA) value of " << delta << " is appropriate" << endl;
	    return(error_code);		
	}
    }
#endif

/*
 * Now convert TOF to required units
 */
    switch(xunit_code_out) {

    case Dimensionless:
	error_code = 3;
	gerr << "UNITS Error: converting to dimensionless X" << endl;
	return(error_code);
	break;

    case ChannelNo:
#if 1
	factor = (xarray[1] - xarray[0]) / 2.0; // Guess halfwidth of bin 
	yarray[0] = yarray[0] * 2.0 * factor;
	earray[0] = earray[0] * 2.0 * factor;
	for(i=1; i<ntc; i++)
	{
	    factor = xarray[i] - xarray[i-1] - factor;
	    yarray[i] = yarray[i] * 2.0 * factor;
	    earray[i] = earray[i] * 2.0 * factor;
	}
#else
	for(i=0; i<ntc; i++)
	{
	    factor = gradient(xarray, ntc, i);
	    yarray[i] = yarray[i] * fabs(factor);
	    earray[i] = earray[i] * fabs(factor);
	}
#endif
	for(i=0; i<ntc; i++)
	{
	    xarray[i] = (g_real)i + 1.0;
	}
	break;

    case TOF:
	for(i=0; i<ntc; i++)
	{
	    xarray[i] = xarray[i] + delta;
	}
	break;

    case Wavelength:
	factor = ltot * C1;
	for(i=0; i<ntc; i++)
	{
	    xarray[i] = xarray[i] / factor;
	    yarray[i] = yarray[i] * factor;
	    earray[i] = earray[i] * factor;
	}
	break;

    case dSpacing:
	factor = 2.0 * ltot * SinTheta * C1;
    	if (factor == 0.0)
    	{
	    error_code = 5;
	    gerr << "UNITS Error: L1+L2 or Sin(TwoTheta) is 0.0" << endl;
	    return(error_code);
    	}
	for(i=0; i<ntc; i++)
	{
	    xarray[i] = xarray[i] / factor;
	    yarray[i] = yarray[i] * factor;
	    earray[i] = earray[i] * factor;
	}
	break;

    case Energy:
	factor = C1 * C1 * ltot * ltot * C2;
	for(i=0; i<ntc; i++)
	{
	    t1 = xarray[i] * xarray[i];
	    yarray[i] = yarray[i] * t1 * xarray[i] / (2.0 * factor);
	    earray[i] = earray[i] * t1 * xarray[i] / (2.0 * factor);
	    xarray[i] = factor / t1;
	}
	break;

    case MomentumTransfer:
	factor = FourPi * C1 * ltot * SinTheta;
	for(i=0; i<ntc; i++)
	{
	    if (xarray[i] == 0.0)
		xarray[i] = (factor > 0.0 ? BIGVAL : -BIGVAL);
	    else
	        xarray[i] = factor / xarray[i];
	    yarray[i] = yarray[i] * factor / (xarray[i] * xarray[i]);
	    earray[i] = earray[i] * factor / (xarray[i] * xarray[i]);
	}
	break;

    case Qsquared:
	factor = FourPi * C1 * ltot * SinTheta;
	factor = factor * factor;
	for(i=0; i<ntc; i++)
	{
	    t1 = xarray[i] * xarray[i];
	    yarray[i] = yarray[i] * t1 * xarray[i] / (2.0 * factor);
	    earray[i] = earray[i] * t1 * xarray[i] / (2.0 * factor);
	    if (xarray[i] == 0.0)
		xarray[i] = BIGVAL;
	    else
	        xarray[i] = factor / t1;
	}
	break;

    case ReciprocalTOF:
	factor = ltot;
	for(i=0; i<ntc; i++)
	{
	    xarray[i] = xarray[i] / factor;
	    yarray[i] = yarray[i] * factor;
	    earray[i] = earray[i] * factor;
	}
	break;

    case DeltaE:
	TOFToEnergy(1.0, emode, efixed, xarray, yarray, earray, ntc, l1, l2);
	break;

    case DeltaE_inWaveNumber:
	TOFToEnergy(MEV2WAVENUMBER, emode, efixed, xarray, yarray, earray, ntc, l1, l2);
	break;


    case PrimaryWavelength:
	if ( emode == 2 )
	{
	    e2 = efixed;
	    t2 = C1 * l2 * sqrt(C2/e2);
	    factor = l1 * C1;
	    for(i=0; i<ntc; i++)
	    {
		t1 = xarray[i] - t2;
		xarray[i] = (( t1 <= 0.0 ) ? ( -BIGVAL ) : ( t1 / factor ));
		yarray[i] = yarray[i] * factor;
		earray[i] = earray[i] * factor;
	    }
	}
	break;

    default:
	error_code = 2;
	gerr << "UNITS Error: invalid output unit code = "  << xunit_code_out << endl;
	return(error_code);		
	break;
    }
    return(error_code);
}


/*
 * 0 success 
 * 1 invalid emode
 */
static int energyToTOF(g_real scale, g_integer emode, g_real efixed,
	g_real xarray[], g_real yarray[], g_real earray[],
	g_integer ntc, g_real l1, g_real l2)
{
    g_real factor, e1, e2, t1, t2, fac2;
    g_integer i;
	if (emode == 1)
	{
	    e1 = efixed;
	    t1 = C1 * l1 * sqrt(C2/e1);
	    factor = C1 * l2 * sqrt(C2);
	    for(i=0; i<ntc; i++)
	    {
		e2 = e1 - xarray[i] / scale;	// xarray[i] / scale --> from whatever to meV
		if (e2 > 0.0)
		{
		    t2 = factor / sqrt(e2);
		    fac2 = 2.0 * scale * factor * factor / (t2 * t2 * t2);
		    yarray[i] = yarray[i] * fac2;
		    earray[i] = earray[i] * fac2;
		    xarray[i] = t1 + t2;
		}
		else
		{
		    xarray[i] = BIGVAL;
		    yarray[i] = 0.0;
		    earray[i] = 0.0;
		}
	    }
	}
	else if (emode == 2)
	{
	    e2 = efixed;
	    t2 = C1 * l2 * sqrt(C2/e2);
	    factor = C1 * l1 * sqrt(C2); 
	    for(i=0; i<ntc; i++)
	    {
		e1 = e2 + xarray[i] / scale;	// xarray[i] / scale --> from whatever to meV
		if (e1 > 0.0)
		{
		    t1 = factor / sqrt(e1);
		    fac2 = 2.0 * scale * factor * factor / (t1 * t1 * t1);
		    yarray[i] = yarray[i] * fac2;
		    earray[i] = earray[i] * fac2;
		    xarray[i] = t1 + t2;
		}
		else
		{
		    xarray[i] = -BIGVAL;
		    yarray[i] = 0.0;
		    earray[i] = 0.0;
		}
	    }
	}
	else
	{
	    return 1;
	}
	return 0;
}

/*
 * 0 success
 * 1 invalid emode
 */
static int TOFToEnergy(g_real scale, g_integer emode, g_real efixed, 
	g_real xarray[], g_real yarray[], g_real earray[],
	g_integer ntc, g_real l1, g_real l2)
{
    g_real factor, e1, e2, t1, t2, fac2;
    g_integer i;
	if (emode == 1)
	{
	    e1 = efixed;
	    t1 = C1 * l1 * sqrt(C2/e1);
	    factor = C1 * C1 * l2 * l2 * C2;
	    for(i=0; i<ntc; i++)
	    {
		t2 = xarray[i] - t1;
		if ( t2 <= 0.0 )
		{
		    xarray[i] = -BIGVAL;
		    yarray[i] = 0.0;
		    earray[i] = 0.0;
		}
		else
		{
		    e2 = factor / (t2 * t2);
		    xarray[i] = (e1 - e2) * scale;	// e1 * scale --> from meV to whatever
		    fac2 = t2 * t2 * t2 / (2.0 * factor * scale);
		    yarray[i] = yarray[i] * fac2;
		    earray[i] = earray[i] * fac2;
		}
	    }
	}
	else if (emode == 2)
	{
	    e2 = efixed;
	    t2 = C1 * l2 * sqrt(C2/e2);
	    factor = C1 * C1 * l1 * l1 * C2;
	    for(i=0; i<ntc; i++)
	    {
		t1 = xarray[i] - t2;
		if ( t1 <= 0.0 )
		{
		    xarray[i] = BIGVAL;
		    yarray[i] = 0.0;
		    earray[i] = 0.0;
		}
		else
		{
		    e1 = factor / (t1 * t1);
		    xarray[i] = (e1 - e2) * scale;	// e1 * scale --> from meV to whatever
		    fac2 = t1 * t1 * t1 / (2.0 * factor * scale);
		    yarray[i] = yarray[i] * fac2;
		    earray[i] = earray[i] * fac2;
		}
	    }
	}
        else 
	{
	    return 1;
	}
    return 0;
}
