/*  Open GENIE, a program for neutron scattering data analysis & display
    ISIS facility, Rutherford Appleton Laboratory, Chilton, Didcot, OXON UK
    Copyright (C) 1996,1997

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA. */

#include "geniedefs.h"
//+++
//
// CLASS: Viewport
//
//
//---
#include <gnustring.h>

#include "Drawable_objects.h"
#include "Viewport.h"
#include "Pginclude.h"
#include "Errors.h"

#include "genievars.h"

static const char* rcsid = "@(#)$Id: viewport.cxx,v 1.12 1998/06/23 16:56:29 faa Exp $";
USE_RCSID(rcsid);

// *** constructor
Viewport::Viewport(float x1,
		   float x2,
		   float y1,
		   float y2,
		   float X1,
		   float X2,
		   float Y1,
		   float Y2,
		   int colour) {

	m_colour = colour;
	
	    // *** set visibility
	    invisible = 0;

	// *** set device co-ordinates
	X_D_origin = x1;
	X_D_limit = x2;
	Y_D_origin = y1;
	Y_D_limit = y2;

	// *** set world co-ordinates
	X_W_origin = X1;
	X_W_limit = X2;
	Y_W_origin = Y1;
	Y_W_limit = Y2;

	// *** set line colour for display list
	line_colour = -1;
}

// *** constructor to copy other viewport
Viewport::Viewport(float x1,
		   float x2,
		   float y1,
		   float y2,
		   Viewport original) {

	    // *** set visibility
	    invisible = 0;

	// *** set device co-ordinates
	X_D_origin = x1;
	X_D_limit = x2;
	Y_D_origin = y1;
	Y_D_limit = y2;

	// *** set world co-ordinates
	X_W_origin = original.X_W_origin;
	X_W_limit = original.X_W_limit;
	Y_W_origin = original.Y_W_origin;
	X_W_limit = original.X_W_limit;

	m_colour = original.m_colour;

	// *** set line colour for display list
	line_colour = -1;
	}


// *** destructor
Viewport::~Viewport() {

	}

// *** set the current viewport
void Viewport::display() {
	static char v_buffer[1024];

	if (X_D_origin == X_D_limit) {
		Errors syserr;
		sprintf(v_buffer, "X window max/min values are equal at %f", X_D_origin);
		syserr.add_error ("Window ",
		    v_buffer,
		    "make X device coordinates different");
		return;
		}

	if (Y_D_origin == Y_D_limit) {
		Errors syserr;
		sprintf(v_buffer, "Y window max/min values are equal at %f", Y_D_origin);
		syserr.add_error ("Window ",
		    v_buffer,
		    "make Y device coordinates different");
		return;
		}

	// *** enforce device coordinates
	pgvport(X_D_origin, X_D_limit, Y_D_origin, Y_D_limit);

	// *** check for 0-0 plotting
	if (X_W_origin == X_W_limit) {
		Errors syserr;
		sprintf(v_buffer, "X axis world max/min values are equal at %f", X_W_origin);
		syserr.add_error ("Window ",
		    v_buffer,
		    "make X world coordinates different");
		return;
		}

	if (Y_W_origin == Y_W_limit) {
		Errors syserr;
		sprintf(v_buffer, "Y axis world max/min values are equal at %f", Y_W_origin);
		syserr.add_error ("Window ",
		    v_buffer,
		    "make Y world coordinates different");
		return;
		}

	// *** enforce world coordinates
	pgwindow(X_W_origin, X_W_limit, Y_W_origin, Y_W_limit);

	// *** colour window if required, by drawing a filled rectangle
        if (m_colour != 0)
	{
	    pgsci(m_colour);
	    pgsfs(1);		// solid
	    pgrect(X_W_origin, X_W_limit, Y_W_origin, Y_W_limit);
	}
}

// *** cursor routine
void Viewport::cursor(float world[],
		      float device[],
		      String& letter) {

	    float X_val,Y_val;
	    float W_ratio;
	    float D_diff;

	    // *** get the cursor position from PGPLOT
	    X_val = world[0];
	    Y_val = world[1];
	    
	// *** enforce device coordinates
	pgvport(X_D_origin, X_D_limit, Y_D_origin, Y_D_limit);

	// *** enforce world coordinates
	pgwindow(X_W_origin, X_W_limit, Y_W_origin, Y_W_limit);

	    PGCURSE_S(X_val,Y_val,letter);
	    if ((const char*)letter == 0) 
		return;
	    letter.upcase();

	    // *** enter the return world coordinates
	    world[0] = X_val;
	    world[1] = Y_val;

	    // *** calculate the device coordinates

	    // *** do X value first
	    W_ratio = (X_val - X_W_origin) / (X_W_limit - X_W_origin);
	    D_diff = X_D_limit - X_D_origin;

	    device[0] = ( W_ratio * D_diff ) + X_D_origin;

	    // *** do Y value next
	    W_ratio = (Y_val - Y_W_origin) / (Y_W_limit - Y_W_origin);
	    D_diff = Y_D_limit - Y_D_origin;

	    device[1] = ( W_ratio * D_diff ) + Y_D_origin;

	    }

// *** given device coordinates, return world
void Viewport::ret_world(float* device,
			    float* world) {
	    float x_val, y_val;

	    // *** do X value first
	    x_val = (device[0] - X_D_origin) / (X_D_limit - X_D_origin);

	    world[0] = (x_val * (X_W_limit - X_W_origin)) + X_W_origin;

	    // *** do Y value
	    y_val = (device[1] - Y_D_origin) / (Y_D_limit - Y_D_origin);

	    world[1] = (y_val * (Y_W_limit - Y_W_origin)) + Y_W_origin;

	}


int Viewport::query(GenieWorkspace& info)
{
    info("TYPE", 1);
    info("DESCRIPTION", "Window");
    info("DXL", X_D_origin);
    info("DXH", X_D_limit);
    info("DYL", Y_D_origin);
    info("DYH", Y_D_limit);
    info("WXL", X_W_origin);
    info("WXH", X_W_limit);
    info("WYL", Y_W_origin);
    info("WYH", Y_W_limit);
    info("COLOUR", m_colour);
    return 0;
}
